/****************************************************************************
 *  Copyright (c) 2018 Oblo Living d.o.o
 *
 *  All Rights Reserved
 *
 *  P R O P R I E T A R Y    &    C O N F I D E N T I A L
 *
 *  -----------------------------------------------------
 *  http://www.obloliving.com/
 *  -----------------------------------------------------
 *
 * \project Oblo WISE Example application - Example application
 * \file LocalDemoDevice.h
 * \brief
 * Example device abstraction class.
 *
 * @Author Igor Stefanovic
 *
 * \notes
 *
 * \history
 *****************************************************************************/
#ifndef LOCAL_DEMO_DEVICE_H
#define LOCAL_DEMO_DEVICE_H

#include "Config.h"
#include "Logger.h"
#include "ssdp/Discovery.h"
#include "messenger/Messages.h"
#include "ipdevice/Property.h"
#include "util/Timer.h"
#include "util/JsonStruct.h"
#include "messenger/Messenger.h"
#include "ipdevice/ActiveDevice.h"
#include "ipdevice/SleepingDevice.h"
#include "ipdevice/Device.h"
#include "PahoLogger.h"

/**
 * @namespace Example
 * @brief Example classes.
 */
namespace Example {

/**
 * @class RPIDemoDevice
 * @brief Raspberry PI demo device.
 * @extends IPDevice::ActiveDevice
 * @implements IPDevice::IDeviceExcludedListener
 * @implements IPDevice::IConnectedListener
 * @implements IPDevice::ITimeChangeListener
 */
class LocalDemoDevice :
        public IPDevice::ActiveDevice,
        public IPDevice::IDeviceExcludedListener,
        public IPDevice::IConnectedListener,
        public IPDevice::ITimeChangeListener,
        public Utility::ILogListener {
public:
    /**
     * @brief Default log level.
     */
    static const Example::Logger::LoggerLevel DEFAULT_LOGGER_LEVEL;

    /**
     * @fn RPIDemoDevice();
     * @brief Constructor.
     */
    LocalDemoDevice();

    /**
     * @fn ~RPIDemoDevice();
     * @brief Destructor.
     */
    ~LocalDemoDevice();

    /**
     * @fn void Start();
     * @brief Starts the raspberry pi demo device.
     */
    void Start();

    /**
     * @fn void DoInclusion();
     * @brief Initiates inclusion procedure.
     */
    void DoInclusion();

    /**
     * @fn void SetConfig(Utility::Config _config);
     * @brief Sets the configuration to the IPDevice::Device.
     * @param _config Config.
     */
    void SetConfig(Utility::Config _config);

    /**
     * @fn void SetData(Utility::Config _data);
     * @brief Sets the data saved in document to the IPDevice::Device if there is any data.
     * @param _data Data.
     */
    void SetData(Utility::Config _data);

    /**
     * @fn void SetDeviceInfomation(IPDevice::DeviceInformation _deviceInformation);
     * @brief Sets the IPDevice::Device information.
     * @param _deviceInformation Information about device.
     */
    void SetDeviceInfomation(IPDevice::DeviceInformation _deviceInformation);

	/**
     * @fn void SetDescProviderUrl(const std::string& _url);
     * @brief Sets url of description provider.
     * @param _url Url.
     */
    void SetDescProviderUrl(const std::string& _url);

    /**
     * @fn void SetNewUptime(Utility::Timer& _timer);
     * @brief Timer callback function that changes the uptime value.
     * @param _timer Timer.
     */
    void SetNewUptime(Utility::Timer& _timer);

    /**
     * @fn void LightStateHandler(unsigned int _groupId, const std::string& _serviceType, const std::string &_propertyName, const Jzon::Node& _newValue);
     * @brief Callback function that is invoked upon state property changing request. IPDevice::Device will change the property.
     * @param _groupId Group id.
     * @param _serviceType Service type.
     * @param _propertyName Property name.
     * @param _newValue New value.
     */
    void LightStateHandler(unsigned int _groupId, const std::string& _serviceType, const std::string &_propertyName, const Jzon::Node& _newValue);

    /**
     * @fn void DimLevelHandler(unsigned int _groupId, const std::string& _serviceType, const std::string &_propertyName, const Jzon::Node& _newValue);
     * @brief Callback function that is invoked upon level property changing request. IPDevice::Device will change the property.
     * @param _groupId Group id.
     * @param _serviceType Service type.
     * @param _propertyName Property name.
     * @param _newValue New value.
     */
    void DimLevelHandler(unsigned int _groupId, const std::string& _serviceType, const std::string &_propertyName, const Jzon::Node& _newValue);

    /**
     * @fn void SimpleAlertStateHandler(unsigned int _groupId, const std::string& _serviceType, const std::string &_propertyName, const Jzon::Node& _newValue);
     * @brief Callback function that is invoked upon state property changing request. IPDevice::Device will change the property.
     * @param _groupId Group id.
     * @param _serviceType Service type.
     * @param _propertyName Property name.
     * @param _newValue New value.
     */
    void SimpleAlertStateHandler(unsigned int _groupId, const std::string& _serviceType, const std::string &_propertyName, const Jzon::Node& _newValue);

    /**
     * @fn void SimpleAlertModeHandler(unsigned int _groupId, const std::string& _serviceType, const std::string &_propertyName, const Jzon::Node& _newValue);
     * @brief Callback function that is invoked upon alert mode property changing request. IPDevice::Device will change the property.
     * @param _groupId Group id.
     * @param _serviceType Service type.
     * @param _propertyName Property name.
     * @param _newValue New value.
     */
    void SimpleAlertModeHandler(unsigned int _groupId, const std::string& _serviceType, const std::string &_propertyName, const Jzon::Node& _newValue);

    /**
     * @fn void SimpleAlertTypeHandler(unsigned int _groupId, const std::string& _serviceType, const std::string &_propertyName, const Jzon::Node& _newValue);
     * @brief Callback function that is invoked upon alert type property changing request. IPDevice::Device will change the property.
     * @param _groupId Group id.
     * @param _serviceType Service type.
     * @param _propertyName Property name.
     * @param _newValue New value.
     */
    void SimpleAlertTypeHandler(unsigned int _groupId, const std::string& _serviceType, const std::string &_propertyName, const Jzon::Node& _newValue);

    /**
     * @fn void DeviceNameHandler(unsigned int _groupId, const std::string& _serviceType, const std::string &_propertyName, const Jzon::Node& _newValue);
     * @brief Callback function that is invoked upon name property changing request. IPDevice::Device will change the property.
     * @param _groupId Group id.
     * @param _serviceType Service type.
     * @param _propertyName Property name.
     * @param _newValue New value.
     */
    void DeviceNameHandler(unsigned int _groupId, const std::string& _serviceType, const std::string &_propertyName, const Jzon::Node& _newValue);

    /**
     * @fn void VolumeLevelHandler(unsigned int _groupId, const std::string& _serviceType, const std::string &_propertyName, const Jzon::Node& _newValue);
     * @brief Callback function that is invoked upon level property changing request. IPDevice::Device will change the property.
     * @param _groupId Group id.
     * @param _serviceType Service type.
     * @param _propertyName Property name.
     * @param _newValue New value.
     */
    void VolumeLevelHandler(unsigned int _groupId, const std::string& _serviceType, const std::string &_propertyName, const Jzon::Node& _newValue);

    /**
     * @fn void LightToggleHandler(unsigned int _groupId, const std::string& _serviceType, const Jzon::Node& _inputList);
     * @brief Callback function that is called upon toggle command invoking request.
     * @param _groupId Group id.
     * @param _serviceType Service type.
     * @param _inputList Input parameters.
     */
    void LightToggleHandler(unsigned int _groupId, const std::string& _serviceType, const Jzon::Node& _inputList);

    /**
     * @fn void DimMoveToHandler(unsigned int _groupId, const std::string& _serviceType, const Jzon::Node& _inputList);
     * @brief Callback function that is called upon moveTo command invoking request.
     * @param _groupId Group id.
     * @param _serviceType Service type.
     * @param _inputList Input parameters.
     */
    void DimMoveToHandler(unsigned int _groupId, const std::string& _serviceType, const Jzon::Node& _inputList);

    /**
     * @fn void DimMoveToHandler(unsigned int _groupId, const std::string& _serviceType, const Jzon::Node& _inputList);
     * @brief Callback function that is called upon stop command invoking request.
     * @param _groupId Group id.
     * @param _serviceType Service type.
     * @param _inputList Input parameters.
     */
    void DimStopHandler(unsigned int _groupId, const std::string& _serviceType, const Jzon::Node& _inputList);

    /**
     * @fn void OnTrace(const std::string &_msg);
     * @brief Callback function that will be called when IPDevice::Device has logs on trace log level to print.
     * @param _msg Log.
     */
    void OnTrace(const std::string &_msg);

    /**
     * @fn void OnDebug(const std::string &_msg);
     * @brief Callback function that will be called when IPDevice::Device has logs on debug log level to print.
     * @param _msg Log.
     */
    void OnDebug(const std::string &_msg);

    /**
     * @fn void OnInfo(const std::string &_msg);
     * @brief Callback function that will be called when IPDevice::Device has logs on info log level to print.
     * @param _msg Log.
     */
    void OnInfo(const std::string &_msg);

    /**
     * @fn void OnWarn(const std::string &_msg);
     * @brief Callback function that will be called when IPDevice::Device has logs on warn log level to print.
     * @param _msg Log.
     */
    void OnWarn(const std::string &_msg);

    /**
     * @fn void OnError(const std::string &_msg);
     * @brief Callback function that will be called when IPDevice::Device has logs on error log level to print.
     * @param _msg Log.
     */
    void OnError(const std::string &_msg);

    /**
     * @fn void OnFatal(const std::string &_msg);
     * @brief Callback function that will be called when IPDevice::Device has logs on error log level to print.
     * @param _msg Log.
     */
    void OnFatal(const std::string &_msg);

    /**
     * @fn void OnDeviceExcluded();
     * @brief Callback function that will be called when IP device is excluded from the hub.
     */
    void OnDeviceExcluded();

    /**
     * @fn void OnConnected(const IPDevice::InclusionInformation& _inclusionResult);
     * @brief Callback function that will be called when IP device is connected to the hub.
     * @param _inclusionResult Information about inclusion process, that might need to be updated.
     */
    void OnConnected(const IPDevice::InclusionInformation& _inclusionResult);

    /**
     * @fn void OnSetTime(unsigned long long _timestamp);
     * @brief Callback function that will be called upon SetTime request arrival.
     * @param _timestamp Timestamp for SetTime request.
     */
    void OnSetTime(unsigned long long _timestamp);

    /**
     * @fn void OnAdjustTime(unsigned long long _diff);
     * @brief Callback function that will be called upon AdjustTime request arrival.
     * @param _diff Timestamp for AdjustTime request.
     */
    void OnAdjustTime(unsigned long long _diff);
private:
    Utility::Timer* m_UptimeTimer;
    Example::Logger m_Logger;
    Example::PahoLogger m_PahoLogger;
    Utility::Config m_Data;
    int m_BootCount;
    bool m_Included;
};

}


#endif // LOCAL_DEMO_DEVICE_H
