#include "LocalDemoDevice.h"

using namespace IPDevice;
using namespace SSDP;
using namespace Messenger;
using namespace Example;
using namespace Utility;

const Example::Logger::LoggerLevel LocalDemoDevice::DEFAULT_LOGGER_LEVEL = Example::Logger::TRACE;

LocalDemoDevice::LocalDemoDevice() : ActiveDevice(), m_Logger("LocalDemoDevice", DEFAULT_LOGGER_LEVEL) {
    Device::SetLogListener(this);
    Device::RegisterConnectedListener(this);
    Device::RegisterDeviceExcludedListener(this);

    // Register services
    // Light service
    LightService* lightService = new LightService();
    lightService->SetStateHandler(SERVICE_PROPERTY_DELEGATE(LocalDemoDevice, LightStateHandler));
    lightService->SetToggleHandler(SERVICE_COMMAND_DELEGATE(LocalDemoDevice, LightToggleHandler));
    AddService(0, lightService);

    // Dim service
    DimService* dimService = new DimService();
    dimService->SetLevelHandler(SERVICE_PROPERTY_DELEGATE(LocalDemoDevice, DimLevelHandler));
    dimService->SetMoveToHandler(SERVICE_COMMAND_DELEGATE(LocalDemoDevice, DimMoveToHandler));
    dimService->SetStopHandler(SERVICE_COMMAND_DELEGATE(LocalDemoDevice, DimStopHandler));
    AddService(0, dimService);

    // Motion detection service
    MotionDetectionService* motionService = new MotionDetectionService();
    AddService(0, motionService);

    // Simple alert service
    SimpleAlertService* simpleAlertService = new SimpleAlertService();
    simpleAlertService->SetStateHandler(SERVICE_PROPERTY_DELEGATE(LocalDemoDevice, SimpleAlertStateHandler));
    simpleAlertService->SetAlertModeHandler(SERVICE_PROPERTY_DELEGATE(LocalDemoDevice, SimpleAlertModeHandler));
    simpleAlertService->SetAlertTypeHandler(SERVICE_PROPERTY_DELEGATE(LocalDemoDevice, SimpleAlertTypeHandler));
    AddService(0, simpleAlertService);

    // Setting the read only property to sound and light
    SetPropertyValueIntern(0, SERVICE_TYPE_SIMPLE_ALERT, PROPERTY_NAME_MODE, Jzon::Node(SimpleAlertService::SOUND_AND_LIGHT), false);

    // Volume service
    VolumeService* volumeService = new VolumeService();
    volumeService->SetLevelHandler(SERVICE_PROPERTY_DELEGATE(LocalDemoDevice, VolumeLevelHandler));
    AddService(0, volumeService);

    // Mandatory services - Device and diagnostics service
    DeviceService* deviceService = new DeviceService();
    deviceService->SetNameHandler(SERVICE_PROPERTY_DELEGATE(LocalDemoDevice, DeviceNameHandler));
    AddService(0, deviceService);

    DiagnosticsService* diagnosticsService = new DiagnosticsService();
    diagnosticsService->UseUptimeProp();
    diagnosticsService->UseBootCountProp();
    AddService(0, diagnosticsService);

    m_UptimeTimer = new Timer(60000, false);
    m_Included = false;
    m_BootCount = 0;
}

LocalDemoDevice::~LocalDemoDevice() {
    Device::SetLogListener(NULL);
    UnregisterDeviceExcludedListener();
    m_UptimeTimer->StopTimer();
    SAFE_DELETE_O(m_UptimeTimer);
}

void LocalDemoDevice::Start() {
    try {
        if (!Device::IsInclusionInfoEmpty()) {
            m_Included = true;
            StartReconnection();
        } else {
            DoInclusion();
        }
    } catch (const ProtocolException& exc) {
        m_Logger.Error(exc.GetMessage());
    }
    m_Logger.Info("Device is ready");
    m_UptimeTimer->StartTimer(Callback<LocalDemoDevice>(*this, &LocalDemoDevice::SetNewUptime));
}

void LocalDemoDevice::DoInclusion() {
    struct timespec timeBegin;
    clock_gettime(CLOCK_MONOTONIC, &timeBegin);
    try {
        InclusionInformation result = StartInclusion();
        if (!result.GetUsername().empty() || !result.GetPassword().empty()) {
            m_Included = true;

            struct timespec timeEnd;
            clock_gettime(CLOCK_MONOTONIC, &timeEnd);

            double accum = (timeEnd.tv_sec - timeBegin.tv_sec) + (timeEnd.tv_nsec - timeBegin.tv_nsec) / 1E9;
            std::stringstream ss;
            ss << "Inclusion time = " << accum << " seconds";
            m_Logger.Debug(ss.str());
        }
    } catch (const BaseException& exc) {
        m_Logger.Error("Inclusion failed, reason: " + exc.GetMessage());
    }
}

void LocalDemoDevice::SetConfig(Config _config) {
    if (_config.HasElements()) {
        m_Logger.SetLogLevel(_config.ExtractStringValue("LogLevel.RPIDevice"));
        ChallengeParameters challengeParam;
        challengeParam.SetChallengeRetry(_config.ExtractIntValue("Challenge.Retry"));
        challengeParam.SetChallengeTimeout(_config.ExtractIntValue("Challenge.Timeout"));
        Device::SetInclusionTimeout(_config.ExtractIntValue("Inclusion.Timeout"));
        Device::SetSearchInterval(_config.ExtractIntValue("Inclusion.SearchInterval"));
        Device::SetChallengeParameters(challengeParam);
        Device::SetKeyFile(_config.ExtractStringValue("Device.KeyPath"));
        Device::SetKeyPasscode(_config.ExtractStringValue("Device.KeyPasscode"));
        Device::SetCertFile(_config.ExtractStringValue("Device.CertPath"));
        Device::SetSecure(_config.ExtractBoolValue("Device.Secured"));

        // For local device
        InclusionInformation prevIncResult;
        prevIncResult.SetInclusionID(_config.ExtractStringValue("Inclusion.InclusionID"));
        prevIncResult.SetID(_config.ExtractIntValue("Inclusion.ID"));
        prevIncResult.SetUsername(_config.ExtractStringValue("Inclusion.Username"));
        prevIncResult.SetPassword(_config.ExtractStringValue("Inclusion.Password"));
        SSDP::HubHolder hub(_config.ExtractStringValue("Inclusion.DomainID"), _config.ExtractStringValue("Inclusion.HomeID"), _config.ExtractStringValue("Inclusion.HubID"), _config.ExtractStringValue("Inclusion.Location"));
        std::string hubTag = hub.GetDomainID() + TOPIC_SEPARATOR + hub.GetHomeID() + TOPIC_SEPARATOR + "hub/" + hub.GetHubID();
        prevIncResult.SetHub(hub);
        prevIncResult.SetHubTag(hubTag);

        Device::SetInclusionInformation(prevIncResult);
    }
}

void LocalDemoDevice::SetData(Config _data) {
    m_Data = _data;
    if (m_Data.HasElements()) {
        InclusionInformation prevIncResult;
        prevIncResult.SetInclusionID(m_Data.ExtractStringValue("InclusionID"));
        prevIncResult.SetID(m_Data.ExtractIntValue("ID"));
        prevIncResult.SetUsername(m_Data.ExtractStringValue("Username"));
        prevIncResult.SetPassword(m_Data.ExtractStringValue("Password"));
        SSDP::HubHolder hub(m_Data.ExtractStringValue("DomainID"), m_Data.ExtractStringValue("HomeID"), m_Data.ExtractStringValue("HubID"), m_Data.ExtractStringValue("Location"));
        std::string hubTag = hub.GetDomainID() + TOPIC_SEPARATOR + hub.GetHomeID() + TOPIC_SEPARATOR + "hub/" + hub.GetHubID();
        prevIncResult.SetHub(hub);
        prevIncResult.SetHubTag(hubTag);
        m_BootCount = m_Data.ExtractIntValue("BootCount", 0);
        if (m_BootCount != 0) {
            ++m_BootCount;
        }
        ActiveDevice::SetPropertyValueIntern(0, SERVICE_TYPE_DIAGNOSTIC, PROPERTY_NAME_BOOT_COUNT, m_BootCount, false);
        m_Data.InsertValue("BootCount", m_BootCount);

        Device::SetInclusionInformation(prevIncResult);
    }
}

void LocalDemoDevice::SetDeviceInfomation(DeviceInformation _deviceInformation) {
    Device::SetDeviceInfomation(_deviceInformation);
    ActiveDevice::SetPropertyValueIntern(0, SERVICE_TYPE_DEVICE, PROPERTY_NAME_NAME, _deviceInformation.GetName(), false);
    ActiveDevice::SetPropertyValueIntern(0, SERVICE_TYPE_DEVICE, PROPERTY_NAME_VERSION, _deviceInformation.GetVersion(), false);
    ActiveDevice::SetPropertyValueIntern(0, SERVICE_TYPE_DEVICE, PROPERTY_NAME_VERSION_NUMBER, _deviceInformation.GetVersionNumber(), false);
    ActiveDevice::SetPropertyValueIntern(0, SERVICE_TYPE_DEVICE, PROPERTY_NAME_PROTOCOL, _deviceInformation.GetProtocolVersion(), false);
}

void LocalDemoDevice::SetDescProviderUrl(const std::string& _url) {
    Device::SetDescriptionProviderUrl(_url);
}

void LocalDemoDevice::SetNewUptime(Timer &_timer) {
    Jzon::Node value = ActiveDevice::GetPropertyValueIntern(0, SERVICE_TYPE_DIAGNOSTIC, PROPERTY_NAME_UP_TIME);
    if (m_Included) {
        ActiveDevice::SetPropertyValueIntern(0, SERVICE_TYPE_DIAGNOSTIC, PROPERTY_NAME_UP_TIME, value.toInt() + 1, true);
    } else {
        ActiveDevice::SetPropertyValueIntern(0, SERVICE_TYPE_DIAGNOSTIC, PROPERTY_NAME_UP_TIME, value.toInt() + 1, false);
    }
}

void LocalDemoDevice::LightStateHandler(unsigned int _groupId, const std::string& _serviceType, const std::string& _propertyName, const Jzon::Node& _newValue) {
    /* TODO: Your code here on change property */
}

void LocalDemoDevice::DimLevelHandler(unsigned int _groupId, const std::string& _serviceType, const std::string& _propertyName, const Jzon::Node& _newValue) {
    /* TODO: Your code here on change property */
}

void LocalDemoDevice::SimpleAlertStateHandler(unsigned int _groupId, const std::string& _serviceType, const std::string& _propertyName, const Jzon::Node& _newValue) {
    /* TODO: Your code here on change property */
}

void LocalDemoDevice::SimpleAlertModeHandler(unsigned int _groupId, const std::string& _serviceType, const std::string& _propertyName, const Jzon::Node& _newValue) {
    /* TODO: Your code here on change property */
}

void LocalDemoDevice::SimpleAlertTypeHandler(unsigned int _groupId, const std::string& _serviceType, const std::string& _propertyName, const Jzon::Node& _newValue) {
    /* TODO: Your code here on change property */
}

void LocalDemoDevice::DeviceNameHandler(unsigned int _groupId, const std::string& _serviceType, const std::string& _propertyName, const Jzon::Node& _newValue) {
    m_Data.InsertStringValue("Name", _newValue.toString());
}

void LocalDemoDevice::VolumeLevelHandler(unsigned int _groupId, const std::string& _serviceType, const std::string& _propertyName, const Jzon::Node& _newValue) {
    /* TODO: Your code here on change property */
}


void LocalDemoDevice::LightToggleHandler(unsigned int _groupId, const std::string &_serviceType, const Jzon::Node &_inputList) {
    Jzon::Node value = ActiveDevice::GetPropertyValueIntern(_groupId, _serviceType, PROPERTY_NAME_STATE);
    if (value.getType() != Jzon::Node::T_INVALID) {
        /* TODO: Your code here on toggle command */
        ActiveDevice::SetPropertyValueIntern(_groupId, _serviceType, PROPERTY_NAME_STATE, !value.toBool(), true);
    }
}

void LocalDemoDevice::DimMoveToHandler(unsigned int _groupId, const std::string &_serviceType, const Jzon::Node &_inputList) {
    /* TODO: Your code here on move to command */
}

void LocalDemoDevice::DimStopHandler(unsigned int _groupId, const std::string &_serviceType, const Jzon::Node &_inputList) {
    /* TODO: Your code here on stop command */
}

void LocalDemoDevice::OnTrace(const std::string &_msg) {
    m_Logger.Trace(_msg);
}

void LocalDemoDevice::OnDebug(const std::string &_msg) {
    m_Logger.Debug(_msg);
}

void LocalDemoDevice::OnInfo(const std::string &_msg) {
    m_Logger.Info(_msg);
}

void LocalDemoDevice::OnWarn(const std::string &_msg) {
    m_Logger.Warn(_msg);
}

void LocalDemoDevice::OnError(const std::string &_msg) {
    m_Logger.Error(_msg);
}

void LocalDemoDevice::OnFatal(const std::string &_msg) {
    m_Logger.Fatal(_msg);
}

void LocalDemoDevice::OnDeviceExcluded() {
    if (m_Included) {
        m_Logger.Info("Device excluded");
        m_Data.CleanAll();
        m_Included = false;
    }
}

void LocalDemoDevice::OnConnected(const InclusionInformation& _inclusionResult) {
    ActivateServices();
    RegisterDeviceExcludedListener(this);
    RegisterTimeChangeListener(this);
    SetLogListener(this);

    SendDeviceRebootedEvent();

    /* Filling config */
    m_Data.InsertStringValue("DomainID", _inclusionResult.GetHub().GetDomainID());
    m_Data.InsertStringValue("HomeID", _inclusionResult.GetHub().GetHomeID());
    m_Data.InsertStringValue("HubID", _inclusionResult.GetHub().GetHubID());
    m_Data.InsertStringValue("Location", _inclusionResult.GetHub().GetLocation());
    m_Data.InsertStringValue("InclusionID", _inclusionResult.GetInclusionID());
    m_Data.InsertStringValue("Username", _inclusionResult.GetUsername());
    m_Data.InsertStringValue("Password", _inclusionResult.GetPassword());
    m_Data.InsertValue("ID", _inclusionResult.GetID());
    m_Data.InsertValue("BootCount", m_BootCount);
}

void LocalDemoDevice::OnSetTime(unsigned long long _timestamp) {
    // TODO: Check this
    std::stringstream ss;
    ss << "Set time request <t=" << _timestamp << ">";
    m_Logger.Debug(ss.str());
    struct timeval timeVal;
    timeVal.tv_sec = _timestamp / 1000;
    timeVal.tv_usec = (_timestamp % 1000) * 1000;
    if (settimeofday(&timeVal, NULL) == -1) {
        m_Logger.Error("Error while setting time <e=" + std::string(std::strerror(errno)) + ">");
    }
}

void LocalDemoDevice::OnAdjustTime(unsigned long long _diff) {
    // TODO: Check this
    std::stringstream ss;
    ss << "Adjust time request <d=" << _diff << ">";
    m_Logger.Debug(ss.str());
    struct timeval timeVal;
    gettimeofday(&timeVal, NULL);
    timeVal.tv_sec += _diff / 1000;
    timeVal.tv_usec += (_diff % 1000) * 1000;
    if (settimeofday(&timeVal, NULL) == -1) {
        m_Logger.Error("Error while adjusting time <e=" + std::string(std::strerror(errno)) + ">");
    }
}
