/****************************************************************************
 *  Copyright (c) 2018 Oblo Living d.o.o
 *
 *  All Rights Reserved
 *
 *  P R O P R I E T A R Y    &    C O N F I D E N T I A L
 *
 *  -----------------------------------------------------
 *  http://www.obloliving.com/
 *  -----------------------------------------------------
 *
 * \project Oblo WISE Example application - Example application
 * \file Config.h
 * \brief
 * Example config abstraction class.
 *
 * @Author Igor Stefanovic
 *
 * \notes
 *
 * \history
 *****************************************************************************/
#ifndef CONFIG_H
#define CONFIG_H

#include "util/Common.h"
#include <fstream>
#include <istream>

/**
 * @namespace Utility
 * @brief Abstraction utility classes.
 */
namespace Utility {

/**
 * @class Config
 * @brief Class for parsing config files.
 */
class DLLExport Config {
public:
    /**
     * @typedef std::map<std::string, std::string> ConfigMap;
     * @brief Map of key and value of configuration.
     */
    typedef std::map<std::string, std::string> ConfigMap;

    /**
     * @fn Config();
     * @brief Constructor.
     */
    Config() {}

    /**
     * @fn Config(const std::string& _fileName);
     * @brief Constructor.
     * @param _fileName Name of the file that will be parsed.
     */
    Config(const std::string& _fileName);

    /**
     * @fn bool HasElements() const;
     * @brief Indication whether the map has elements.
     * @return Indication.
     */
    bool HasElements() const;

    /**
     * @fn int ExtractStringValue(const std::string& _key, const std::string& _def = "") const;
     * @brief Extract value from map.
     * @param _key Key of value.
     * @param _def Default value.
     * @return Value as string.
     */
    std::string ExtractStringValue(const std::string& _key, const std::string& _def = "") const;

    /**
     * @fn int ExtractIntValue(const std::string& _key, int _def = 0) const;
     * @brief Extract value from map.
     * @param _key Key of value.
     * @param _def Default value.
     * @return Value as int.
     */
    int ExtractIntValue(const std::string& _key, int _def = 0) const;

    /**
     * @fn float ExtractFloatValue(const std::string& _key, float _def = 0.0) const;
     * @brief Extract value from map.
     * @param _key Key of value.
     * @param _def Default value.
     * @return Value as float.
     */
    float ExtractFloatValue(const std::string& _key, float _def = 0.0) const;

    /**
     * @fn bool ExtractBoolValue(const std::string& _key, bool _def = false) const;
     * @brief Extract value from map.
     * @param _key Key of value.
     * @param _def Default value.
     * @return Value as bool.
     */
    bool ExtractBoolValue(const std::string& _key, bool _def = false) const;

    /**
     * @fn void InsertValue(const std::string& _key, const std::string& _value);
     * @brief Insert value in map and write it to file.
     * @param _key Key of the value.
     * @param _value Value as string.
     */
    void InsertStringValue(const std::string& _key, const std::string& _value);

    /**
     * @fn void InsertValue(const std::string& _key, T _value);
     * @brief Insert value in map and write it to file.
     * @param _key Key of the value.
     * @param _value Value as class T.
     */
    template <class T>
    void InsertValue(const std::string& _key, T _value) {
        std::ofstream ofstr(m_FileName.c_str(), std::ios_base::out | std::ios_base::trunc);
        std::stringstream ss;
        ss << _value;
        m_ConfigMap[_key] = ss.str();

        WriteAll(ofstr);

        ofstr.close();
    }

    /**
     * @fn void Parse();
     * @brief Parse the configuration file.
     */
    void Parse();

    /**
     * @fn void CleanAll();
     * @brief Empty the map and the file.
     */
    void CleanAll();

    /**
     * @brief Delimiter inside the configuration file.
     */
    static const std::string DELIMITER;
private:
    void WriteAll(std::ofstream& _ofstr) const;

    ConfigMap m_ConfigMap;
    std::string m_FileName;
};

}


#endif // CONFIG_H
