/****************************************************************************
 *  Copyright (c) 2018 Oblo Living d.o.o
 *
 *  All Rights Reserved
 *
 *  P R O P R I E T A R Y    &    C O N F I D E N T I A L
 *
 *  -----------------------------------------------------
 *  http://www.obloliving.com/
 *  -----------------------------------------------------
 *
 * \project Oblo WISE Example application - Example application
 * \file Logger.h
 * \brief
 * Example logger abstraction class.
 *
 * @Author Igor Stefanovic
 *
 * \notes
 *
 * \history
 *****************************************************************************/
#ifndef EXAMPLE_LOGGER_H
#define EXAMPLE_LOGGER_H

#include "humblelogging/api.h"
#include <sys/time.h>
#include <sstream>
#include <iostream>

/**
 * @namespace Example
 * @brief Example classes.
 */
namespace Example {

/**
 * @class Logger
 * @brief Implementation of logger.
 */
class Logger {
public:
    /**
     * @enum LoggerLevel
     * @brief Possible log levels.
     */
    enum LoggerLevel {
        OFF = 0,
        FATAL = 1,
        ERROR = 2,
        WARN = 3,
        INFO = 4,
        DEBUG = 5,
        TRACE = 6,
    };

    /**
     * @fn Logger(const std::string& _name, LoggerLevel _logLevel = TRACE);
     * @brief Constructor.
     * @param _name Name of logger.
     * @param _logLevel Log level.
     */
    Logger(const std::string& _name, LoggerLevel _logLevel = TRACE);

    /**
     * @fn ~Logger();
     * @brief Destructor.
     */
    ~Logger();

    /**
     * @fn void SetLogLevel(const std::string& _logLevel);
     * @brief Setter for log level field.
     * @param _logLevel Log level as string.
     */
    void SetLogLevel(const std::string& _logLevel);

    /**
     * @fn void SetLogLevel(LoggerLevel _logLevel);
     * @brief Setter for log level field.
     * @param _logLevel Log level as LoggerLevel enum.
     */
    void SetLogLevel(LoggerLevel _logLevel);

    /**
     * @fn void Fatal(const std::string& _msg);
     * @brief Fatal log level logging.
     * @param _msg Message.
     */
    void Fatal(const std::string& _msg);

    /**
     * @fn void Error(const std::string& _msg);
     * @brief Error log level logging.
     * @param _msg Message.
     */
    void Error(const std::string& _msg);

    /**
     * @fn void Warn(const std::string& _msg);
     * @brief Warn log level logging.
     * @param _msg Message.
     */
    void Warn(const std::string& _msg);

    /**
     * @fn void Info(const std::string& _msg);
     * @brief Info log level logging.
     * @param _msg Message.
     */
    void Info(const std::string& _msg);

    /**
     * @fn void Debug(const std::string& _msg);
     * @brief Debug log level logging.
     * @param _msg Message.
     */
    void Debug(const std::string& _msg);

    /**
     * @fn void Trace(const std::string& _msg);
     * @brief Trace log level logging.
     * @param _msg Message.
     */
    void Trace(const std::string& _msg);
private:
    std::string m_Name;
    LoggerLevel m_LogLevel;
    humble::logging::Logger* m_Logger;
};

/**
 * @class LogFormatter
 * @brief Custom formatter for logging messages.
 * @extends humble::logging::Formatter
 */
class LogFormatter : public humble::logging::Formatter {
public:
    /**
     * @fn LogFormatter();
     * @brief Constructor.
     */
    LogFormatter();

    /**
     * @fn LogFormatter(const LogFormatter& _other);
     * @brief Constructor.
     * @param _other Copy object.
     */
    LogFormatter(const LogFormatter& _other);

    /**
     * @fn virtual ~LogFormatter();
     * @brief Destructor.
     */
    virtual ~LogFormatter();

    /**
     * @brief Creates a object.
     * @returns Object.
     */
    virtual humble::logging::Formatter* copy() const;

    /**
     * @fn virtual std::string format(const humble::logging::LogEvent &_logEvent) const;
     * @brief Formats logging messages.
     * @param _logEvent Log object containing the timestamp, message, ...
     * @return Formated log.
     */
    virtual std::string format(const humble::logging::LogEvent &_logEvent) const;
};

class FlushConsoleAppender : public humble::logging::ConsoleAppender {
public:
    FlushConsoleAppender();
    ~FlushConsoleAppender();
    virtual void log(const humble::logging::LogEvent &logEvent);
};

/**
 * @class LoggerHelper
 * @brief Logger helper class.
 */
class LoggerHelper {
public:
    /**
     * @fn static void Initialize(Logger::LoggerLevel _globalLevel, const std::string& _fileName, bool _immediate, int _maxRoll, unsigned long _maxFileSize);
     * @brief Initialize logger.
     * @param _globalLevel Log level.
     * @param _fileName File name.
     * @param _immediate Indication whether the logs should be immediately flushed in file.
     * @param _maxRoll Max files.
     * @param _maxFileSize Max file size.
     */
    static void Initialize(Logger::LoggerLevel _globalLevel, const std::string& _fileName, bool _immediate, int _maxRoll, unsigned long _maxFileSize);
};

/**
 * @class LogEventWithName
 * @brief The LogEventWithName class
 */
class LogEventWithName :
        public humble::logging::LogEvent {
public:
    /**
     * @fn LogEventWithName();
     * @brief Constructor.
     */
    LogEventWithName(int level, const std::string &message, int line, const std::string &file, const std::string &func, const std::string& _name);

    /**
     * @fn const std::string& GetName() const;
     * @brief Getting the name of the logger.
     * @return Returns name of logger.
     */
    const std::string& GetName() const;
private:
    std::string m_Name;
};

#define HUMBLE_LOG_NAME(L,S,LL,N) \
  do { \
    if (L.wouldLog(LL)) { \
      LogEventWithName le(LL, S, __LINE__, __FILE__, std::string(), N); \
      L.log(le); \
    } \
  } \
  while (false)

}

#endif // EXAMPLE_LOGGER_H
