/****************************************************************************
 *  Copyright (c) 2018 Oblo Living d.o.o
 *
 *  All Rights Reserved
 *
 *  P R O P R I E T A R Y    &    C O N F I D E N T I A L
 *
 *  -----------------------------------------------------
 *  http://www.obloliving.com/
 *  -----------------------------------------------------
 *
 * \project Oblo WISE Example application - Example application
 * \file RPIDemoDevice.h
 * \brief
 * Example device abstraction class.
 *
 * @Author Igor Stefanovic
 *
 * \notes
 *
 * \history
 *****************************************************************************/
#ifndef RPI_DEMO_DEVICE_H
#define RPI_DEMO_DEVICE_H

#include "Config.h"
#include "Logger.h"
#include "ssdp/Discovery.h"
#include "messenger/Messages.h"
#include "ipdevice/Property.h"
#include "util/Timer.h"
#include "util/JsonStruct.h"
#include "messenger/Messenger.h"
#include "ipdevice/ActiveDevice.h"
#include "ipdevice/SleepingDevice.h"
#include "ipdevice/Device.h"
#include "PahoLogger.h"

/**
 * @namespace Example
 * @brief Example classes.
 */
namespace Example {
#ifdef ON_RPI
/**
 * @class Led
 * @brief Abstraction of the inclusion led indicator.
 * @extends Utility::Runnable
 */
class Led : public Utility::Runnable {
public:
    /**
     * @brief LED value for wiringPi library when connected to the pin GPIO27 and any ground pin.
     */
    static const int RED_LED;

    /**
     * @fn Led();
     * @brief Constructor.
     */
    Led();

    /**
     * @fn ~Led();
     * @brief Destructor.
     */
    ~Led();

    /**
     * @fn void Start();
     * @brief Starts the blinking of the LED.
     */
    void Start();

    /**
     * @fn void Stop();
     * @brief Stops the blinking of the LED.
     */
    void Stop();

    /**
     * @fn void Run();
     * @brief Changes the state of the LED.
     */
    void Run();
private:
    Utility::Thread m_Thread;
    bool m_ShouldRun;
    bool m_Terminated;
};
#endif

/**
 * @class RPIDemoDevice
 * @brief Raspberry PI demo device.
 * @extends IPDevice::ActiveDevice
 * @implements IPDevice::IDeviceExcludedListener
 * @implements IPDevice::IConnectedListener
 * @implements IPDevice::ITimeChangeListener
 */
class RPIDemoDevice :
        public IPDevice::ActiveDevice,
        public IPDevice::IDeviceExcludedListener,
        public IPDevice::IConnectedListener,
        public IPDevice::ITimeChangeListener,
        public Utility::ILogListener
#ifdef ON_RPI
        , public Utility::Runnable
#endif
{
public:
    /**
     * @brief Default log level.
     */
    static const Example::Logger::LoggerLevel DEFAULT_LOGGER_LEVEL;
#ifdef ON_RPI
    /**
     * @brief LED value for wiringPi library when connected to the pin GPIO17 and any 3.3v pin.
     */
    static const int GREEN_LED;

    /**
     * @brief BUTTON value for wiringPi library when connected to the pin GPIO18 and any 3.3v pin.
     */
    static const int BUTTON;
#endif
    /**
     * @fn RPIDemoDevice();
     * @brief Constructor.
     */
    RPIDemoDevice();

    /**
     * @fn ~RPIDemoDevice();
     * @brief Destructor.
     */
    ~RPIDemoDevice();

    /**
     * @fn void Start();
     * @brief Starts the raspberry pi demo device.
     */
    void Start();

    /**
     * @fn void DoInclusion();
     * @brief Initiates inclusion procedure.
     */
    void DoInclusion();

    /**
     * @fn void SetConfig(Utility::Config _config);
     * @brief Sets the configuration to the IPDevice::Device.
     * @param _config Config.
     */
    void SetConfig(Utility::Config _config);

    /**
     * @fn void SetData(Utility::Config _data);
     * @brief Sets the data saved in document to the IPDevice::Device if there is any data.
     * @param _data Data.
     */
    void SetData(Utility::Config _data);

    /**
     * @fn void SetDeviceInfomation(IPDevice::DeviceInformation _deviceInformation);
     * @brief Sets the IPDevice::Device information.
     * @param _deviceInformation Information about device.
     */
    void SetDeviceInfomation(IPDevice::DeviceInformation _deviceInformation);

	/**
     * @fn void SetDescProviderUrl(const std::string& _url);
     * @brief Sets url of description provider.
     * @param _url Url.
     */
    void SetDescProviderUrl(const std::string& _url);

    /**
     * @fn void SetNewTemperatureValue(Utility::Timer& _timer);
     * @brief Timer callback function that changes the temperature value.
     * @param _timer Timer.
     */
    void SetNewTemperatureValue(Utility::Timer& _timer);

    /**
     * @fn void SetNewUptime(Utility::Timer& _timer);
     * @brief Timer callback function that changes the uptime value.
     * @param _timer Timer.
     */
    void SetNewUptime(Utility::Timer& _timer);

    /**
     * @fn void StatePropertyChangeHandler(unsigned int _groupId, const std::string& _serviceType, const std::string &_propertyName, const Jzon::Node& _newValue);
     * @brief Callback function that is invoked upon state property changing request. IPDevice::Device will change the property.
     * @param _groupId Group id.
     * @param _serviceType Service type.
     * @param _propertyName Property name.
     * @param _newValue New value.
     */
    void StatePropertyChangeHandler(unsigned int _groupId, const std::string& _serviceType, const std::string &_propertyName, const Jzon::Node& _newValue);

    /**
     * @fn void NamePropertyChangeHandler(unsigned int _groupId, const std::string& _serviceType, const std::string &_propertyName, const Jzon::Node& _newValue);
     * @brief Callback function that is invoked upon name property changing request. IPDevice::Device will change the property.
     * @param _groupId Group id.
     * @param _serviceType Service type.
     * @param _propertyName Property name.
     * @param _newValue New value.
     */
    void NamePropertyChangeHandler(unsigned int _groupId, const std::string& _serviceType, const std::string &_propertyName, const Jzon::Node& _newValue);

    /**
     * @fn void ToggleHandler(unsigned int _groupId, const std::string& _serviceType, const Jzon::Node& _inputList);
     * @brief Callback function that is called upon toggle command invoking request.
     * @param _groupId Group id.
     * @param _serviceType Service type.
     * @param _inputList Input parameters.
     */
    void ToggleHandler(unsigned int _groupId, const std::string& _serviceType, const Jzon::Node& _inputList);

    /**
     * @fn void ProxyCmdHandler(unsigned int _groupId, const std::string& _serviceType, const Jzon::Node& _inputList);
     * @brief Callback function that is called upon proxy command invoking request.
     * @param _groupId Group id.
     * @param _serviceType Service type.
     * @param _inputList Input parameters.
     */
    void ProxyCmdHandler(unsigned int _groupId, const std::string& _serviceType, const Jzon::Node& _inputList);

    /**
     * @fn void SetConfigParamHandler(unsigned int _groupId, const std::string& _serviceType, const Jzon::Node& _inputList);
     * @brief Callback function that is called upon proxy command invoking request.
     * @param _groupId Group id.
     * @param _serviceType Service type.
     * @param _inputList Input parameters.
     */
    void SetConfigParamHandler(unsigned int _groupId, const std::string& _serviceType, const Jzon::Node& _inputList);

    /**
     * @fn void OnTrace(const std::string &_msg);
     * @brief Callback function that will be called when IPDevice::Device has logs on trace log level to print.
     * @param _msg Log.
     */
    void OnTrace(const std::string &_msg);

    /**
     * @fn void OnDebug(const std::string &_msg);
     * @brief Callback function that will be called when IPDevice::Device has logs on debug log level to print.
     * @param _msg Log.
     */
    void OnDebug(const std::string &_msg);

    /**
     * @fn void OnInfo(const std::string &_msg);
     * @brief Callback function that will be called when IPDevice::Device has logs on info log level to print.
     * @param _msg Log.
     */
    void OnInfo(const std::string &_msg);

    /**
     * @fn void OnWarn(const std::string &_msg);
     * @brief Callback function that will be called when IPDevice::Device has logs on warn log level to print.
     * @param _msg Log.
     */
    void OnWarn(const std::string &_msg);

    /**
     * @fn void OnError(const std::string &_msg);
     * @brief Callback function that will be called when IPDevice::Device has logs on error log level to print.
     * @param _msg Log.
     */
    void OnError(const std::string &_msg);

    /**
     * @fn  void OnFatal(const std::string& _msg);
     * @brief Used to report fatal messages.
     * @param _msg Log message.
     */
     void OnFatal(const std::string& _msg);

    /**
     * @fn void OnDeviceExcluded();
     * @brief Callback function that will be called when IP device is excluded from the hub.
     */
    void OnDeviceExcluded();

    /**
     * @fn void OnConnected(const IPDevice::InclusionInformation& _inclusionResult);
     * @brief Callback function that will be called when IP device is connected to the hub.
     * @param _inclusionResult Information about inclusion process, that might need to be updated.
     */
    void OnConnected(const IPDevice::InclusionInformation& _inclusionResult);

    /**
     * @fn void OnSetTime(unsigned long long _timestamp);
     * @brief Callback function that will be called upon SetTime request arrival.
     * @param _timestamp Timestamp for SetTime request.
     */
    void OnSetTime(unsigned long long _timestamp);

    /**
     * @fn void OnAdjustTime(unsigned long long _diff);
     * @brief Callback function that will be called upon AdjustTime request arrival.
     * @param _diff Timestamp for AdjustTime request.
     */
    void OnAdjustTime(unsigned long long _diff);

#ifdef ON_RPI
    /**
     * @fn void Run();
     * @brief Function which reads the data from button and acts accordingly.
     */
    void Run();
#endif
private:
    Utility::Timer* m_TemperatureTimer;
    Utility::Timer* m_UptimeTimer;
    Example::Logger m_Logger;
    Example::PahoLogger m_PahoLogger;
    Utility::Config m_Data;
    int m_BootCount;
    bool m_Included;
#ifdef ON_RPI
    Utility::Thread m_Thread;
    int m_Counter;
    bool m_ButtonDown;
    bool m_ShouldRun;
#endif
};

}


#endif // RPI_DEMO_DEVICE_H
