#include "RPIDemoDevice.h"

using namespace IPDevice;
using namespace SSDP;
using namespace Messenger;
using namespace Example;
using namespace Utility;

const Example::Logger::LoggerLevel RPIDemoDevice::DEFAULT_LOGGER_LEVEL = Example::Logger::INFO;

#ifdef ON_RPI
#include <wiringPi.h>

const int Led::RED_LED = 2;
const int RPIDemoDevice::GREEN_LED = 0;
const int RPIDemoDevice::BUTTON = 1;
Led::Led() {
    if (wiringPiSetup() != -1) {
        pinMode(RED_LED, OUTPUT);
        digitalWrite(RED_LED, LOW);
    }
    m_ShouldRun = false;
    m_Terminated = false;
}

Led::~Led() {
    Stop();
}

void Led::Start() {
    m_ShouldRun = true;
    m_Terminated = false;
    m_Thread.Start(*this);
}

void Led::Stop() {
    if (!m_Terminated) {
        m_ShouldRun = false;
        m_Terminated = true;
        m_Thread.Join();
    }
}

void Led::Run() {
    bool high = true;
    while(m_ShouldRun) {
        if (high) {
            digitalWrite(RED_LED, LOW);
            high = false;
        } else {
            digitalWrite(RED_LED, HIGH);
            high = true;
        }

        usleep(300000);
    }
    digitalWrite(RED_LED, LOW);
}
#endif

RPIDemoDevice::RPIDemoDevice() : ActiveDevice(), m_Logger("RPIDemoDevice", DEFAULT_LOGGER_LEVEL) {
#ifdef ON_RPI
    if (wiringPiSetup() != -1) {
        pinMode(GREEN_LED, OUTPUT);
        pinMode(BUTTON, INPUT);
        pullUpDnControl(BUTTON, PUD_DOWN);
        digitalWrite(GREEN_LED, HIGH);
        m_ButtonDown = (digitalRead(BUTTON) == LOW);
        m_Counter = 0;
    }

    m_ShouldRun = true;
    m_Thread.Start(*this);
#endif
    Device::RegisterConnectedListener(this);
    Device::SetLogListener(this);
    Device::RegisterDeviceExcludedListener(this);

    SwitchService* switchService = new SwitchService();
    switchService->SetStateHandler(SERVICE_PROPERTY_DELEGATE(RPIDemoDevice, StatePropertyChangeHandler));
    switchService->SetToggleHandler(SERVICE_COMMAND_DELEGATE(RPIDemoDevice, ToggleHandler));
    AddService(0, switchService);

    TemperatureService* temperatureService = new TemperatureService();
    AddService(0, temperatureService);

    // Mandatory
    DeviceService* deviceService = new DeviceService();
    deviceService->SetNameHandler(SERVICE_PROPERTY_DELEGATE(RPIDemoDevice, NamePropertyChangeHandler));
    AddService(0, deviceService);

    DiagnosticsService* diagnosticsService = new DiagnosticsService();
    diagnosticsService->UseUptimeProp();
    diagnosticsService->UseBootCountProp();
    AddService(0, diagnosticsService);

    BasicService* basicService = new BasicService();
    basicService->SetProxyCmdHandler(SERVICE_COMMAND_DELEGATE(RPIDemoDevice, ProxyCmdHandler));
    AddService(0, basicService);

    ConfigurationService* configService = new ConfigurationService();
    configService->SetSetConfigParamCmdHandler(SERVICE_COMMAND_DELEGATE(RPIDemoDevice, SetConfigParamHandler));
    AddService(0, configService);

    m_TemperatureTimer = new Timer(20000, false);
    m_UptimeTimer = new Timer(60000, false);
    m_Included = false;
    m_BootCount = 0;
}

RPIDemoDevice::~RPIDemoDevice() {
#ifdef ON_RPI
    m_ShouldRun = false;
    m_Thread.Join();
#endif
    Device::SetLogListener(NULL);
    UnregisterDeviceExcludedListener();
    m_TemperatureTimer->StopTimer();
    SAFE_DELETE_O(m_TemperatureTimer);
    m_UptimeTimer->StopTimer();
    SAFE_DELETE_O(m_UptimeTimer);
}

void RPIDemoDevice::Start() {
    if (!Device::IsInclusionInfoEmpty()) {
        try {
            m_Included = true;
            StartReconnection();
        } catch (const ProtocolException& exc) {
            m_Logger.Error(exc.GetMessage());
        }
    }
#ifndef ON_RPI
    else {
         DoInclusion();
    }
#endif
    m_Logger.Info("Device is ready");
    m_TemperatureTimer->StartTimer(Callback<RPIDemoDevice>(*this, &RPIDemoDevice::SetNewTemperatureValue));
    m_UptimeTimer->StartTimer(Callback<RPIDemoDevice>(*this, &RPIDemoDevice::SetNewUptime));
}

void RPIDemoDevice::DoInclusion() {
    struct timespec timeBegin;
    clock_gettime(CLOCK_MONOTONIC, &timeBegin);
#ifdef ON_RPI
    Led l;
#endif
    try {
#ifdef ON_RPI
        l.Start();
        InclusionInformation result = StartInclusion();
        l.Stop();
#else
        InclusionInformation result = StartInclusion();
#endif
        if (!result.GetUsername().empty() || !result.GetPassword().empty()) {
            m_Included = true;

            struct timespec timeEnd;
            clock_gettime(CLOCK_MONOTONIC, &timeEnd);

            double accum = (timeEnd.tv_sec - timeBegin.tv_sec) + (timeEnd.tv_nsec - timeBegin.tv_nsec) / 1E9;
            std::stringstream ss;
            ss << "Inclusion time = " << accum << " seconds";
            m_Logger.Debug(ss.str());
        }
    } catch (const BaseException& exc) {
        m_Logger.Error("Inclusion failed, reason: " + exc.GetMessage());
#ifdef ON_RPI
        l.Stop();
#endif
    }
}

void RPIDemoDevice::SetConfig(Config _config) {
    if (_config.HasElements()) {
        m_Logger.SetLogLevel(_config.ExtractStringValue("LogLevel.RPIDevice"));
        ChallengeParameters challengeParam;
        challengeParam.SetChallengeRetry(_config.ExtractIntValue("Challenge.Retry"));
        challengeParam.SetChallengeTimeout(_config.ExtractIntValue("Challenge.Timeout"));
        Device::SetInclusionTimeout(_config.ExtractIntValue("Inclusion.Timeout"));
        Device::SetSearchInterval(_config.ExtractIntValue("Inclusion.SearchInterval"));
        Device::SetChallengeParameters(challengeParam);
        Device::SetKeyFile(_config.ExtractStringValue("Device.KeyPath"));
        Device::SetKeyPasscode(_config.ExtractStringValue("Device.KeyPasscode"));
        Device::SetCertFile(_config.ExtractStringValue("Device.CertPath"));
        Device::SetSecure(_config.ExtractBoolValue("Device.Secured"));
    }
}

void RPIDemoDevice::SetData(Config _data) {
    m_Data = _data;
    if (m_Data.HasElements()) {
        InclusionInformation prevIncResult;
        prevIncResult.SetInclusionID(m_Data.ExtractStringValue("InclusionID"));
        prevIncResult.SetID(m_Data.ExtractIntValue("ID"));
        prevIncResult.SetUsername(m_Data.ExtractStringValue("Username"));
        prevIncResult.SetPassword(m_Data.ExtractStringValue("Password"));
        SSDP::HubHolder hub(m_Data.ExtractStringValue("DomainID"), m_Data.ExtractStringValue("HomeID"), m_Data.ExtractStringValue("HubID"), m_Data.ExtractStringValue("Location"));
        std::string hubTag = hub.GetDomainID() + TOPIC_SEPARATOR + hub.GetHomeID() + TOPIC_SEPARATOR + "hub/" + hub.GetHubID();
        prevIncResult.SetHub(hub);
        prevIncResult.SetHubTag(hubTag);
        m_BootCount = m_Data.ExtractIntValue("BootCount", 0);
        if (m_BootCount != 0) {
            ++m_BootCount;
        }
        ActiveDevice::SetPropertyValueIntern(0, SERVICE_TYPE_DIAGNOSTIC, PROPERTY_NAME_BOOT_COUNT, m_BootCount, false);
        m_Data.InsertValue("BootCount", m_BootCount);

        Device::SetInclusionInformation(prevIncResult);
    }
}

void RPIDemoDevice::SetDeviceInfomation(DeviceInformation _deviceInformation) {
    Device::SetDeviceInfomation(_deviceInformation);
    ActiveDevice::SetPropertyValueIntern(0, SERVICE_TYPE_DEVICE, PROPERTY_NAME_NAME, _deviceInformation.GetName(), false);
    ActiveDevice::SetPropertyValueIntern(0, SERVICE_TYPE_DEVICE, PROPERTY_NAME_VERSION, _deviceInformation.GetVersion(), false);
    ActiveDevice::SetPropertyValueIntern(0, SERVICE_TYPE_DEVICE, PROPERTY_NAME_VERSION_NUMBER, _deviceInformation.GetVersionNumber(), false);
    ActiveDevice::SetPropertyValueIntern(0, SERVICE_TYPE_DEVICE, PROPERTY_NAME_PROTOCOL, _deviceInformation.GetProtocolVersion(), false);
}

void RPIDemoDevice::SetDescProviderUrl(const std::string& _url) {
    Device::SetDescriptionProviderUrl(_url);
}

void RPIDemoDevice::SetNewTemperatureValue(Timer& _timer) {
    if (m_Included) {
        ActiveDevice::SetPropertyValueIntern(0, SERVICE_TYPE_TEMPERATURE, PROPERTY_NAME_VALUE, (rand() % (40 + 1 + 40) - 40), true);
    } else {
        ActiveDevice::SetPropertyValueIntern(0, SERVICE_TYPE_TEMPERATURE, PROPERTY_NAME_VALUE, (rand() % (40 + 1 + 40) - 40), false);
    }
}

void RPIDemoDevice::SetNewUptime(Timer &_timer) {
    Jzon::Node value = ActiveDevice::GetPropertyValueIntern(0, SERVICE_TYPE_DIAGNOSTIC, PROPERTY_NAME_UP_TIME);
    if (m_Included) {
        ActiveDevice::SetPropertyValueIntern(0, SERVICE_TYPE_DIAGNOSTIC, PROPERTY_NAME_UP_TIME, value.toInt() + 1, true);
    } else {
        ActiveDevice::SetPropertyValueIntern(0, SERVICE_TYPE_DIAGNOSTIC, PROPERTY_NAME_UP_TIME, value.toInt() + 1, false);
    }
}

void RPIDemoDevice::StatePropertyChangeHandler(unsigned int _groupId, const std::string& _serviceType, const std::string& _propertyName, const Jzon::Node& _newValue) {
    /* Your code here: on change property */
#ifdef ON_RPI
    if (_newValue.toBool()) {
        digitalWrite(GREEN_LED, LOW);
    } else {
        digitalWrite(GREEN_LED, HIGH);
    }
#endif
}

void RPIDemoDevice::NamePropertyChangeHandler(unsigned int _groupId, const std::string& _serviceType, const std::string& _propertyName, const Jzon::Node& _newValue) {
    /* Your code here: on change property */
    m_Data.InsertValue("Name", _newValue.toString());
}

void RPIDemoDevice::ToggleHandler(unsigned int _groupId, const std::string &_serviceType, const Jzon::Node &_inputList) {
    Jzon::Node value = ActiveDevice::GetPropertyValueIntern(_groupId, _serviceType, PROPERTY_NAME_STATE);
    if (value.getType() != Jzon::Node::T_INVALID) {
#ifdef ON_RPI
        if (!value.toBool()) {
            digitalWrite(GREEN_LED, LOW);
        } else {
            digitalWrite(GREEN_LED, HIGH);
        }
#endif
        ActiveDevice::SetPropertyValueIntern(_groupId, _serviceType, PROPERTY_NAME_STATE, !value.toBool(), true);
    }
}

void RPIDemoDevice::ProxyCmdHandler(unsigned int _groupId, const std::string& _serviceType, const Jzon::Node& _inputList) {
    // TODO: Implement
    SendGenericEvent("EventReturnsCmdParameters", new JsonStruct(_inputList));
}

void RPIDemoDevice::SetConfigParamHandler(unsigned int _groupId, const std::string& _serviceType, const Jzon::Node& _inputList) {
    // TODO: Implement
    JsonStruct json(_inputList);
    m_Logger.Info(json.ToString());
}

void RPIDemoDevice::OnTrace(const std::string &_msg) {
    // std::cout << _msg << std::endl;
    m_Logger.Trace(_msg);
}

void RPIDemoDevice::OnDebug(const std::string &_msg) {
    // std::cout << _msg << std::endl;
    m_Logger.Debug(_msg);
}

void RPIDemoDevice::OnInfo(const std::string &_msg) {
    // std::cout << _msg << std::endl;
    m_Logger.Info(_msg);
}

void RPIDemoDevice::OnWarn(const std::string &_msg) {
    // std::cout << _msg << std::endl;
    m_Logger.Warn(_msg);
}

void RPIDemoDevice::OnError(const std::string &_msg) {
    // std::cout << _msg << std::endl;
    m_Logger.Error(_msg);
}

void RPIDemoDevice::OnFatal(const std::string &_msg) {
    // std::cout << _msg << std::endl;
    m_Logger.Fatal(_msg);
}

void RPIDemoDevice::OnDeviceExcluded() {
    if (m_Included) {
        m_Logger.Info("Device excluded");
        m_Data.CleanAll();
        m_Included = false;
    }
}

void RPIDemoDevice::OnConnected(const InclusionInformation& _inclusionResult) {
    ActivateServices();
    RegisterDeviceExcludedListener(this);
    RegisterTimeChangeListener(this);
    SetLogListener(this);

    SendDeviceRebootedEvent();

    /* Filling config */
    m_Data.InsertValue("DomainID", _inclusionResult.GetHub().GetDomainID());
    m_Data.InsertValue("HomeID", _inclusionResult.GetHub().GetHomeID());
    m_Data.InsertValue("HubID", _inclusionResult.GetHub().GetHubID());
    m_Data.InsertValue("Location", _inclusionResult.GetHub().GetLocation());
    m_Data.InsertValue("InclusionID", _inclusionResult.GetInclusionID());
    m_Data.InsertValue("Username", _inclusionResult.GetUsername());
    m_Data.InsertValue("Password", _inclusionResult.GetPassword());
    std::stringstream sst;
    sst << _inclusionResult.GetID();
    m_Data.InsertValue("ID", sst.str());
    m_Data.InsertValue("BootCount", ItoS(m_BootCount));
}

void RPIDemoDevice::OnSetTime(unsigned long long _timestamp) {
    std::stringstream ss;
    ss << "Set time request <t=" << _timestamp << ">";
    m_Logger.Debug(ss.str());
    struct timeval timeVal;
    timeVal.tv_sec = _timestamp / 1000;
    timeVal.tv_usec = (_timestamp % 1000) * 1000;
    if (settimeofday(&timeVal, NULL) == -1) {
        m_Logger.Error("Error while setting time <e=" + std::string(std::strerror(errno)) + ">");
    }
}

void RPIDemoDevice::OnAdjustTime(unsigned long long _diff) {
    std::stringstream ss;
    ss << "Adjust time request <d=" << _diff << ">";
    m_Logger.Debug(ss.str());
    struct timeval timeVal;
    gettimeofday(&timeVal, NULL);
    timeVal.tv_sec += _diff / 1000;
    timeVal.tv_usec += (_diff % 1000) * 1000;
    if (settimeofday(&timeVal, NULL) == -1) {
        m_Logger.Error("Error while adjusting time <e=" + std::string(std::strerror(errno)) + ">");
    }
}

#ifdef ON_RPI
double CalculateDifference(struct timespec timeBegin, struct timespec timeEnd) {
   double accum = (timeEnd.tv_sec - timeBegin.tv_sec) + (timeEnd.tv_nsec - timeBegin.tv_nsec) / 1E9;
   return accum;
}

void RPIDemoDevice::Run() {
    struct timespec timeFirst;
    struct timespec timeSecond;

    while(m_ShouldRun) {
        if (!m_ButtonDown && digitalRead(BUTTON) == LOW) {
            m_ButtonDown = true;
            if (!m_Included) {
                if (m_Counter == 0) {
                        clock_gettime(CLOCK_MONOTONIC, &timeFirst);
                        clock_gettime(CLOCK_MONOTONIC, &timeSecond);
                } else if (m_Counter == 1) {
                        clock_gettime(CLOCK_MONOTONIC, &timeSecond);
                }

                ++m_Counter;

                if (m_Counter == 2 && (CalculateDifference(timeFirst, timeSecond) < 0.5 && CalculateDifference(timeFirst, timeSecond) > 0.1)) {
                        m_Counter = 0;
                        DoInclusion();
                } else if (m_Counter == 2 && !(CalculateDifference(timeFirst, timeSecond) < 0.5 && CalculateDifference(timeFirst, timeSecond) > 0.1)) {
                        m_Counter = 0;
                }
            } else {
                Jzon::Node value = ActiveDevice::GetPropertyValueIntern(0, SERVICE_TYPE_SWITCH, PROPERTY_NAME_STATE);
                if (!value.toBool()) {
                    digitalWrite(GREEN_LED, LOW);
                } else {
                    digitalWrite(GREEN_LED, HIGH);
                }

                ActiveDevice::SetPropertyValueIntern(0, SERVICE_TYPE_SWITCH, PROPERTY_NAME_STATE, !value.toBool(), true);
            }
        } else if (m_ButtonDown && digitalRead(BUTTON) == HIGH) {
            m_ButtonDown = false;
        }
        usleep(10000);
    }
}

#endif
