﻿/****************************************************************************
 *  Copyright (c) 2018 Oblo Living d.o.o
 *
 *  All Rights Reserved
 *
 *  P R O P R I E T A R Y    &    C O N F I D E N T I A L
 *
 *  -----------------------------------------------------
 *  http://www.obloliving.com/
 *  -----------------------------------------------------
 *
 * \project Oblo WISE SDK - SDK library
 * \file Command.h
 * \brief
 * WISE Command class.
 *
 * @Author Igor Stefanovic
 *
 * \notes
 *
 * \history
 *****************************************************************************/
#ifndef COMMAND_H
#define COMMAND_H

#include "util/Common.h"
#include "util/JsonStruct.h"
#include "ipdevice/Property.h"
#include "util/Exceptions.h"

/**
 * @namespace IPDevice
 * @brief IP device functionality.
 */
namespace IPDevice {

class ServiceCommandDelegate;
class Command;

#define SERVICE_COMMAND_DELEGATE(tp, m) \
    ServiceCommandDelegate::CreateDelegate<tp, & tp :: m>(this)

/**
 * @class ServiceCommandDelegate
 * @brief Delegate class for command handling.
 */
class ServiceCommandDelegate {
public:
    /**
     * @typedef Utility::SharedPointer<ServiceCommandDelegate> Ptr
     * @brief Shared pointer to delegate.
     */
    typedef Utility::SharedPointer<ServiceCommandDelegate> Ptr;
    /**
     * @fn ServiceCommandDelegate();
     * @brief Constructor.
     */
    ServiceCommandDelegate()
        : m_Object(0)
        , m_Stub(0)
    {}

    template <class T, void (T::*TMethod)(unsigned int _groupId, const std::string& _serviceType, const Jzon::Node& _inputList)>
    /**
     * @fn static ServiceCommandDelegate::Ptr CreateDelegate(T* _object);
     * @brief Creates command delegate.
     * @param _object Object pointer.
     * @return Shared pointer to a command delegate.
     */
    static ServiceCommandDelegate::Ptr CreateDelegate(T* _object)
    {
        ServiceCommandDelegate::Ptr d = new ServiceCommandDelegate();
        d->m_Object = _object;
        d->m_Stub = &MethodStub<T, TMethod>;
        return d;
    }

    /**
     * @fn void operator()(unsigned int _groupId, const std::string& _serviceType, const Jzon::Node& _inputList) const;
     * @brief operator ().
     * @param _groupId ID of the Group object.
     * @param _serviceType Type of the Service object.
     * @param _inputList Input parameters for Command.
     */
    void operator()(unsigned int _groupId, const std::string& _serviceType, const Jzon::Node& _inputList) const
    {
        (*m_Stub)(m_Object, _groupId, _serviceType, _inputList);
    }

private:
    typedef void (*StubType)(void* m_Object, unsigned int _groupId, const std::string& _serviceType, const Jzon::Node& _inputList);

    void* m_Object;
    StubType m_Stub;

    template <class T, void (T::*TMethod)(unsigned int _groupId, const std::string& _serviceType, const Jzon::Node& _inputList)>
    static void MethodStub(void* _object, unsigned int _groupId, const std::string& _serviceType, const Jzon::Node& _inputList)
    {
        T* p = static_cast<T*>(_object);
        (p->*TMethod)(_groupId, _serviceType, _inputList);
    }
};

/**
 * @class InputParam
 * @brief Abstraction class of Command Input parameters.
 */
class DLLExport InputParam {
public:
    /**
     * @enum InputParamType
     * @brief Possible types for Command Input parameters.
     */
    enum InputParamType { NUMBER = 0,
                        TEXT = 1,
                        BOOL = 2,
                        RANGE = 3,
                        ATTRIBUTE = 4,
                        COLOR = 5,
                        ANY = 6};

    /**
     * @fn InputParam();
     * @brief Constructor.
     */
    InputParam() {}

    /**
     * @fn ~InputParam();
     * @brief Destructor.
     */
    ~InputParam() {}

    /**
     * @fn InputParam(const std::string& _name, InputParamType _type);
     * @brief Constructor.
     * @param _name Name of the Input parameter;
     * @param _type Type of the Input parameter;
     */
    InputParam(const std::string& _name, InputParamType _type);

    /**
     * @fn void SetName(const std::string& _name);
     * @brief Setter for Name field.
     * @param _name New name to be set.
     */

    void SetName(const std::string& _name);
    /**
     * @fn void SetType(InputParamType _type);
     * @brief Setter for Type field.
     * @param _type New type to be set.
     */
    void SetType(InputParamType _type);

    /**
     * @fn InputParam &AddOption(Jzon::Node _option)
     * @brief In case that Input parameter is ATTRIBUTE type, this function defines all options.
     * @param _option Option.
     * @return Reference to this object.
     */
    InputParam& AddOption(Jzon::Node _option);

    /**
     * @fn InputParam &AddRange(Jzon::Node _min, Jzon::Node _max);
     * @brief In case that Input parameter is RANGE type, this function defines a range.
     * @param _min Minimal value of range.
     * @param _max Maximal value of range.
     * @return Reference to this object.
     */
    InputParam& AddRange(Jzon::Node _min, Jzon::Node _max);

    /**
     * @fn std::vector<Jzon::Node> GetOptions() const;
     * @brief Getter for Options field.
     * @return Vector of Options.
     */
    const std::vector<Jzon::Node>& GetOptions() const;

    /**
     * @fn std::string GetName() const;
     * @brief Getter for Name field.
     * @return Current Name.
     */
    std::string GetName() const;

    /**
     * @fn InputParamType GetType() const;
     * @brief Getter for Type field.
     * @return Current Type.
     */
    InputParamType GetType() const;

    /**
     * @fn Jzon::Node GetMin() const;
     * @brief Getter for minimal value of range.
     * @return Jzon::Node object which holds minimal value.
     */
    Jzon::Node GetMin() const;

    /**
     * @fn Jzon::Node GetMax() const;
     * @brief Getter for maximal value of range.
     * @return Jzon::Node object which holds maximal value.
     */
    Jzon::Node GetMax() const;

    /**
     * @fn void ToJson(Utility::JsonStruct& _json) const;
     * @brief Fill Utility::JsonStruct object with information about Input parameters.
     * @param _json Utility::JsonStruct to be filled.
     */
    void ToJson(Utility::JsonStruct& _json) const;
private:
    std::string m_Name;
    InputParamType m_Type;
    std::vector<Jzon::Node> m_Options;
};

/**
 * @class Command
 * @brief Abstraction class for Command object.
 */
class DLLExport Command {
public:
    /**
     * @fn Command();
     * @brief Constructor.
     */
    Command() {}

    /**
     * @fn ~Command();
     * @brief Destructor.
     */
    ~Command() {}

    /**
     * @fn Command(const std::string _name, ServiceCommandDelegate::Ptr _function);
     * @brief Constructor.
     * @param _name Name of the Command object.
     * @param _function Shared pointer to delegate function.
     */
    Command(const std::string _name, ServiceCommandDelegate::Ptr _function);

    /**
     * @fn std::string GetName() const;
     * @brief Getter for Name field.
     * @return Current Name.
     */
    std::string GetName() const;

    /**
     * @fn void SetName(const std::string& _name);
     * @brief Setter for Name field.
     * @param _name New type to be set.
     */
    void SetName(const std::string& _name);

    /**
     * @fn Command& AddInputParam(const InputParam& _inputParam);
     * @brief Add input parameters for Command object.
     * @param _inputParam Input parameter object.
     * @return Reference to this object.
     */
    Command& AddInputParam(const InputParam& _inputParam);

    /**
     * @fn void SetCommandFunction(ServiceCommandDelegate::Ptr _function);
     * @brief Setter for shared pointer to delegate function.
     * @param _function Shared pointer to delegate function.
     */
    void SetCommandFunction(ServiceCommandDelegate::Ptr _function);

    /**
     * @fn ServiceCommandDelegate::Ptr GetCommandFunction() const;
     * @brief Getter for shared pointer to delegate function.
     * @return Shared pointer to delegate function.
     */
    ServiceCommandDelegate::Ptr GetCommandFunction() const;

    /**
     * @fn bool HasParamAny() const
     * @brief Getter for has any param
     * @return Indication whether this command holds param type ANY
     */
    bool HasParamAny() const;

    /**
     * @fn const std::vector<InputParam>& GetInputParams() const
     * @brief Getter for vector of input params
     * @return Vector of input param objects
     */
    const std::vector<InputParam>& GetInputParams() const;

    /**
     * @fn void ToJson(Utility::JsonStruct& _json) const;
     * @brief Fills Utility::JsonStruct object with information about Input parameters.
     * @param _json Utility::JsonStruct to be filled.
     */
    void ToJson(Utility::JsonStruct& _json) const;
private:
    std::string m_Name;
    ServiceCommandDelegate::Ptr m_Function;
    bool m_HasParamAny;
    std::vector<InputParam> m_InputParams;
};
}

#endif // COMMAND_H
