/****************************************************************************
 *  Copyright (c) 2018 Oblo Living d.o.o
 *
 *  All Rights Reserved
 *
 *  P R O P R I E T A R Y    &    C O N F I D E N T I A L
 *
 *  -----------------------------------------------------
 *  http://www.obloliving.com/
 *  -----------------------------------------------------
 *
 * \project Oblo WISE SDK - SDK library
 * \file Device.h
 * \brief
 * WISE IP Device class.
 *
 * @Author Igor Stefanovic
 *
 * \notes
 *
 * \history
 *****************************************************************************/
#ifndef DEVICE_H
#define DEVICE_H

#include "ssdp/Discovery.h"
#include "ssdp/SSDPClient.h"
#include "messenger/Messenger.h"
#include "ipdevice/Service.h"
#include "ipdevice/Group.h"
#include "util/Logger.h"
#include <openssl/crypto.h>
#include <openssl/aes.h>
#include <openssl/sha.h>
#include <openssl/evp.h>
#include <openssl/rsa.h>
#include <openssl/bio.h>
#include <openssl/pem.h>
#include <openssl/err.h>
#include <openssl/opensslv.h>
#if __cplusplus > 199711L
    #include <regex>
    #define REGEX_OBJ std::regex
    #define REGEX_MATCH std::regex_match
#else
    #include <util/Regex.h>
    #define REGEX_OBJ Utility::Regex
    #define REGEX_MATCH Utility::Regex::Match
#endif

/**
 * @namespace IPDevice
 * @brief IP device functionality.
 */
namespace IPDevice {

class Device;
class InclusionInformation;

/**
 * @interface IDeviceExcludedListener
 * @brief Interface for device excluded.
 */
class DLLExport IDeviceExcludedListener {
public:
    /**
     * @fn virtual void OnDeviceExcluded() = 0;
     * @brief Function to be called when IP device gets excluded from network.
     */
    virtual void OnDeviceExcluded() = 0;
};


/**
 * @interface IConnectedListener
 * @brief Interface for device connected.
 */
class DLLExport IConnectedListener {
public:
    /**
     * @fn virtual void OnConnected(const InclusionInformation& _inclusionResult) = 0;
     * @brief Function to be called when IP device connects to the Gateway.
     * @param _inclusionResult Result of the inclusion process.
     */
    virtual void OnConnected(const InclusionInformation& _inclusionResult) = 0;
};


/**
 * @interface ITimeChangeListener
 * @brief Interface for SetTime and AdjustTime invocation.
 */
class DLLExport ITimeChangeListener {
public:
    /**
     * @fn virtual void OnSetTime(unsigned long long _timestamp) = 0;
     * @brief Function to be called when SetTime is invoked by the Gateway.
     * @param _timestamp Timestamp in milliseconds.
     */
    virtual void OnSetTime(unsigned long long _timestamp) = 0;
    /**
     * @fn virtual void OnAdjustTime(unsigned long long _diff) = 0;
     * @brief Function to be called when AdjustTime is invoked by the Gateway.
     * @param _diff Timestamp in milliseconds.
     */
    virtual void OnAdjustTime(unsigned long long _diff) = 0;
};

/**
 * @class DeviceInformation
 * @brief Class which holds the information about the IP device.
 */
class DLLExport DeviceInformation {
public:
    /**
     * @fn DeviceInformation();
     * @brief Constructor.
     */
    DeviceInformation();

    /**
     * @fn void SetName(const std::string& _name);
     * @brief Setter for Name field.
     * @param _name New name to be set.
     */
    void SetName(const std::string& _name);

    /**
     * @fn void SetManufacturer(const std::string& _manufacturer);
     * @brief Setter for Manufacturer field.
     * @param _manufacturer New manufacturer name to be set.
     */
    void SetManufacturer(const std::string& _manufacturer);

    /**
     * @fn void SetIdentifier(const std::string& _identifier);
     * @brief Setter for Identifier field.
     * @param _identifier New identifier to be set.
     */
    void SetIdentifier(const std::string& _identifier);

    /**
     * @fn void SetModel(const std::string& _model);
     * @brief Setter for Model field.
     * @param _model New model to be set.
     */
    void SetModel(const std::string& _model);

    /**
     * @fn void SetVersion(const std::string& _version);
     * @brief Setter for Version field.
     * @param _version New version to be set.
     */
    void SetVersion(const std::string& _version);

    /**
     * @fn void SetProtocolVersion(int _protocolVersion);
     * @brief Setter for ProtocolVersion field.
     * @param _protocolVersion New protocol version to be set.
     */
    void SetProtocolVersion(int _protocolVersion);

    /**
     * @fn void SetMAC(const std::string& _mac);
     * @brief Setter for MAC field.
     * @param _mac New MAC to be set.
     */
    void SetMAC(const std::string& _mac);

    /**
     * @fn void SetVersionNumber(int _versionNumber);
     * @brief Setter for VersionNumber field.
     * @param _versionNumber New version number to be set.
     */
    void SetVersionNumber(int _versionNumber);

    /**
     * @fn std::string GetName() const;
     * @brief Getter for Name field.
     * @return Current Name.
     */
    std::string GetName() const;

    /**
     * @fn std::string GetManufacturer() const;
     * @brief Getter for Manufacturer field.
     * @return Current Manufacturer.
     */
    std::string GetManufacturer() const;

    /**
     * @fn std::string GetIdentifier() const;
     * @brief Getter for Identifier field.
     * @return Current Identifier.
     */
    std::string GetIdentifier() const;

    /**
     * @fn std::string GetModel() const;
     * @brief Getter for Model field.
     * @return Current Model.
     */
    std::string GetModel() const;

    /**
     * @fn std::string GetVersion() const;
     * @brief Getter for Version field.
     * @return Current Version.
     */
    std::string GetVersion() const;

    /**
     * @fn int GetProtocolVersion() const;
     * @brief Getter for ProtocolVersion field.
     * @return Current ProtocolVersion.
     */
    int GetProtocolVersion() const;

    /**
     * @fn std::string GetMAC() const;
     * @brief Getter for MAC field.
     * @return Current MAC.
     */
    std::string GetMAC() const;

    /**
     * @fn int GetVersionNumber() const;
     * @brief Getter for VersionNumber field.
     * @return Current VersionNumber.
     */
    int GetVersionNumber() const;
private:
    std::string m_Name;
    std::string m_Manufacturer;
    std::string m_Identifier;
    std::string m_Model;
    std::string m_Version;
    int m_ProtocolVersion;
    std::string m_MAC;
    int m_VersionNumber;
};

/**
 * @class ChallengeParameters
 * @brief Class which holds the parameters for challenge procedure.
 */
class DLLExport ChallengeParameters {
public:
    /**
     * @fn ChallengeParameters();
     * @brief Constructor.
     */
    ChallengeParameters();

    /**
     * @fn unsigned int GetChallengeRetry() const;
     * @brief Getter for ChallengerRetry field.
     * @return Current ChallengeRetry.
     */
    unsigned int GetChallengeRetry() const;

    /**
     * @fn long GetChallengeTimeout() const;
     * @brief Getter for ChallengeTimeout field.
     * @return Current ChallengeTimeout.
     */
    long GetChallengeTimeout() const;

    /**
     * @fn std::string GetChallengeCode() const;
     * @brief Getter for ChallengeCode field.
     * @return Current ChallengeCode.
     */
    std::string GetChallengeCode() const;

    /**
     * @fn void SetChallengeRetry(unsigned int _challengeRetry);
     * @brief Setter for ChallengeRetry field.
     * @param _challengeRetry New challenge retry number to be set.
     */
    void SetChallengeRetry(unsigned int _challengeRetry);

    /**
     * @fn void SetChallengeTimeout(long _challengeTimeout);
     * @brief Setter for ChallengeTimeout field.
     * @param _challengeTimeout New challenge timeout to be set.
     */
    void SetChallengeTimeout(long _challengeTimeout);

    /**
     * @fn void SetChallengeCode(const std::string& _challengeCode);
     * @brief Setter for ChallengeCode field.
     * @param _challengeCode New challenge code to be set.
     */
    void SetChallengeCode(const std::string& _challengeCode);
private:
    unsigned int m_ChallengeRetry;
    long m_ChallengeTimeout;
    std::string m_ChallengeCode;
};

/**
 * @class InclusionInformation
 * @brief Class which holds the information about previous inclusion process.
 */
class DLLExport InclusionInformation {
public:
    /**
     * @fn InclusionInformation();
     * @brief Constructor.
     */
    InclusionInformation();

    /**
     * @fn void SetUsername(const std::string& _username);
     * @brief Setter for Username field.
     * @param _username New username to be set.
     */
    void SetUsername(const std::string& _username);

    /**
     * @fn void SetPassword(const std::string& _password);
     * @brief Setter for Password field.
     * @param _password New password to be set.
     */
    void SetPassword(const std::string& _password);

    /**
     * @fn void SetHub(const SSDP::HubHolder& _hub);
     * @brief Setter for Hub field.
     * @param _hub New hub to be set.
     */
    void SetHub(const SSDP::HubHolder& _hub);

    /**
     * @fn void SetInclusionID(const std::string& _inclusionID);
     * @brief Setter for InclusionID field.
     * @param _inclusionID New inclusion id to be set.
     */
    void SetInclusionID(const std::string& _inclusionID);

    /**
     * @fn void SetID(int _id);
     * @brief Setter for ID field.
     * @param _id New id to be set.
     */
    void SetID(int _id);

    /**
     * @fn void SetHubTag(const std::string& _hubTag);
     * @brief Setter for HubTag field.
     * @param _hubTag New hub tag to be set.
     */
    void SetHubTag(const std::string& _hubTag);

    /**
     * @fn std::string GetUsername() const;
     * @brief Getter for Username field.
     * @return Current Username;
     */
    std::string GetUsername() const;

    /**
     * @fn std::string GetPassword() const;
     * @brief Getter for Password field.
     * @return Current Password;
     */
    std::string GetPassword() const;

    /**
     * @fn SSDP::HubHolder GetHub() const;
     * @brief Getter for Hub field.
     * @return Current Hub;
     */
    SSDP::HubHolder GetHub() const;

    /**
     * @fn std::string GetInclusionID() const;
     * @brief Getter for InclusionID field.
     * @return Current InclusionID;
     */
    std::string GetInclusionID() const;

    /**
     * @fn int GetID() const;
     * @brief Getter for ID field.
     * @return Current ID;
     */
    int GetID() const;

    /**
     * @fn std::string GetHubTag() const;
     * @brief Getter for HubTag field.
     * @return Current HubTag;
     */
    std::string GetHubTag() const;

    /**
     * @fn bool IsEmpty() const;
     * @brief Check whether the inclusion information object isn't filled with information.
     * @return Indication whether the inclusion information object is filled.
     */
    bool IsEmpty() const;
private:
    std::string m_Username;
    std::string m_Password;
    SSDP::HubHolder m_Hub;
    std::string m_InclusionID;
    int m_ID;
    std::string m_HubTag;
};

/**
 * @class RequestHandler
 * @extends Utility::Logger
 * @brief Class which distributes the incoming requests to callback functions.
 */
class DLLExport RequestHandler : public Utility::Logger {
public:
    /**
     * @typedef void (Device::*RespondFunctionPtr)(const Messenger::RequestMessage::Ptr, Messenger::ResponseMessage::Ptr);
     * @brief Typedef of function.
     */
    typedef void (Device::*RespondFunctionPtr)(const Messenger::RequestMessage::Ptr, Messenger::ResponseMessage::Ptr);

    /**
     * @typedef void (Device::*RespondServiceFunctionPtr)(unsigned int _groupId, const std::string& _serviceType, const Messenger::RequestMessage::Ptr, Messenger::ResponseMessage::Ptr);
     * @brief Typedef of function.
     */
    typedef void (Device::*RespondServiceFunctionPtr)(unsigned int _groupId, const std::string& _serviceType, const Messenger::RequestMessage::Ptr, Messenger::ResponseMessage::Ptr);

    /**
     * @typedef std::pair<std::string, RespondFunctionPtr> RequestHandlerObj;
     * @brief Pair of name of the request on Device topic and shared pointer to callback function.
     */
    typedef std::pair<std::string, RespondFunctionPtr> RequestHandlerObj;

    /**
     * @typedef std::pair<std::string, RespondServiceFunctionPtr> RequestServiceHandlerObj;
     * @brief Pair of name of the request on Service topic and shared pointer to callback function.
     */
    typedef std::pair<std::string, RespondServiceFunctionPtr> RequestServiceHandlerObj;

    /**
     * @typedef std::map<std::string, RespondFunctionPtr> RequestHandlers;
     * @brief Map of callback functions for requests on Device topics.
     */
    typedef std::map<std::string, RespondFunctionPtr> RequestHandlers;

    /**
     * @typedef std::map<std::string, RespondServiceFunctionPtr> ServiceRequestHandlers;
     * @brief Map of callback functions for requests on Service topics.
     */
    typedef std::map<std::string, RespondServiceFunctionPtr> ServiceRequestHandlers;

    /**
     * @fn RequestHandler(Device *_obj);
     * @brief Constructor.
     * @param _obj Object of Device class.
     */
    RequestHandler(Device *_obj);
protected:
    /**
     * @fn virtual void AddRequestHandler(const std::string& _commandName, const RespondFunctionPtr _func);
     * @brief Adding the pair of name of the request and shared pointer to callback function.
     * @param _commandName Name of the request.
     * @param _func Shared pointer to callback function to be invoked on incoming request.
     */
    virtual void AddRequestHandler(const std::string& _commandName, const RespondFunctionPtr _func);

    /**
     * @fn virtual void CallRequestHandler(const std::string& _commandName, const Messenger::RequestMessage::Ptr _req, Messenger::ResponseMessage::Ptr _rsp);
     * @brief Invokes the callback function of the request defined by _commandName parameter.
     * @param _commandName Name of the request.
     * @param _req Shared pointer to request message.
     * @param _rsp Shared pointer to response message.
     */
    virtual void CallRequestHandler(const std::string& _commandName, const Messenger::RequestMessage::Ptr _req, Messenger::ResponseMessage::Ptr _rsp);

    /**
     * @fn virtual void AddServiceRequestHandler(const std::string& _commandName, const RespondServiceFunctionPtr _func);
     * @brief AddServiceRequestHandler
     * @param _commandName Name of the request.
     * @param _func Shared pointer to callback function to be invoked on incoming request.
     */
    virtual void AddServiceRequestHandler(const std::string& _commandName, const RespondServiceFunctionPtr _func);

    /**
     * @fn virtual void CallServiceRequestHandler(unsigned int _groupId, const std::string& _serviceType, const std::string& _commandName, const Messenger::RequestMessage::Ptr _req, Messenger::ResponseMessage::Ptr _rsp);
     * @brief CallServiceRequestHandler
     * @param _groupId ID of the Group object.
     * @param _serviceType Type of the Service object.
     * @param _commandName Name of the request.
     * @param _req Shared pointer to request message.
     * @param _rsp Shared pointer to response message.
     */
    virtual void CallServiceRequestHandler(unsigned int _groupId, const std::string& _serviceType, const std::string& _commandName, const Messenger::RequestMessage::Ptr _req, Messenger::ResponseMessage::Ptr _rsp);
private:
    RequestHandlers m_RequestHandlers;
    ServiceRequestHandlers m_ServiceRequestHandlers;
    Utility::Mutex m_Mutex;
    Device* m_Obj;
};

/**
 * @class Device
 * @extends RequestHandler
 * @implements SSDP::ISearchListener
 * @implements Messenger::IResponder
 * @implements Messenger::IConnectionListener
 * @brief Abstraction class for IP device. Contains full implementation of IP device from WISE specification.
 */
class DLLExport Device :
        public SSDP::ISearchListener,
        public Messenger::IResponder,
        public RequestHandler,
        public Messenger::IConnectionListener {
public:
    /**
     * @class ConnectionResurrector
     * @extends Utility::Runnable
     * @brief The ConnectionResurrector class
     */
    class ConnectionResurrector: public Utility::Runnable {
    public:
        /**
         * @fn ConnectionResurrector(Device& _device);
         * @brief Constructor.
         * @param _device Reference to Device object.
         */
        ConnectionResurrector(Device& _device);

        /**
         * @fn ~ConnectionResurrector();
         * @brief Destructor.
         */
        ~ConnectionResurrector();

        /**
         * @fn void Start();
         * @brief Starts the connection resurrector thread.
         */
        void Start();

        /**
         * @fn void Run();
         * @brief Thread function which establish connection after it is lost.
         */
        void Run();

        /**
         * @fn void Stop();
         * @brief Stops the connection resurrector thread.
         */
        void Stop();
    private:
        bool m_ShouldRun;
        Utility::Thread m_Thread;
        Device& m_Device;
        bool m_Terminated;
    };

    /**
     * @fn Device();
     * @brief Constructor.
     */
    Device();

    /**
     * @fn virtual ~Device();
     * @brief Destructor.
     */
    virtual ~Device();

    /**
     * @fn InclusionInformation StartInclusion();
     * @brief Starts the inclusion process.
     * @return Information about inclusion process.
     * @throws Utility::TimeoutException
     * @throws Utility::ProtocolException
     * @throws Utility::BaseException
     */
    InclusionInformation StartInclusion();

    /**
     * @fn void StartReconnection();
     * @brief Initiates the reconnection process.
     */
    void StartReconnection();

    /**
     * @fn void StartDiscoverySearch();
     * @brief Start sending SSDP MSEARCH messages and parsing responses from WISE hub.
     */
    void StartDiscoverySearch();

    /**
     * @fn void StartDiscoveryListen();
     * @brief Start listening for SSDP NOTIFY messages from WISE hub.
     */
    void StartDiscoveryListen();

    /**
     * @fn void StopDiscovery();
     * @brief Stops the discovery process, whether the sending MSEARCH or listening for NOTIFY messages.
     */
    void StopDiscovery();

    /**
     * @fn void GetDescription(const std::string& _url);
     * @brief Get description from url for HUBLESS mode.
     * @param _url Url.
     */
    void GetDescription(const std::string& _url);

    /**
     * @fn void InitialConnect();
     * @brief Establishes connection with default username: client and password: tmpclient.
     * @throws Utility::ProtocolException
     */
    void InitialConnect();

    /**
     * @fn void FinalConnect();
     * @brief Establishes connection with username and password provided during the inclusion process.
     * @throws Utility::ProtocolException
     * @throws Utility::BaseException
     */
    void FinalConnect();

    /**
     * @fn void SendInclusionRequest();
     * @brief Sends the inclusion request defined in WISE specification.
     * @throws Utility::ProtocolException
     */
    void SendInclusionRequest();

    /**
     * @fn void SendChallengeReq();
     * @brief Sends the challenge request which initiates authentication process.
     * @throws Utility::ProtocolException
     */
    void SendChallengeReq();

    /**
     * @fn void SendCredentialsReq();
     * @brief Sends the credential request which requests username and password from WISE hub.
     * @throws Utility::ProtocolException
     */
    void SendCredentialsReq();

    /**
     * @fn void ActivateServices();
     * @brief Forms topics for requests on Service topic, and subscribes for such requests.
     */
    void ActivateServices();

    /**
     * @fn void WaitForHub(long _timeout = -1);
     * @brief Waiter on SSDP MSEARCH or SSDP NOTIFY message to detect a WISE hub.
     * @param _timeout Defines the time of waiting in seconds.
     * @throws Utility::TimeoutException
     */
    void WaitForHub(long _timeout = -1);

    /**
     * @fn void WaitForJoinConfirmation();
     * @brief Waiter for JoinConfirmation request. Time for waiting is set to 5 seconds.
     * @throws Utility::TimeoutException
     */
    void WaitForJoinConfirmation();

    /**
     * @fn void AddService(unsigned int _groupId, BaseService* _service);
     * @brief Adds service to specified Group object. If Group objects doesnt exist, creates one.
     * @param _groupId ID of the Group object in which service will be added.
     * @param _service Pointer to service object.
     */
    void AddService(unsigned int _groupId, BaseService* _service);

    /**
     * @fn void SetCertFile(const std::string& _cert);
     * @brief Setter for the path to certificate file.
     * @param _cert Path to certificate file.
     */
    void SetCertFile(const std::string& _cert);

    /**
     * @fn void SetSecure(bool _secure);
     * @brief Setting a secure connection.
     * @param _secure Should be secured.
     */
    void SetSecure(bool _secure);

    /**
     * @fn void SetKeyFile(const std::string& _key);
     * @brief Setter for the path to private key file.
     * @param _key Path to private key file.
     */
    void SetKeyFile(const std::string& _key);

    /**
     * @fn void SetKeyPasscode(const std::string& _passcode);
     * @brief Setter for the private key passcode.
     * @param _passcode Private key passcode.
     */
    void SetKeyPasscode(const std::string& _passcode);

    /**
     * @fn void SetChallengeParameters(ChallengeParameters _challengeParams);
     * @brief Setter for challenge parameters.
     * @param _challengeParams Challenge parameters.
     */
    void SetChallengeParameters(ChallengeParameters _challengeParams);

    /**
     * @fn void SetDeviceInfomation(DeviceInformation _deviceInformation);
     * @brief Setter for device information.
     * @param _deviceInformation Device information.
     */
    void SetDeviceInfomation(DeviceInformation _deviceInformation);

    /**
     * @fn void SetInclusionInformation(InclusionInformation _prevInclusionInformation);
     * @brief Setter for inclusion information.
     * @param _prevInclusionInformation Inclusion information.
     */
    void SetInclusionInformation(InclusionInformation _prevInclusionInformation);

    /**
     * @fn bool IsInclusionInfoEmpty() const;
     * @brief Checks whether the inclusion information object is filled.
     * @return Indication whether the inclusion information object is filled.
     */
    bool IsInclusionInfoEmpty() const;

    /**
     * @fn void SetInclusionTimeout(unsigned int _inclusionTimeout);
     * @brief Setter for timeout of the inclusion process.
     * @param _inclusionTimeout Timeout of the inclusion process in seconds.
     */
    void SetInclusionTimeout(long _inclusionTimeout);

    /**
     * @fn void SetSearchInterval(long _searchInterval);
     * @brief Setter for interval for SSDP MSEARCH messages.
     * @param _searchInterval Interval for SSDP MSEARCH messages in seconds.
     */
    void SetSearchInterval(long _searchInterval);

    /**
     * @fn static std::string GetVersion();
     * @brief Provides the version of the WISE SDK.
     * @return Version of the WISE SDK.
     */
    static std::string GetVersion();

    /**
     * @fn static std::string GetPlatform();
     * @brief Provides the platform for which WISE SDK is built for.
     * @return Platform for which WISE SDK is built for.
     */
    static std::string GetPlatform();

    /**
     * @fn static std::string GetType();
     * @brief Provides the type of the application which uses WISE SDK.
     * @return Type of the application which uses WISE SDK.
     */
    static std::string GetType();

    /**
     * @fn void SetDescriptionProviderUrl(const std::string& _url);
     * @brief Setter description provider url.
     * @param _url Url to be set.
     */
    void SetDescriptionProviderUrl(const std::string& _url);

    /**
     * @fn virtual void OnHubDetected(const SSDP::HubHolder &_hub);
     * @brief Callback function that will be called when hub is discovered using SSDP MSEARCH or NOTIFY messages.
     * @param _hub Object of the Hub with hub information.
     */
    virtual void OnHubDetected(const SSDP::HubHolder &_hub);

    /**
     * @fn virtual const Messenger::ResponseMessage::Ptr Respond(const Messenger::RequestMessage::Ptr _request);
     * @brief This function is called when Hub sends a request to Device topic.
     * @param _request Shared pointer to request message.
     * @return Shared pointer to response message.
     */
    virtual const Messenger::ResponseMessage::Ptr Respond(const Messenger::RequestMessage::Ptr _request);

    /**
     * @fn virtual const Messenger::ResponseMessage::Ptr ServiceRespond(unsigned int _groupId, const std::string& _serviceType, const Messenger::RequestMessage::Ptr _request);
     * @brief This function is called when Hub sends a request to Service topic.
     * @param _groupId ID of the Group object.
     * @param _serviceType Type of the Service object.
     * @param _request Shared pointer to request message.
     * @return Shared pointer to response message.
     */
    virtual const Messenger::ResponseMessage::Ptr ServiceRespond(unsigned int _groupId, const std::string& _serviceType, const Messenger::RequestMessage::Ptr _request);

    /**
     * @fn virtual void OnConnectionLost(const std::string& _cause);
     * @brief Callback function that will be called connection lost is detected.
     * @param _cause Cause of the connection lost. Can be empty.
     */
    virtual void OnConnectionLost(const std::string& _cause);
protected:
    /**
     * @brief Default username for InitialConnect. Value is "tmpclient".
     */
    static const std::string DEFAULT_USERNAME;

    /**
     * @brief Default username for InitialConnect. Value is "client".
     */
    static const std::string DEFAULT_PASSWORD;

    /**
     * @brief Default inclusion timeout for InitialConnect. Value is 180 seconds.
     */
    static const unsigned int DEFAULT_INCLUSION_TIMEOUT;

    /**
     * @brief Topic separator. Value is "/".
     */
    static const std::string TOPIC_SEPARATOR;

    /**
     * @fn virtual void GetDeviceInfo(const Messenger::RequestMessage::Ptr _req, Messenger::ResponseMessage::Ptr _resp);
     * @brief Respond to GetDeviceInfo request providing information about Device object, Group object and Service object.
     * @param _req Shared pointer to request message.
     * @param _resp Shared pointer to response message.
     */
    virtual void GetDeviceInfo(const Messenger::RequestMessage::Ptr _req, Messenger::ResponseMessage::Ptr _resp);

    /**
     * @fn virtual void GetState(const Messenger::RequestMessage::Ptr _req, Messenger::ResponseMessage::Ptr _resp);
     * @brief Respond to GetState request providing changes since time defined in parameter of message.
     * @param _req Shared pointer to request message.
     * @param _resp Shared pointer to response message.
     */
    virtual void GetState(const Messenger::RequestMessage::Ptr _req, Messenger::ResponseMessage::Ptr _resp);

    /**
     * @fn virtual void SetTime(const Messenger::RequestMessage::Ptr _req, Messenger::ResponseMessage::Ptr _resp);
     * @brief Respond to SetTime request providing an interface for time changing.
     * @param _req Shared pointer to request message.
     * @param _resp Shared pointer to response message.
     */
    virtual void SetTime(const Messenger::RequestMessage::Ptr _req, Messenger::ResponseMessage::Ptr _resp);

    /**
     * @fn virtual void AdjustTime(const Messenger::RequestMessage::Ptr _req, Messenger::ResponseMessage::Ptr _resp);
     * @brief Respond to AdjustTime request providing an interface for time adjusting.
     * @param _req Shared pointer to request message.
     * @param _resp Shared pointer to response message.
     */
    virtual void AdjustTime(const Messenger::RequestMessage::Ptr _req, Messenger::ResponseMessage::Ptr _resp);

    /**
     * @fn virtual void VerifyIdentity(const Messenger::RequestMessage::Ptr _req, Messenger::ResponseMessage::Ptr _resp);
     * @brief Respond to VerifyIdentity request as defined in WISE specification.
     * @param _req Shared pointer to request message.
     * @param _resp Shared pointer to response message.
     */
    virtual void VerifyIdentity(const Messenger::RequestMessage::Ptr _req, Messenger::ResponseMessage::Ptr _resp);

    /**
     * @fn virtual void GetGroupList(const Messenger::RequestMessage::Ptr _req, Messenger::ResponseMessage::Ptr _resp);
     * @brief Respond to GetGroupList request providing a list of all Group objects supported by Device object.
     * @param _req Shared pointer to request message.
     * @param _resp Shared pointer to response message.
     */
    virtual void GetGroupList(const Messenger::RequestMessage::Ptr _req, Messenger::ResponseMessage::Ptr _resp);

    /**
     * @fn virtual void GetServiceList(const Messenger::RequestMessage::Ptr _req, Messenger::ResponseMessage::Ptr _resp);
     * @brief Respond to GetServiceList request providing a list of all Service objects supported by specified Group object.
     * @param _req Shared pointer to request message.
     * @param _resp Shared pointer to response message.
     */
    virtual void GetServiceList(const Messenger::RequestMessage::Ptr _req, Messenger::ResponseMessage::Ptr _resp);

    /**
     * @fn virtual void GetService(const Messenger::RequestMessage::Ptr _req, Messenger::ResponseMessage::Ptr _resp);
     * @brief Respond to GetService request providing a Service object supported by specified Group object.
     * @param _req Shared pointer to request message.
     * @param _resp Shared pointer to response message.
     */
    virtual void GetService(const Messenger::RequestMessage::Ptr _req, Messenger::ResponseMessage::Ptr _resp);

    /**
     * @fn virtual void JoinConfirmation(const Messenger::RequestMessage::Ptr _req, Messenger::ResponseMessage::Ptr _resp);
     * @brief Respond to JoinConfirmation request as defined in WISE specification.
     * @param _req Shared pointer to request message.
     * @param _resp Shared pointer to response message.
     */
    void JoinConfirmation(const Messenger::RequestMessage::Ptr _req, Messenger::ResponseMessage::Ptr _resp);

    /**
     * @fn virtual void GetPropertyList(unsigned int _groupId, const std::string& _serviceType, const Messenger::RequestMessage::Ptr _req, Messenger::ResponseMessage::Ptr _resp);
     * @brief Respond to GetPropertyList request providing a list of all Property objects defined in Service object.
     * @param _groupId ID of the Group object.
     * @param _serviceType Type of the Service object.
     * @param _req Shared pointer to request message.
     * @param _resp Shared pointer to response message.
     */
    virtual void GetPropertyList(unsigned int _groupId, const std::string& _serviceType, const Messenger::RequestMessage::Ptr _req, Messenger::ResponseMessage::Ptr _resp);


    /**
     * @fn virtual void GetPropertyValue(unsigned int _groupId, const std::string& _serviceType, const Messenger::RequestMessage::Ptr _req, Messenger::ResponseMessage::Ptr _resp);
     * @brief Respond to GetPropertyValue request providing a value of Property object of Service object specified in parameters of the request.
     * @param _groupId ID of the Group object.
     * @param _serviceType Type of the Service object.
     * @param _req Shared pointer to request message.
     * @param _resp Shared pointer to response message.
     */
    virtual void GetPropertyValue(unsigned int _groupId, const std::string& _serviceType, const Messenger::RequestMessage::Ptr _req, Messenger::ResponseMessage::Ptr _resp);

    /**
     * @fn virtual void SetPropertyValue(unsigned int _groupId, const std::string& _serviceType, const Messenger::RequestMessage::Ptr _req, Messenger::ResponseMessage::Ptr _resp);
     * @brief Respond to SetPropertyValue request and changes value of Property object of Service object specified in parameters of the request.
     * @param _groupId ID of the Group object.
     * @param _serviceType Type of the Service object.
     * @param _req Shared pointer to request message.
     * @param _resp Shared pointer to response message.
     */
    virtual void SetPropertyValue(unsigned int _groupId, const std::string& _serviceType, const Messenger::RequestMessage::Ptr _req, Messenger::ResponseMessage::Ptr _resp);

    /**
     * @fn virtual void GetCommandList(unsigned int _groupId, const std::string& _serviceType, const Messenger::RequestMessage::Ptr _req, Messenger::ResponseMessage::Ptr _resp);
     * @brief Respond to GetCommandList request providing a list of all Command object defined in Service object.
     * @param _groupId ID of the Group object.
     * @param _serviceType Type of the Service object.
     * @param _req Shared pointer to request message.
     * @param _resp Shared pointer to response message.
     */
    virtual void GetCommandList(unsigned int _groupId, const std::string& _serviceType, const Messenger::RequestMessage::Ptr _req, Messenger::ResponseMessage::Ptr _resp);

    /**
     * @fn virtual void ExecuteCommand(unsigned int _groupId, const std::string& _serviceType, const Messenger::RequestMessage::Ptr _req, Messenger::ResponseMessage::Ptr _resp);
     * @brief Respond to ExecuteCommand request and invokes a command specified by parameters of the request.
     * @param _groupId ID of the Group object.
     * @param _serviceType Type of the Service object.
     * @param _req Shared pointer to request message.
     * @param _resp Shared pointer to response message.
     */
    virtual void ExecuteCommand(unsigned int _groupId, const std::string& _serviceType, const Messenger::RequestMessage::Ptr _req, Messenger::ResponseMessage::Ptr _resp);

    /**
     * @fn virtual bool UpdatePropertyAttributeList(unsigned int _groupId, const std::string& _serviceType, const std::string& _name, const std::vector<Jzon::Node>& _attributeList)
     * @brief Sets new value of the Property object specified by the parameters.
     * @param _groupId ID of the Group object.
     * @param _serviceType Type of the Service object.
     * @param _name Name of the Property object.
     * @param _attributeList New attribute list for Property object.
     * @return Indication whether updating was a success.
     */
    virtual bool UpdatePropertyAttributeList(unsigned int _groupId, const std::string& _serviceType, const std::string& _name, const std::vector<Jzon::Node>& _attributeList);

    /**
     * @fn virtual void SetPropertyValueIntern(unsigned int _groupId, const std::string& _serviceType, const std::string& _name, const Jzon::Node& _value, bool _sendEvent);
     * @brief Sets new value of the Property object specified by the parameters.
     * @param _groupId ID of the Group object.
     * @param _serviceType Type of the Service object.
     * @param _name Name of the Property object.
     * @param _value Value of the Property object.
     * @param _sendEvent Indication whether the PropertyChanged event will be sent.
     */
    virtual void SetPropertyValueIntern(unsigned int _groupId, const std::string& _serviceType, const std::string& _name, const Jzon::Node& _value, bool _sendEvent);

    /**
     * @fn virtual Jzon::Node GetPropertyValueIntern(unsigned int _groupId, const std::string& _serviceType, const std::string& _name);
     * @brief Provides value of the Property object specified by the parameters.
     * @param _groupId ID of the Group object.
     * @param _serviceType Type of the Service object.
     * @param _name Name of the Property object.
     * @return Current value of the Property object inside of Jzon::Node object.
     */
    virtual Jzon::Node GetPropertyValueIntern(unsigned int _groupId, const std::string& _serviceType, const std::string& _name);

    /**
     * @fn void SendPropertyChangedEvent(const std::string& _name, const Jzon::Node& _value, unsigned int _groupId, const std::string &_serviceType);
     * @brief Sends property changed event to subscribed users.
     * @param _name Name of the Property object.
     * @param _value Value of the Property object.
     * @param _groupId ID of the Group object.
     * @param _serviceType Type of the Service object.
     */
    void SendPropertyChangedEvent(const std::string& _name, const Jzon::Node& _value, unsigned int _groupId, const std::string &_serviceType);

    /**
     * @fn void SendDeviceRebootedEvent();
     * @brief Sends the DeviceRebooted event.
     */
    void SendDeviceRebootedEvent();

    /**
     * @fn void SendGenericEvent();
     * @param _name Name of the generic event
     * @param _params Params of the generic event
     * @brief Sends the Generic event.
     */
    void SendGenericEvent(const std::string &_name, const Utility::JsonStruct::Ptr _params);

    /**
     * @fn void DisconnectWithoutStsMessage();
     * @brief Disconnects from the Hub withoud sending Status message.
     */
    void DisconnectWithoutStsMessage();

    /**
     * @fn void RegisterDeviceExcludedListener(IDeviceExcludedListener* _deviceExcludedListener);
     * @brief Registers IDeviceExcludedListener object.
     * @param _deviceExcludedListener Pointer to IDeviceExcludedListener object.
     */
    void RegisterDeviceExcludedListener(IDeviceExcludedListener* _deviceExcludedListener);

    /**
     * @fn void UnregisterDeviceExcludedListener();
     * @brief Unregisters IDeviceExcludedListener object.
     */
    void UnregisterDeviceExcludedListener();

    /**
     * @fn void SetLogListener(Utility::ILogListener* _logListener);
     * @brief Registers Utility::ILogListener object.
     * @param _logListener Pointer to Utility::ILogListener object.
     */
    void SetLogListener(Utility::ILogListener* _logListener);

    /**
     * @fn void RegisterConnectedListener(IConnectedListener* _connectedListener);
     * @brief Registers IConnectedListener object.
     * @param _connectedListener Pointer to IConnectedListener object.
     */
    void RegisterConnectedListener(IConnectedListener* _connectedListener);

    /**
     * @fn void UnregisterConnectedListener();
     * @brief Unregisters IConnectedListener object.
     */
    void UnregisterConnectedListener();

    /**
     * @fn void RegisterTimeChangeListener(ITimeChangeListener* _timeChangeListener);
     * @brief Registers ITimeChangeListener object.
     * @param _timeChangeListener Pointer to ITimeChangeListener object.
     */
    void RegisterTimeChangeListener(ITimeChangeListener* _timeChangeListener);

    /**
     * @fn void UnregisterTimeChangeListener();
     * @brief Unregisters ITimeChangeListener object.
     */
    void UnregisterTimeChangeListener();

    /**
     * @fn void ToJson(Utility::JsonStruct& _toFill, unsigned long long _timestamp = 0, bool _stripped = false) const;
     * @brief Fills Utility::JsonStruct object with information about Device object.
     * @param _toFill Utility::JsonStruct to be filled.
     * @param _timestamp Timestamp that defines which Property objects will be a part of information. Check Device::GetState.
     * @param _stripped Whether the Property object will have full information or only the Name and the Value.
     */
    void ToJson(Utility::JsonStruct& _toFill, unsigned long long _timestamp = 0, bool _stripped = false) const;

    /**
     * @fn void Destroy();
     * @brief Disconnects and destroys current Messenger::CMessenger object.
     */
    void Destroy();
private:
    void InclusionTimerExpired(Utility::Timer& _timer);
    bool CheckInputParamFormat(const Jzon::Node& _param, const InputParam& _inputParam, Messenger::ResponseMessage::Ptr _resp) const;
    bool CheckPropertyFormat(const Jzon::Node& _value, const Property& _property, Messenger::ResponseMessage::Ptr _resp = NULL) const;

    SSDP::Discovery m_DiscoveryObj;
    Utility::SynchronizedObject m_SyncSSDP;
    Utility::SynchronizedObject m_SyncJoin;
    Messenger::CMessenger* m_Messenger;
    bool m_Stopped;
    std::map<unsigned int, Group*> m_GroupList;
    Utility::Mutex m_ConnMutex;
    DeviceInformation m_DeviceInfo;
    InclusionInformation m_InclusionResult;
    ChallengeParameters m_ChallengeParameters;
    IDeviceExcludedListener* m_DeviceExcludedListener;
    IConnectedListener* m_ConnectedListener;
    ITimeChangeListener* m_TimeChangeListener;
    std::string m_CertFile;
    std::string m_KeyFile;
    std::string m_KeyPasscode;
    std::string m_DescProviderUrl;
    bool m_InclusionInitiated;
    Device::ConnectionResurrector m_ConnResurector;
    // Seconds
    long m_InclusionTimeout;
    Utility::SynchronizedObject m_ConnSync;
    Utility::Timer* m_InclusionTimer;
    bool m_InclusionTimerExpired;
    bool m_Secured;
};

}

#endif // DEVICE_H
