/****************************************************************************
 *  Copyright (c) 2018 Oblo Living d.o.o
 *
 *  All Rights Reserved
 *
 *  P R O P R I E T A R Y    &    C O N F I D E N T I A L
 *
 *  -----------------------------------------------------
 *  http://www.obloliving.com/
 *  -----------------------------------------------------
 *
 * \project Oblo WISE SDK - SDK library
 * \file Group.h
 * \brief
 * WISE Group class.
 *
 * @Author Igor Stefanovic
 *
 * \notes
 *
 * \history
 *****************************************************************************/
#ifndef GROUP_H
#define GROUP_H

#include "util/Common.h"
#include "ipdevice/Service.h"

/**
 * @namespace IPDevice
 * @brief IP device functionality.
 */
namespace IPDevice {

/**
 * @class Group
 * @brief Abstraction class for Group object.
 */
class DLLExport Group {
public:
    /**
     * @typedef std::map<std::string, BaseService*>::const_iterator ConstIterator;
     * @brief Typedef for ConstIterator of std::map.
     */
    typedef std::map<std::string, BaseService*>::const_iterator ConstIterator;

    /**
     * @fn Group(unsigned int _id);
     * @brief Constructor.
     * @param _id ID of the Group object.
     */
    Group(unsigned int _id);

    /**
     * @fn ~Group();
     * @brief Destructor.
     */
    ~Group();

    /**
     * @fn void AddService(BaseService* _baseService);
     * @brief Adds Service object to the Group object.
     * @param _baseService Pointer to Service object.
     */
    void AddService(BaseService* _baseService);

    /**
     * @fn void ToJson(Utility::JsonStruct& _json, unsigned long long _timestamp = 0, bool _stripped = false) const;
     * @brief Fills Utility::JsonStruct object with information about Group object.
     * @param _json Utility::JsonStruct to be filled.
     * @param _timestamp Timestamp that defines which Property objects will be a part of information. Check Device::GetState.
     * @param _stripped Whether the Property object will have full information or only the Name and the Value.
     */
    void ToJson(Utility::JsonStruct& _json, unsigned long long _timestamp = 0, bool _stripped = false) const;

    /**
     * @fn unsigned int GetGroupId() const;
     * @brief Getter for ID field.
     * @return Current ID.
     */
    unsigned int GetGroupId() const;

    /**
     * @fn BaseService* FindServiceByType(const std::string& _serviceType) const;
     * @brief Provides the Service object supported by this Group object specified with parameter.
     * @param _serviceType Type of Service object.
     * @return Pointer to Service object.
     */
    BaseService* FindServiceByType(const std::string& _serviceType) const;

    /**
     * @fn ConstIterator Begin() const
     * @brief Returns a iterator for begin of the map of Service object.
     * @return Iterator for map.
     */
    ConstIterator Begin() const;

    /**
     * @fn ConstIterator End() const
     * @brief Returns a iterator for end of the map of Service object.
     * @return Iterator for map.
     */
    ConstIterator End() const;
private:
    unsigned int m_Id;
    std::map<std::string, BaseService*> m_ServiceList;
};
}




#endif // GROUP_H
