/****************************************************************************
 *  Copyright (c) 2018 Oblo Living d.o.o
 *
 *  All Rights Reserved
 *
 *  P R O P R I E T A R Y    &    C O N F I D E N T I A L
 *
 *  -----------------------------------------------------
 *  http://www.obloliving.com/
 *  -----------------------------------------------------
 *
 * \project Oblo WISE SDK - SDK library
 * \file Property.h
 * \brief
 * WISE Property class.
 *
 * @Author Igor Stefanovic
 *
 * \notes
 *
 * \history
 *****************************************************************************/
#ifndef PROPERTY_H
#define PROPERTY_H

#include "util/Common.h"
#include "util/SharedPointer.h"
#include "jzon/Jzon.h"
#include "util/JsonStruct.h"
#include "util/Mutex.h"
#include "util/ScopedLock.h"

/**
 * @namespace IPDevice
 * @brief IP device functionality.
 */
namespace IPDevice {

class ServicePropertyDelegate;
class Property;
class Options;

#define SERVICE_PROPERTY_DELEGATE(tp, m) \
    ServicePropertyDelegate::CreateDelegate<tp, & tp :: m>(this)

/**
 * @class ServicePropertyDelegate
 * @brief Delegate class for property handling.
 */
class ServicePropertyDelegate {
public:
    /**
     * @typedef Utility::SharedPointer<ServicePropertyDelegate> Ptr
     * @brief Shared pointer to delegate.
     */
    typedef Utility::SharedPointer<ServicePropertyDelegate> Ptr;

    /**
     * @fn ServicePropertyDelegate();
     * @brief Constructor.
     */
    ServicePropertyDelegate()
        : m_Object(0)
        , m_Stub(0)
    {}

    template <class T, void (T::*TMethod)(unsigned int _groupId, const std::string& _serviceType, const std::string& _propertyName, const Jzon::Node& _NewValue)>
    /**
     * @fn static ServicePropertyDelegate::Ptr CreateDelegate(T* _object);
     * @brief Creates property delegate.
     * @param _object Object pointer.
     * @return Shared pointer to a request delegate.
     */
    static ServicePropertyDelegate::Ptr CreateDelegate(T* _object)
    {
        ServicePropertyDelegate::Ptr d = new ServicePropertyDelegate();
        d->m_Object = _object;
        d->m_Stub = &MethodStub<T, TMethod>;
        return d;
    }

    /**
     * @fn void operator()(unsigned int _groupId, const std::string& _serviceType, const std::string& _propertyName, const Jzon::Node& _NewValue);
     * @brief operator ().
     * @param _groupId ID of the Group object.
     * @param _serviceType Type of the Service object.
     * @param _propertyName Name of the Property object.
     * @param _NewValue Value of the Property object.
     */
    void operator()(unsigned int _groupId, const std::string& _serviceType, const std::string& _propertyName, const Jzon::Node& _NewValue) const
    {
        (*m_Stub)(m_Object, _groupId, _serviceType, _propertyName, _NewValue);
    }

private:
    typedef void (*StubType)(void* m_Object, unsigned int _groupId, const std::string& _serviceType, const std::string& _propertyName, const Jzon::Node& _NewValue);

    void* m_Object;
    StubType m_Stub;

    template <class T, void (T::*TMethod)(unsigned int _groupId, const std::string& _serviceType, const std::string& _propertyName, const Jzon::Node& _NewValue)>
    static void MethodStub(void* _object, unsigned int _groupId, const std::string& _serviceType, const std::string& _propertyName, const Jzon::Node& _NewValue)
    {
        T* p = static_cast<T*>(_object);
        (p->*TMethod)(_groupId, _serviceType, _propertyName, _NewValue);
    }
};

/**
 * @class Property
 * @brief Abstraction class of Property object.
 */
class DLLExport Property {
public:
    /**
     * @enum PropertyType
     * @brief Possible types for Property object.
     */
    enum PropertyType { NUMBER = 0,
                        TEXT = 1,
                        BOOL = 2,
                        COLOR = 3,
                        RANGE = 4,
                        ATTRIBUTE = 5,
                        CHANGABLE_ATTRIBUTE = 6 };

    /**
     * @brief Array of string values of types of Property object.
     */
    static const std::string m_TypeNames[];
    /**
     * @brief Max option list size.
     */
    static const unsigned int MAX_OPTION_SIZE;

    /**
     * @fn Property();
     * @brief Constructor.
     */
    Property() {}

    /**
     * @fn virtual ~Property();
     * @brief Destructor.
     */
    virtual ~Property() {}

    /**
     * @fn Property(const std::string& _name, Jzon::Node _initialValue, bool _readOnly, bool _used, PropertyType _type, ServicePropertyDelegate* _function);
     * @brief Constructor.
     * @param _name Name of the Property object.
     * @param _initialValue Initial value of the Property object.
     * @param _readOnly Access rights for Property object.
     * @param _used Defines whether the Property object is optional.
     * @param _type Type of the Property object.
     * @param _function Shared pointer to delegate function.
     */
    Property(const std::string& _name, Jzon::Node _initialValue, bool _readOnly, bool _used, PropertyType _type, ServicePropertyDelegate* _function);

    /**
     * @fn std::string GetName() const;
     * @brief Getter for Name field.
     * @return Current Name.
     */
    std::string GetName() const;

    /**
     * @fn Jzon::Node GetValue();
     * @brief Getter for Value field.
     * @return Current Value.
     */
    Jzon::Node GetValue();

    /**
     * @fn bool IsReadOnly() const;
     * @brief Getter for ReadOnly field.
     * @return Indication of ReadOnly.
     */
    bool IsReadOnly() const;

    /**
     * @fn bool IsUsed() const;
     * @brief Getter for Used field.
     * @return Indication of Used.
     */
    bool IsUsed() const;

    /**
     * @fn std::string GetStringType() const;
     * @brief Getter for Type field.
     * @return Current Type.
     */
    std::string GetStringType() const;

    /**
     * @fn PropertyType GetType() const;
     * @brief Getter for Type field.
     * @return Current Type.
     */
    PropertyType GetType() const;

    /**
     * @fn std::vector<Jzon::Node> GetOptions() const;
     * @brief Getter for Options field.
     * @return Vector of Options.
     */
    const std::vector<Jzon::Node>& GetOptions() const;

    /**
     * @fn ServicePropertyDelegate::Ptr GetPropertyFunction() const;
     * @brief Getter for shared pointer to delegate function.
     * @return Shared pointer to delegate function.
     */
    ServicePropertyDelegate::Ptr GetPropertyFunction() const;

    /**
     * @fn unsigned long long GetTimestamp() const;
     * @brief Getter for timestamp of last change of Property object.
     * @return Current Timestamp.
     */
    unsigned long long GetTimestamp() const;

    /**
     * @fn void SetName(const std::string& _name);
     * @brief Setter for Name field.
     * @param _name New name to be set.
     */
    void SetName(const std::string& _name);

    /**
     * @fn void SetValue(Jzon::Node _value);
     * @brief Setter for Value field and updates Timestamp field.
     * @param _value New value to be set.
     */
    void SetValue(Jzon::Node _value);

    /**
     * @fn void SetReadOnly(bool _readOnly);
     * @brief Setter for ReadOnly field.
     * @param _readOnly New read only to be set.
     */
    void SetReadOnly(bool _readOnly);

    /**
     * @fn void SetUsed(bool _used);
     * @brief Setter for Used field.
     * @param _used New used to be set.
     */
    void SetUsed(bool _used);

    /**
     * @fn void SetType(PropertyType _type)
     * @brief Setter for Type field.
     * @param _type New type to be set.
     */
    void SetType(PropertyType _type);

    /**
     * @fn void SetOptions(const std::vector<Jzon::Node>& _options)
     * @brief Setting option list.
     * @param _options New option list to be set.
     */
    void SetOptions(const std::vector<Jzon::Node>& _options);

    /**
     * @fn void SetPropertyFunction(ServicePropertyDelegate::Ptr _function);
     * @brief Setter for shared pointer to delegate function.
     * @param _function Shared pointer to delegate function.
     */
    void SetPropertyFunction(ServicePropertyDelegate::Ptr _function);

    /**
     * @fn void SetTimestamp();
     * @brief Setter for Timestamp field. Timestamp will be set to the current time.
     */
    void SetTimestamp();

    /**
     * @fn Property& AddOption(Jzon::Node _value);
     * @brief Adds potential options for Property object.
     * @param _value Option.
     * @return Reference to this object.
     */
    Property& AddOption(Jzon::Node _value);

    /**
     * @fn void ToJson(Utility::JsonStruct& _json, bool _stripped = false) const;
     * @brief Fills Utility::JsonStruct object with information about Property object.
     * @param _json Utility::JsonStruct to be filled.
     * @param _stripped Whether the Property object will have full information or only the Name and the Value.
     */
    void ToJson(Utility::JsonStruct& _json, bool _stripped = false) const;
private:
    std::string m_Name;
    Jzon::Node m_Value;
    bool m_ReadOnly;
    bool m_Used;
    PropertyType m_Type;
    ServicePropertyDelegate::Ptr m_Function;
    std::vector<Jzon::Node> m_Options;
    unsigned long long m_Timestamp;
    Utility::Mutex m_Mutex;
};

}


#endif // PROPERTY_H
