/****************************************************************************
 *  Copyright (c) 2018 Oblo Living d.o.o
 *
 *  All Rights Reserved
 *
 *  P R O P R I E T A R Y    &    C O N F I D E N T I A L
 *
 *  -----------------------------------------------------
 *  http://www.obloliving.com/
 *  -----------------------------------------------------
 *
 * \project Oblo WISE SDK - SDK library
 * \file Service.h
 * \brief
 * WISE Service class.
 *
 * @Author Igor Stefanovic
 *
 * \notes
 *
 * \history
 *****************************************************************************/
#ifndef SERVICE_H
#define SERVICE_H

#include "util/Common.h"
#include "ipdevice/StrServiceTable.h"
#include "ipdevice/Property.h"
#include "ipdevice/Command.h"
#include "util/Exceptions.h"

/**
 * @namespace IPDevice
 * @brief IP device functionality.
 */
namespace IPDevice {

/**
 * @class BaseService
 * @brief The base class of service implementation.
 */
class DLLExport BaseService {
public:
    /**
     * @typedef std::map<std::string, Property>::const_iterator PropertyIterator;
     * @brief Iterator for map of Property objects.
     */
    typedef std::map<std::string, Property>::const_iterator PropertyIterator;

    /**
     * @typedef std::map<std::string, Command>::const_iterator CommandIterator;
     * @brief Iterator for map of Command objects.
     */
    typedef std::map<std::string, Command>::const_iterator CommandIterator;

    /**
     * @fn BaseService(const std::string& _type);
     * @brief Constructor.
     * @param _type Service type.
     */
    BaseService(const std::string& _type);

    /**
     * @fn virtual ~BaseService();
     * @brief Destructor.
     */
    virtual ~BaseService() {}

    /**
     * @fn std::string GetServiceType() const;
     * @brief Getter for service type field.
     * @return Service type.
     */
    std::string GetServiceType() const;

    /**
     * @fn void AddProperty(const Property &_property);
     * @brief Add property to the service.
     * @param _property Property object.
     */
    void AddProperty(const Property &_property);

    /**
     * @fn void AddCommand(const Command &_command);
     * @brief Add command to the service.
     * @param _command Command object.
     */
    void AddCommand(const Command &_command);

    /**
     * @fn Property& FindPropertyByName(const std::string& _name);
     * @brief Find property by name in service.
     * @param _name Name of the property.
     * @return Property object.
     */
    Property& FindPropertyByName(const std::string& _name);

    /**
     * @fn Command& FindCommandByName(const std::string& _name);
     * @brief Find command by name in service.
     * @param _name Name of the command.
     * @return Command object.
     */
    Command& FindCommandByName(const std::string& _name);

    /**
     * @fn void SetPropertyFunction(const std::string& _property, ServicePropertyDelegate::Ptr _function);
     * @brief Setter for function to be invoked on property changing.
     * @param _property Property name.
     * @param _function Function.
     */
    void SetPropertyFunction(const std::string& _property, ServicePropertyDelegate::Ptr _function);

    /**
     * @fn void SetCommandFunction(const std::string& _command, ServiceCommandDelegate::Ptr _function);
     * @brief Setter for function to be invoked on command invoking.
     * @param _command Command name.
     * @param _function Function.
     */
    void SetCommandFunction(const std::string& _command, ServiceCommandDelegate::Ptr _function);

    /**
     * @fn void ToJson(Utility::JsonStruct& _json, unsigned long long _timestamp = 0, bool _stripped = false) const;
     * @brief Fill the JSON with service informations.
     * @param _json Json to be filled.
     * @param _timestamp Timestamp that defines which properties will be a part of information. Check Device::GetState.
     * @param _stripped Indication whether the full property information will be provided.
     */
    void ToJson(Utility::JsonStruct& _json, unsigned long long _timestamp = 0, bool _stripped = false) const;

    /**
     * @fn PropertyIterator BeginProperty() const;
     * @brief Provides the begin iterator of the map of properties.
     * @return Map iterator.
     */
    PropertyIterator BeginProperty() const;

    /**
     * @fn PropertyIterator EndProperty() const;
     * @brief Provides the end iterator of the map of properties.
     * @return Map iterator.
     */
    PropertyIterator EndProperty() const;

    /**
     * @fn CommandIterator BeginCommand() const;
     * @brief Provides the begin iterator of the map of commands.
     * @return Map iterator.
     */
    CommandIterator BeginCommand() const;

    /**
     * @fn CommandIterator EndCommand() const;
     * @brief Provides the end iterator of the map of commands.
     * @returns Map iterator.
     */
    CommandIterator EndCommand() const;
private:
    std::string m_Type;
    std::map<std::string, Property> m_PropertyList;
    std::map<std::string, Command> m_CommandList;
};

/**
 * @class DeviceService
 * @brief DeviceService class implementation.
 * @extends BaseService
 */
class DLLExport DeviceService : public BaseService {
public:
    /**
     * @enum ConnectionType
     * @brief Connection type enumeration.
     */
    enum ConnectionType {
        HUB = 0,
        HUBLESS
    };

    /**
     * @fn DeviceService();
     * @brief Constructor.
     */
    DeviceService() : BaseService(SERVICE_TYPE_DEVICE) {
        AddProperty(Property(PROPERTY_NAME_CONNECTION_TYPE, Jzon::Node(0), true, true, Property::ATTRIBUTE, NULL)
                    .AddOption("HUB")
                    .AddOption("HUBLESS"));
        AddProperty(Property(PROPERTY_NAME_SLEEPING_TIMEOUT, Jzon::Node(0.0), true, true, Property::NUMBER, NULL));
        AddProperty(Property(PROPERTY_NAME_PROTOCOL, Jzon::Node(0.0), true, true, Property::NUMBER, NULL));
        AddProperty(Property(PROPERTY_NAME_VERSION, Jzon::Node(""), true, true, Property::TEXT, NULL));
        AddProperty(Property(PROPERTY_NAME_VERSION_NUMBER, Jzon::Node(0.0), true, true, Property::NUMBER, NULL));
        AddProperty(Property(PROPERTY_NAME_NAME, Jzon::Node(""), false, true, Property::TEXT, NULL));
    }

    /**
     * @fn void SetNameHandler(ServicePropertyDelegate::Ptr _function);
     * @brief Registers the callback function on name property change.
     * @param _function Callback function.
     */
    void SetNameHandler(ServicePropertyDelegate::Ptr _function) {
        SetPropertyFunction(PROPERTY_NAME_NAME, _function);
    }

    /**
     * @fn static BaseService* Create();
     * @brief Creates an DeviceService object.
     * @return Pointer to BaseService object.
     */
    static BaseService* Create() {
        return new DeviceService();
    }
};

/**
 * @class BatteryService
 * @brief BatteryService class implementation.
 * @extends BaseService
 */
class DLLExport BatteryService : public BaseService {
public:
    /**
     * @fn BatteryService();
     * @brief Constructor.
     */
    BatteryService() : BaseService(SERVICE_TYPE_BATTERY) {
        AddProperty(Property(PROPERTY_NAME_LEVEL, Jzon::Node(100.0), true, true, Property::RANGE, NULL)
                    .AddOption(Jzon::Node(0.0))
                    .AddOption(Jzon::Node(100.0)));
        AddProperty(Property(PROPERTY_NAME_CRITICAL, Jzon::Node(false), true, true, Property::BOOL, NULL));
        AddProperty(Property(PROPERTY_NAME_VOLTAGE, Jzon::Node(0.0), true, false, Property::NUMBER, NULL));
    }

    /**
     * @fn void UseVoltageProp();
     * @brief Sets the optional voltage property for usage.
     */
    void UseVoltageProp() {
        Property& p = FindPropertyByName(PROPERTY_NAME_VOLTAGE);
        if (!p.IsUsed()) {
            p.SetUsed(true);
        }
    }

    /**
     * @fn static BaseService* Create();
     * @brief Creates an BatteryService object.
     * @return Pointer to BaseService object.
     */
    static BaseService* Create() {
        return new BatteryService();
    }
};

/**
 * @class SwitchService
 * @brief SwitchService class implementation.
 * @extends BaseService
 */
class DLLExport SwitchService : public BaseService {
public:
    /**
     * @fn SwitchService();
     * @brief Constructor.
     */
    SwitchService() : BaseService(SERVICE_TYPE_SWITCH) {
        AddProperty(Property(PROPERTY_NAME_STATE, Jzon::Node(false), false, true, Property::BOOL, NULL));
        AddCommand(Command(COMMAND_NAME_TOGGLE, NULL));
    }

    /**
     * @fn void SetStateHandler(ServicePropertyDelegate::Ptr _function);
     * @brief Registers the callback function on state property change.
     * @param _function Callback function.
     */
    void SetStateHandler(ServicePropertyDelegate::Ptr _function) {
        SetPropertyFunction(PROPERTY_NAME_STATE, _function);
    }

    /**
     * @fn void SetToggleHandler(ServiceCommandDelegate::Ptr _function);
     * @brief Registers the callback function on toggle command invocation.
     * @param _function Callback function.
     */
    void SetToggleHandler(ServiceCommandDelegate::Ptr _function) {
        SetCommandFunction(COMMAND_NAME_TOGGLE, _function);
    }

    /**
     * @fn static BaseService* Create();
     * @brief Creates an SwitchService object.
     * @return Pointer to BaseService object.
     */
    static BaseService* Create() {
        return new SwitchService();
    }
};

/**
 * @class LightService
 * @brief LightService class implementation.
 * @extends BaseService
 */
class DLLExport LightService : public BaseService {
public:
    /**
     * @fn LightService();
     * @brief Constructor.
     */
    LightService() : BaseService(SERVICE_TYPE_LIGHT) {
        AddProperty(Property(PROPERTY_NAME_STATE, Jzon::Node(false), false, true, Property::BOOL, NULL));
        AddCommand(Command(COMMAND_NAME_TOGGLE, NULL));
    }

    /**
     * @fn void SetStateHandler(ServicePropertyDelegate::Ptr _function);
     * @brief Registers the callback function on state property change.
     * @param _function Callback function.
     */
    void SetStateHandler(ServicePropertyDelegate::Ptr _function) {
        SetPropertyFunction(PROPERTY_NAME_STATE, _function);
    }

    /**
     * @fn void SetToggleHandler(ServiceCommandDelegate::Ptr _function);
     * @brief Registers the callback function on toggle command invocation.
     * @param _function Callback function.
     */
    void SetToggleHandler(ServiceCommandDelegate::Ptr _function) {
        SetCommandFunction(COMMAND_NAME_TOGGLE, _function);
    }

    /**
     * @fn static BaseService* Create();
     * @brief Creates an LightService object.
     * @return Pointer to BaseService object.
     */
    static BaseService* Create() {
        return new LightService();
    }
};

/**
 * @class TemperatureService
 * @brief TemperatureService class implementation.
 * @extends BaseService
 */
class DLLExport TemperatureService : public BaseService {
public:
    /**
     * @fn TemperatureService();
     * @brief Constructor.
     */
    TemperatureService() : BaseService(SERVICE_TYPE_TEMPERATURE) {
        AddProperty(Property(PROPERTY_NAME_MAX, Jzon::Node(50.0), true, true, Property::NUMBER, NULL));
        AddProperty(Property(PROPERTY_NAME_MIN, Jzon::Node(-30.0), true, true, Property::NUMBER, NULL));
        AddProperty(Property(PROPERTY_NAME_VALUE, Jzon::Node(20.0), true, true, Property::RANGE, NULL)
                    .AddOption(FindPropertyByName(PROPERTY_NAME_MIN).GetValue())
                    .AddOption(FindPropertyByName(PROPERTY_NAME_MAX).GetValue()));
    }

    /**
     * @fn static BaseService* Create();
     * @brief Creates an TemperatureService object.
     * @return Pointer to BaseService object.
     */
    static BaseService* Create() {
        return new TemperatureService();
    }
};

/**
 * @class LevelControlService
 * @brief LevelControlService class implementation.
 * @extends BaseService
 */
class DLLExport LevelControlService : public BaseService {
public:
    /**
     * @fn LevelControlService();
     * @brief Constructor.
     */
    LevelControlService() : BaseService(SERVICE_TYPE_LEVEL_CONTROL) {
        AddProperty(Property(PROPERTY_NAME_LEVEL, Jzon::Node(100.0), false, true, Property::RANGE, NULL)
                    .AddOption(Jzon::Node(0.0))
                    .AddOption(Jzon::Node(100.0)));
        AddCommand(Command(COMMAND_NAME_MOVE_TO, NULL)
                   .AddInputParam(InputParam(PROPERTY_NAME_LEVEL, InputParam::RANGE).AddRange(Jzon::Node(0), Jzon::Node(100)))
                   .AddInputParam(InputParam(COMMAND_PARAM_NAME_TRANSITION_TIME, InputParam::RANGE).AddRange(Jzon::Node(0), Jzon::Node(3600))));
        AddCommand(Command(COMMAND_NAME_STOP, NULL));
    }

    /**
     * @fn void SetLevelHandler(ServicePropertyDelegate::Ptr _function);
     * @brief Registers the callback function on level property change.
     * @param _function Callback function.
     */
    void SetLevelHandler(ServicePropertyDelegate::Ptr _function) {
        SetPropertyFunction(PROPERTY_NAME_LEVEL, _function);
    }

    /**
     * @fn void SetMoveToHandler(ServiceCommandDelegate::Ptr _function);
     * @brief Registers the callback function on move to command invocation.
     * @param _function Callback function.
     */
    void SetMoveToHandler(ServiceCommandDelegate::Ptr _function) {
        SetCommandFunction(COMMAND_NAME_MOVE_TO, _function);
    }

    /**
     * @fn void SetStopHandler(ServiceCommandDelegate::Ptr _function);
     * @brief Registers the callback function on stop command invocation.
     * @param _function Callback function.
     */
    void SetStopHandler(ServiceCommandDelegate::Ptr _function) {
        SetCommandFunction(COMMAND_NAME_STOP, _function);
    }

    /**
     * @fn static BaseService* Create();
     * @brief Creates an LevelControlService object.
     * @return Pointer to BaseService object.
     */
    static BaseService* Create() {
        return new LevelControlService();
    }
};

/**
 * @class DimService
 * @brief DimService class implementation.
 * @extends BaseService
 */
class DLLExport DimService : public BaseService {
public:
    /**
     * @fn DimService();
     * @brief Constructor.
     */
    DimService() : BaseService(SERVICE_TYPE_DIM) {
        AddProperty(Property(PROPERTY_NAME_LEVEL, Jzon::Node(100.0), false, true, Property::RANGE, NULL)
                    .AddOption(Jzon::Node(0.0))
                    .AddOption(Jzon::Node(100.0)));
        AddCommand(Command(COMMAND_NAME_MOVE_TO, NULL)
                   .AddInputParam(InputParam(PROPERTY_NAME_LEVEL, InputParam::RANGE).AddRange(Jzon::Node(0), Jzon::Node(100)))
                   .AddInputParam(InputParam(COMMAND_PARAM_NAME_TRANSITION_TIME, InputParam::RANGE).AddRange(Jzon::Node(0), Jzon::Node(3600))));
        AddCommand(Command(COMMAND_NAME_STOP, NULL));
    }

    /**
     * @fn void SetLevelHandler(ServicePropertyDelegate::Ptr _function);
     * @brief Registers the callback function on level property change.
     * @param _function Callback function.
     */
    void SetLevelHandler(ServicePropertyDelegate::Ptr _function) {
        SetPropertyFunction(PROPERTY_NAME_LEVEL, _function);
    }

    /**
     * @fn void SetMoveToHandler(ServiceCommandDelegate::Ptr _function);
     * @brief Registers the callback function on move to command invocation.
     * @param _function Callback function.
     */
    void SetMoveToHandler(ServiceCommandDelegate::Ptr _function) {
        SetCommandFunction(COMMAND_NAME_MOVE_TO, _function);
    }

    /**
     * @fn void SetStopHandler(ServiceCommandDelegate::Ptr _function);
     * @brief Registers the callback function on stop command invocation.
     * @param _function Callback function.
     */
    void SetStopHandler(ServiceCommandDelegate::Ptr _function) {
        SetCommandFunction(COMMAND_NAME_STOP, _function);
    }

    /**
     * @fn static BaseService* Create();
     * @brief Creates an DimService object.
     * @return Pointer to BaseService object.
     */
    static BaseService* Create() {
        return new DimService();
    }
};

/**
 * @class ColorTemperatureService
 * @brief ColorTemperatureService class implementation.
 * @extends BaseService
 */
class DLLExport ColorTemperatureService : public BaseService {
public:
    /**
     * @fn ColorTemperatureService();
     * @brief Constructor.
     */
    ColorTemperatureService() : BaseService(SERVICE_TYPE_COLOR_TEMPERATURE) {
        AddProperty(Property(PROPERTY_NAME_COLOR_TEMPERATURE, Jzon::Node(4500.0), false, true, Property::RANGE, NULL)
                    .AddOption(Jzon::Node(2700.0))
                    .AddOption(Jzon::Node(6500.0)));
        AddCommand(Command(COMMAND_NAME_MOVE_TO, NULL)
                   .AddInputParam(InputParam(PROPERTY_NAME_COLOR_TEMPERATURE, InputParam::RANGE).AddRange(Jzon::Node(2700.0), Jzon::Node(6500.0)))
                   .AddInputParam(InputParam(COMMAND_PARAM_NAME_TRANSITION_TIME, InputParam::RANGE).AddRange(Jzon::Node(0), Jzon::Node(3600))));
        AddCommand(Command(COMMAND_NAME_STOP, NULL));
    }

    /**
     * @fn void SetColorTemperatureHandler(ServicePropertyDelegate::Ptr _function);
     * @brief Registers the callback function on color temperature property change.
     * @param _function Callback function.
     */
    void SetColorTemperatureHandler(ServicePropertyDelegate::Ptr _function) {
        SetPropertyFunction(PROPERTY_NAME_COLOR_TEMPERATURE, _function);
    }

    /**
     * @fn void SetMoveToHandler(ServiceCommandDelegate::Ptr _function);
     * @brief Registers the callback function on move to command invocation.
     * @param _function Callback function.
     */
    void SetMoveToHandler(ServiceCommandDelegate::Ptr _function) {
        SetCommandFunction(COMMAND_NAME_MOVE_TO, _function);
    }

    /**
     * @fn void SetStopHandler(ServiceCommandDelegate::Ptr _function);
     * @brief Registers the callback function on stop command invocation.
     * @param _function Callback function.
     */
    void SetStopHandler(ServiceCommandDelegate::Ptr _function) {
        SetCommandFunction(COMMAND_NAME_STOP, _function);
    }

    /**
     * @fn static BaseService* Create();
     * @brief Creates an ColorTemperatureService object.
     * @return Pointer to BaseService object.
     */
    static BaseService* Create() {
        return new ColorTemperatureService();
    }
};

/**
 * @class ColorService
 * @brief ColorService class implementation.
 * @extends BaseService
 */
class DLLExport ColorService : public BaseService {
public:
    /**
     * @fn ColorService();
     * @brief Constructor.
     */
    ColorService() : BaseService(SERVICE_TYPE_COLOR) {
        AddProperty(Property(PROPERTY_NAME_COLOR, Jzon::Node("FFFFFFFF"), false, true, Property::COLOR, NULL));
        AddCommand(Command(COMMAND_NAME_MOVE_TO, NULL)
                   .AddInputParam(InputParam(PROPERTY_NAME_COLOR, InputParam::COLOR))
                   .AddInputParam(InputParam(COMMAND_PARAM_NAME_TRANSITION_TIME, InputParam::RANGE).AddRange(Jzon::Node(0), Jzon::Node(3600))));
        AddCommand(Command(COMMAND_NAME_STOP, NULL));
    }

    /**
     * @fn void SetColorHandler(ServicePropertyDelegate::Ptr _function);
     * @brief Registers the callback function on color property change.
     * @param _function Callback function.
     */
    void SetColorHandler(ServicePropertyDelegate::Ptr _function) {
        SetPropertyFunction(PROPERTY_NAME_COLOR, _function);
    }

    /**
     * @fn void SetMoveToHandler(ServiceCommandDelegate::Ptr _function);
     * @brief Registers the callback function on move to command invocation.
     * @param _function Callback function.
     */
    void SetMoveToHandler(ServiceCommandDelegate::Ptr _function) {
        SetCommandFunction(COMMAND_NAME_MOVE_TO, _function);
    }

    /**
     * @fn void SetStopHandler(ServiceCommandDelegate::Ptr _function);
     * @brief Registers the callback function on stop command invocation.
     * @param _function Callback function.
     */
    void SetStopHandler(ServiceCommandDelegate::Ptr _function) {
        SetCommandFunction(COMMAND_NAME_STOP, _function);
    }

    /**
     * @fn static BaseService* Create();
     * @brief Creates an ColorService object.
     * @return Pointer to BaseService object.
     */
    static BaseService* Create() {
        return new ColorService();
    }
};

/**
 * @class AlarmService
 * @brief AlarmService class implementation.
 * @extends BaseService
 */
class DLLExport AlarmService : public BaseService {
public:
    /**
     * @fn AlarmService();
     * @brief Constructor.
     */
    AlarmService() : BaseService(SERVICE_TYPE_ALARM) {
        AddProperty(Property(PROPERTY_NAME_STATE, Jzon::Node(false), true, true, Property::BOOL, NULL));
    }

    /**
     * @fn static BaseService* Create();
     * @brief Creates an AlarmService object.
     * @return Pointer to BaseService object.
     */
    static BaseService* Create() {
        return new AlarmService();
    }
};

/**
 * @class HumidityService
 * @brief HumidityService class implementation.
 * @extends BaseService
 */
class DLLExport HumidityService : public BaseService {
public:
    /**
     * @fn HumidityService();
     * @brief Constructor.
     */
    HumidityService() : BaseService(SERVICE_TYPE_HUMIDITY) {
        AddProperty(Property(PROPERTY_NAME_VALUE, Jzon::Node(0.0), true, true, Property::RANGE, NULL)
                    .AddOption(Jzon::Node(0.0))
                    .AddOption(Jzon::Node(100.0)));
    }

    /**
     * @fn static BaseService* Create();
     * @brief Creates an HumidityService object.
     * @return Pointer to BaseService object.
     */
    static BaseService* Create() {
        return new HumidityService();
    }
};

/**
 * @class GasService
 * @brief GasService class implementation.
 * @extends BaseService
 */
class DLLExport GasService : public BaseService {
public:
    /**
     * @enum GasType
     * @brief Gas types supported by device.
     */
    enum GasType {
        OTHER = 0,
        CO,
        CO2,
        COMBUSTIBLE
    };

    /**
     * @fn GasService();
     * @brief Constructor.
     */
    GasService() : BaseService(SERVICE_TYPE_GAS) {
        AddProperty(Property(PROPERTY_NAME_STATE, Jzon::Node(false), true, true, Property::BOOL, NULL));
        AddProperty(Property(PROPERTY_NAME_TYPE, Jzon::Node(OTHER), true, true, Property::ATTRIBUTE, NULL)
                    .AddOption("OTHER")
                    .AddOption("CO")
                    .AddOption("CO2")
                    .AddOption("COMBUSTIBLE"));
    }

    /**
     * @fn static BaseService* Create();
     * @brief Creates an GasService object.
     * @return Pointer to BaseService object.
     */
    static BaseService* Create() {
        return new GasService();
    }
};

/**
 * @class PowerMeteringService
 * @brief PowerMeteringService class implementation.
 * @extends BaseService
 */
class DLLExport PowerMeteringService : public BaseService {
public:
    /**
     * @fn PowerMeteringService();
     * @brief Constructor.
     */
    PowerMeteringService() : BaseService(SERVICE_TYPE_POWER_METERING) {
        AddProperty(Property(PROPERTY_NAME_POWER_CONSUMPTION, Jzon::Node(0.0), true, true, Property::NUMBER, NULL));
        AddProperty(Property(PROPERTY_NAME_ACTIVE_POWER, Jzon::Node(0.0), true, true, Property::NUMBER, NULL));
        AddProperty(Property(PROPERTY_NAME_REACTIVE_POWER, Jzon::Node(0.0), true, true, Property::NUMBER, NULL));
        AddProperty(Property(PROPERTY_NAME_POWER_FACTOR, Jzon::Node(0.0), true, true, Property::NUMBER, NULL));
        AddProperty(Property(PROPERTY_NAME_VOLTAGE, Jzon::Node(0.0), true, true, Property::NUMBER, NULL));
        AddProperty(Property(PROPERTY_NAME_RMS_CURRENT, Jzon::Node(0.0), true, true, Property::NUMBER, NULL));
        AddProperty(Property(PROPERTY_NAME_APPARENT_POWER, Jzon::Node(0.0), true, true, Property::NUMBER, NULL));
    }

    /**
     * @fn static BaseService* Create();
     * @brief Creates an PowerMeteringService object.
     * @return Pointer to BaseService object.
     */
    static BaseService* Create() {
        return new PowerMeteringService();
    }
};

/**
 * @class ShutterService
 * @brief ShutterService class implementation.
 * @extends BaseService
 */
class DLLExport ShutterService : public BaseService {
public:
    /**
     * @fn ShutterService();
     * @brief Constructor.
     */
    ShutterService() : BaseService(SERVICE_TYPE_SHUTTER) {
        AddProperty(Property(PROPERTY_NAME_CALIBRATE, Jzon::Node(false), false, true, Property::BOOL, NULL));
        AddProperty(Property(PROPERTY_NAME_LEVEL, Jzon::Node(0.0), false, true, Property::RANGE, NULL)
                    .AddOption(Jzon::Node(0.0))
                    .AddOption(Jzon::Node(100.0)));
        AddProperty(Property(PROPERTY_NAME_DEFAULT_LEVEL, Jzon::Node(0.0), false, true, Property::RANGE, NULL)
                    .AddOption(Jzon::Node(0.0))
                    .AddOption(Jzon::Node(100.0)));
        AddProperty(Property(PROPERTY_NAME_ANGLE, Jzon::Node(0.0), false, true, Property::RANGE, NULL)
                    .AddOption(Jzon::Node(0.0))
                    .AddOption(Jzon::Node(100.0)));
    }

    /**
     * @fn void SetCalibrateHandler(ServicePropertyDelegate::Ptr _function);
     * @brief Registers the callback function on calibrate property change.
     * @param _function Callback function.
     */
    void SetCalibrateHandler(ServicePropertyDelegate::Ptr _function) {
        SetPropertyFunction(PROPERTY_NAME_CALIBRATE, _function);
    }

    /**
     * @fn void SetLevelHandler(ServicePropertyDelegate::Ptr _function);
     * @brief Registers the callback function on level property change.
     * @param _function Callback function.
     */
    void SetLevelHandler(ServicePropertyDelegate::Ptr _function) {
        SetPropertyFunction(PROPERTY_NAME_LEVEL, _function);
    }

    /**
     * @fn void SetDefaultLevelHandler(ServicePropertyDelegate::Ptr _function);
     * @brief Registers the callback function on default level property change.
     * @param _function Callback function.
     */
    void SetDefaultLevelHandler(ServicePropertyDelegate::Ptr _function) {
        SetPropertyFunction(PROPERTY_NAME_DEFAULT_LEVEL, _function);
    }

    /**
     * @fn void SetAngleHandler(ServicePropertyDelegate::Ptr _function);
     * @brief Registers the callback function on angle property change.
     * @param _function Callback function.
     */
    void SetAngleHandler(ServicePropertyDelegate::Ptr _function) {
        SetPropertyFunction(PROPERTY_NAME_ANGLE, _function);
    }

    /**
     * @fn void SetMoveToDefaultHandler(ServiceCommandDelegate::Ptr _function);
     * @brief Registers the callback function on move to default command invocation.
     * @param _function Callback function.
     */
    void SetMoveToDefaultHandler(ServiceCommandDelegate::Ptr _function) {
        SetCommandFunction(COMMAND_NAME_MOVE_TO_DEFAULT, _function);
    }

    /**
     * @fn static BaseService* Create();
     * @brief Creates an ShutterService object.
     * @return Pointer to BaseService object.
     */
    static BaseService* Create() {
        return new ShutterService();
    }
};

/**
 * @class ThermostatService
 * @brief ThermostatService class implementation.
 * @extends BaseService
 */
class DLLExport ThermostatService : public BaseService {
public:
    /**
     * @enum Mode
     * @brief Possible modes for thermostat device.
     */
    enum Mode {
        AUTO_MD = 0,
        HEATING = 1,
        COOLING = 2
    };

    /**
     * @enum Operation
     * @brief Possible operations for thermostat device.
     */
    enum Operation {
        MANUAL = 0,
        AUTO_OP = 1
    };

    /**
     * @fn ThermostatService();
     * @brief Constructor.
     */
    ThermostatService() : BaseService(SERVICE_TYPE_THERMOSTAT) {
        AddProperty(Property(PROPERTY_NAME_MAX, Jzon::Node(40.0), true, true, Property::NUMBER, NULL));
        AddProperty(Property(PROPERTY_NAME_MIN, Jzon::Node(-10.0), true, true, Property::NUMBER, NULL));
        AddProperty(Property(PROPERTY_NAME_SETPOINT, Jzon::Node(20.0), false, true, Property::RANGE, NULL)
                    .AddOption(FindPropertyByName(PROPERTY_NAME_MAX).GetValue())
                    .AddOption(FindPropertyByName(PROPERTY_NAME_MIN).GetValue()));
        AddProperty(Property(PROPERTY_NAME_MODE, Jzon::Node(AUTO_MD), false, false, Property::ATTRIBUTE, NULL)
                    .AddOption("AUTO_MD")
                    .AddOption("HEATING")
                    .AddOption("COOLING"));
        AddProperty(Property(PROPERTY_NAME_OPERATION, Jzon::Node(MANUAL), false, false, Property::ATTRIBUTE, NULL)
                    .AddOption("MANUAL")
                    .AddOption("AUTO_OP"));
    }

    /**
     * @fn void SetPointHandler(ServicePropertyDelegate::Ptr _function);
     * @brief Registers the callback function on setpoint property change.
     * @param _function Callback function.
     */
    void SetPointHandler(ServicePropertyDelegate::Ptr _function) {
        SetPropertyFunction(PROPERTY_NAME_SETPOINT, _function);
    }

    /**
     * @fn void SetModeHandler(ServicePropertyDelegate::Ptr _function);
     * @brief Registers the callback function on mode property change.
     * @param _function Callback function.
     */
    void SetModeHandler(ServicePropertyDelegate::Ptr _function) {
        SetPropertyFunction(PROPERTY_NAME_MODE, _function);
    }

    /**
     * @fn void SetOperationHandler(ServicePropertyDelegate::Ptr _function);
     * @brief Registers the callback function on operation property change.
     * @param _function Callback function.
     */
    void SetOperationHandler(ServicePropertyDelegate::Ptr _function) {
        SetPropertyFunction(PROPERTY_NAME_OPERATION, _function);
    }

    /**
     * @fn void UseModeProp();
     * @brief Sets the optional mode property for usage.
     */
    void UseModeProp() {
        Property& p = FindPropertyByName(PROPERTY_NAME_MODE);
        if (!p.IsUsed()) {
            p.SetUsed(true);
        }
    }

    /**
     * @fn void UseOperationProp();
     * @brief Sets the optional operation property for usage.
     */
    void UseOperationProp() {
        Property& p = FindPropertyByName(PROPERTY_NAME_OPERATION);
        if (!p.IsUsed()) {
            p.SetUsed(true);
        }
    }

    /**
     * @fn static BaseService* Create();
     * @brief Creates an ThermostatService object.
     * @return Pointer to BaseService object.
     */
    static BaseService* Create() {
        return new ThermostatService();
    }
};

/**
 * @class DiagnosticsService
 * @brief DiagnosticsService class implementation.
 * @extends BaseService
 */
class DLLExport DiagnosticsService : public BaseService {
public:
    /**
     * @fn DiagnosticsService();
     * @brief Constructor.
     */
    DiagnosticsService() : BaseService(SERVICE_TYPE_DIAGNOSTIC) {
        AddProperty(Property(PROPERTY_NAME_UP_TIME, Jzon::Node(0), true, false, Property::NUMBER, NULL));
        AddProperty(Property(PROPERTY_NAME_BOOT_COUNT, Jzon::Node(0), true, false, Property::NUMBER, NULL));
        AddProperty(Property(PROPERTY_NAME_RSSI, Jzon::Node(0), true, false, Property::RANGE, NULL)
                    .AddOption(Jzon::Node(-120))
                    .AddOption(Jzon::Node(0)));
    }

    /**
     * @fn void UseUptimeProp();
     * @brief Sets the optional uptime property for usage.
     */
    void UseUptimeProp() {
        Property& p = FindPropertyByName(PROPERTY_NAME_UP_TIME);
        if (!p.IsUsed()) {
            p.SetUsed(true);
        }
    }

    /**
     * @fn void UseBootCountProp();
     * @brief Sets the optional boot count property for usage.
     */
    void UseBootCountProp() {
        Property& p = FindPropertyByName(PROPERTY_NAME_BOOT_COUNT);
        if (!p.IsUsed()) {
            p.SetUsed(true);
        }
    }

    /**
     * @fn void UseRSSIProp();
     * @brief Sets the optional rssi property for usage.
     */
    void UseRSSIProp() {
        Property& p = FindPropertyByName(PROPERTY_NAME_RSSI);
        if (!p.IsUsed()) {
            p.SetUsed(true);
        }
    }

    /**
     * @fn static BaseService* Create();
     * @brief Creates an DiagnosticsService object.
     * @return Pointer to BaseService object.
     */
    static BaseService* Create() {
        return new DiagnosticsService();
    }
};

/**
 * @class MotionDetectionService
 * @brief MotionDetectionService class implementation.
 * @extends BaseService
 */
class DLLExport MotionDetectionService : public BaseService {
public:
    /**
     * @fn MotionDetectionService();
     * @brief Constructor.
     */
    MotionDetectionService() : BaseService(SERVICE_TYPE_MOTION) {
        AddProperty(Property(PROPERTY_NAME_STATE, Jzon::Node(false), true, true, Property::BOOL, NULL));
        AddProperty(Property(PROPERTY_NAME_DEBOUNCING_PERIOD, Jzon::Node(0), false, false, Property::NUMBER, NULL));
    }


    /**
     * @fn void SetDebouncePeriodHandler(ServicePropertyDelegate::Ptr _function);
     * @brief Registers the callback function on debouncing period property change.
     * @param _function Callback function.
     */
    void SetDebouncePeriodHandler(ServicePropertyDelegate::Ptr _function) {
        Property& p = FindPropertyByName(PROPERTY_NAME_DEBOUNCING_PERIOD);
        if (p.IsUsed()) {
            SetPropertyFunction(PROPERTY_NAME_DEBOUNCING_PERIOD, _function);
        }
    }

    /**
     * @fn void UseDebouncingPeriodProp();
     * @brief Sets the optional debouncing period property for usage.
     */
    void UseDebouncingPeriodProp() {
        Property& p = FindPropertyByName(PROPERTY_NAME_DEBOUNCING_PERIOD);
        if (!p.IsUsed()) {
            p.SetUsed(true);
        }
    }

    /**
     * @fn static BaseService* Create();
     * @brief Creates an MotionDetectionService object.
     * @return Pointer to BaseService object.
     */
    static BaseService* Create() {
        return new MotionDetectionService();
    }
};

/**
 * @class ContactService
 * @brief ContactService class implementation.
 * @extends BaseService
 */
class DLLExport ContactService : public BaseService {
public:
    /**
     * @fn ContactService();
     * @brief Constructor.
     */
    ContactService() : BaseService(SERVICE_TYPE_CONTACT) {
        AddProperty(Property(PROPERTY_NAME_STATE, Jzon::Node(false), true, true, Property::BOOL, NULL));
    }

    /**
     * @fn static BaseService* Create();
     * @brief Creates an ContactService object.
     * @return Pointer to BaseService object.
     */
    static BaseService* Create() {
        return new ContactService();
    }
};

/**
 * @class FloodService
 * @brief FloodService class implementation.
 * @extends BaseService
 */
class DLLExport FloodService : public BaseService {
public:
    /**
     * @fn FloodService();
     * @brief Constructor.
     */
    FloodService() : BaseService(SERVICE_TYPE_FLOOD) {
        AddProperty(Property(PROPERTY_NAME_STATE, Jzon::Node(false), true, true, Property::BOOL, NULL));
    }

    /**
     * @fn static BaseService* Create();
     * @brief Creates an FloodService object.
     * @return Pointer to BaseService object.
     */
    static BaseService* Create() {
        return new FloodService();
    }
};

/**
 * @class SmokeService
 * @brief SmokeService class implementation.
 * @extends BaseService
 */
class DLLExport SmokeService : public BaseService {
public:
    /**
     * @fn SmokeService();
     * @brief Constructor.
     */
    SmokeService() : BaseService(SERVICE_TYPE_SMOKE) {
        AddProperty(Property(PROPERTY_NAME_STATE, Jzon::Node(false), true, true, Property::BOOL, NULL));
    }

    /**
     * @fn static BaseService* Create();
     * @brief Creates an SmokeService object.
     * @return Pointer to BaseService object.
     */
    static BaseService* Create() {
        return new SmokeService();
    }
};

/**
 * @class SimpleAlertService
 * @brief SimpleAlertService class implementation.
 * @extends BaseService
 */
class DLLExport SimpleAlertService : public BaseService {
public:
    /**
     * @enum AlertMode
     * @brief Alert modes.
     */
    enum AlertMode {
        SOUND = 0,
        LIGHT = 1,
        SOUND_AND_LIGHT = 2,
    };

    /**
     * @enum Alert
     * @brief Alert types.
     */
    enum AlertLight {
        NOTIFICATION = 0,
        ALARM = 1,
        WARNING = 2,
        EMERGENCY = 3
    };

    /**
     * @fn SimpleAlertService();
     * @brief Constructor.
     */
    SimpleAlertService() : BaseService(SERVICE_TYPE_SIMPLE_ALERT) {
        AddProperty(Property(PROPERTY_NAME_STATE, Jzon::Node(false), false, true, Property::BOOL, NULL));
        AddProperty(Property(PROPERTY_NAME_MODE, Jzon::Node(SOUND), false, true, Property::ATTRIBUTE, NULL)
                    .AddOption("SOUND")
                    .AddOption("LIGHT")
                    .AddOption("SOUND_AND_LIGHT"));
        AddProperty(Property(PROPERTY_NAME_ALERT, Jzon::Node(NOTIFICATION), false, true, Property::ATTRIBUTE, NULL)
                    .AddOption("NOTIFICATION")
                    .AddOption("ALARM")
                    .AddOption("WARNING")
                    .AddOption("EMERGENCY"));

    }

    /**
     * @fn void SetStateHandler(ServicePropertyDelegate::Ptr _function);
     * @brief Registers the callback function on state property change.
     * @param _function Callback function.
     */
    void SetStateHandler(ServicePropertyDelegate::Ptr _function) {
        SetPropertyFunction(PROPERTY_NAME_STATE, _function);
    }

    /**
     * @fn void SetAlertModeHandler(ServicePropertyDelegate::Ptr _function);
     * @brief Registers the callback function on alert mode property change.
     * @param _function Callback function.
     */
    void SetAlertModeHandler(ServicePropertyDelegate::Ptr _function) {
        SetPropertyFunction(PROPERTY_NAME_MODE, _function);
    }

    /**
     * @fn void SetAlertTypeHandler(ServicePropertyDelegate::Ptr _function);
     * @brief Registers the callback function on alert type property change.
     * @param _function Callback function.
     */
    void SetAlertTypeHandler(ServicePropertyDelegate::Ptr _function) {
        SetPropertyFunction(PROPERTY_NAME_ALERT, _function);
    }

    /**
     * @fn static BaseService* Create();
     * @brief Creates an SimpleAlertService object.
     * @return Pointer to BaseService object.
     */
    static BaseService* Create() {
        return new SimpleAlertService();
    }
};

/**
 * @class VolumeService
 * @brief VolumeService class implementation.
 * @extends BaseService
 */
class DLLExport VolumeService : public BaseService {
public:
    /**
     * @fn VolumeService();
     * @brief Constructor.
     */
    VolumeService() : BaseService(SERVICE_TYPE_VOLUME) {
        AddProperty(Property(PROPERTY_NAME_LEVEL, Jzon::Node(100.0), false, true, Property::RANGE, NULL)
                    .AddOption(Jzon::Node(0.0))
                    .AddOption(Jzon::Node(100.0)));
    }

    /**
     * @fn void SetLevelHandler(ServicePropertyDelegate::Ptr _function);
     * @brief Registers the callback function on level property change.
     * @param _function Callback function.
     */
    void SetLevelHandler(ServicePropertyDelegate::Ptr _function) {
        SetPropertyFunction(PROPERTY_NAME_LEVEL, _function);
    }

    /**
     * @fn static BaseService* Create();
     * @brief Creates an VolumeService object.
     * @return Pointer to BaseService object.
     */
    static BaseService* Create() {
        return new VolumeService();
    }
};

/**
 * @class SystemMonitorService
 * @brief SystemMonitorService class implementation.
 * @extends BaseService
 */
class DLLExport SystemMonitorService : public BaseService {
public:
    /**
     * @fn SystemMonitorService();
     * @brief Constructor.
     */
    SystemMonitorService() : BaseService(SERVICE_TYPE_SYSTEM_MONITOR) {
        AddProperty(Property(PROPERTY_NAME_CPU_LOAD, Jzon::Node(0), true, true, Property::RANGE, NULL)
                    .AddOption(Jzon::Node(0))
                    .AddOption(Jzon::Node(100)));
        AddProperty(Property(PROPERTY_NAME_RAM_LOAD, Jzon::Node(0), true, true, Property::RANGE, NULL)
                    .AddOption(Jzon::Node(0))
                    .AddOption(Jzon::Node(100)));
        AddProperty(Property(PROPERTY_NAME_MEMORY_LOAD, Jzon::Node(0), true, true, Property::RANGE, NULL)
                    .AddOption(Jzon::Node(0))
                    .AddOption(Jzon::Node(100)));
    }

    /**
     * @fn static BaseService* Create();
     * @brief Creates an SystemMonitorService object.
     * @return Pointer to BaseService object.
     */
    static BaseService* Create() {
        return new SystemMonitorService();
    }
};

/**
 * @class VMControlService
 * @brief VMControlService class implementation.
 * @extends BaseService
 */
class DLLExport VMControlService : public BaseService {
public:
    /**
     * @enum State
     * @brief States.
     */
    enum State {
        IDLE = 0,
        CREATING,
        REMOVING,
        SUSPENDING,
        SUSPENDED,
        STARTING,
        RESTARTING,
        RESUMING,
        STARTED,
        STOPPING,
        STOPPED,
        ERROR
    };

    /**
     * @enum Command
     * @brief Commands.
     */
    enum Command {
        NONE = 0,
        START,
        STOP,
        SUSPEND,
        RESUME,
        RESTART,
        REMOVE
    };

    /**
     * @fn VMControlService();
     * @brief Constructor.
     */
    VMControlService() : BaseService(SERVICE_TYPE_VM_CONTROL) {
        AddProperty(Property(PROPERTY_NAME_STATE, Jzon::Node(IDLE), true, true, Property::ATTRIBUTE, NULL)
                    .AddOption(Jzon::Node("IDLE"))
                    .AddOption(Jzon::Node("CREATING"))
                    .AddOption(Jzon::Node("REMOVING"))
                    .AddOption(Jzon::Node("SUSPENDING"))
                    .AddOption(Jzon::Node("SUSPENDED"))
                    .AddOption(Jzon::Node("STARTING"))
                    .AddOption(Jzon::Node("RESTARTING"))
                    .AddOption(Jzon::Node("RESUMING"))
                    .AddOption(Jzon::Node("STARTED"))
                    .AddOption(Jzon::Node("STOPPING"))
                    .AddOption(Jzon::Node("STOPPED"))
                    .AddOption(Jzon::Node("ERROR")));
        AddProperty(Property(PROPERTY_NAME_COMMAND, Jzon::Node(NONE), false, true, Property::ATTRIBUTE, NULL)
                    .AddOption(Jzon::Node("NONE"))
                    .AddOption(Jzon::Node("START"))
                    .AddOption(Jzon::Node("STOP"))
                    .AddOption(Jzon::Node("SUSPEND"))
                    .AddOption(Jzon::Node("RESUME"))
                    .AddOption(Jzon::Node("RESTART"))
                    .AddOption(Jzon::Node("REMOVE")));
        AddProperty(Property(PROPERTY_NAME_MESSAGE, Jzon::Node(""), false, true, Property::TEXT, NULL));
    }

    /**
     * @fn void SetVMCommandHandler(ServiceCommandDelegate::Ptr _function);
     * @brief Registers the callback function on command property changed.
     * @param _function Callback function.
     */
    void SetVMCommandHandler(ServicePropertyDelegate::Ptr _function) {
        SetPropertyFunction(PROPERTY_NAME_COMMAND, _function);
    }

    /**
     * @fn void SetMessageHandler(ServiceCommandDelegate::Ptr _function);
     * @brief Registers the callback function on message property changed.
     * @param _function Callback function.
     */
    void SetMessageHandler(ServicePropertyDelegate::Ptr _function) {
        SetPropertyFunction(PROPERTY_NAME_MESSAGE, _function);
    }

    /**
     * @fn static BaseService* Create();
     * @brief Creates an VMControlService object.
     * @return Pointer to BaseService object.
     */
    static BaseService* Create() {
        return new VMControlService();
    }
};

/**
 * @class BasicService
 * @brief BasicService class implementation.
 * @extends BaseService
 */
class BasicService : public BaseService {
public:
    /**
     * @fn BasicService();
     * @brief Constructor.
     */
    BasicService() : BaseService(SERVICE_TYPE_BASIC) {
        AddCommand(Command(COMMAND_NAME_PROXY_CMD, NULL)
                   .AddInputParam(InputParam(COMMAND_PARAM_ANY, InputParam::ANY))
                   .AddInputParam(InputParam(COMMAND_PARAM_CMD_NAME, InputParam::TEXT)));
    }

    /**
     * @fn void SetProxyCmdHandler(ServiceCommandDelegate::Ptr _function);
     * @brief Registers the callback function on proxy command invocation.
     * @param _function Callback function.
     */
    void SetProxyCmdHandler(ServiceCommandDelegate::Ptr _function) {
        SetCommandFunction(COMMAND_NAME_PROXY_CMD, _function);
    }

    /**
     * @fn static BaseService* Create();
     * @brief Creates an BasicService object.
     * @return Pointer to BaseService object.
     */
    static BaseService* Create() {
        return new BasicService();
    }
};

/**
 * @class ConfigurationService
 * @brief ConfigurationService class implementation.
 * @extends BaseService
 */
class ConfigurationService : public BaseService {
public:
    /**
     * @fn ConfigurationService();
     * @brief Constructor.
     */
    ConfigurationService() : BaseService(SERVICE_TYPE_CONFIGURATION) {
        AddProperty(Property(PROPERTY_NAME_NAME, Jzon::Node(""), false, true, Property::TEXT, NULL));
        AddProperty(Property(PROPERTY_NAME_TYPE, Jzon::Node(Property::NUMBER), false, true, Property::ATTRIBUTE, NULL)
                    .AddOption("NUMBER")
                    .AddOption("TEXT")
                    .AddOption("BOOL")
                    .AddOption("COLOR"));
        AddProperty(Property(PROPERTY_NAME_VALUE, Jzon::Node(""), false, true, Property::TEXT, NULL));
        AddProperty(Property(PROPERTY_NAME_SET, Jzon::Node(false), false, true, Property::BOOL, NULL));
        AddProperty(Property(PROPERTY_NAME_GET, Jzon::Node(false), false, true, Property::BOOL, NULL));
        AddCommand(Command(COMMAND_NAME_SET_CONFIG_PARAM, NULL)
                   .AddInputParam(InputParam(PROPERTY_NAME_NAME, InputParam::TEXT))
                   .AddInputParam(InputParam(PROPERTY_NAME_TYPE, InputParam::ATTRIBUTE)
                                  .AddOption("NUMBER")
                                  .AddOption("TEXT")
                                  .AddOption("BOOL")
                                  .AddOption("COLOR"))
                   .AddInputParam(InputParam(PROPERTY_NAME_VALUE, InputParam::TEXT)));
        AddCommand(Command(COMMAND_NAME_GET_CONFIG_PARAM, NULL)
                   .AddInputParam(InputParam(PROPERTY_NAME_NAME, InputParam::TEXT)));
    }

    /**
     * @fn void SetNamePropHandler(ServicePropertyDelegate::Ptr _function);
     * @brief Registers the callback function on level property change.
     * @param _function Callback function.
     */
    void SetNamePropHandler(ServicePropertyDelegate::Ptr _function) {
        SetPropertyFunction(PROPERTY_NAME_NAME, _function);
    }

    /**
     * @fn void SetTypePropHandler(ServicePropertyDelegate::Ptr _function);
     * @brief Registers the callback function on level property change.
     * @param _function Callback function.
     */
    void SetTypePropHandler(ServicePropertyDelegate::Ptr _function) {
        SetPropertyFunction(PROPERTY_NAME_TYPE, _function);
    }

    /**
     * @fn void SetValuePropHandler(ServicePropertyDelegate::Ptr _function);
     * @brief Registers the callback function on level property change.
     * @param _function Callback function.
     */
    void SetValuePropHandler(ServicePropertyDelegate::Ptr _function) {
        SetPropertyFunction(PROPERTY_NAME_VALUE, _function);
    }

    /**
     * @fn void SetSetPropHandler(ServicePropertyDelegate::Ptr _function);
     * @brief Registers the callback function on level property change.
     * @param _function Callback function.
     */
    void SetSetPropHandler(ServicePropertyDelegate::Ptr _function) {
        SetPropertyFunction(PROPERTY_NAME_SET, _function);
    }

    /**
     * @fn void SetGetPropHandler(ServicePropertyDelegate::Ptr _function);
     * @brief Registers the callback function on level property change.
     * @param _function Callback function.
     */
    void SetGetPropHandler(ServicePropertyDelegate::Ptr _function) {
        SetPropertyFunction(PROPERTY_NAME_GET, _function);
    }


    /**
     * @fn void SetSetConfigParamCmdHandler(ServiceCommandDelegate::Ptr _function);
     * @brief Registers the callback function on proxy command invocation.
     * @param _function Callback function.
     */
    void SetSetConfigParamCmdHandler(ServiceCommandDelegate::Ptr _function) {
        SetCommandFunction(COMMAND_NAME_SET_CONFIG_PARAM, _function);
    }

    /**
     * @fn void SetGetConfigParamCmdHandler(ServiceCommandDelegate::Ptr _function);
     * @brief Registers the callback function on proxy command invocation.
     * @param _function Callback function.
     */
    void SetGetConfigParamCmdHandler(ServiceCommandDelegate::Ptr _function) {
        SetCommandFunction(COMMAND_NAME_GET_CONFIG_PARAM, _function);
    }

    /**
     * @fn static BaseService* Create();
     * @brief Creates an BasicService object.
     * @return Pointer to BaseService object.
     */
    static BaseService* Create() {
        return new ConfigurationService();
    }
};

}

#endif // SERVICE_H
