/****************************************************************************
 *  Copyright (c) 2018 Oblo Living d.o.o
 *
 *  All Rights Reserved
 *
 *  P R O P R I E T A R Y    &    C O N F I D E N T I A L
 *
 *  -----------------------------------------------------
 *  http://www.obloliving.com/
 *  -----------------------------------------------------
 *
 * \project Oblo WISE SDK - SDK library
 * \file SleepingDevice.h
 * \brief
 * WISE Sleeping device class.
 *
 * @Author Igor Stefanovic
 *
 * \notes
 *
 * \history
 *****************************************************************************/
#ifndef SLEEPING_DEVICE_H
#define SLEEPING_DEVICE_H

#include "util/Common.h"
#include "ipdevice/Device.h"

/**
 * @namespace IPDevice
 */
namespace IPDevice {

/**
 * @class SleepingDevice
 * @extends Device
 * @brief Implementation of SleepingDevice type of IP device.
 */
class DLLExport SleepingDevice : protected Device {
public:
    /**
     * @enum State
     * @brief Sleeping device states defined by WISE specification
     */
    enum State {
        SLEEP,
        ALIVE
    };

    /**
     * @fn SleepingDevice(long _sleepingTimeout);
     * @brief Constructor.
     * @param _sleepingTimeout Sleeping timeout parameter defining time spent in State::SLEEP state.
     */
    SleepingDevice(long _sleepingTimeout);

    /**
     * @fn virtual ~SleepingDevice();
     * @brief Destructor.
     */
    virtual ~SleepingDevice();

    /**
     * @brief Time spent in State::ALIVE state. Value is 10 seconds.
     */
    static const int ALIVE_TIME = 10000;
protected:
    /**
     * @fn virtual void Start();
     * @brief Starting Sleeping device.
     */
    virtual void Start();

    /**
     * @fn virtual void Stop();
     * @brief Stopping Sleeping device.
     */
    virtual void Stop();

    /**
     * @fn virtual void GetDeviceInfo(const Messenger::RequestMessage::Ptr _req, Messenger::ResponseMessage::Ptr _resp);
     * @brief Respond to GetDeviceInfo request providing information about Device object, Group object and Service object.
     * @param _req Shared pointer to request message.
     * @param _resp Shared pointer to response message.
     */
    void GetDeviceInfo(const Messenger::RequestMessage::Ptr _req, Messenger::ResponseMessage::Ptr _resp);

    /**
     * @fn virtual void GetState(const Messenger::RequestMessage::Ptr _req, Messenger::ResponseMessage::Ptr _resp);
     * @brief Respond to GetState request providing changes since time defined in parameter of message.
     * @param _req Shared pointer to request message.
     * @param _resp Shared pointer to response message.
     */
    void GetState(const Messenger::RequestMessage::Ptr _req, Messenger::ResponseMessage::Ptr _resp);

    /**
     * @fn virtual void SetTime(const Messenger::RequestMessage::Ptr _req, Messenger::ResponseMessage::Ptr _resp);
     * @brief Respond to SetTime request providing an interface for time changing.
     * @param _req Shared pointer to request message.
     * @param _resp Shared pointer to response message.
     */
    void SetTime(const Messenger::RequestMessage::Ptr _req, Messenger::ResponseMessage::Ptr _resp);

    /**
     * @fn virtual void AdjustTime(const Messenger::RequestMessage::Ptr _req, Messenger::ResponseMessage::Ptr _resp);
     * @brief Respond to AdjustTime request providing an interface for time adjusting.
     * @param _req Shared pointer to request message.
     * @param _resp Shared pointer to response message.
     */
    void AdjustTime(const Messenger::RequestMessage::Ptr _req, Messenger::ResponseMessage::Ptr _resp);

    /**
     * @fn virtual void VerifyIdentity(const Messenger::RequestMessage::Ptr _req, Messenger::ResponseMessage::Ptr _resp);
     * @brief Respond to VerifyIdentity request as defined in WISE specification.
     * @param _req Shared pointer to request message.
     * @param _resp Shared pointer to response message.
     */
    void VerifyIdentity(const Messenger::RequestMessage::Ptr _req, Messenger::ResponseMessage::Ptr _resp);

    /**
     * @fn virtual void GetGroupList(const Messenger::RequestMessage::Ptr _req, Messenger::ResponseMessage::Ptr _resp);
     * @brief Respond to GetGroupList request providing a list of all Group objects supported by Device object.
     * @param _req Shared pointer to request message.
     * @param _resp Shared pointer to response message.
     */
    void GetGroupList(const Messenger::RequestMessage::Ptr _req, Messenger::ResponseMessage::Ptr _resp);

    /**
     * @fn virtual void GetServiceList(const Messenger::RequestMessage::Ptr _req, Messenger::ResponseMessage::Ptr _resp);
     * @brief Respond to GetServiceList request providing a list of all Service objects supported by specified Group object.
     * @param _req Shared pointer to request message.
     * @param _resp Shared pointer to response message.
     */
    void GetServiceList(const Messenger::RequestMessage::Ptr _req, Messenger::ResponseMessage::Ptr _resp);

    /**
     * @fn virtual void GetService(const Messenger::RequestMessage::Ptr _req, Messenger::ResponseMessage::Ptr _resp);
     * @brief Respond to GetService request providing a Service object supported by specified Group object.
     * @param _req Shared pointer to request message.
     * @param _resp Shared pointer to response message.
     */
    void GetService(const Messenger::RequestMessage::Ptr _req, Messenger::ResponseMessage::Ptr _resp);

    /**
     * @fn virtual void GetPropertyList(unsigned int _groupId, const std::string& _serviceType, const Messenger::RequestMessage::Ptr _req, Messenger::ResponseMessage::Ptr _resp);
     * @brief Respond to GetPropertyList request providing a list of all Property objects defined in Service object.
     * @param _groupId ID of the Group object.
     * @param _serviceType Type of the Service object.
     * @param _req Shared pointer to request message.
     * @param _resp Shared pointer to response message.
     */
    void GetPropertyList(unsigned int _groupId, const std::string& _serviceType, const Messenger::RequestMessage::Ptr _req, Messenger::ResponseMessage::Ptr _resp);

    /**
     * @fn virtual void SetPropertyValue(unsigned int _groupId, const std::string& _serviceType, const Messenger::RequestMessage::Ptr _req, Messenger::ResponseMessage::Ptr _resp);
     * @brief Respond to SetPropertyValue request and changes value of Property object of Service object specified in parameters of the request.
     * @param _groupId ID of the Group object.
     * @param _serviceType Type of the Service object.
     * @param _req Shared pointer to request message.
     * @param _resp Shared pointer to response message.
     */
    void SetPropertyValue(unsigned int _groupId, const std::string& _serviceType, const Messenger::RequestMessage::Ptr _req, Messenger::ResponseMessage::Ptr _resp);

    /**
     * @fn virtual void GetPropertyValue(unsigned int _groupId, const std::string& _serviceType, const Messenger::RequestMessage::Ptr _req, Messenger::ResponseMessage::Ptr _resp);
     * @brief Respond to GetPropertyValue request providing a value of Property object of Service object specified in parameters of the request.
     * @param _groupId ID of the Group object.
     * @param _serviceType Type of the Service object.
     * @param _req Shared pointer to request message.
     * @param _resp Shared pointer to response message.
     */
    void GetPropertyValue(unsigned int _groupId, const std::string& _serviceType, const Messenger::RequestMessage::Ptr _req, Messenger::ResponseMessage::Ptr _resp);

    /**
     * @fn virtual void GetCommandList(unsigned int _groupId, const std::string& _serviceType, const Messenger::RequestMessage::Ptr _req, Messenger::ResponseMessage::Ptr _resp);
     * @brief Respond to GetCommandList request providing a list of all Command object defined in Service object.
     * @param _groupId ID of the Group object.
     * @param _serviceType Type of the Service object.
     * @param _req Shared pointer to request message.
     * @param _resp Shared pointer to response message.
     */
    void GetCommandList(unsigned int _groupId, const std::string& _serviceType, const Messenger::RequestMessage::Ptr _req, Messenger::ResponseMessage::Ptr _resp);

    /**
     * @fn virtual void ExecuteCommand(unsigned int _groupId, const std::string& _serviceType, const Messenger::RequestMessage::Ptr _req, Messenger::ResponseMessage::Ptr _resp);
     * @brief Respond to ExecuteCommand request and invokes a command specified by parameters of the request.
     * @param _groupId ID of the Group object.
     * @param _serviceType Type of the Service object.
     * @param _req Shared pointer to request message.
     * @param _resp Shared pointer to response message.
     */
    void ExecuteCommand(unsigned int _groupId, const std::string& _serviceType, const Messenger::RequestMessage::Ptr _req, Messenger::ResponseMessage::Ptr _resp);

    /**
     * @fn virtual bool UpdatePropertyAttributeList(unsigned int _groupId, const std::string& _serviceType, const std::string& _name, const std::vector<Jzon::Node>& _attributeList)
     * @brief Sets new value of the Property object specified by the parameters.
     * @param _groupId ID of the Group object.
     * @param _serviceType Type of the Service object.
     * @param _name Name of the Property object.
     * @param _attributeList New attribute list for Property object.
     * @return Indication whether updating was a success.
     */
    bool UpdatePropertyAttributeList(unsigned int _groupId, const std::string& _serviceType, const std::string& _name, const std::vector<Jzon::Node>& _attributeList);

    /**
     * @fn virtual void SetPropertyValueIntern(unsigned int _groupId, const std::string& _serviceType, const std::string& _name, const Jzon::Node& _value, bool _sendEvent);
     * @brief Sets new value of the Property object specified by the parameters.
     * @param _groupId ID of the Group object.
     * @param _serviceType Type of the Service object.
     * @param _name Name of the Property object.
     * @param _value Value of the Property object.
     * @param _sendEvent Indication whether the PropertyChanged event will be sent.
     */
    void SetPropertyValueIntern(unsigned int _groupId, const std::string& _serviceType, const std::string& _name, const Jzon::Node& _value, bool _sendEvent);

    /**
     * @fn virtual Jzon::Node GetPropertyValueIntern(unsigned int _groupId, const std::string& _serviceType, const std::string& _name);
     * @brief Provides value of the Property object specified by the parameters.
     * @param _groupId ID of the Group object.
     * @param _serviceType Type of the Service object.
     * @param _name Name of the Property object.
     * @return Current value of the Property object inside of Jzon::Node object.
     */
    Jzon::Node GetPropertyValueIntern(unsigned int _groupId, const std::string &_serviceType, const std::string &_name);

    /**
     * @fn void OnStateChange(Utility::Timer& _timer);
     * @brief Callback function that will be invoked after timer expiration to change the state of Sleeping device.
     * @param _timer Timer that invoked this callback.
     */
    void OnStateChange(Utility::Timer& _timer);
private:
    long m_SleepingTimeout;
    Utility::Timer* m_Timer;
    State m_State;
    Utility::Mutex m_Mutex;
};

}

#endif // SLEEPING_DEVICE_H
