/****************************************************************************
 *  Copyright (c) 2018 Oblo Living d.o.o
 *
 *  All Rights Reserved
 *
 *  P R O P R I E T A R Y    &    C O N F I D E N T I A L
 *
 *  -----------------------------------------------------
 *  http://www.obloliving.com/
 *  -----------------------------------------------------
 *
 * \project Oblo WISE SDK - SDK library
 * \file Client.h
 * \brief
 * MQTT client abstraction classes.
 *
 * @Author Igor Stefanovic
 *
 * \notes
 *
 * \history
 *****************************************************************************/
#ifndef CLIENT_H
#define CLIENT_H

#include "util/Common.h"
#include "util/Mutex.h"
#include "util/Exceptions.h"
#include "messenger/Network.h"
#include "messenger/NetworkProxy.h"
#include "messenger/MQTTClient.h"
#include <algorithm>

/**
 * @namespace Messenger
 * @brief Implementation of MQTT Client.
 */
namespace Messenger {

/**
 * @class AsyncClient
 * @brief The AsyncClient class.
 * @extends ClientImpl
 * @implements INetworkProxyListener
 */
class DLLExport AsyncClient : public ClientImpl, public INetworkProxyListener {
public:
    /**
     * @fn AsyncClient(const std::string& _clientId, const std::string& _serverUri,
                IDeliveryListener* _deliveryListener, Utility::ILogListener* _logListener = NULL);
     * @brief Constructor.
     * @param _clientId MQTT Client ID.
     * @param _serverUri MQTT Server URI.
     * @param _deliveryListener Listener for arrived messages.
     * @param _logListener Log listener.
     */
    AsyncClient(const std::string& _clientId, const std::string& _serverUri,
                IDeliveryListener* _deliveryListener, Utility::ILogListener* _logListener = NULL);

    /**
     * @fn virtual ~AsyncClient();
     * @brief Destructor.
     */
    virtual ~AsyncClient();

    /**
     * @fn void SetUsername(const std::string& _username);
     * @brief Setter for username. Must be called before Connect().
     * @param _username Username to set.
     */
    void SetUsername(const std::string& _username);

    /**
     * @fn void SetPassword(const std::string& _password);
     * @brief Setter for password. Must be called before Connect().
     * @param _password Password to set.
     */
    void SetPassword(const std::string& _password);

    /**
     * @fn void SetCleanSession(bool _cleanSession);
     * @brief Setter for clean session value. Must be called before Connect().
     * @param _cleanSession Clean session on disconnect.
     */
    void SetCleanSession(bool _cleanSession);

    /**
     * @fn void SetKeepAliveInterval(int _keepAliveInterval);
     * @brief Setter for keep alive interval. Must be called before Connect().
     * @param _keepAliveInterval Interval within ping to server will be set.
     */
    void SetKeepAliveInterval(int _keepAliveInterval);

    /**
     * @fn void SetConnectionTimeoutInterval(int _timeout);
     * @brief Setter for connection timeout interval. Must be called before Connect().
     * @param _timeout Interval to wait connection to complete.
     */
    void SetConnectionTimeoutInterval(int _timeout);

    /**
     * @fn void SetStatusConfiguration(const std::string& _topic, const Message::Ptr _msg);
     * @brief Configure message to be sent after client is disconnected.
     * @param _topic MQTT topic.
     * @param _msg MQTT message.
     */
    void SetStatusConfiguration(const std::string& _topic, const Message::Ptr _msg);

    /**
     * @fn void SetSecureConfiguration(const std::string& _trustStore, const std::string& _keyStore, const std::string& _privateKey, const std::string& _privateKeyPassword);
     * @brief Set secure configuration. Connect must be called again, to apply this configuration.
     * @param _trustStore Secure connection trust store.
     * @param _keyStore Secure connection key store.
     * @param _privateKey Secure connection private key.
     * @param _privateKeyPassword Secure connection pasword for a provate key.
     */
    void SetSecureConfiguration(const std::string& _trustStore, const std::string& _keyStore, const std::string& _privateKey, const std::string& _privateKeyPassword);

    /**
     * @fn bool IsCleanSession() const;
     * @brief Indicates whether session is cleaned on disconnect.
     * @return Clean session value.
     */
    bool IsCleanSession() const;

    /**
     * @fn void DisableStatusMessage();
     * @brief Disables MQTT last will message.
     */
    void DisableStatusMessage();

    /**
     * @fn void RegisterConnectionListener(IConnectionListener* _connectionListener);
     * @brief Function used to register for events about network status.
     * @param _connectionListener Network connection listener.
     */
    void RegisterConnectionListener(IConnectionListener* _connectionListener);

    /**
     * @fn void UnregisterConnectionListener();
     * @brief Unregisters network connection listener.
     */
    void UnregisterConnectionListener();

    /**
     * @fn virtual void Connect();
     * @brief Connects to messaging server.
     */
    virtual void Connect();

    /**
     * @fn virtual void Disconnect(long _timeout = 0);
     * @brief Disconnects from a messaging server, if it was previusly connected.
     * @param _timeout Timeout to disconnect.
     */
    virtual void Disconnect(long _timeout = 0);

    /**
     * @fn void Publish(const std::string &_topic, const Message::Ptr _msg);
     * @brief Asynchronous publishing.
     * @param _topic Topic.
     * @param _msg MQTT message.
     */
    void Publish(const std::string &_topic, const Message::Ptr _msg);
protected:
    /**
     * @fn virtual void OnMessageArrived(const std::string& _topic, const Message::Ptr _msg);
     * @brief Method called when new message is arrived. It is redirected to delivery listener.
     * @param _topic Topic.
     * @param _msg MQTT message.
     */
    virtual void OnMessageArrived(const std::string& _topic, const Message::Ptr _msg);

    /**
     * @fn virtual void OnConnectionLost(const std::string& _cause);
     * @brief Redirects connection lost to connection listener.
     * @param _cause Connection lost cause.
     */
    virtual void OnConnectionLost(const std::string& _cause);

    /**
     * @fn virtual void OnPublishRequired(const std::string& _topic, const Message::Ptr _msg);
     * @brief Publishes messages.
     * @param _topic Topic.
     * @param _msg MQTT message.
     */
    virtual void OnPublishRequired(const std::string& _topic, const Message::Ptr _msg);

    /**
     * @fn virtual void OnPublishFailed(int _code, const std::string& _topic, const Message::Ptr _msg);
     * @brief Used to report publish failure to delivery listener.
     * @param _code MQTT error code.
     * @param _topic Topic.
     * @param _msg MQTT message.
     */
    virtual void OnPublishFailed(int _code, const std::string& _topic, const Message::Ptr _msg);
private:
    IDeliveryListener* m_DeliveryListener;
    IConnectionListener* m_ConnectionListener;
    Utility::ILogListener* m_LogListener;

    ConnectConfig::Ptr m_ConnConfig;
    SSLConfig::Ptr m_SSLConfig;
    StatusConfig::Ptr m_StatusConfig;

    AsyncNetworkProxy m_NetworkProxy;
    Utility::Mutex m_PublishMutex;
};

/**
 * @class BuffClient
 * @brief The BuffClient class.
 * @extends AsyncClient
 */
class DLLExport BuffClient : public AsyncClient {
public:
    /**
     * @fn BuffClient(const std::string& _clientId, const std::string& _serverUri,
               IDeliveryListener& _deliveryListener, unsigned int _maxQueueSize = 100,
               Utility::ILogListener* _logListener = NULL);
     * @brief Constructor.
     * @param _clientId Client ID.
     * @param _serverUri Server URI.
     * @param _deliveryListener Delivery listener.
     * @param _maxQueueSize Maximum number of publish messages stored to the buffer.
     * @param _logListener Log listener.
     */
    BuffClient(const std::string& _clientId, const std::string& _serverUri,
               IDeliveryListener& _deliveryListener, unsigned int _maxQueueSize = 100,
               Utility::ILogListener* _logListener = NULL);

    /**
     * @fn virtual void Connect();
     * @brief If basic connect succeeds it recovers subscriptions and failed publish messages.
     */
    virtual void Connect();

    /**
     * @fn void Subscribe(const std::string &_topic, int _qos);
     * @brief If client is not connected, it stores subscriptions and will try to complete the procedure on next connect.
     * @param _topic Topic to subscribe.
     * @param _qos QoS level.
     */
    void Subscribe(const std::string &_topic, int _qos);

    /**
     * @fn void Subscribe(const TopicCollection& _topics, const QoSCollection& _qoss);
     * @brief If client is not connected, it stores subscriptions and will try to complete the procedure on next connect.
     * @param _topics Topics to subscribe.
     * @param _qoss QoS levels.
     */
    void Subscribe(const TopicCollection& _topics, const QoSCollection& _qoss);

    /**
     * @fn virtual void Unsubscribe(const TopicCollection &_topics);
     * @brief Does basic unsubscribe and removes stored subscriptions.
     * @param _topics Topics.
     */
    void Unsubscribe(const TopicCollection &_topics);

    /**
     * @fn virtual void Unsubscribe(const std::string& _topic);
     * @brief Does basic unsubscribe and removes stored subscription.
     * @param _topic Topic.
     */
    void Unsubscribe(const std::string& _topic);

    /**
     * @fn void ImmediatePublish(const std::string& _topic, const Message::Ptr _msg);
     * @brief Publish without pushing the message to the queue.
     * @param _topic Topic.
     * @param _msg MQTT message.
     */
    void ImmediatePublish(const std::string& _topic, const Message::Ptr _msg);

    /**
     * @fn void IgnoreSessionPresent(bool _ignore);
     * @brief Control subscriptions if a session is present on the broker. Session present flag will be checked and if it is false the subscription will be applied. Otherwise, subscription will be skipped.
     * @param _ignore Indication how to handle the session present flag.
     */
    void IgnoreSessionPresent(bool _ignore);
private:
    typedef std::pair<std::string, int> SubElement;
    typedef std::deque<SubElement> SubscriptionBuffer;

    virtual void OnPublishFailed(int _code, const std::string &_topic, const Message::Ptr _msg);
    virtual void StoreSubscription(const TopicCollection& _topics, const QoSCollection& _qoss);
    virtual void StoreUnsubscription(const TopicCollection& _topics);

    void RemoveSubscription(const TopicCollection& _topics);
    void RemoveUnsubscription(const TopicCollection& _topics);
    void Remove(SubscriptionBuffer& _buffer, const TopicCollection& _topics);
    void PublishStore();
    void SubscribeStore();
    void UnsubscribeStore();
    SubscriptionBuffer m_Subs;
    Utility::Mutex m_SubMutex;

    SubscriptionBuffer m_Unsubs;
    Utility::Mutex m_UnsubMutex;

    typedef std::queue<PublishHolder::Ptr> PublicationBuffer;
    PublicationBuffer m_Publications;
    Utility::Mutex m_PublishMutex;

    const unsigned int m_MaxQueueSize;
    bool m_IgnoreSessionPresent;

};


}


#endif // CLIENT_H
