/****************************************************************************
 *  Copyright (c) 2018 Oblo Living d.o.o
 *
 *  All Rights Reserved
 *
 *  P R O P R I E T A R Y    &    C O N F I D E N T I A L
 *
 *  -----------------------------------------------------
 *  http://www.obloliving.com/
 *  -----------------------------------------------------
 *
 * \project Oblo WISE SDK - SDK library
 * \file MQTTClient.h
 * \brief
 * PAHO client abstraction classes.
 *
 * @Author Igor Stefanovic
 *
 * \notes
 *
 * \history
 *****************************************************************************/
#ifndef MQTT_CLIENT_H
#define MQTT_CLIENT_H

#include "util/Common.h"
#include "util/Logger.h"
#include "util/SharedPointer.h"
#include "util/Mutex.h"
#include "util/ScopedLock.h"
#include "util/Exceptions.h"
#include "messenger/MQTTMessage.h"
#include "messenger/MQTTConfiguration.h"

extern "C" {
    #include "paho/MQTTClient.h"
}

/**
 * @namespace Messenger
 * @brief Implementation of MQTT Client.
 */
namespace Messenger {

/**
 * @interface IDeliveryToken
 * @brief Delivery listener.
 */
class IDeliveryToken {
public:
    /**
     * @fn virtual Message GetMessage() const = 0;
     * @brief Getter for message field.
     * @return Message.
     */
    virtual Message GetMessage() const = 0;
};

/**
 * @interface IStatusListener
 * @brief Status listener.
 */
class IStatusListener {
public:
    /**
     * @fn virtual void ConnectionLost(const std::string& _cause) = 0;
     * @brief Function that will be called after connection lost is detected.
     * @param _cause Reason for connection lost.
     */
    virtual void ConnectionLost(const std::string& _cause) = 0;

    /**
     * @fn virtual void MessageArrived(const std::string& _topic, const Message::Ptr _msg) = 0;
     * @brief Function that will be called when message arrived.
     * @param _topic Topic.
     * @param _msg MQTT message.
     */
    virtual void MessageArrived(const std::string& _topic, const Message::Ptr _msg) = 0;

    /**
     * @fn virtual void DeliveryComplete(const IDeliveryToken& _token) = 0;
     * @brief Function that will be called when message was successfully delivered.
     * @param _token Token.
     */
    virtual void DeliveryComplete(const IDeliveryToken& _token) = 0;

    /**
     * @fn virtual void PublishFailed(int _code, const std::string& _topic, const Message::Ptr _msg) = 0;
     * @brief Function that will be called if publish faield.
     * @param _code Error code.
     * @param _topic Topic.
     * @param _msg MQTT message.
     */
    virtual void PublishFailed(int _code, const std::string& _topic, const Message::Ptr _msg) = 0;
};

/**
 * @class ClientImpl
 * @brief Wrapper class for Paho MQTT client library.
 */
class DLLExport ClientImpl {
public:
    /**
     * @typedef Utility::SharedPointer<ClientImpl> Ptr;
     * @brief Shared pointer to ClientImpl object.
     */
    typedef Utility::SharedPointer<ClientImpl> Ptr;

    /**
     * @typedef std::vector<std::string> TopicCollection;
     * @brief Array of topics.
     */
    typedef std::vector<std::string> TopicCollection;

    /**
     * @typedef std::vector<int> QoSCollection;
     * @brief Array of QoSs.
     */
    typedef std::vector<int> QoSCollection;

    /**
     * @fn ClientImpl(const std::string& _serverUri, const std::string& _clientId);
     * @brief Constructor.
     * @param _serverUri URI of a MQTT broker. Start with <tcp://> for TCP connections, or <ssl://> for secure connections.
     * @param _clientId Client identifier.
     */
    ClientImpl(const std::string& _serverUri, const std::string& _clientId);

    /**
     * @fn virtual ~ClientImpl();
     * @brief Destructor.
     */
    virtual ~ClientImpl();

    /**
     * @fn virtual void Connect(ConnectConfig::Ptr _config);
     * @brief Connect Connects to messaging server.
     * @param _config Connection config.
     * @throws Utility::ProtocolException
     */
    virtual void Connect(ConnectConfig::Ptr _config);

    /**
     * @fn bool isSessionPresent() const;
     * @brief Indicates whether session is present.
     * @return Indication.
     */
    bool isSessionPresent() const;

    /**
     * @fn virtual void Disconnect(long _timeout = 0);
     * @brief Disconnects from a messaging server, if it was previusly connected.
     * @param _timeout Timeout to disconnect.
     * @throws Utility::ProtocolException
     */
    virtual void Disconnect(long _timeout = 0);

    /**
     * @fn std::string GetClientId() const;
     * @brief Getter for client id.
     * @return Client id.
     */
    std::string GetClientId() const;

    /**
     * @fn std::string GetServerUri() const;
     * @brief Getter for server uri.
     * @return Server uri.
     */
    std::string GetServerUri() const;

    /**
     * @fn bool IsConnected() const;
     * @brief Checks whether the client is connected.
     * @return Indication.
     */
    bool IsConnected() const;

    /**
     * @fn virtual void Publish(const std::string& _topic, const Message::Ptr _msg);
     * @brief Asynchronous publishing.
     * @param _topic Topic.
     * @param _msg MQTT message.
     */
    virtual void Publish(const std::string& _topic, const Message::Ptr _msg);

    /**
     * @fn virtual void Subscribe(const TopicCollection& _topics, const QoSCollection& _qoss);
     * @brief If client is not connected, it stores subscriptions and will try to complete the procedure on next connect.
     * @param _topics Topics to subscribe.
     * @param _qoss QoS levels.
     * @throws Utility::ProtocolException
     */
    virtual void Subscribe(const TopicCollection& _topics, const QoSCollection& _qoss);

    /**
     * @fn virtual void Subscribe(const std::string& _topic, int _qos);
     * @brief If client is not connected, it stores subscriptions and will try to complete the procedure on next connect.
     * @param _topic Topic to subscribe.
     * @param _qos QoS level.
     * @throws Utility::ProtocolException
     */
    virtual void Subscribe(const std::string& _topic, int _qos);

    /**
     * @fn virtual void Unsubscribe(const TopicCollection& _topics);
     * @brief Does basic unsubscribe and removes stored subscriptions.
     * @param _topics Topics.
     * @throws Utility::ProtocolException
     */
    virtual void Unsubscribe(const TopicCollection& _topics);

    /**
     * @fn virtual void Unsubscribe(const std::string& _topic);
     * @brief Does basic unsubscribe and removes stored subscription.
     * @param _topic Topic.
     * @throws Utility::ProtocolException
     */
    virtual void Unsubscribe(const std::string& _topic);

    /**
     * @fn void RegisterListener(IStatusListener* _listener);
     * @brief Registers status listener.
     * @param _listener Listener.
     * @throws Utility::ProtocolException
     */
    void RegisterListener(IStatusListener* _listener);
private:
    static void OnConnectionLost(void* _context, char* _cause);
    static int OnMessageArrived(void* _context, char* _topic, int _topicLen, MQTTClient_message* _message);
    static void OnDeliveryComplete(void* _context, MQTTClient_token _token);

    IStatusListener* GetStatusListener();

    Utility::Mutex m_Mutex;
    MQTTClient m_Client;
    std::string m_ServerUri;
    std::string m_ClientId;
    IStatusListener* m_StatusListener;
    bool m_SessionPresent;
};

}

#endif // MQTT_CLIENT_H
