/****************************************************************************
 *  Copyright (c) 2018 Oblo Living d.o.o
 *
 *  All Rights Reserved
 *
 *  P R O P R I E T A R Y    &    C O N F I D E N T I A L
 *
 *  -----------------------------------------------------
 *  http://www.obloliving.com/
 *  -----------------------------------------------------
 *
 * \project Oblo WISE SDK - SDK library
 * \file MQTTConfiguration.h
 * \brief
 * PAHO configuration abstraction classes.
 *
 * @Author Igor Stefanovic
 *
 * \notes
 *
 * \history
 *****************************************************************************/
#ifndef MQTT_CONFIGURATION_H
#define MQTT_CONFIGURATION_H

#include "util/Common.h"
#include "messenger/MQTTMessage.h"

extern "C" {
    #include "paho/MQTTClient.h"
}

/**
 * @namespace Messenger
 * @brief Implementation of MQTT Client.
 */
namespace Messenger {

/**
 * @class StatusConfig
 * @brief The StatusConfig class
 */
class DLLExport StatusConfig {
public:
    /**
     * @typedef Utility::SharedPointer<StatusConfig> Ptr;
     * @brief Shared pointer to StatusConfig object.
     */
    typedef Utility::SharedPointer<StatusConfig> Ptr;

    /**
     * @fn StatusConfig();
     * @brief Constructor.
     */
    StatusConfig() {}

    /**
     * @fn virtual ~StatusConfig();
     * @brief Destructor.
     */
    virtual ~StatusConfig() {}

    /**
     * @fn std::string GetTopic() const;
     * @brief Getter for topic field.
     * @return Topic.
     */
    std::string GetTopic() const;

    /**
     * @fn void SetTopic(const std::string& _topic);
     * @brief Setter for topic field.
     * @param _topic Topic.
     */
    void SetTopic(const std::string& _topic);

    /**
     * @fn Message::Ptr GetMessage() const;
     * @brief Getter for message field.
     * @return MQTT message.
     */
    Message::Ptr GetMessage() const;

    /**
     * @fn void SetMessage(const Message::Ptr _msg);
     * @brief Setter for message field.
     * @param _msg MQTT message.
     */
    void SetMessage(const Message::Ptr _msg);

    /**
     * @fn MQTTClient_willOptions* PackConfig() const;
     * @brief Packs configuration to Paho C library option format.
     * @return Config.
     */
    MQTTClient_willOptions* PackConfig() const;

    /**
     * @fn static void FreeConfig(MQTTClient_willOptions* _config);
     * @brief Dealocates memory that is allocated for this configuration.
     * @param _config Config.
     */
    static void FreeConfig(MQTTClient_willOptions* _config);
private:
    StatusConfig(const StatusConfig&);
    StatusConfig& operator=(const StatusConfig&);

    std::string m_Topic;
    Message::Ptr m_Message;
};

/**
 * @class SSLConfig
 * @brief MQTT SSL/TLS configuration.
 */
class DLLExport SSLConfig {
public:
    /**
     * @typedef Utility::SharedPointer<SSLConfig> Ptr;
     * @brief Shared pointer to SSLConfig object.
     */
    typedef Utility::SharedPointer<SSLConfig> Ptr;

    /**
     * @fn SSLConfig();
     * @brief Constructor.
     */
    SSLConfig();

    /**
     * @fn std::string GetTrustStore() const;
     * @brief Getter for trust store field.
     * @return Trust store.
     */
    std::string GetTrustStore() const;

    /**
     * @fn void SetTrustStore(const std::string& _trustStore);
     * @brief Setter for trust store field.
     * @param _trustStore Trust store.
     */
    void SetTrustStore(const std::string& _trustStore);

    /**
     * @fn std::string GetKeyStore() const;
     * @brief Getter for key store field.
     * @return Key Store.
     */
    std::string GetKeyStore() const;

    /**
     * @fn void SetKeyStore(const std::string& _keyStore);
     * @brief Setter for key store field.
     * @param _keyStore Key store.
     */
    void SetKeyStore(const std::string& _keyStore);

    /**
     * @fn std::string GetPrivateKey() const;
     * @brief Getter for private key field.
     * @return Private key.
     */
    std::string GetPrivateKey() const;

    /**
     * @fn void SetPrivateKey(const std::string& _privateKey);
     * @brief Setter for private key field.
     * @param _privateKey Private key.
     */
    void SetPrivateKey(const std::string& _privateKey);

    /**
     * @fn std::string GetPrivateKeyPassword() const;
     * @brief Getter for private key password field.
     * @return Private key password.
     */
    std::string GetPrivateKeyPassword() const;

    /**
     * @fn void SetPrivateKeyPassword(const std::string& _privateKeyPassword);
     * @brief Setter for private key password field.
     * @param _privateKeyPassword Private key password.
     */
    void SetPrivateKeyPassword(const std::string& _privateKeyPassword);

    /**
     * @fn bool GetVerifyServer() const;
     * @brief Getter for verify server field.
     * @return Verify server indication.
     */
    bool GetVerifyServer() const;

    /**
     * @fn void SetVerifyServer(bool _verifyServer);
     * @brief Setter for verify server field.
     * @param _verifyServer Verify server indication.
     */
    void SetVerifyServer(bool _verifyServer);

    /**
     * @fn MQTTClient_SSLOptions* PackConfig() const;
     * @brief Packs configuration to Paho C library option format.
     * @return Config
     */
    MQTTClient_SSLOptions* PackConfig() const;

    /**
     * @fn static void FreeConfig(MQTTClient_SSLOptions* _config);
     * @brief Dealocates memory that is allocated for this configuration.
     * @param _config Config.
     */
    static void FreeConfig(MQTTClient_SSLOptions* _config);
private:
    SSLConfig(const SSLConfig&);
    SSLConfig& operator=(const SSLConfig&);

    std::string m_TrustStore;
    std::string m_KeyStore;
    std::string m_PrivateKey;
    std::string m_PrivateKeyPassword;
    bool m_VerifyServer;
};

/**
 * @class ConnectConfig
 * @brief MQTT connection configuration.
 */
class DLLExport ConnectConfig {
public:
    /**
     * @typedef Utility::SharedPointer<ConnectConfig> Ptr;
     * @brief Shared pointer to ConnectConfig object.
     */
    typedef Utility::SharedPointer<ConnectConfig> Ptr;

    /**
     * @fn ConnectConfig(const std::string& _username = "", const std::string& _password = "",
                  bool _cleanSession = true, int _keepAliveInterval = 60,
                  int _timeout = 5, int _retryinterval = 20,
                  bool _reliable = false);
     * @brief Constructor.
     * @param _username MQTT username.
     * @param _password MQTT password.
     * @param _cleanSession MQTT clean session flag.
     * @param _keepAliveInterval MQTT keepalive interval.
     * @param _timeout MQTT connection timeout.
     * @param _retryinterval MQTT operation retry interval.
     * @param _reliable Paho reliable flag.
     */
    ConnectConfig(const std::string& _username = "", const std::string& _password = "",
                  bool _cleanSession = true, int _keepAliveInterval = 60,
                  int _timeout = 5, int _retryinterval = 20,
                  bool _reliable = false);
    /**
     * @fn virtual ~ConnectConfig();
     * @brief Destructor.
     */
    virtual ~ConnectConfig() {}

    /**
     * @fn int GetKeepAliveInterval() const;
     * @brief Getter for keep alive interval field.
     * @return Keep alive interval.
     */
    int GetKeepAliveInterval() const;

    /**
     * @fn void SetKeepAliveInterval(int _keepAliveInterval);
     * @brief Setter for keep alive interval field.
     * @param _keepAliveInterval Keep alive interval.
     */
    void SetKeepAliveInterval(int _keepAliveInterval);

    /**
     * @fn bool GetCleanSession() const;
     * @brief Getter for clean session field.
     * @return Clean session.
     */
    bool GetCleanSession() const;

    /**
     * @fn void SetCleanSession(bool _cleanSession);
     * @brief Setter for clean session field.
     * @param _cleanSession Clean session.
     */
    void SetCleanSession(bool _cleanSession);

    /**
     * @fn bool IsReliable() const;
     * @brief Getter for reliable field.
     * @return Reliable.
     */
    bool IsReliable() const;

    /**
     * @fn void SetReliable(bool _reliable);
     * @brief Setter for reliable field.
     * @param _reliable Reliable.
     */
    void SetReliable(bool _reliable);

    /**
     * @fn StatusConfig::Ptr GetStatus() const;
     * @brief Getter for status field.
     * @return Status.
     */
    StatusConfig::Ptr GetStatus() const;

    /**
     * @fn void SetStatus(const StatusConfig::Ptr& _status);
     * @brief Setter for status field.
     * @param _status Status.
     */
    void SetStatus(const StatusConfig::Ptr& _status);

    /**
     * @fn std::string GetUsername() const;
     * @brief Getter for username field.
     * @return Username.
     */
    std::string GetUsername() const;

    /**
     * @fn void SetUsername(const std::string& _username);
     * @brief Setter for username field.
     * @param _username Username.
     */
    void SetUsername(const std::string& _username);

    /**
     * @fn std::string GetPassword() const;
     * @brief Getter for password field.
     * @return Password.
     */
    std::string GetPassword() const;

    /**
     * @fn void SetPassword(const std::string& _password);
     * @brief Setter for password field.
     * @param _password Password.
     */
    void SetPassword(const std::string& _password);

    /**
     * @fn int GetTimeout() const;
     * @brief Getter for timeout field.
     * @return Timeout.
     */
    int GetTimeout() const;

    /**
     * @fn void SetTimeout(int _timeout);
     * @brief Setter for timeout field.
     * @param _timeout Timeout.
     */
    void SetTimeout(int _timeout);

    /**
     * @fn int GetRetryInterval() const;
     * @brief Getter for retry interval field.
     * @return Retry interval.
     */
    int GetRetryInterval() const;

    /**
     * @fn void SetRetryInterval(int _retryInterval);
     * @brief Setter for retry interval field.
     * @param _retryInterval Retry interval.
     */
    void SetRetryInterval(int _retryInterval);

    /**
     * @fn SSLConfig::Ptr GetSsl() const;
     * @brief Getter for SSL config field.
     * @return SSL config.
     */
    SSLConfig::Ptr GetSsl() const;

    /**
     * @fn void SetSsl(const SSLConfig::Ptr& _ssl);
     * @brief Setter for SSL config field.
     * @param _ssl SSL config.
     */
    void SetSsl(const SSLConfig::Ptr& _ssl);

    /**
     * @fn MQTTClient_connectOptions* PackConfig() const;
     * @brief Packs configuration to Paho C library option format.
     * @return Config.
     */
    MQTTClient_connectOptions* PackConfig() const;

    /**
     * @fn static void FreeConfig(MQTTClient_connectOptions* _config);
     * @brief Dealocates memory that is allocated for this configuration.
     * @param _config Config.
     */
    static void FreeConfig(MQTTClient_connectOptions* _config);
private:
    ConnectConfig(const ConnectConfig&);
    ConnectConfig& operator=(const ConnectConfig&);

    int m_KeepAliveInterval;
    bool m_CleanSession;
    bool m_Reliable;
    StatusConfig::Ptr m_Status;
    std::string m_Username;
    std::string m_Password;
    int m_Timeout;
    int m_RetryInterval;
    SSLConfig::Ptr m_SSL;
};

}

#endif // MQTT_CONFIGURATION_H
