/****************************************************************************
 *  Copyright (c) 2018 Oblo Living d.o.o
 *
 *  All Rights Reserved
 *
 *  P R O P R I E T A R Y    &    C O N F I D E N T I A L
 *
 *  -----------------------------------------------------
 *  http://www.obloliving.com/
 *  -----------------------------------------------------
 *
 * \project Oblo WISE SDK - SDK library
 * \file MQTTMessage.h
 * \brief
 * PAHO Message abstraction classes.
 *
 * @Author Igor Stefanovic
 *
 * \notes
 *
 * \history
 *****************************************************************************/
#ifndef MQTT_MESSAGE_H
#define MQTT_MESSAGE_H

#include "util/Common.h"
#include "util/SharedPointer.h"

extern "C" {
    #include "paho/MQTTClient.h"
}

/**
 * @namespace Messenger
 * @brief Implementation of MQTT Client.
 */
namespace Messenger {

/**
 * @class Message
 * @brief MQTT message wrapper class.
 */
class DLLExport Message {
public:
    /**
     * @typedef Utility::SharedPointer<Message> Ptr;
     * @brief Shared pointer to Message object.
     */
    typedef Utility::SharedPointer<Message> Ptr;

    /**
     * @enum QosLevel
     * @brief QoS level enumeration.
     */
    enum QosLevel {
        QOS0 = 0,
        QOS1,
        QOS2
    };

    /**
     * @fn Message();
     * @brief Constructor.
     */
    Message();

    /**
     * @fn Message(const std::string& _payload);
     * @brief Constructor.
     * @param _payload Payload of the MQTT message.
     */
    Message(const std::string& _payload);

    /**
     * @fn Message(const Message& _msg);
     * @brief Constructor.
     * @param _msg Copy object.
     */
    Message(const Message& _msg);

    /**
     * @fn virtual ~Message();
     * @brief Destructor.
     */
    virtual ~Message();

    /**
     * @fn std::string GetPayload() const;
     * @brief Getter for payload field.
     * @return Payload.
     */
    std::string GetPayload() const;

    /**
     * @fn void SetPayload(const std::string& _payload);
     * @brief Setter for payload field.
     * @param _payload Payload.
     */
    void SetPayload(const std::string& _payload);

    /**
     * @fn QosLevel GetQoS() const;
     * @brief Getter for QoS level field.
     * @return QoS level.
     */
    QosLevel GetQoS() const;

    /**
     * @fn void SetQoS(QosLevel _qos);
     * @brief Setter for QoS level field.
     * @param _qos QoS level.
     */
    void SetQoS(QosLevel _qos);

    /**
     * @fn bool IsRetained() const;
     * @brief Getter for retained field.
     * @return Retained.
     */
    bool IsRetained() const;

    /**
     * @fn void SetRetained(bool _ret);
     * @brief Setter for retained field.
     * @param _ret Retained.
     */
    void SetRetained(bool _ret);

    /**
     * @fn bool IsDuplicated() const;
     * @brief Getter for duplicated field.
     * @return Duplicated.
     */
    bool IsDuplicated() const;

    /**
     * @fn void SetDuplicated(bool _dup);
     * @brief Getter for duplicated field.
     * @param _dup Duplicated.
     */
    void SetDuplicated(bool _dup);

    /**
     * @fn void WriteMessage(MQTTClient_message** _msg) const;
     * @brief Writes a message to a Paho MQTT message.
     * @param _msg Paho MQTT message.
     */
    void WriteMessage(MQTTClient_message** _msg) const;

    /**
     * @fn static void FreeMessage(MQTTClient_message** _msg);
     * @brief Dealocates memory that is allocated for a MQTT message.
     * @param _msg Paho MQTT message.
     */
    static void FreeMessage(MQTTClient_message** _msg);

    /**
     * @fn static Message::Ptr ParseMessage(const MQTTClient_message* _msg);
     * @brief  Parses Paho message to corresponding format.
     * @param _msg Message to parse.
     * @return Formated message.
     */
    static Message::Ptr ParseMessage(const MQTTClient_message* _msg);
private:
    std::string m_Payload;
    QosLevel m_QoS;
    bool m_Retained;
    bool m_Duplicated;
};
}


#endif // MQTT_MESSAGE_H
