/****************************************************************************
 *  Copyright (c) 2018 Oblo Living d.o.o
 *
 *  All Rights Reserved
 *
 *  P R O P R I E T A R Y    &    C O N F I D E N T I A L
 *
 *  -----------------------------------------------------
 *  http://www.obloliving.com/
 *  -----------------------------------------------------
 *
 * \project Oblo WISE SDK - SDK library
 * \file Messages.h
 * \brief
 * MQTT Message abstraction classes.
 *
 * @Author Igor Stefanovic
 *
 * \notes
 *
 * \history
 *****************************************************************************/
#ifndef MESSAGES_H
#define MESSAGES_H

#include "util/Common.h"
#include "util/SynchronizedObject.h"
#include "util/SharedPointer.h"
#include "util/JsonStruct.h"
#include "util/Exceptions.h"
#include <iostream>
#include <cstdlib>

/**
 * @namespace Messenger
 * @brief Implementation of MQTT Client.
 */
namespace Messenger {

class ResponseMessage;
class RequestMessage;

/**
 * @class RequestDelegate
 * @brief Delegate used to report new response messages.
 */
class DLLExport RequestDelegate {
public:
    /**
     * @typedef Utility::SharedPointer<RequestDelegate> Ptr;
     * @brief Shared pointer to RequestDelegate object.
     */
    typedef Utility::SharedPointer<RequestDelegate> Ptr;
    /**
     * @fn RequestDelegate();
     * @brief Constructor.
     */
    RequestDelegate();

    template<class C, void (C::*CSuccessMethod)(unsigned int _requestId, const ResponseMessage& _response), void (C::*CFailedMethod)(unsigned int _requestId)>
    /**
     * @fn static RequestDelegate::Ptr CreateDelegate(C* _object, bool _autoDelete = false);
     * @brief Create requst delegate.
     * @param _object Object pointer.
     * @param _autoDelete Delete _pObject after a delegation is completed.
     * @return Shared pointer to a request delegate.
     */
    static RequestDelegate::Ptr CreateDelegate(C* _object, bool _autoDelete = false) {
        RequestDelegate::Ptr d(new RequestDelegate());
        d->m_Object = _object;
        d->m_StubSucc = &MethodStubSucc<C, CSuccessMethod>;
        d->m_StubFail = &MethodStubFail<C, CFailedMethod>;
        d->m_AutoDelete = _autoDelete;
        return d;
    }

    /**
     * @fn void ReportSuccess(unsigned int _requestId, const ResponseMessage& _response) const;
     * @brief Report success.
     * @param _requestId Request ID.
     * @param _response A response reference.
     */
    void ReportSuccess(unsigned int _requestId, const ResponseMessage& _response) const;

    /**
     * @fn void ReportFailure(unsigned int _requestId) const;
     * @brief Report failure.
     * @param _requestId Request ID.
     */
    void ReportFailure(unsigned int _requestId) const;
private:
    typedef void (*StubSuccType)(void* _object, unsigned int _requestId, const ResponseMessage& _response, bool _autoDelete);
    typedef void (*StubFailType)(void* _object, unsigned int _requestId, bool _autoDelete);

    void* m_Object;
    StubSuccType m_StubSucc;
    StubFailType m_StubFail;
    bool m_AutoDelete;

    template <class C, void(C::*CMethodSucc)(unsigned int _requestId, const ResponseMessage& _response)>
    static void MethodStubSucc(void* _object, unsigned int _requestId, const ResponseMessage& _response, bool _autoDelete) {
        C* p = static_cast<C*>(_object);

        (p->*CMethodSucc)(_requestId, _response);

        if (_autoDelete) {
            SAFE_DELETE_O(p);
        }
    }

    template <class C, void(C::*CMethodFail)(unsigned int _requestId)>
    static void MethodStubFail(void* _object, unsigned int _requestId, bool _autoDelete) {
        C* p = static_cast<C*>(_object);

        (p->*CMethodFail)(_requestId);

        if (_autoDelete) {
            SAFE_DELETE_O(p);
        }
    }
};

/**
 * @class MessageImpl
 * @brief Implementation of basic functionalities for messages.
 */
class DLLExport MessageImpl {
public:
    /**
     * @typedef Utility::SharedPointer<MessageImpl> Ptr;
     * @brief Shared pointer to MessageImpl object.
     */
    typedef Utility::SharedPointer<MessageImpl> Ptr;

    /**
     * @fn MessageImpl();
     * @brief Constructor.
     */
    MessageImpl();

    /**
     * @fn MessageImpl(const MessageImpl& _obj);
     * @brief Constructor.
     * @param _obj Copy object.
     */
    MessageImpl(const MessageImpl& _obj);

    /**
     * @fn virtual ~MessageImpl();
     * @brief Destructor.
     */
    virtual ~MessageImpl() {}

    /**
     * @fn unsigned int GetId() const;
     * @brief Getter for ID of the message.
     * @return ID.
     */
    unsigned int GetId() const;

    /**
     * @fn const Utility::JsonStruct::Ptr GetParams() const;
     * @brief Getter for parameters of the message.
     * @return Shared pointer to parameters.
     */
    const Utility::JsonStruct::Ptr GetParams() const;

    /**
     * @fn unsigned long long GetTimestamp() const;
     * @brief Getter for timestamp when message is created.
     * @return Timestamp.
     */
    unsigned long long GetTimestamp() const;

    /**
     * @fn void SetId(unsigned int _id);
     * @brief Setter for ID of the message.
     * @param _id ID to be set.
     */
    void SetId(unsigned int _id);

    /**
     * @fn void SetParams(Utility::JsonStruct::Ptr _params);
     * @brief Setter for parameters of the message.
     * @param _params Shared pointer to parameters to be set.
     */
    void SetParams(Utility::JsonStruct::Ptr _params);

    /**
     * @fn void SetTimestamp();
     * @brief Setter for timestamp when message is created. Called in constructor.
     */
    void SetTimestamp();

    /**
     * @fn void SetTimestamp(unsigned long long _timestamp);
     * @brief Setter for timestamp.
     * @param _timestamp Timestamp to be set.
     */
    void SetTimestamp(unsigned long long _timestamp);

    /**
     * @fn virtual std::string ToString() const;
     * @brief Convert message to string.
     * @return String message.
     */
    virtual std::string ToString() const;
protected:
    /**
     * @fn static void Parse(MessageImpl::Ptr _messageToFill, Utility::JsonStruct::Ptr _messageToParse);
     * @brief Parse message to fill the fields of MessageImpl.
     * @param _messageToFill Message to be filled.
     * @param _messageToParse Message to be parsed.
     */
    static void Parse(MessageImpl::Ptr _messageToFill, Utility::JsonStruct::Ptr _messageToParse);
    friend class CMessenger;

    /**
     * @fn virtual void InflateJson(Utility::JsonStruct& _obj) const = 0;
     * @brief Abstract method used to inflate fields from a subclass into a JSON object.
     * @param _obj Object.
     */
    virtual void InflateJson(Utility::JsonStruct& _obj) const = 0;

    /**
     * @brief Parameter field. Value is "params".
     */
    static const std::string PARAMS;

    /**
     * @brief Id field. Value is "uid".
     */
    static const std::string ID;

    /**
     * @brief Timestamp field. Value is "ts".
     */
    static const std::string TS;
private:
    Utility::JsonStruct::Ptr m_Params;
    unsigned int m_Id;
    unsigned long long m_Timestamp;
};

/**
 * @class EventMessage
 * @brief Wrapper for messages about events arriving from objects included in communication.
 * @extends MessageImpl
 */
class DLLExport EventMessage : public MessageImpl {
public:
    /**
     * @typedef Utility::SharedPointer<EventMessage> Ptr;
     * @brief Shared pointer to EventMessage object.
     */
    typedef Utility::SharedPointer<EventMessage> Ptr;

    /**
     * @fn EventMessage();
     * @brief Constructor.
     */
    EventMessage();

    /**
     * @fn EventMessage(const EventMessage& _event);
     * @brief Constructor.
     * @param _event Copy object.
     */
    EventMessage(const EventMessage& _event);

    /**
     * @fn std::string GetName() const;
     * @brief Getter for name of the event message.
     * @return Name.
     */
    std::string GetName() const;

    /**
     * @fn unsigned long long GetExpiration() const;
     * @brief Getter for expiration timestamp of the event message.
     * @return Expiration timestamp.
     */
    unsigned long long GetExpiration() const;

    /**
     * @fn void SetName(const std::string& _name);
     * @brief Setter for name of the event message.
     * @param _name Name to be set.
     */
    void SetName(const std::string& _name);

    /**
     * @fn void SetExpiration(unsigned long long _expiration);
     * @brief Setter for expiration of the event message.
     * @param _expiration Expiration timestamp to be set.
     */
    void SetExpiration(unsigned long long _expiration);

    /**
     * @fn static EventMessage::Ptr Parse(Utility::JsonStruct::Ptr _message);
     * @brief Parse message to fill the fields of EventMessage.
     * @param _message Message to be parsed.
     * @return Message to be filled.
     */
    static EventMessage::Ptr Parse(Utility::JsonStruct::Ptr _message);
protected:
    /**
     * @fn virtual void InflateJson(Utility::JsonStruct& _obj) const;
     * @brief Method to inflate class members to JSON object.
     * @param _obj Object.
     */
    virtual void InflateJson(Utility::JsonStruct& _obj) const;
    /**
     * @brief Name field. Value is "name".
     */
    static const std::string NAME;

    /**
     * @brief Expiration timeout field. Value is "ets".
     */
    static const std::string ETS;
private:
    std::string m_Name;
    unsigned long long m_Expiration;
};

/**
 * @class ResponseMessage
 * @brief The ResponseMessage class
 * @extends MessageImpl
 */
class DLLExport ResponseMessage : public MessageImpl {
public:
    /**
     * @typedef Utility::SharedPointer<ResponseMessage> Ptr;
     * @brief Shared pointer to ResponseMessage object.
     */
    typedef Utility::SharedPointer<ResponseMessage> Ptr;

    /**
     * @fn ResponseMessage();
     * @brief Constructor.
     */
    ResponseMessage();

    /**
     * @fn ResponseMessage(const ResponseMessage& _response);
     * @brief Constructor.
     * @param _response Copy object.
     */
    ResponseMessage(const ResponseMessage& _response);

    /**
     * @fn int GetCode() const;
     * @brief Getter for code of the response message.
     * @return Code.
     */
    int GetCode() const;

    /**
     * @fn std::string GetMessage() const;
     * @brief Getter for message of the response message.
     * @return Message.
     */
    std::string GetMessage() const;

    /**
     * @fn void SetCode(int _code);
     * @brief Setter for code of the response message.
     * @param _code Code to be set.
     */
    void SetCode(int _code);

    /**
     * @fn void SetMessage(const std::string& _message);
     * @brief Setter for message of the response message.
     * @param _message Message to be set.
     */
    void SetMessage(const std::string& _message);

    /**
     * @fn static ResponseMessage::Ptr Parse(Utility::JsonStruct::Ptr _message);
     * @brief Parse message to fill the fields of ResponseMessage.
     * @param _message Message to be parsed.
     * @return Message to be filled.
     */
    static ResponseMessage::Ptr Parse(Utility::JsonStruct::Ptr _message);
protected:
    friend class CMessenger;
    friend class User;
    /**
     * @fn virtual void InflateJson(Utility::JsonStruct& _obj) const;
     * @brief Method to inflate class members to JSON object.
     * @param _obj Object.
     */
    virtual void InflateJson(Utility::JsonStruct& _obj) const;

    /**
     * @brief Code field. Value is "code".
     */
    static const std::string CODE;

    /**
     * @brief Message field. Value is "message".
     */
    static const std::string MESSAGE;
private:
    int m_Code;
    std::string m_Message;
};

/**
 * @interface ICompletionListener
 * @brief The ICompletionListener class.
 */
class ICompletionListener {
public:
    /**
     * @fn virtual void OnRequestCompleted(unsigned int _requestId) = 0;
     * @brief Callback to be called on request completion.
     * @param _requestId ID of request message.
     */
    virtual void OnRequestCompleted(unsigned int _requestId) = 0;
};

/**
 * @class RequestMessage
 * @brief Wrapper for request messages.
 * @extends MessageImpl
 */
class DLLExport RequestMessage : public MessageImpl {
public:
    /**
     * @typedef Utility::SharedPointer<RequestMessage> Ptr;
     * @brief Shared pointer to RequestMessage object.
     */
    typedef Utility::SharedPointer<RequestMessage> Ptr;

    /**
     * @fn RequestMessage();
     * @brief Constructor.
     */
    RequestMessage();

    /**
     * @fn RequestMessage(const RequestMessage& _obj);
     * @brief Constructor.
     * @param _obj Copy object.
     */
    RequestMessage(const RequestMessage& _obj);

    /**
     * @fn std::string GetSender() const;
     * @brief Getter for sender of the request message.
     * @return Sender.
     */
    std::string GetSender() const;

    /**
     * @fn std::string GetName() const;
     * @brief Getter for name of the request message.
     * @return Name.
     */
    std::string GetName() const;

    /**
     * @fn unsigned long long GetExpiration() const;
     * @brief Getter for expiration timeout of the request message.
     * @return Expiration timeout.
     */
    unsigned long long GetExpiration() const;

    /**
     * @fn void SetSender(const std::string& _sender);
     * @brief Setter for sender of the request message.
     * @param _sender Sender to be set.
     */
    void SetSender(const std::string& _sender);

    /**
     * @fn void SetName(const std::string& _name);
     * @brief Setter for the name of the request message.
     * @param _name Name to be set.
     */
    void SetName(const std::string& _name);

    /**
     * @fn void SetExpiration(unsigned long long _expiration);
     * @brief Setter for expiration timeout of the request message.
     * @param _expiration Expiration timeout to be set.
     */
    void SetExpiration(unsigned long long _expiration);

    /**
     * @fn void RegisterCompletionListener(ICompletionListener* _completionListener);
     * @brief Registers listener on information about request status.
     * @param _completionListener Listener.
     */
    void RegisterCompletionListener(ICompletionListener* _completionListener);

    /**
     * @fn virtual void ReportResponse(const ResponseMessage::Ptr _response);
     * @brief Report the response of the request.
     * @param _response Response that arrived.
     */
    virtual void ReportResponse(const ResponseMessage::Ptr _response);

    /**
     * @fn virtual void ReportFailure();
     * @brief Report failure.
     */
    virtual void ReportFailure();

    /**
     * @fn static RequestMessage::Ptr Parse(Utility::JsonStruct::Ptr _message);
     * @brief Parse message to fill the fields of RequestMessage.
     * @param _message Message to be parsed.
     * @return Message to be filled.
     */
    static RequestMessage::Ptr Parse(Utility::JsonStruct::Ptr _message);
protected:
    friend class CMessenger;
    friend class User;

    /**
     * @fn virtual void InflateJson(Utility::JsonStruct& _obj) const;
     * @brief Method to inflate class members to JSON object.
     * @param _obj Object.
     */
    virtual void InflateJson(Utility::JsonStruct& _obj) const;

    /**
     * @brief Sender field. Value is "sender".
     */
    static const std::string SENDER;

    /**
     * @brief Name field. Value is "name".
     */
    static const std::string NAME;

    /**
     * @brief Expiration timeout field. Value is "ets".
     */
    static const std::string ETS;
private:
    std::string m_Sender;
    std::string m_Name;
    unsigned long long m_Expiration;
    ICompletionListener* m_CompletionListener;
};

/**
 * @class SyncRequest
 * @brief Class used to handle synchronized requests inside a messenger.
 * @extends RequestMessage
 */
class SyncRequest : public RequestMessage {
public:
    /**
     * @typedef Utility::SharedPointer<SyncRequest> Ptr;
     * @brief Shared pointer to SyncRequest object.
     */
    typedef Utility::SharedPointer<SyncRequest> Ptr;

    /**
     * @fn SyncRequest(const RequestMessage& _request);
     * @brief Constructor.
     * @param _request RequestMessage object.
     */
    SyncRequest(const RequestMessage& _request);

    /**
     * @fn const ResponseMessage::Ptr GetResponse() const;
     * @brief Getter for response of the sync request message.
     * @return Shared pointer to response message.
     */
    const ResponseMessage::Ptr GetResponse() const;
protected:
    friend class User;

    /**
     * @fn void Wait(unsigned int _timeout);
     * @brief Waits for response.
     * @param _timeout Time to wait.
     */
    void Wait(unsigned int _timeout);

    /**
     * @fn virtual void ReportResponse(const ResponseMessage::Ptr _response);
     * @brief Report the response of the request.
     * @param _response Response that arrived.
     */
    virtual void ReportResponse(const ResponseMessage::Ptr _response);

    /**
     * @fn virtual void ReportFailure();
     * @brief Report failure.
     */
    virtual void ReportFailure();
private:
    SyncRequest();
    SyncRequest(const SyncRequest &_request);

    ResponseMessage::Ptr m_Response;
    Utility::SynchronizedObject m_SyncObject;
};

/**
 * @class AsyncRequest
 * @brief Class used to handle asynchronous requests inside a messenger.
 * @extends RequestMessage
 */
class AsyncRequest : public RequestMessage {
public:
    /**
     * @typedef Utility::SharedPointer<AsyncRequest> Ptr;
     * @brief Shared pointer to AsyncRequest object.
     */
    typedef Utility::SharedPointer<AsyncRequest> Ptr;

    /**
     * @fn AsyncRequest(const RequestMessage& _request, RequestDelegate::Ptr _delegate);
     * @brief Constructor.
     * @param _request Request message.
     * @param _delegate Delegate to inform when a response arrives.
     */
    AsyncRequest(const RequestMessage& _request, RequestDelegate::Ptr _delegate);
protected:
    /**
     * @fn virtual void ReportResponse(const ResponseMessage::Ptr _response);
     * @brief Report the response of the request.
     * @param _response Response that arrived.
     */
    virtual void ReportResponse(const ResponseMessage::Ptr _response);

    /**
     * @fn virtual void ReportFailure();
     * @brief Report failure.
     */
    virtual void ReportFailure();
private:
    const RequestDelegate::Ptr m_RequestDelegate;
};

/**
 * @class StatusMessage
 * @brief Wrapper for messages about status arriving from objects included in communication.
 * @extends MessageImpl
 */
class DLLExport StatusMessage : public MessageImpl {
public:
    /**
     * @typedef Utility::SharedPointer<StatusMessage> Ptr;
     * @brief Shared pointer to StatusMessage object.
     */
    typedef Utility::SharedPointer<StatusMessage> Ptr;

    /**
     * @fn StatusMessage();
     * @brief Constructor.
     */
    StatusMessage();

    /**
     * @fn StatusMessage(const StatusMessage& _status);
     * @brief Constructor.
     * @param _status Copy object.
     */
    StatusMessage(const StatusMessage& _status);

    /**
     * @fn bool GetStatus() const;
     * @brief Getter for status of the status message.
     * @return Status.
     */
    bool GetStatus() const;

    /**
     * @fn void SetStatus(bool _online);
     * @brief Setter for status of the status message.
     * @param _online Status to be set.
     */
    void SetStatus(bool _online);

    /**
     * @fn static std::string GenerateWillPayload();
     * @brief Generates the payload of status message.
     * @return Payload as string.
     */
    static std::string GenerateWillPayload();

    /**
     * @fn static StatusMessage::Ptr Parse(Utility::JsonStruct::Ptr _message);
     * @brief Parse message to fill the fields of StatusMessage.
     * @param _message Message to be parsed.
     * @return Message to be filled.
     */
    static StatusMessage::Ptr Parse(Utility::JsonStruct::Ptr _message);
protected:
    /**
     * @fn virtual void InflateJson(Utility::JsonStruct& _obj) const;
     * @brief Method to inflate class members to JSON object.
     * @param _obj Object.
     */
    virtual void InflateJson(Utility::JsonStruct& _obj) const;

    /**
     * @brief Online field. Value is "online".
     */
    static const std::string ONLINE;
private:
    bool m_Online;
};

}
#endif // MESSAGES_H
