/****************************************************************************
 *  Copyright (c) 2018 Oblo Living d.o.o
 *
 *  All Rights Reserved
 *
 *  P R O P R I E T A R Y    &    C O N F I D E N T I A L
 *
 *  -----------------------------------------------------
 *  http://www.obloliving.com/
 *  -----------------------------------------------------
 *
 * \project Oblo WISE SDK - SDK library
 * \file Client.h
 * \brief
 * MQTT Messenger class.
 *
 * @Author Igor Stefanovic
 *
 * \notes
 *
 * \history
 *****************************************************************************/
#ifndef MESSENGER_H
#define MESSENGER_H

#include "util/Common.h"
#include "util/SharedPointer.h"
#include "util/Timer.h"
#include "util/Logger.h"
#include "messenger/Messages.h"
#include "messenger/Client.h"
#include "messenger/User.h"
#include <map>
#include <algorithm>

/**
 * @namespace Messenger
 * @brief Implementation of MQTT Client.
 */
namespace Messenger {

/**
 * @interface IResponder
 * @brief Used to get responses to request messages.
 */
class IResponder {
public:
    /**
     * @fn virtual const ResponseMessage::Ptr Respond(const RequestMessage::Ptr _request) = 0;
     * @brief Function which will be called when new requests arrive.
     * @param _request Request message.
     * @return Response to request message.
     */
    virtual const ResponseMessage::Ptr Respond(const RequestMessage::Ptr _request) = 0;

    /**
     * @fn virtual const ResponseMessage::Ptr ServiceRespond(unsigned int _groupId, const std::string& _serviceType, const RequestMessage::Ptr _request) = 0;
     * @brief Function which will be called when new requests arrive to a service.
     * @param _groupId Group ID.
     * @param _serviceType Service type.
     * @param _request Request message.
     * @return Response to request message.
     */
    virtual const ResponseMessage::Ptr ServiceRespond(unsigned int _groupId, const std::string& _serviceType, const RequestMessage::Ptr _request) = 0;
};

/**
 * @class Cleaner
 * @brief Request cleaner implementation.
 */
class Cleaner {
protected:
    /**
     * @class CleanupElement
     * @brief Wrapper for a cleanup element.
     */
    class CleanupElement {
    public:
        /**
         * @typedef Utility::SharedPointer<CleanupElement> Ptr;
         * @brief Shared pointer to CleanupElement object.
         */
        typedef Utility::SharedPointer<CleanupElement> Ptr;

        /**
         * @fn CleanupElement(unsigned int _id, int _timeout);
         * @brief Constructor.
         * @param _id ID of element to clean.
         * @param _timeout Time when cleaning should be done.
         */
        CleanupElement(unsigned int _id, int _timeout);

        /**
         * @fn CleanupElement(const CleanupElement& _elem);
         * @brief Constructor.
         * @param _elem Copy object.
         */
        CleanupElement(const CleanupElement& _elem);

        /**
         * @brief Id of the request.
         */
        unsigned int m_Id;
        /**
         * @brief Timeout.
         */
        int m_Timeout;
    };

    /**
     * @fn Cleaner();
     * @brief Constructor.
     */
    Cleaner();

    /**
     * @fn Cleaner(const Cleaner& _cleaner);
     * @brief Constructor.
     * @param _cleaner Copy object.
     */
    Cleaner(const Cleaner& _cleaner);

    /**
     * @fn void StopCleaning();
     * @brief Stop cleaning process.
     */
    void StopCleaning();

    /**
     * @fn void ScheduleCleanup(unsigned int _id, unsigned int _time);
     * @brief Schedule cleaning for a request.
     * @param _id ID element to clean.
     * @param _time Relative time to clean request in seconds.
     */
    void ScheduleCleanup(unsigned int _id, unsigned int _time);

    /**
     * @fn void OnCleanup(Utility::Timer& _timer);
     * @brief Callback called when async requests should be cleaned.
     * @param _timer Timer.
     */
    void OnCleanup(Utility::Timer& _timer);

    /**
     * @fn void CleanupAll();
     * @brief Cleanup all pending requests.
     */
    void CleanupAll();

    /**
     * @fn virtual void CleanRequest(unsigned int _reqId) = 0;
     * @brief Removes a request from the queue.
     * @param _reqId Request identifier.
     */
    virtual void CleanRequest(unsigned int _reqId) = 0;
private:
    typedef std::list<CleanupElement::Ptr> CleanupList;

    CleanupList m_CleanupList;
    Utility::Mutex m_Mutex;
    Utility::Timer* m_Timer;
};

/**
 * @class CMessenger
 * @brief The CMessenger class
 * @extends Utility::Logger
 * @extends Cleaner
 * @implements IDeliveryListener
 */
class DLLExport CMessenger: public Utility::Logger, public IDeliveryListener, protected Cleaner {
public:
    /**
     * @fn CMessenger(const std::string& _clientId, const std::string& _tag,
               const std::string& _username = "", const std::string& _password = "",
               const std::string& _uri = "tcp://localhost:1883");
     * @brief Constructor.
     * @param _clientId Id of the user.
     * @param _tag Tag to be used to identify as an user.
     * @param _username Username used to authorize as an user.
     * @param _password Password used to authorize as an user.
     * @param _uri URI of a messaging server.
     */
    CMessenger(const std::string& _clientId, const std::string& _tag,
               const std::string& _username = "", const std::string& _password = "",
               const std::string& _uri = "tcp://localhost:1883");

    /**
     * @fn virtual ~CMessenger();
     * @brief Destructor.
     */
    virtual ~CMessenger();

    /**
     * @fn void SetSecureConfiguration(const std::string& _trustStore, const std::string& _keyStore,
                                const std::string& _privateKey, const std::string& _privateKeyPassword);
     * @brief Set secure configuration. Connect must be called again, to apply this configuration.
     * @param _trustStore Secure connection trust store.
     * @param _keyStore Secure connection key store.
     * @param _privateKey Secure connection private key.
     * @param _privateKeyPassword Secure connection pasword for a provate key.
     */
    void SetSecureConfiguration(const std::string& _trustStore, const std::string& _keyStore,
                                const std::string& _privateKey, const std::string& _privateKeyPassword);
    /**
     * @fn void Connect();
     * @brief Connects to messaging server.
     * @throws Utility::ProtocolException
     */
    void Connect();

    /**
     * @fn void Reconnect();
     * @brief Reconnects to messaging server. Combination of disconnect and connect force clean session.
     */
    void Reconnect();

    /**
     * @fn void Disconnect();
     * @brief Disconnects from a messaging server, if it was previously connected.
     */
    void Disconnect();

    /**
     * @fn void DisconnectWithoutStsMessage();
     * @brief Disconnects from a messaging server, if it was previously connected without sending status message.
     */
    void DisconnectWithoutStsMessage();

    /**
     * @fn bool IsConnected() const;
     * @brief Checks whether the client is connected.
     * @return Indication.
     */
    bool IsConnected() const;

    /**
     * @fn void IgnoreSessionPresent(bool _ignore);
     * @brief Control subscriptions if a session is present on the broker. Session present flag will be checked and if it is false the subscription will be applied. Otherwise, subscription will be skipped.
     * @param _ignore Indication how to handle the session present flag.
     */
    void IgnoreSessionPresent(bool _ignore);

    /**
     * @fn void RegisterResponder(IResponder* _responder);
     * @brief Registers a responder for requests arriving to messenger.
     * @param _responder Responder interface.
     * @throws Utility::ProtocolException
     */
    void RegisterResponder(IResponder* _responder);

    /**
     * @fn void UnregisterResponder();
     * @brief Unregisters from previusly defined responder.
     * @throws Utility::ProtocolException
     */
    void UnregisterResponder();

    /**
     * @fn void ActivateService(unsigned int _groupId, const std::string& _serviceType);
     * @brief Registers for requests arriving on a service. Responder needs to be registered to receive corresponding messages.
     * @param _groupId Group id.
     * @param _serviceType Service type.
     * @throws Utility::ProtocolException
     */
    void ActivateService(unsigned int _groupId, const std::string& _serviceType);

    /**
     * @fn void DeactivateService(unsigned int _groupId, const std::string& _serviceType);
     * @brief Unregisters from previously activated service.
     * @param _groupId Group id.
     * @param _serviceType Service type.
     * @throws Utility::ProtocolException
     */
    void DeactivateService(unsigned int _groupId, const std::string& _serviceType);

    /**
     * @fn void SendEvent(const EventMessage& _event, unsigned int _groupId, const std::string& _serviceType="");
     * @brief Sends event about messenger user.
     * @param _event Event message.
     * @param _groupId Group id.
     * @param _serviceType Service type.
     * @throws Utility::ProtocolException
     */
    void SendEvent(const EventMessage& _event, unsigned int _groupId, const std::string& _serviceType="");

    /**
     * @fn User::Ptr GetUser(const std::string& _tag);
     * @brief Returns a messaging user for an easier interaction.
     * @param _tag User tag.
     * @return Messaging user.
     */
    User::Ptr GetUser(const std::string& _tag);

    /**
     * @fn void RegisterConnectionListener(IConnectionListener* _connectionListener);
     * @brief Function used to register for events about network status.
     * @param _connectionListener Network connection listener.
     */
    void RegisterConnectionListener(IConnectionListener* _connectionListener);

    /**
     * @fn void UnregisterConnectionListener();
     * @brief Unregisters network connection listener.
     */
    void UnregisterConnectionListener();

    /**
     * @fn void RegisterDefaultListener(IEventListener* _listener);
     * @brief Register event listener to call if there is no listener registered to a specific topic. This can be used for wildcard subscriptions.
     * @param _listener Event and status listener.
     */
    void RegisterDefaultListener(IEventListener* _listener);

    /**
     * @fn void RegisterLogListener(ILogListener *_logListener);
     * @brief Registers log listener.
     * @param _logListener Log listener interface.
     */
    void RegisterLogListener(Utility::ILogListener *_logListener);

    /**
     * @fn void UnregisterLogListener();
     * @brief Cleans a log listener.
     */
    void UnregisterLogListener();

    /**
     * @brief Request message topic element. Value is "/req".
     */
    static const std::string REQUEST_TOPIC_ELEMENT;

    /**
     * @brief Response message topic element. Value is "/rsp".
     */
    static const std::string RESPONSE_TOPIC_ELEMENT;

    /**
     * @brief Event message topic element. Value is "/evt".
     */
    static const std::string EVENT_TOPIC_ELEMENT;

    /**
     * @brief Status message topic element. Value is "/sts".
     */
    static const std::string STS_TOPIC_ELEMENT;

    /**
     * @brief Topic element separator. Value is "/".
     */
    static const std::string TOPIC_SEPARATOR;

    /**
     * @brief Empty string. Value is "".
     */
    static const std::string EMPTY_STRING;
protected:
    friend class User;

    /**
     * @fn void RegisterForEvents(IEventListener* _eventListener, const std::string& _tag,
                           unsigned int _groupId, const std::string& _serviceType="");
     * @brief Registers for events arriving from user with a tag.
     * @param _eventListener Interface to call when a event arrives.
     * @param _tag User tag.
     * @param _groupId Group id.
     * @param _serviceType Service type.
     * @throws Utility::ProtocolException
     */
    void RegisterForEvents(IEventListener* _eventListener, const std::string& _tag,
                           unsigned int _groupId, const std::string& _serviceType="");

    /**
     * @fn void UnregisterFromEvents(IEventListener* _eventListener, const std::string& _tag,
                              unsigned int _groupId, const std::string& _serviceType="");
     * @brief Unregisters from events bound to a messenger user.
     * @param _eventListener Interface to call when a event arrives.
     * @param _tag User tag.
     * @param _groupId Group id.
     * @param _serviceType Service type.
     * @throws Utility::ProtocolException
     */
    void UnregisterFromEvents(IEventListener* _eventListener, const std::string& _tag,
                              unsigned int _groupId, const std::string& _serviceType="");

    /**
     * @fn void RegisterForStatusMsg(IEventListener* _eventListener, const std::string& _tag);
     * @brief Registers for status messages arriving from user with a tag.
     * @param _eventListener Interface to call when a event arrives.
     * @param _tag User tag.
     * @throws Utility::ProtocolException
     */
    void RegisterForStatusMsg(IEventListener* _eventListener, const std::string& _tag);

    /**
     * @fn void UnregisterFromStatusMsg(IEventListener* _eventListener, const std::string& _tag);
     * @brief Unregisters from status messages bound to a messenger user.
     * @param _eventListener Interface to call when a LWM arrives.
     * @param _tag User tag.
     * @throws Utility::ProtocolException
     */
    void UnregisterFromStatusMsg(IEventListener* _eventListener, const std::string& _tag);

    /**
     * @fn void Request(RequestMessage::Ptr _request, const std::string& _tag,
                 unsigned int _groupId, const std::string& _serviceType="");
     * @brief Sends a request to an user identified with a tag.
     * @param _request Request message.
     * @param _tag User tag.
     * @param _groupId Group id.
     * @param _serviceType Service type.
     * @throws Utility::ProtocolException
     * @throws Utility::InvalidStateException
     */
    void Request(RequestMessage::Ptr _request, const std::string& _tag,
                 unsigned int _groupId, const std::string& _serviceType="");

    /**
     * @fn virtual void CleanRequest(unsigned int _reqId);
     * @brief Removes a request from the queue.
     * @param _reqId Request identifier.
     */
    virtual void CleanRequest(unsigned int _reqId);
private:
    void Initialize();

    void CleanRequests();
    void ProcessEvent(const std::string& _senderTag, const Utility::JsonStruct::Ptr _payload);
    void ProcessStatus(const std::string& _senderTag, const Utility::JsonStruct::Ptr _payload);
    void ProcessRequest(const Utility::JsonStruct::Ptr _payload, unsigned int _groupId, const std::string& _serviceType);
    void ProcessFailedRequest(const Utility::JsonStruct::Ptr _payload);
    void ProcessResponse(const Utility::JsonStruct::Ptr _payload);
    void SendResponse(const std::string& _receiverTag, const ResponseMessage::Ptr _response);

    virtual void OnMessageArrived(const std::string &_topic, const Message::Ptr _msg);
    virtual void OnPublishFailed(int _code, const std::string &_topic, const Message::Ptr _msg);

    Utility::JsonStruct::Ptr ParseObject(const std::string& _object);
    typedef std::list<IEventListener*> EventListenersList;
    typedef std::map<std::string, EventListenersList> EventListenersMap;
    typedef std::map<unsigned int, RequestMessage::Ptr> RequestsMap;

    const std::string m_UserTag;
    bool m_ProcessingMsg;

    IResponder* m_Responder;
    Utility::Mutex m_EventStatusMutex;
    EventListenersMap m_EventListeners;
    EventListenersMap m_StatusListeners;
    Utility::Mutex m_RequestMutex;
    RequestsMap m_Requests;
    Utility::Event m_ProcessMsgSync;

    IEventListener* m_DefaultListener;
    BuffClient m_Client;
};
}


#endif // MESSENGER_H
