/****************************************************************************
 *  Copyright (c) 2018 Oblo Living d.o.o
 *
 *  All Rights Reserved
 *
 *  P R O P R I E T A R Y    &    C O N F I D E N T I A L
 *
 *  -----------------------------------------------------
 *  http://www.obloliving.com/
 *  -----------------------------------------------------
 *
 * \project Oblo WISE SDK - SDK library
 * \file NetworkProxy.h
 * \brief
 * Message queue classes.
 *
 * @Author Igor Stefanovic
 *
 * \notes
 *
 * \history
 *****************************************************************************/
#ifndef NETWORK_PROXY_H
#define NETWORK_PROXY_H

#include "util/Common.h"
#include "util/SharedPointer.h"
#include "messenger/SynchronizedRunner.h"
#include "messenger/MQTTMessage.h"
#include "messenger/MQTTClient.h"
#include "messenger/MessageHolder.h"

/**
 * @namespace Messenger
 * @brief Implementation of MQTT Client.
 */
namespace Messenger {

/**
 * @interface IDeliveryListener
 * @brief Delivery listener.
 */
class IDeliveryListener {
public:
    /**
     * @fn virtual void OnMessageArrived(const std::string& _topic, const Message::Ptr _msg) = 0;
     * @brief Function that will be called upon message arrival.
     * @param _topic Topic.
     * @param _msg MQTT message.
     */
    virtual void OnMessageArrived(const std::string& _topic, const Message::Ptr _msg) = 0;

    /**
     * @fn virtual void OnPublishFailed(int _code, const std::string& _topic, const Message::Ptr _msg) = 0;
     * @brief Function that will be called on failed publish.
     * @param _code Error code.
     * @param _topic Topic.
     * @param _msg MQTT message.
     */
    virtual void OnPublishFailed(int _code, const std::string& _topic, const Message::Ptr _msg) = 0;
};

/**
 * @interface IPublishListener
 * @brief Publish listener.
 */
class IPublishListener {
public:
    /**
     * @fn virtual void OnPublishRequired(const std::string& _topic, const Message::Ptr _msg) = 0;
     * @brief OnPublishRequired
     * @param _topic
     * @param _msg
     */
    virtual void OnPublishRequired(const std::string& _topic, const Message::Ptr _msg) = 0;
};

/**
 * @interface IConnListener
 * @brief Connection listener.
 */
class IConnListener {
public:
    /**
     * @fn virtual void OnConnectionLost(const std::string& _cause) = 0;
     * @brief Function that will be called on connection lost.
     * @param _cause Connection lost cause.
     */
    virtual void OnConnectionLost(const std::string& _cause) = 0;
};

/**
 * @interface INetworkProxyListener
 * @brief Group of listeners.
 * @implements IDeliveryListener
 * @implements IPublishListener
 * @implements IConnListener
 */
class INetworkProxyListener:
    public IDeliveryListener,
    public IPublishListener,
    public IConnListener {
};


/**
 * @class MessageDelegate
 * @brief Delegate class for message handling.
 */
class DLLExport MessageDelegate {
public:
    /**
     * @typedef Utility::SharedPointer<MessageDelegate> Ptr;
     * @brief Shared pointer to MessageDelegate object.
     */
    typedef Utility::SharedPointer<MessageDelegate> Ptr;

    /**
     * @fn MessageDelegate();
     * @brief Constructor.
     */
    MessageDelegate() :
        m_Object(NULL),
        m_Stub(NULL) {
    }

    template <class T, void (T::*TMethod)(const std::string& _topic, const Message& _msg)>
    /**
     * @fn static MessageDelegate::Ptr CreateDelegate(T* _object);
     * @brief Creates message delegate.
     * @param _object Object pointer.
     * @return Shared pointer to a request delegate.
     */
    static MessageDelegate::Ptr CreateDelegate(T* _object) {
        MessageDelegate::Ptr d(new MessageDelegate());
        d->m_Object = _object;
        d->m_Stub = &MethodStub<T, TMethod>;
        return d;
    }

    /**
     * @fn void operator()(const std::string& _topic, const Message& _msg) const;
     * @brief operator ().
     * @param _topic Topic.
     * @param _msg MQTT message.
     */
    void operator()(const std::string& _topic, const Message& _msg) const {
        (*m_Stub)(m_Object, _topic, _msg);
    }

private:
    typedef void (*StubType)(void* m_Object, const std::string& _topic, const Message& _msg);

    void* m_Object;
    StubType m_Stub;

    template <class T, void (T::*TMethod)(const std::string& _topic, const Message& _msg)>
    static void MethodStub(void* _object, const std::string& _topic, const Message& _msg) {
        T* p = static_cast<T*>(_object);
        (p->*TMethod)(_topic, _msg);
    }
};

/**
 * @class ConnectionRunner
 * @brief Class for checking the connection.
 * @extends SynchronizedRunner
 */
class DLLExport ConnectionRunner : public SynchronizedRunner {
public:
    /**
     * @fn ConnectionRunner(IConnListener& _connListener);
     * @brief Constructor.
     * @param _connListener Listener.
     */
    ConnectionRunner(IConnListener& _connListener);

    /**
     * @fn virtual void Reinitialize();
     * @brief Reinitialize synchronized runner.
     */
    virtual void Reinitialize();

    /**
     * @fn void Report(const std::string& _cause);
     * @brief Reports the connection lost.
     * @param _cause Connection lost cause.
     */
    void Report(const std::string& _cause);
protected:
    /**
     * @fn virtual void DoJob();
     * @brief Invokes the callback when connection lost is detected.
     */
    virtual void DoJob();
private:
    static const std::string NO_LOST;
    std::string m_Cause;
    IConnListener& m_ConnListener;
};

/**
 * @class AsyncNetworkProxy
 * @brief The AsyncNetworkProxy class
 * @implements IStatusListener
 * @implements IQueueProcessor<PublishHolder::Ptr>
 * @implements IQueueProcessor<RecvHolder::Ptr>
 */
class DLLExport AsyncNetworkProxy: public IStatusListener, public IQueueProcessor<PublishHolder::Ptr>, public IQueueProcessor<RecvHolder::Ptr> {
public:
    /**
     * @fn AsyncNetworkProxy(INetworkProxyListener& _listener);
     * @brief Constructor.
     * @param _listener Listener.
     */
    AsyncNetworkProxy(INetworkProxyListener& _listener);

    /**
     * @fn ~AsyncNetworkProxy();
     * @brief Destructor.
     */
    ~AsyncNetworkProxy();

    /**
     * @fn void Publish(const std::string& _topic, const Message::Ptr _msg);
     * @brief Pushes the MQTT message to queue for processing.
     * @param _topic Topic.
     * @param _msg MQTT message.
     */
    void Publish(const std::string& _topic, const Message::Ptr _msg);
private:
    virtual void ConnectionLost(const std::string& _cause);
    virtual void MessageArrived(const std::string& _topic, const Message::Ptr _msg);
    virtual void DeliveryComplete(const IDeliveryToken &_token);
    virtual void PublishFailed(int _code, const std::string &_topic, const Message::Ptr _msg);
    virtual void OnProcess(PublishHolder::Ptr _object);
    virtual void OnProcess(RecvHolder::Ptr _object);

    INetworkProxyListener& m_Listener;
    Utility::Thread m_SendingThread;
    Utility::Thread m_RecvThread;
    Utility::Thread m_ConnThread;
    QueueRunner<PublishHolder::Ptr> m_SendingQueue;
    QueueRunner<RecvHolder::Ptr> m_RecvQueue;
    ConnectionRunner m_ConnRunner;
    bool m_Stopped;
};










}


#endif // NETWORK_PROXY_H
