/****************************************************************************
 *  Copyright (c) 2018 Oblo Living d.o.o
 *
 *  All Rights Reserved
 *
 *  P R O P R I E T A R Y    &    C O N F I D E N T I A L
 *
 *  -----------------------------------------------------
 *  http://www.obloliving.com/
 *  -----------------------------------------------------
 *
 * \project Oblo WISE SDK - SDK library
 * \file SynchronizedRunner.h
 * \brief
 * Syncronized Runner classes.
 *
 * @Author Igor Stefanovic
 *
 * \notes
 *
 * \history
 *****************************************************************************/
#ifndef SYNCHRONIZED_RUNNER_H
#define SYNCHRONIZED_RUNNER_H

#include "util/Common.h"
#include "util/Thread.h"
#include "util/SynchronizedObject.h"

/**
 * @namespace Messenger
 * @brief Implementation of MQTT Client.
 */
namespace Messenger {

/**
 * @class SynchronizedRunner
 * @brief Synchronous. thread implementation class.
 * @extends Utility::Runnable
 */
class DLLExport SynchronizedRunner : public Utility::Runnable {
public:
    /**
     * @fn SynchronizedRunner();
     * @brief Constructor.
     */
    SynchronizedRunner();

    /**
     * @fn virtual ~SynchronizedRunner();
     * @brief Destructor.
     */
    virtual ~SynchronizedRunner();

    /**
     * @fn virtual void Reinitialize();
     * @brief Resets the runner.
     */
    virtual void Reinitialize();

    /**
     * @fn virtual void Stop();
     * @brief Stops the runner.
     */
    virtual void Stop();
protected:
    /**
     * @fn void Pause();
     * @brief Pause the runner.
     */
    void Pause();

    /**
     * @fn void Continue();
     * @brief Continue.
     */
    void Continue();

    /**
     * @fn void Lock();
     * @brief Lock the mutex.
     */
    void Lock();

    /**
     * @fn void Unlock();
     * @brief Unlock the mutex.
     */
    void Unlock();

    /**
     * @fn bool IsRunning() const;
     * @brief Indication whether the runner is running.
     * @return Indication.
     */
    bool IsRunning() const;

    /**
     * @fn virtual void DoJob() = 0;
     * @brief Function that will be called on running.
     */
    virtual void DoJob() = 0;
private:
    virtual void Run();
    bool m_Running;
    Utility::SynchronizedObject m_Sync;
    Utility::Mutex m_Mutex;
};

template <class C>
/**
 * @interface IQueueProcessor
 * @brief Queue processor interface.
 */
class IQueueProcessor {
public:
    /**
     * @fn virtual void OnProcess(C _object) = 0;
     * @brief Function that will be called for processing elements in queue.
     * @param _object Element in queue.
     */
    virtual void OnProcess(C _object) = 0;
};

template <class C>
/**
 * @class QueueRunner
 * @brief Queue for processing elements.
 * @extends SynchronizedRunner
 */
class DLLExport QueueRunner : public SynchronizedRunner {
public:
    /**
     * @fn QueueRunner(IQueueProcessor<C>& _processor);
     * @brief Constructor.
     * @param _processor Queue processor.
     */
    QueueRunner(IQueueProcessor<C>& _processor) : m_Processor(_processor) {

    }

    /**
     * @fn virtual ~QueueRunner()
     * @brief Destructor.
     */
    virtual ~QueueRunner() {}

    /**
     * @fn void Add(C _element)
     * @brief Add new element to the queue.
     * @param _element Element to be added.
     */
    void Add(C _element) {
        Lock();
        m_Queue.push(_element);
        Unlock();
        Continue();
    }
protected:
    /**
     * @fn virtual void DoJob()
     * @brief Processing the elements.
     */
    virtual void DoJob() {
        Lock();

        if (m_Queue.empty()) {
            Unlock();
            Pause();
        } else {
            while (!m_Queue.empty() && IsRunning()) {
                C elem = m_Queue.front();
                m_Queue.pop();
                Unlock();
                m_Processor.OnProcess(elem);
                Lock();
            }

            Unlock();
        }
    }
private:
    typedef std::queue<C> Queue;
    Queue m_Queue;
    IQueueProcessor<C>& m_Processor;
};


}



#endif // SYNCHRONIZED_RUNNER_H
