/****************************************************************************
 *  Copyright (c) 2018 Oblo Living d.o.o
 *
 *  All Rights Reserved
 *
 *  P R O P R I E T A R Y    &    C O N F I D E N T I A L
 *
 *  -----------------------------------------------------
 *  http://www.obloliving.com/
 *  -----------------------------------------------------
 *
 * \project Oblo WISE SDK - SDK library
 * \file User.h
 * \brief
 * MQTT User abstraction classes.
 *
 * @Author Igor Stefanovic
 *
 * \notes
 *
 * \history
 *****************************************************************************/
#ifndef USER_H
#define USER_H

#include "util/Common.h"
#include "util/Timer.h"
#include "util/Exceptions.h"
#include "messenger/Messages.h"

/**
 * @namespace Messenger
 * @brief Implementation of MQTT Client.
 */
namespace Messenger {

/**
 * @brief Default request timeout. Value is 5.
 */
const unsigned int DEFAULT_REQUEST_TIMEOUT = 5;

/**
 * @interface IEventListener
 * @brief Used to report new events from messaging objects.
 */
class IEventListener {
public:
    /**
     * @fn virtual void OnEvent(const std::string& _senderTag, const EventMessage::Ptr _msg) = 0;
     * @brief Called when an event arrives.
     * @param _senderTag Sender's tag.
     * @param _msg Event message.
     */
    virtual void OnEvent(const std::string& _senderTag, const EventMessage::Ptr _msg) = 0;

    /**
     * @fn virtual void OnStatusMessage(const std::string& _senderTag, const StatusMessage::Ptr _msg) = 0;
     * @brief Called when a status message arrives.
     * @param _senderTag Sender's tag.
     * @param _msg Status message.
     */
    virtual void OnStatusMessage(const std::string& _senderTag, const StatusMessage::Ptr _msg) = 0;
};

class User;

/**
 * @class ServiceFunctionality
 * @brief The wrapper for messaging user services.
 */
class DLLExport ServiceFunctionality {
public:
    /**
     * @typedef Utility::SharedPointer<ServiceFunctionality> Ptr;
     * @brief Shared pointer to ServiceFunctionality object.
     */
    typedef Utility::SharedPointer<ServiceFunctionality> Ptr;

    /**
     * @fn void RegisterForEvents(IEventListener* _eventListener);
     * @brief Registers for events from user wrapped inside current object. Use NULL to get events to default listener.
     * @param _eventListener Event listener.
     * @throws Utility::InvalidStateException
     */
    void RegisterForEvents(IEventListener* _eventListener);

    /**
     * @fn void UnregisterFromEvents();
     * @brief Unregisters from events, if previously registered.
     * @throws Utility::InvalidStateException
     */
    void UnregisterFromEvents();

    /**
     * @fn const ResponseMessage::Ptr Request(const RequestMessage &_request, unsigned int _timeout = DEFAULT_REQUEST_TIMEOUT);
     * @brief Sends a synchronous request to an user.
     * @param _request Request message.
     * @param _timeout Time to wait for a response in seconds.
     * @return Response to a request.
     * @throws Utility::InvalidStateException
     */
    const ResponseMessage::Ptr Request(const RequestMessage &_request, unsigned int _timeout = DEFAULT_REQUEST_TIMEOUT);

    /**
     * @fn void Request(const RequestMessage& _request, const RequestDelegate::Ptr _requestDelegate, unsigned int _timeout = DEFAULT_REQUEST_TIMEOUT);
     * @brief Sends an asynchronous request to an user.
     * @param _request Request message.
     * @param _requestDelegate Delegate to call when response arrives.
     * @param _timeout Time to wait for a response in seconds.
     * @throws Utility::InvalidStateException
     */
    void Request(const RequestMessage& _request, const RequestDelegate::Ptr _requestDelegate, unsigned int _timeout = DEFAULT_REQUEST_TIMEOUT);
private:
    friend class User;

    ServiceFunctionality(unsigned int _groupId, const std::string& _serviceType);

    void SetOwner(User* _user);
    unsigned int m_GroupId;
    std::string m_ServiceType;
    User* m_User;
};

class CMessenger;

/**
 * @class User
 * @brief Parties included in network communication.
 */
class DLLExport User {
public:
    /**
     * @typedef Utility::SharedPointer<User> Ptr;
     * @brief Shared pointer to User object.
     */
    typedef Utility::SharedPointer<User> Ptr;

    /**
     * @fn User(const std::string& _tag);
     * @brief Constructor.
     * @param _tag User's tag.
     */
    User(const std::string& _tag);

    /**
     * @fn User(const User& _obj);
     * @brief Constructor.
     * @param _obj Copy object.
     */
    User(const User& _obj);

    /**
     * @fn virtual ~User();
     * @brief Destructor.
     */
    virtual ~User();

    /**
     * @fn std::string GetTag() const;
     * @brief Getter for tag field.
     * @return User's tag.
     */
    std::string GetTag() const;

    /**
     * @fn void RegisterForEvents(IEventListener* _eventListener);
     * @brief Registers for events from user wrapped inside current object. Use NULL to get events to default listener.
     * @param _eventListener Event listener.
     * @throws Utiltiy::InvalidStateException
     */
    void RegisterForEvents(IEventListener* _eventListener);

    /**
     * @fn void UnregisterFromEvents();
     * @brief Unregisters from events, if previously registered.
     * @throws Utiltiy::InvalidStateException
     */
    void UnregisterFromEvents();

    /**
     * @fn void RegisterForStatusMsg(IEventListener* _statusMsgListener);
     * @brief Registers for status messages from user wrapped inside current object. Use NULL to get events to default listener.
     * @param _statusMsgListener Event listener.
     * @throws Utiltiy::InvalidStateException
     */
    void RegisterForStatusMsg(IEventListener* _statusMsgListener);

    /**
     * @fn void UnregisterFromStatusMsg();
     * @brief Unregisters from status message, if previously registered.
     * @throws Utiltiy::InvalidStateException
     */
    void UnregisterFromStatusMsg();

    /**
     * @fn const ResponseMessage::Ptr Request(const RequestMessage& _request, unsigned int _timeout = DEFAULT_REQUEST_TIMEOUT);
     * @brief Sends a synchronous request to an user.
     * @param _request Request message.
     * @param _timeout Time to wait for a response in seconds.
     * @return Response to a request.
     * @throws Utility::TimeoutException
     * @throws Utility::NullPointerException
     * @throws Utility::InvalidStateException
     */
    const ResponseMessage::Ptr Request(const RequestMessage& _request, unsigned int _timeout = DEFAULT_REQUEST_TIMEOUT);

    /**
     * @fn void Request(const RequestMessage &_request, const RequestDelegate::Ptr _requestDelegate, unsigned int _timeout = DEFAULT_REQUEST_TIMEOUT);
     * @brief Sends an asynchronous request to an user.
     * @param _request Request message.
     * @param _requestDelegate Delegate to call when response arrives.
     * @param _timeout Time to wait for a response in seconds.
     * @throws Utility::InvalidArgumentException
     * @throws Utility::InvalidStateException
     */
    void Request(const RequestMessage &_request, const RequestDelegate::Ptr _requestDelegate, unsigned int _timeout = DEFAULT_REQUEST_TIMEOUT);

    /**
     * @fn void SendResponse(unsigned int _requestId, ResponseMessage::Ptr _response);
     * @brief Sends a response to a messaging user.
     * @param _requestId Identifier of a matching request.
     * @param _response Response message.
     * @throws Utility::NullPointerException
     * @throws Utility::InvalidStateException
     */
    void SendResponse(unsigned int _requestId, ResponseMessage::Ptr _response);

    /**
     * @fn ServiceFunctionality::Ptr GetServiceFunctionality(unsigned int _groupId, const std::string& _serviceType);
     * @brief Getter for an user's service.
     * @param _groupId Group id.
     * @param _serviceType Service type.
     * @return Shared pointer to service functionality.
     */
    ServiceFunctionality::Ptr GetServiceFunctionality(unsigned int _groupId, const std::string& _serviceType);
protected:
    friend class CMessenger;
    friend class ServiceFunctionality;

    /**
     * @fn void SetMessenger(CMessenger* _messenger);
     * @brief Setter for messenger.
     * @param _messenger Messenger.
     */
    void SetMessenger(CMessenger* _messenger);
private:
    void RegisterForEvents(IEventListener *_eventListener, unsigned int _groupId, const std::string& _serviceType);
    void UnregisterFromEvents(unsigned int _groupId, const std::string& _serviceType);

    const ResponseMessage::Ptr Request(const RequestMessage& _request, unsigned int _groupId, const std::string& _serviceType, unsigned int _timeout = DEFAULT_REQUEST_TIMEOUT);
    void Request(const RequestMessage& _request, const RequestDelegate::Ptr _requestDelegate, unsigned int _groupId, const std::string& _serviceType, unsigned int _timeout = DEFAULT_REQUEST_TIMEOUT);

    std::string m_Tag;
    CMessenger* m_Messenger;
    IEventListener* m_EventListener;
    IEventListener* m_StatusMsgListener;
};

}

#endif // USER_H
