/****************************************************************************
 *  Copyright (c) 2018 Oblo Living d.o.o
 *
 *  All Rights Reserved
 *
 *  P R O P R I E T A R Y    &    C O N F I D E N T I A L
 *
 *  -----------------------------------------------------
 *  http://www.obloliving.com/
 *  -----------------------------------------------------
 *
 * \project Oblo WISE SDK - SDK library
 * \file Discovery.h
 * \brief
 * Discovery process abstraction class.
 *
 * @Author Igor Stefanovic
 *
 * \notes
 *
 * \history
 *****************************************************************************/
#ifndef DISCOVERY_H
#define DISCOVERY_H

#include "util/Mutex.h"
#include "util/Timer.h"
#include "util/ScopedLock.h"
#include "ssdp/SSDPClient.h"

/**
 * @namespace SSDP
 * @brief SSDP client implementation
 */
namespace SSDP {

/**
 * @class Discovery
 * @brief Class which offers SSDPClient API to IPDevice::Device object.
 */
class DLLExport Discovery {
public:
    /**
     * @brief Default MSEARCH interval set to 5 seconds.
     */
    static int DEFAULT_MSEARCH_INTERVAL;

    /**
     * @fn Discovery();
     * @brief Constructor.
     */
    Discovery();

    /**
     * @fn ~Discovery();
     * @brief Destructor.
     */
    ~Discovery();

    /**
     * @fn void StartSearch();
     * @brief Starts sending SSDP MSEARCH messages.
     */
    void StartSearch();

    /**
     * @fn void StartListen();
     * @brief Starts listening SSDP NOTIFY messages.
     * @param _hubId Hub id.
     */
    void StartListen(const std::string& _hubId);

    /**
     * @fn void Stop();
     * @brief Stops all searching/listening.
     */
    void Stop();

    /**
     * @fn void SetSearchInterval(long _searchInterval);
     * @brief Sets the search interval.
     * @param _searchInterval New search interval.
     */
    void SetSearchInterval(long _searchInterval);

    /**
     * @fn void RegisterListener(ISearchListener* _listener);
     * @brief Register ISearchListener object with name.
     * @param _name Name of ISearchListener object.
     * @param _listener Pointer to ISearchListener object.
     */
    void RegisterListener(ISearchListener* _listener);

    /**
     * @fn void UnregisterListener();
     * @brief Unregisters ISearchListener object.
     * @param _name Name of the ISearchListener object.
     */
    void UnregisterListener();

    /**
     * @fn void RegisterLogListener(Utility::ILogListener* _logListener);
     * @brief Registers Utility::ILogListener object.
     * @param _logListener Pointer to Utility::ILogListener object.
     */
    void RegisterLogListener(Utility::ILogListener* _logListener);

    /**
     * @fn void UnregisterLogListener();
     * @brief Unregisters Utility::ILogListener object.
     */
    void UnregisterLogListener();
private:
    Utility::Mutex m_Mutex;
    SSDPClient m_SSDPClient;
};

}


#endif // DISCOVERY_H
