/****************************************************************************
 *  Copyright (c) 2018 Oblo Living d.o.o
 *
 *  All Rights Reserved
 *
 *  P R O P R I E T A R Y    &    C O N F I D E N T I A L
 *
 *  -----------------------------------------------------
 *  http://www.obloliving.com/
 *  -----------------------------------------------------
 *
 * \project Oblo WISE SDK - SDK library
 * \file SSDPClient.h
 * \brief
 * SSDP client abstraction classes.
 *
 * @Author Igor Stefanovic
 *
 * \notes
 *
 * \history
 *****************************************************************************/
#ifndef SSDP_CLIENT_H
#define SSDP_CLIENT_H

#include <sys/time.h>
#include "util/Mutex.h"
#include "util/Event.h"
#include "util/ScopedLock.h"
#include "util/Thread.h"
#include "util/Common.h"
#include "util/Timer.h"
#include "util/HTTPClient.h"
#include "util/Logger.h"
#include "util/SynchronizedObject.h"
#include "ssdp/Socket.h"
#include "rapidxml/rapidxml.hpp"
#include <map>

/**
 * @namespace SSDP
 * @brief SSDP client implementation.
 */
namespace SSDP {
class HubHolder;

/**
 * @interface ISearchListener
 * @brief Interface to a hub detection.
 */
class ISearchListener {
public:
    /**
     * @fn virtual void OnHubDetected(const HubHolder& _hub) = 0;
     * @brief Callback to be called when Hub is detected.
     * @param _hub HubHolder object to be filled.
     */
    virtual void OnHubDetected(const HubHolder& _hub) = 0;
};

/**
 * @class SSDPClient
 * @brief The SSDPClient class
 * @extends Utility::Runnable
 * @extends Utility::Logger
 */
class DLLExport SSDPClient : Utility::Runnable, public Utility::Logger {
public:
    /**
     * @enum Mode
     * @brief Mode of the SSDP client.
     */
    enum Mode {
        SEARCH = 0,
        LISTEN
    };

    /**
     * @fn SSDPClient(long _searchInterval);
     * @brief Constructor.
     * @param _searchInterval Period of MSEARCH message sending.
     */
    SSDPClient(long _searchInterval);

    /**
     * @fn ~SSDPClient();
     * @brief Destructor.
     */
    ~SSDPClient();

    /**
     * @fn void StartSSDP(Mode _mode, const std::string& _hubId);
     * @brief Start the SSDP client.
     * @param _mode Mode in which SSDP client will be started.
     * @param _hubId Hub id.
     */
    void StartSSDP(Mode _mode, const std::string& _hubId);

    /**
     * @fn void StopSSDP();
     * @brief Stop the SSDP client.
     */
    void StopSSDP();

    /**
     * @fn void SetSearchInterval(long _searchInterval);
     * @brief Setter for search interval field.
     * @param _searchInterval Search interval to be set.
     */
    void SetSearchInterval(long _searchInterval);

    /**
     * @fn void RegisterListener(ISearchListener *_listener);
     * @brief Function to register for when hub is discovered on network.
     * @param _name Name of the listener.
     * @param _listener Listener.
     */
    void RegisterListener(ISearchListener *_listener);

    /**
     * @fn void UnregisterListener();
     * @brief Unregister listener for discover.
     * @param _name Name of the listener.
     */
    void UnregisterListener();

    /**
     * @fn void RegisterLogListener(Utility::ILogListener *_logListener);
     * @brief Registers log listener.
     * @param _logListener Listener.
     */
    void RegisterLogListener(Utility::ILogListener *_logListener);

    /**
     * @fn void UnregisterLogListener();
     * @brief Unregister log listener.
     */
    void UnregisterLogListener();

private:
    static const int SSDP_PORT;
    static const int MAX_BUFFER_LENGTH;
    static const std::string HUB_URN;
    static const int TIME_FOR_DATA;

    ISearchListener* m_SearchListener;
    Mode m_Mode;
    Utility::Thread m_Thread;
    Socket* m_Socket;
    Utility::Timer* m_SearchTimer;
    Utility::Mutex m_Mutex;
    bool m_ShouldRun;
    bool m_IsTerminated;
    bool m_Processing;
    std::string m_HubId;
    void ParseDeviceXml(char *_xml, const std::string& _location);
    void SendMSearch(Utility::Timer& _timer);
    void ProcessMessage(const std::string& _msg);
    void Run();
};

/**
 * @class HubHolder
 * @brief Holder class for information about hub after discovery.
 */
class DLLExport HubHolder {
public:
    /**
     * @fn HubHolder();
     * @brief Constructor.
     */
    HubHolder();

    /**
     * @fn HubHolder(const std::string& _domainId, const std::string& _homeId, const std::string& _hubId, const std::string& _location);
     * @brief Constructor.
     * @param _domainId Domain id. Check WISE specification.
     * @param _homeId Home id. Check WISE specification.
     * @param _hubId Hub id. Check WISE specification.
     * @param _location Location of the hub.
     */
    HubHolder(const std::string& _domainId, const std::string& _homeId, const std::string& _hubId, const std::string& _location);

    /**
     * @fn std::string GetDomainID() const;
     * @brief Getter for domain id.
     * @return Domain id.
     */
    std::string GetDomainID() const;

    /**
     * @fn std::string GetHomeID() const;
     * @brief Getter for home id.
     * @return Home id.
     */
    std::string GetHomeID() const;

    /**
     * @fn std::string GetHubID() const;
     * @brief Getter for hub id.
     * @return Hub id.
     */
    std::string GetHubID() const;

    /**
     * @fn std::string GetLocation() const;
     * @brief Getter for location of the hub.
     * @return Location.
     */
    std::string GetLocation() const;

    /**
     * @fn void SetDomainID(const std::string& _domainId);
     * @brief Setter for domain id.
     * @param _projectId Domain id to be set.
     */
    void SetDomainID(const std::string& _domainId);

    /**
     * @fn void SetHomeID(const std::string& _homeId);
     * @brief Setter for home id.
     * @param _homeId Home id to be set.
     */
    void SetHomeID(const std::string& _homeId);

    /**
     * @fn void SetHubID(const std::string& _hubId);
     * @brief Setter for hub id.
     * @param _hubId Hub id to be set.
     */
    void SetHubID(const std::string& _hubId);

    /**
     * @fn void SetLocation(const std::string& _location);
     * @brief Setter for location of the hub.
     * @param _location Location to be set.
     */
    void SetLocation(const std::string& _location);
private:
    std::string m_DomainID;
    std::string m_HomeID;
    std::string m_HubID;
    std::string m_Location;
};

}
#endif // SSDP_CLIENT_H
