/****************************************************************************
 *  Copyright (c) 2018 Oblo Living d.o.o
 *
 *  All Rights Reserved
 *
 *  P R O P R I E T A R Y    &    C O N F I D E N T I A L
 *
 *  -----------------------------------------------------
 *  http://www.obloliving.com/
 *  -----------------------------------------------------
 *
 * \project Oblo WISE SDK - SDK library
 * \file Socket.h
 * \brief
 * Socket abstraction class.
 *
 * @Author Igor Stefanovic
 *
 * \notes
 *
 * \history
 *****************************************************************************/
#ifndef SOCKET_H
#define SOCKET_H

#include <sys/socket.h>
#include <fcntl.h>
#include "ssdp/SocketAddress.h"
#include "util/Exceptions.h"

/**
 * @namespace SSDP
 * @brief SSDP client implementation
 */
namespace SSDP {

/**
 * @class Socket
 * @brief Abstraction class for Socket.
 */
class DLLExport Socket {
public:
    /**
     * @fn Socket();
     * @brief Constructor.
     */
    Socket();

    /**
     * @fn Socket(const SocketAddress& _address);
     * @brief Constructor
     * @param _address
     */
    Socket(const SocketAddress& _address);

    /**
     * @fn ~Socket();
     * @brief Destructor.
     */
    ~Socket();

    /**
     * @fn void BindSocket();
     * @brief Binds to a socket.
     */
    void BindSocket();

    /**
     * @fn void ReuseSocket(bool _reuse);
     * @brief Enable/disable an option for socket reuse.
     * @param _reuse Indication whether this option will be enabled or disabled.
     */
    void ReuseSocket(bool _reuse);

    /**
     * @fn void JoinGroup(const SocketAddress& _address);
     * @brief Joins the group specified with parameter.
     * @param _address IP address to group.
     */
    void JoinGroup(const SocketAddress& _address);

    /**
     * @fn void IncludeLoopback(bool _include);
     * @brief Enable/disable an option for loopback.
     * @param _include Indication whether this option will be enabled or disabled.
     */
    void IncludeLoopback(bool _include);

    /**
     * @fn void SetNonBlocking(bool _blocking);
     * @brief Enable/disable a non blocking option.
     * @param _blocking Indication whether this option will be enabled or disabled.
     */
    void SetNonBlocking(bool _blocking);

    /**
     * @fn int OnDataRead(unsigned long _timeToWait);
     * @brief Waits on data to arrive on socket.
     * @param _timeToWait Time to wait for the data.
     * @return Response code which defines whether there is data to be read, whether the timeout expired or whether there is an error about data.
     */
    int OnDataRead(unsigned long _timeToWait);

    /**
     * @fn int RecvDataGram(void* _buffer, int _bufferLen, std::string& _sourceAddress, unsigned short& _sourcePort);
     * @brief Reads the data arrived on socket.
     * @param _buffer Buffer which will be filled on read.
     * @param _bufferLen Length of read bytes.
     * @param _sourceAddress Address from which the data arrived.
     * @param _sourcePort Port from which the data arrived.
     * @return Returns whether the data was successfully read or not.
     */
    int RecvDataGram(void* _buffer, int _bufferLen, std::string& _sourceAddress, unsigned short& _sourcePort);

    /**
     * @fn void SendMulticast(const void* _buffer, int _bufferLen, unsigned short _foreignPort);
     * @brief Send data to multicast group.
     * @param _buffer Buffer which contains the data.
     * @param _bufferLen Length of bytes to be sent.
     * @param _foreignPort Port on which the data will be sent.
     */
    void SendMulticast(const void* _buffer, int _bufferLen, unsigned short _foreignPort);

    /**
     * @fn int GetSocketFD() const;
     * @brief Getter for socket file descriptor.
     * @return Socket file descriptor.
     */
    int GetSocketFD() const;

    /**
     * @fn void Close();
     * @brief Closes the socket.
     */
    void Close();

    /**
     * @brief Multicast address set to 239.255.255.250.
     */
    static const std::string MULTICAST_ADDR;

    /**
     * @brief Multicast port set to 1900.
     */
    static const int MULTICAST_PORT;
private:
    void SendDataGram(const void* _buffer, int _bufferLen, const std::string& _foreignAddress, unsigned short _foreignPort);
    void Send(const void* _buffer, int _bufferLen);
    int Recv(void* _buffer, int _bufferLen);

    int m_Socket;
    int m_Family;
    SocketAddress m_Addr;
    bool m_Closed;

};

}


#endif // SOCKET_H
