/****************************************************************************
 *  Copyright (c) 2018 Oblo Living d.o.o
 *
 *  All Rights Reserved
 *
 *  P R O P R I E T A R Y    &    C O N F I D E N T I A L
 *
 *  -----------------------------------------------------
 *  http://www.obloliving.com/
 *  -----------------------------------------------------
 *
 * \project Oblo WISE SDK - SDK library
 * \file Exceptions.h
 * \brief
 * Exceptions abstraction classes.
 *
 * @Author Igor Stefanovic
 *
 * \notes
 *
 * \history
 *****************************************************************************/
#ifndef EXCEPTIONS_H
#define EXCEPTIONS_H

#include "util/Common.h"

/**
 * @namespace Utility
 * @brief Abstraction utility classes.
 */
namespace Utility {

/**
 * @class BaseException
 * @brief Base wrapper class for std::exception.
 * @extends std::exception
 */
class DLLExport BaseException : public std::exception {
public:
    /**
     * @fn BaseException(const std::string& _msg, int _code = 0);
     * @brief Constructor.
     * @param _msg Message of the exception.
     * @param _code Code of the exception
     */
    BaseException(const std::string& _msg, int _code = 0);

    /**
     * @fn BaseException(const std::string _name, const std::string& _msg, int _code = 0);
     * @brief Constructor.
     * @param _name Name of the exception.
     * @param _msg Message of the exception.
     * @param _code Code of the exception.
     */
    BaseException(const std::string _name, const std::string& _msg, int _code = 0);

    /**
     * @fn BaseException(const BaseException& _baseException);
     * @brief Constructor
     * @param _baseException Object.
     */
    BaseException(const BaseException& _baseException);

    /**
     * @fn ~BaseException() throw();
     * @brief Destructor.
     */
    ~BaseException() throw();

    /**
     * @fn virtual const char* what() const throw();
     * @brief what function from std::exception.
     * @return Name of the exception.
     */
    virtual const char* what() const throw();

    /**
     * @fn std::string GetName() const;
     * @brief Getter for Name field.
     * @return Name of the exception.
     */
    std::string GetName() const;

    /**
     * @fn std::string GetMessage() const;
     * @brief Getter for Message field.
     * @return Message of the exception.
     */
    std::string GetMessage() const;

    /**
     * @fn void SetMessage(const std::string& _message);
     * @brief Setter for Message field.
     * @param _message Message to be set.
     */
    void SetMessage(const std::string& _message);

    /**
     * @fn int GetCode() const;
     * @brief Getter for Code field.
     * @return Code of the exception.
     */
    int GetCode() const;

    /**
     * @fn void SetCode(int _code);
     * @brief Setter for Code field.
     * @param _code Code to be set.
     */
    void SetCode(int _code);

    /**
     * @fn virtual void Rethrow() const;
     * @brief Rethrows the exception.
     */
    virtual void Rethrow() const;

    /**
     * @brief Name of the base exception. Value is "BaseException".
     */
    static const std::string BASE_EXCEPTION;
private:
    std::string m_Name;
    std::string m_Message;
    int m_Code;
};


/**
 * @class ProtocolException
 * @brief The protocol exception.
 * @extends Utility::BaseException
 */
class DLLExport ProtocolException : public BaseException {
public:
    /**
     * @brief Name of the protocol exception. Value is "ProtocolException".
     */
    static const std::string PROTOCOL_EXCEPTION;

    /**
     * @fn ProtocolException(const std::string& _msg, int _code = 0);
     * @brief Constructor.
     * @param _msg Message of the exception.
     * @param _code Code of the exception.
     */
    ProtocolException(const std::string& _msg, int _code = 0) : BaseException(PROTOCOL_EXCEPTION, _msg, _code) {
    }
};

/**
 * @class ParseException
 * @brief The parsing exception.
 * @extends Utility::BaseException
 */
class DLLExport ParseException : public BaseException {
public:
    /**
     * @brief Name of the parse exception. Value is "ParseException".
     */
    static const std::string PARSE_EXCEPTION;

    /**
     * @fn ParseException(const std::string& _msg, int _code = 0);
     * @brief Constructor.
     * @param _msg Message of the exception.
     * @param _code Code of the exception.
     */
    ParseException(const std::string& _msg, int _code = 0) : BaseException(PARSE_EXCEPTION, _msg, _code) {
    }
};

/**
 * @class TimeoutException
 * @brief The timeout exception.
 * @extends Utility::BaseException
 */
class DLLExport TimeoutException : public BaseException {
public:
    /**
     * @brief Name of the timeout exception. Value is "TimeoutException".
     */
    static const std::string TIMEOUT_EXCEPTION;

    /**
     * @fn TimeoutException(const std::string& _msg, int _code = 0);
     * @brief Constructor.
     * @param _msg Message of the exception.
     * @param _code Code of the exception.
     */
    TimeoutException(const std::string& _msg, int _code = 0) : BaseException(TIMEOUT_EXCEPTION, _msg, _code) {
    }
};

/**
 * @class NullPointerException
 * @brief The null pointer exception.
 * @extends Utility::BaseException
 */
class DLLExport NullPointerException : public BaseException {
public:
    /**
     * @brief Name of the null pointer exception. Value is "NullPointerException".
     */
    static const std::string NULLPOINTER_EXCEPTION;

    /**
     * @fn NullPointerException(const std::string& _msg, int _code = 0);
     * @brief Constructor.
     * @param _msg Message of the exception.
     * @param _code Code of the exception.
     */
    NullPointerException(const std::string& _msg, int _code = 0) : BaseException(NULLPOINTER_EXCEPTION, _msg, _code) {
    }
};

/**
 * @class InvalidArgumentException
 * @brief The invalid argument exception.
 * @extends Utility::BaseException
 */
class DLLExport InvalidArgumentException : public BaseException {
public:
    /**
     * @brief Name of the invalid argument exception. Value is "InvalidArgumentException".
     */
    static const std::string INVALID_ARGUMENT_EXCEPTION;

    /**
     * @fn InvalidArgumentException(const std::string& _msg, int _code = 0);
     * @brief Constructor.
     * @param _msg Message of the exception.
     * @param _code Code of the exception.
     */
    InvalidArgumentException(const std::string& _msg, int _code = 0) : BaseException(INVALID_ARGUMENT_EXCEPTION, _msg, _code) {
    }
};

/**
 * @class InvalidStateException
 * @brief The invalid state exception.
 * @extends Utility::BaseException
 */
class DLLExport InvalidStateException : public BaseException {
public:
    /**
     * @brief Name of the invalid state exception. Value is "InvalidStateException".
     */
    static const std::string INVALID_STATE_EXCEPTION;

    /**
     * @fn InvalidStateException(const std::string& _msg, int _code = 0);
     * @brief Constructor.
     * @param _msg Message of the exception.
     * @param _code Code of the exception.
     */
    InvalidStateException(const std::string& _msg, int _code = 0) : BaseException(INVALID_STATE_EXCEPTION, _msg, _code) {
    }
};

/**
 * @class RuntimeException
 * @brief The runtime exception.
 * @extends Utility::BaseException
 */
class DLLExport RuntimeException : public BaseException {
public:
    /**
     * @brief Name of the runtime exception. Value is "RuntimeException".
     */
    static const std::string RUNTIME_EXCEPTION;

    /**
     * @fn RuntimeException(const std::string& _msg, int _code = 0);
     * @brief Constructor.
     * @param _msg Message of the exception.
     * @param _code Code of the exception.
     */
    RuntimeException(const std::string& _msg, int _code = 0) : BaseException(RUNTIME_EXCEPTION, _msg, _code) {
    }
};

/**
 * @class NotFoundException
 * @brief The not found exception.
 * @extends Utility::BaseException
 */
class DLLExport NotFoundException : public BaseException {
public:
    /**
     * @brief Name of the not found exception. Value is "NotFoundException".
     */
    static const std::string NOTFOUND_EXCEPTION;

    /**
     * @fn NotFoundException(const std::string& _msg, int _code = 0);
     * @brief Constructor.
     * @param _msg Message of the exception.
     * @param _code Code of the exception.
     */
    NotFoundException(const std::string& _msg, int _code = 0) : BaseException(NOTFOUND_EXCEPTION, _msg, _code) {
    }
};

}


#endif // EXCEPTIONS_H
