/****************************************************************************
 *  Copyright (c) 2018 Oblo Living d.o.o
 *
 *  All Rights Reserved
 *
 *  P R O P R I E T A R Y    &    C O N F I D E N T I A L
 *
 *  -----------------------------------------------------
 *  http://www.obloliving.com/
 *  -----------------------------------------------------
 *
 * \project Oblo WISE SDK - SDK library
 * \file CppCurl.h
 * \brief
 * cURL wrapper class.
 *
 * @Author Igor Stefanovic
 *
 * \notes
 *
 * \history
 *****************************************************************************/
#ifndef CPP_CURL_H
#define CPP_CURL_H

#include "util/Common.h"
#include "util/Exceptions.h"

extern "C"
{
    #include <curl/curl.h>
}

/**
 * @namespace Utility
 * @brief Abstraction utility classes.
 */
namespace Utility {

/**
 * @class HTTPClient
 * @brief The implementation of HTTPClient with GET request.
 */
class HTTPClient {
public:
    /**
     * @fn HTTPClient(const std::string _url);
     * @brief Constructor.
     * @param _url URL to HTTP server.
     * @throws Utility::BaseException
     */
    HTTPClient(const std::string _url);

    /**
     * @fn ~HTTPClient();
     * @brief Destructor.
     */
    ~HTTPClient();

    /**
     * @fn std::string Fetch();
     * @brief Sends the HTTP GET request to URL.
     * @return HTTP response as a string.
     * @throws Utility::BaseException
     */
    std::string Fetch();
protected:
    /**
     * @fn void SetOptions();
     * @brief CURL options for HTTP GET request.
     * @throws Utility::BaseException
     */
    void SetOptions();

    /**
     * @fn void SendGetRequest();
     * @brief Perform CURL HTTP request.
     * @throws Utility::BaseException
     */
    void SendGetRequest();

    /**
     * @fn static size_t WriteDataCallback(void* _ptr, size_t _size, size_t _nmemb, void* _pInstance);
     * @brief Callback for HTTP response to be set to CURL library.
     * @param _ptr Pointer to buffer.
     * @param _size Size of buffer.
     * @param _nmemb Size.
     * @param _pInstance Instance which invokes the HTTP GET request.
     * @return Number of bytes of HTTP response.
     */
    static size_t WriteDataCallback(void* _ptr, size_t _size, size_t _nmemb, void* _pInstance);

    /**
     * @fn size_t WriteData(void* _ptr, size_t _size, size_t _nmemb);
     * @brief Function to be invoked on HTTP Response.
     * @param _ptr Pointer to buffer.
     * @param _size Size of buffer.
     * @param _nmemb Size.
     * @return Number of bytes of HTTP response.
     */
    size_t WriteData(void* _ptr, size_t _size, size_t _nmemb);
private:
    CURL* m_Handle;
    std::string m_URL;
    std::string m_Data;
};
}


#endif // CPP_CURL_H
