/****************************************************************************
 *  Copyright (c) 2018 Oblo Living d.o.o
 *
 *  All Rights Reserved
 *
 *  P R O P R I E T A R Y    &    C O N F I D E N T I A L
 *
 *  -----------------------------------------------------
 *  http://www.obloliving.com/
 *  -----------------------------------------------------
 *
 * \project Oblo WISE SDK - SDK library
 * \file JsonStruct.h
 * \brief
 * JZON abstraction class.
 *
 * @Author Igor Stefanovic
 *
 * \notes
 *
 * \history
 *****************************************************************************/
#ifndef JSON_STRUCT_H
#define JSON_STRUCT_H

#include "util/SharedPointer.h"
#include "util/Common.h"
#include "jzon/Jzon.h"

/**
 * @namespace Utility
 * @brief Abstraction utility classes.
 */
namespace Utility {

/**
 * @class JsonStruct
 * @brief Abstraction class for JSON objects.
 */
class DLLExport JsonStruct {
public:
    /**
     * @typedef SharedPointer<JsonStruct> Ptr;
     * @brief Shared pointer of JsonStruct object.
     */
    typedef SharedPointer<JsonStruct> Ptr;

    /**
     * @fn JsonStruct();
     * @brief Constructor.
     */
    JsonStruct();

    /**
     * @fn JsonStruct(const JsonStruct& _jsonStruct);
     * @brief Constructor.
     * @param _jsonStruct Copy object.
     */
    JsonStruct(const JsonStruct& _jsonStruct);

    /**
     * @fn JsonStruct(const std::string& _json);
     * @brief Constructor.
     * @param _json Stringified JSON object.
     */
    JsonStruct(const std::string& _json);

    /**
     * @fn JsonStruct(const Jzon::Node& _jzonNode);
     * @brief Constructor.
     * @param _jzonNode Jzon Node object.
     */
    JsonStruct(const Jzon::Node& _jzonNode);

    /**
     * @fn ~JsonStruct();
     * @brief Destructor.
     */
    ~JsonStruct();

    template<typename T>
    /**
     * @fn void Insert(const std::string& _key, const T& _value);
     * @brief Inserting a new value to JSON object.
     * @param _key Key of new value.
     * @param _value New value.
     */
    void Insert(const std::string& _key, const T& _value) {
        m_Node.add(_key, _value);
    }

    template<typename T>
    /**
     * @fn void Insert(const std::string& _key, const std::vector<T>& _array)
     * @brief Inserting a new array to JSON object.
     * @param _key Key of new array.
     * @param _array New array.
     */
    void Insert(const std::string& _key, const std::vector<T>& _array) {
        Jzon::Node array = Jzon::array();
        typename std::vector<T>::const_iterator itr = _array.begin();
        for (; itr != _array.end(); ++itr) {
            array.add(*itr);
        }
        m_Node.add(_key, array);
    }

    /**
     * @fn void Insert(const std::string &_key, const JsonStruct& _json);
     * @brief Inserting a new JsonStruct object to JSON object.
     * @param _key Key of new JsonStruct object.
     * @param _json New JsonStruct object.
     */
    void Insert(const std::string &_key, const JsonStruct& _json);

    /**
     * @fn JsonStruct::Ptr ExtractJson(const std::string& _key);
     * @brief Extracting a JsonStruct object from JSON object by key.
     * @param _key Key of JsonStruct object inside JSON object.
     * @return Shared pointer to JsonStruct object.
     */
    JsonStruct::Ptr ExtractJson(const std::string& _key);

    /**
     * @fn Jzon::Node Extract(const std::string& _key) const;
     * @brief Extracting a value from JSON object by key.
     * @param _key Key of object to be extracted.
     * @return Object which holds the value.
     */
    Jzon::Node Extract(const std::string& _key) const;

    /**
     * @fn std::vector<Jzon::Node> ExtractArray(const std::string &_key) const;
     * @brief Extracting an array from JSON object by key.
     * @param _key Key of array.
     * @return Vector representation of array.
     */
    std::vector<Jzon::Node> ExtractArray(const std::string &_key) const;

    /**
     * @fn bool IsEmpty() const;
     * @brief Indicates whether the JsonStruct object has elements.
     * @return Indication.
     */
    bool IsEmpty() const;

    /**
     * @fn void Parse(const std::string& _json);
     * @brief Parses the stringified JSON object into JsonStruct object.
     * @param _json Stringified JSON object.
     */
    void Parse(const std::string& _json);

    /**
     * @fn Jzon::Node GetNode() const;
     * @brief Getter for Node field.
     * @return Node field.
     */
    Jzon::Node GetNode() const;

    /**
     * @fn std::string ToString() const;
     * @brief Converts JsonStruct object to string.
     * @return Converted string.
     */
    std::string ToString() const;
private:
    Jzon::Node m_Node;
};

}

#endif // JSON_STRUCT_H
