/****************************************************************************
 *  Copyright (c) 2018 Oblo Living d.o.o
 *
 *  All Rights Reserved
 *
 *  P R O P R I E T A R Y    &    C O N F I D E N T I A L
 *
 *  -----------------------------------------------------
 *  http://www.obloliving.com/
 *  -----------------------------------------------------
 *
 * \project Oblo WISE SDK - SDK library
 * \file Logger.h
 * \brief
 * Logger interface.
 *
 * @Author Igor Stefanovic
 *
 * \notes
 *
 * \history
 *****************************************************************************/
#ifndef LOGGER_H
#define LOGGER_H

#include "util/Common.h"
#include "util/Mutex.h"
#include "util/ScopedLock.h"

/**
 * @namespace Utility
 * @brief Abstraction utility classes.
 */
namespace Utility {

#define LOG(lvl, msg) { \
    std::stringstream ss; \
    ss << msg; \
    this->On##lvl(ss.str());\
}

#define SSLOG(lvl, ss) { \
    this->On##lvl(ss.str());\
}

/**
 * @interface ILogListener
 * @brief Interface which is used to report logging messages to.
 */
class DLLExport ILogListener {
public:
    /**
     * @fn virtual void OnTrace(const std::string& _msg) = 0;
     * @brief Used to report trace messages. Trace messages have the lowest log level.
     * @param _msg Log message.
     */
    virtual void OnTrace(const std::string& _msg) = 0;

    /**
     * @fn virtual void OnDebug(const std::string& _msg) = 0;
     * @brief Used to report debug messages.
     * @param _msg Log message.
     */
    virtual void OnDebug(const std::string& _msg) = 0;

    /**
     * @fn virtual void OnInfo(const std::string& _msg) = 0;
     * @brief Used to report information messages.
     * @param _msg Log message.
     */
    virtual void OnInfo(const std::string& _msg) = 0;

    /**
     * @fn virtual void OnWarn(const std::string& _msg) = 0;
     * @brief Used to report warning messages.
     * @param _msg Log message.
     */
    virtual void OnWarn(const std::string& _msg) = 0;

    /**
     * @fn virtual void OnError(const std::string& _msg) = 0;
     * @brief Used to report error messages.
     * @param _msg Log message.
     */
    virtual void OnError(const std::string& _msg) = 0;

    /**
     * @fn virtual void OnFatal(const std::string& _msg) = 0;
     * @brief Used to report fatal messages.
     * @param _msg Log message.
     */
    virtual void OnFatal(const std::string& _msg) = 0;
};

/**
 * @class Logger
 * @brief The logger implementation.
 * @implements Utility::ILogListener
 */
class DLLExport Logger {
public:
    /**
     * @fn Logger();
     * @brief Constructor
     */
    Logger();

    /**
     * @fn void RegisterLogListener(ILogListener* _logListener);
     * @brief Registers log listener.
     * @param _logListener Log listener interface.
     */
    void RegisterLogListener(ILogListener* _logListener);

    /**
     * @fn void UnregisterLogListener();
     * @brief Cleans a log listener.
     */
    void UnregisterLogListener();

    /**
     * @fn void OnTrace(const std::string& _msg);
     * @brief Handles trace messages.
     * @param _msg Message.
     */
    void OnTrace(const std::string& _msg);

    /**
     * @fn void OnDebug(const std::string& _msg);
     * @brief Handles debug messages.
     * @param _msg Message.
     */
    void OnDebug(const std::string& _msg);

    /**
     * @fn void OnInfo(const std::string& _msg);
     * @brief Handles information messages.
     * @param _msg Message.
     */
    void OnInfo(const std::string& _msg);

    /**
     * @fn void OnWarn(const std::string& _msg);
     * @brief Handles warning messages.
     * @param _msg Message.
     */
    void OnWarn(const std::string& _msg);

    /**
     * @fn void OnError(const std::string& _msg);
     * @brief Handles error messages.
     * @param _msg Message.
     */
    void OnError(const std::string& _msg);

    /**
     * @fn void OnFatal(const std::string& _msg);
     * @brief Handles fatal messages.
     * @param _msg Message.
     */
    void OnFatal(const std::string& _msg);
private:
    ILogListener* m_LogListener;
};

}



#endif // LOGGER_H
