/****************************************************************************
 *  Copyright (c) 2018 Oblo Living d.o.o
 *
 *  All Rights Reserved
 *
 *  P R O P R I E T A R Y    &    C O N F I D E N T I A L
 *
 *  -----------------------------------------------------
 *  http://www.obloliving.com/
 *  -----------------------------------------------------
 *
 * \project Oblo WISE SDK - SDK library
 * \file SharedPointer.h
 * \brief
 * Utility Shared pointer class.
 *
 * @Author Igor Stefanovic
 *
 * \notes
 *
 * \history
 *****************************************************************************/
#ifndef SHARED_PTR_H
#define SHARED_PTR_H

#include "util/Common.h"

namespace Utility {

/**
 * @class ReferenceCounter
 * @brief Counter of the pointer references.
 */
class ReferenceCounter {
public:
    /**
     * @fn ReferenceCounter();
     * @brief Constructor.
     */
    ReferenceCounter() : m_Counter(0) {}

    /**
     * @fn void AddRef();
     * @brief Atomically increase the counter.
     */
    void AddRef() {
        __sync_add_and_fetch(&m_Counter, 1);;
    }

    /**
     * @fn unsigned int Release();
     * @brief Atomically decrease the counter.
     * @return Value of the counter after decrease.
     */
    unsigned int Release() {
        return __sync_sub_and_fetch(&m_Counter, 1);
    }
private:
    unsigned int m_Counter;
};

template <class C, class RC = ReferenceCounter>
/**
 * @class SharedPointer
 * @brief Implementation of shared pointer functionality.
 */
class DLLExport SharedPointer {
public:
    /**
     * @fn SharedPointer();
     * @brief Constructor.
     */
    SharedPointer() :
        m_DataPtr(0),
        m_Counter(new RC()) {
        m_Counter->AddRef();
    }

    /**
     * @fn SharedPointer(C* _ptr);
     * @brief Constructor.
     * @param _ptr Pointer to the data.
     */
    SharedPointer(C* _ptr) :
        m_DataPtr(_ptr),
        m_Counter(new RC()) {
        m_Counter->AddRef();
    }

    /**
     * @fn SharedPointer(const SharedPointer<C, RC>& _sharedPtr);
     * @brief Constructor.
     * @param _sharedPtr Copy object.
     */
    SharedPointer(const SharedPointer<C, RC>& _sharedPtr) :
        m_DataPtr(_sharedPtr.m_DataPtr),
        m_Counter(_sharedPtr.m_Counter) {
        m_Counter->AddRef();
    }

    template<class Other>
    /**
     * @fn SharedPointer(const SharedPointer<Other, RC>& _sharedPtr)
     * @brief Constructor.
     * @param _sharedPtr Copy object with different pointer data.
     */
    SharedPointer(const SharedPointer<Other, RC>& _sharedPtr) :
        m_DataPtr(const_cast<Other*>(_sharedPtr.Get())),
        m_Counter(const_cast<RC*>(_sharedPtr.GetCounter())) {
        m_Counter->AddRef();
    }

    /**
     * @fn ~SharedPointer();
     * @brief Destructor.
     */
    ~SharedPointer() {
        if (m_Counter->Release() == 0) {
            delete m_Counter;
            m_Counter = 0;
            delete m_DataPtr;
            m_DataPtr = 0;
        }
    }

    /**
     * @fn C* operator -> ();
     * @brief Dereference.
     * @return Dereferenced data pointer.
     */
    C* operator -> () {
        return Deref();
    }

    /**
     * @fn const C* operator -> () const;
     * @brief Dereference.
     * @return Const dereferenced data pointer.
     */
    const C* operator -> () const {
        return Deref();
    }

    /**
     * @fn C& operator * ();
     * @brief Dereference.
     * @return Reference of dereferenced data pointer.
     */
    C& operator * () {
        return *Deref();
    }

    /**
     * @fn const C& operator * () const;
     * @brief Dereference.
     * @return Const reference of data pointer.
     */
    const C& operator * () const {
        return *Deref();
    }

    /**
     * @fn  operator C* ();
     * @brief Returns the data pointer.
     * @return Data pointer.
     */
    operator C* () {
        return m_DataPtr;
    }

    /**
     * @fn operator const C* () const;
     * @brief Returns the const data pointer.
     * @return Const data pointer.
     */
    operator const C* () const {
        return m_DataPtr;
    }

    /**
     * @fn const C* Get() const;
     * @brief Getter for data pointer.
     * @return Const data pointer.
     */
    const C* Get() const {
        return m_DataPtr;
    }

    /**
     * @fn C* Get();
     * @brief Getter for data pointer.
     * @return Data pointer.
     */
    C* Get() {
        return m_DataPtr;
    }

    /**
     * @fn const RC* GetCounter() const;
     * @brief Getter for counter pointer.
     * @return Const counter pointer.
     */
    const RC* GetCounter() const {
        return m_Counter;
    }

    /**
     * @fn RC* GetCounter();
     * @brief Getter for counter pointer.
     * @return Counter pointer.
     */
    RC* GetCounter() {
        return m_Counter;
    }

    /**
     * @fn SharedPointer<C, RC>& operator = (const SharedPointer<C, RC>& _sharedPtr);
     * @brief Assignment operator.
     * @param _sharedPtr Shared pointer object to be assigned.
     * @return New object after assignment.
     */
    SharedPointer<C, RC>& operator = (const SharedPointer<C, RC>& _sharedPtr) {
        if (this != &_sharedPtr) {
            if (m_Counter->Release() == 0) {
                delete m_DataPtr;
                delete m_Counter;
            }

            m_DataPtr = _sharedPtr.m_DataPtr;
            m_Counter = _sharedPtr.m_Counter;
            m_Counter->AddRef();
        }

        return *this;
    }

    /**
     * @fn bool IsNull() const;
     * @brief Checks whether the data pointer is NULL.
     * @return Indication.
     */
    bool IsNull() const {
        return m_DataPtr == 0;
    }
private:
    C* Deref() const {
        return m_DataPtr;
    }
    C* m_DataPtr;
    RC* m_Counter;
};

}

#endif // SHARED_PTR_H
