/****************************************************************************
 *  Copyright (c) 2018 Oblo Living d.o.o
 *
 *  All Rights Reserved
 *
 *  P R O P R I E T A R Y    &    C O N F I D E N T I A L
 *
 *  -----------------------------------------------------
 *  http://www.obloliving.com/
 *  -----------------------------------------------------
 *
 * \project Oblo WISE SDK - SDK library
 * \file Thread.h
 * \brief
 * Utility Thread abstraction class.
 *
 * @Author Igor Stefanovic
 *
 * \notes
 *
 * \history
 *****************************************************************************/
#ifndef THREAD_H
#define THREAD_H

#include "util/Common.h"
#include "util/SharedPointer.h"

/**
 * @namespace Utility
 * @brief Abstraction utility classes.
 */
namespace Utility {

/**
 * @class Runnable
 * @brief Runnable class.
 */
class DLLExport Runnable {
public:
    /**
     * @typedef SharedPointer<Runnable> Ptr;
     * @brief Shared pointer to Runnable object.
     */
    typedef SharedPointer<Runnable> Ptr;

    /**
     * @fn Runnable();
     * @brief Constructor.
     */
    Runnable() {}

    /**
     * @fn virtual ~Runnable();
     * @brief Destructor.
     */
    virtual ~Runnable() {}

    /**
     * @fn virtual void Run() = 0;
     * @brief Function of the thread.
     */
    virtual void Run() = 0;
};

/**
 * @class RunnableHolder
 * @brief The holder class for Runnable object.
 * @extends Utility::Runnable
 */
class RunnableHolder : public Runnable {
public:
    /**
     * @fn RunnableHolder(Runnable& _target);
     * @brief Constructor.
     * @param _target Runnable object.
     */
    RunnableHolder(Runnable& _target) : m_Target(_target) {}

    /**
     * @fn ~RunnableHolder()
     * @brief Destrictor.
     */
    ~RunnableHolder() {}

    /**
     * @fn void Run();
     * @brief Thread function.
     */
    void Run();
private:
    Runnable& m_Target;
};

/**
 * @class Thread
 * @brief Abstraction thread class.
 */
class DLLExport Thread {
public:
    /**
     * @fn Thread();
     * @brief Constructor.
     */
    Thread();

    /**
     * @fn virtual ~Thread();
     * @brief Destructor.
     */
    virtual ~Thread();

    /**
     * @fn void Start(Runnable& _runnable);
     * @brief Starts the thread function.
     * @param _runnable Runnable which holds the thread function.
     */
    void Start(Runnable& _runnable);

    /**
     * @fn void Join();
     * @brief Used to join running thread.
     */
    void Join();

    /**
     * @fn pthread_t GetID() const;
     * @brief Getter for ID field.
     * @return ID field.
     */
    pthread_t GetID() const;
private:
    void CreateThread();
    static void* ThreadFunction(void* ptr);
    pthread_t m_TID;
    SharedPointer<Runnable> m_Runnable;
    bool m_Terminated;
};

}


#endif
