/****************************************************************************
 *  Copyright (c) 2018 Oblo Living d.o.o
 *
 *  All Rights Reserved
 *
 *  P R O P R I E T A R Y    &    C O N F I D E N T I A L
 *
 *  -----------------------------------------------------
 *  http://www.obloliving.com/
 *  -----------------------------------------------------
 *
 * \project Oblo WISE SDK - SDK library
 * \file Timer.h
 * \brief
 * Utility Timer class.
 *
 * @Author Igor Stefanovic
 *
 * \notes
 *
 * \history
 *****************************************************************************/
#ifndef TIMER_H
#define TIMER_H

#include "util/Common.h"
#include "util/Mutex.h"
#include "util/Thread.h"
#include "util/ScopedLock.h"
#include "util/SynchronizedObject.h"

/**
 * @namespace Utility
 * @brief Abstraction utility classes.
 */
namespace Utility {

class Timer;

/**
 * @class AbstractCallback
 * @brief The AbstractCallback class
 */
class DLLExport AbstractCallback {
public:
    /**
     * @fn AbstractCallback()
     * @brief Constructor.
     */
    AbstractCallback() {}

    /**
     * @fn AbstractCallback(const AbstractCallback& _cb)
     * @brief Constructor.
     * @param _cb Copy object.
     */
    AbstractCallback(const AbstractCallback& _cb) {}

    /**
     * @fn virtual ~AbstractCallback()
     * @brief Destructor
     */
    virtual ~AbstractCallback() {}

    /**
     * @fn virtual void Invoke(Timer& _timer) const = 0;
     * @brief Invoke timer function.
     * @param _timer Timer.
     */
    virtual void Invoke(Timer& _timer) const = 0;

    /**
     * @fn virtual AbstractCallback* Clone() const = 0;
     * @brief Cloning object.
     * @return Cloned object.
     */
    virtual AbstractCallback* Clone() const = 0;
};

template <class C>
/**
 * @class Callback
 * @brief The Callback class
 * @extends Utility::AbstractCallback
 */
class DLLExport Callback : public AbstractCallback {
public:
    /**
     * @typedef void (C::*Cb)(Timer&);
     * @brief Callback function.
     */
    typedef void (C::*Cb)(Timer&);
    /**
     * @fn Callback(C& _object, Cb _method);
     * @brief Constructor.
     * @param _object Object field.
     * @param _method Method field.
     */
    Callback(C& _object, Cb _method) : m_Object(&_object), m_Method(_method) {
    }

    /**
     * @fn Callback(const Callback& _callback)
     * @brief Constructor
     * @param _callback Copy object.
     */
    Callback(const Callback& _callback): m_Object(_callback.m_Object), m_Method(_callback.m_Method) {
    }

    /**
     * @fn ~Callback()
     * @brief Destructor.
     */
    ~Callback() {}

    /**
     * @fn void Invoke(Timer& _timer) const;
     * @brief Function which invokes registered callback.
     * @param _timer Timer.
     */
    void Invoke(Timer& _timer) const {
        (m_Object->*m_Method)(_timer);
    }

    /**
     * @fn AbstractCallback* Clone() const;
     * @brief Cloning Callback object.
     * @return Cloned object.
     */
    AbstractCallback* Clone() const {
        return new Callback(*this);
    }
private:
    C* m_Object;
    Cb m_Method;
};

/**
 * @class Timer
 * @brief The Timer class
 * @extends Utility::Runnable
 */
class DLLExport Timer : Runnable {
public:
    /**
     * @fn Timer(long _timerInterval, bool _nowStart = true, int _repeatTimes = -1);
     * @brief Constructor.
     * @param _timerInterval Timer interval when callback function will be invoked.
     * @param _nowStart Defines whether callback function will be invoked immediately.
     * @param _repeatTimes Repeat times. If its set to -1, the callback function will be invoked without stopping.
     */
    Timer(long _timerInterval, bool _nowStart = true, int _repeatTimes = -1);

    /**
     * @fn ~Timer();
     * @brief Destructor.
     */
    ~Timer();

    /**
     * @fn void StartTimer(const AbstractCallback& _method);
     * @brief Starting the timer.
     * @param _method Method which will be invoked on timer interval.
     */
    void StartTimer(const AbstractCallback& _method);

    /**
     * @fn void StopTimer();
     * @brief Stopping the timer.
     * @note To stop the timer from timer callback, call RestartTimer(0)
     */
    void StopTimer();

    /**
     * @fn void RestartTimer(long _timerInterval, bool _nowStart = true, int _repeatTimes = -1);
     * @brief Setting new parameters to timer which actually restarts the timer.
     * @param _timerInterval New timer interval.
     * @param _nowStart Defines whether callback function will be invoked immediately.
     * @param _repeatTimes Repeat times. If its set to -1, the callback function will be invoked without stopping.
     */
    void RestartTimer(long _timerInterval, bool _nowStart = true, int _repeatTimes = -1);

    /**
     * @fn void SkipInvocation(bool _skip);
     * @brief Defines whether the timer should skip invocation.
     * @param _skip Indication.
     */
    void SkipInvocation(bool _skip);

    /**
     * @fn bool ShouldSkip();
     * @brief Defines whether the skip indicator is set.
     * @return Skip value.
     */
    bool ShouldSkip();

    /**
     * @fn void SetTimeout(long _timerInterval);
     * @brief Setter for timer interval field.
     * @param _timerInterval Timer interval to be set.
     */
    void SetTimeout(long _timerInterval);

    /**
     * @fn void SetRepeatTimes(int _repeatTimes);
     * @brief Setter for repeat times field.
     * @param _repeatTimes Repeat times to be set.
     */
    void SetRepeatTimes(int _repeatTimes);
private:
    Mutex m_Mutex;
    SynchronizedObject m_Sync;
    Thread m_Thread;
    long m_TimerInterval;
    bool m_ShouldRun;
    bool m_Terminated;
    bool m_NowStart;
    int m_RepeatTimes;
    bool m_Skip;
    AbstractCallback* m_Callback;
    void Run();
};



}

#endif // TIMER_H
