#include "ipdevice/Device.h"

using namespace Utility;
using namespace SSDP;
using namespace Messenger;
using namespace IPDevice;

const std::string Device::DEFAULT_USERNAME = "client";
const std::string Device::DEFAULT_PASSWORD = "tmpclient";
const unsigned int Device::DEFAULT_INCLUSION_TIMEOUT = 180;
const std::string Device::TOPIC_SEPARATOR = "/";

Device::ConnectionResurrector::ConnectionResurrector(Device &_device) : m_ShouldRun(false), m_Device(_device), m_Terminated(true) {
}

Device::ConnectionResurrector::~ConnectionResurrector() {
    if (!m_Terminated) {
        m_ShouldRun = false;
        m_Device.m_ConnSync.Notify();
        m_Thread.Join();
    }
}

void Device::ConnectionResurrector::Start() {
    if (m_Terminated) {
        m_ShouldRun = true;
        m_Thread.Start(*this);
        m_Terminated = false;
    }
}

void Device::ConnectionResurrector::Stop() {
    m_ShouldRun = false;
    if (!m_Terminated) {
        m_Device.m_ConnSync.Notify();
        m_Thread.Join();
        m_Terminated = true;
    }
}

void Device::ConnectionResurrector::Run() {
    while (m_ShouldRun) {
        m_Device.m_ConnSync.Wait();
        if (!m_ShouldRun) {
            break;
        }
        m_Device.Destroy();
        try {
            m_Device.StartReconnection();
        } catch (const BaseException& exc) {
        }
    }
}

Device::Device() : RequestHandler(this),
    m_DiscoveryObj(),
    m_Messenger(NULL),
    m_Stopped(true),
    m_DeviceInfo(),
    m_InclusionResult(),
    m_DeviceExcludedListener(NULL),
    m_ConnectedListener(NULL),
    m_TimeChangeListener(NULL),
    m_InclusionInitiated(false),
    m_ConnResurector(*this),
    m_InclusionTimeout(DEFAULT_INCLUSION_TIMEOUT),
    m_InclusionTimerExpired(false),
    m_Secured(true) {

    OpenSSL_add_all_algorithms(); // For encryption

    AddRequestHandler("getDeviceInfo", &Device::GetDeviceInfo);
    AddRequestHandler("getState", &Device::GetState);
    AddRequestHandler("setTime", &Device::SetTime);
    AddRequestHandler("adjustTime", &Device::AdjustTime);
    AddRequestHandler("verifyIdentity", &Device::VerifyIdentity);
    AddRequestHandler("getGroupList", &Device::GetGroupList);
    AddRequestHandler("getServiceList", &Device::GetServiceList);
    AddRequestHandler("getService", &Device::GetService);
    AddRequestHandler("joinConfirmation", &Device::JoinConfirmation);

    AddServiceRequestHandler("getPropertyList", &Device::GetPropertyList);
    AddServiceRequestHandler("setPropertyValue", &Device::SetPropertyValue);
    AddServiceRequestHandler("getPropertyValue", &Device::GetPropertyValue);
    AddServiceRequestHandler("getCommandList", &Device::GetCommandList);
    AddServiceRequestHandler("executeCommand", &Device::ExecuteCommand);

    m_ConnResurector.Start();
    m_InclusionTimer = new Timer(m_InclusionTimeout * 1000, false);
    m_DescProviderUrl = "";
}

Device::~Device() {
    m_ConnResurector.Stop();
    m_InclusionTimer->StopTimer();
    SAFE_DELETE_O(m_InclusionTimer);
    EVP_cleanup();
    StopDiscovery();
    Destroy();
    std::map<unsigned int, Group*>::iterator itr = m_GroupList.begin();
    for (; itr != m_GroupList.end(); ++itr) {
        SAFE_DELETE_O(itr->second);
    }
}

InclusionInformation Device::StartInclusion() {
    LOG(Info, "Inclusion started");
    m_InclusionInitiated = true;
    m_InclusionTimerExpired = false;
    m_InclusionTimer->StartTimer(Callback<Device>(*this, &Device::InclusionTimerExpired));

    try {
        if (m_DescProviderUrl.empty()) {
            m_SyncSSDP.Notify();
            StartDiscoverySearch();
            m_InclusionInitiated = false;
            WaitForHub(m_InclusionTimeout);
        } else {
            GetDescription(m_DescProviderUrl);
        }

        if (!m_InclusionTimerExpired) {
            InitialConnect();
            SendInclusionRequest();
            WaitForJoinConfirmation();

            SendChallengeReq();
            SendCredentialsReq();

            Destroy();
            FinalConnect();

            LOG(Info, "Inclusion finished");
            m_InclusionTimer->StopTimer();
        }
    } catch (const TimeoutException& exc) {
        std::stringstream ss;
        ss << exc.GetMessage() << " <c=" << exc.GetCode() << ">";
        SSLOG(Error, ss);
        Destroy();
        throw TimeoutException(exc);
    } catch (const ProtocolException& exc) {
        std::stringstream ss;
        ss << exc.GetMessage() << " <c=" << exc.GetCode() << ">";
        SSLOG(Error, ss);
        Destroy();
        throw ProtocolException(exc);
    } catch (const BaseException& exc) {
        std::stringstream ss;
        ss << exc.GetMessage() << " <c=" << exc.GetCode() << ">";
        SSLOG(Error, ss);
        Destroy();
        throw BaseException(exc);
    }
    return m_InclusionResult;
}

void Device::StartReconnection() {
    LOG(Info, "Reconnection started");

    bool finalFailed = false;
    bool notAuthorize = false;
    int retry = 3;
    for (int i = 0; i < retry; i++) {
        try {
            FinalConnect();
            finalFailed = false;
            notAuthorize = false;
            break;
        } catch (const ProtocolException& exc) {
            if (exc.GetCode() == MQTT_NOT_AUTHORIZED) {
                LOG(Info, "Not authorized to connect");
                notAuthorize = true;
                break;
            } else {
                finalFailed = true;
                sleep(5);
            }
        } catch (const BaseException& exc) {
            LOG(Error, exc.GetMessage());
            return;
        }
    }

    if (notAuthorize) {
        try {
            InitialConnect();
            SendChallengeReq();
            SendCredentialsReq();
            Destroy();
            FinalConnect();
        } catch (const ProtocolException& exc) {
            if (exc.GetCode() == WISE_NOT_AUTHORIZED) {
                LOG(Info, "Excluded from the network");
                Destroy();
                m_DeviceExcludedListener ? m_DeviceExcludedListener->OnDeviceExcluded() : void();
            } else {
                finalFailed = true;
            }
        } catch (const BaseException& exc) {
            LOG(Error, exc.GetMessage());
            return;
        }
    }

    if (finalFailed) {
        StartDiscoveryListen();
        WaitForHub();
        if (!m_InclusionInitiated) {
            StartReconnection();
        } else {
            return;
        }
    }
}

void Device::StartDiscoveryListen() {
    m_Stopped = false;
    m_DiscoveryObj.RegisterListener(this);
    m_DiscoveryObj.StartListen(m_InclusionResult.GetHub().GetHubID());
}

void Device::StartDiscoverySearch() {
    m_Stopped = false;
    m_DiscoveryObj.RegisterListener(this);
    m_DiscoveryObj.StartSearch();
}

void Device::OnHubDetected(const SSDP::HubHolder &_hub) {
    m_InclusionResult.SetHub(_hub);
    m_InclusionResult.SetHubTag(_hub.GetDomainID() + TOPIC_SEPARATOR + _hub.GetHomeID() + TOPIC_SEPARATOR + "hub/" + _hub.GetHubID());
    m_SyncSSDP.Notify();
}

void Device::StopDiscovery() {
    if (!m_Stopped) {
        m_SyncSSDP.Notify();
        m_DiscoveryObj.UnregisterListener();
        m_DiscoveryObj.Stop();
        m_Stopped = true;
    }
}

void Device::GetDescription(const std::string& _url) {
    std::stringstream out;
    std::stringstream ss;
    ss << _url << "/wise" << "/" << m_DeviceInfo.GetManufacturer() << "/" << m_DeviceInfo.GetModel() << "/" << m_DeviceInfo.GetIdentifier();
    LOG(Info, "Get description from provider <u=" + ss.str() + ">");
    try {
        HTTPClient httpClient(ss.str());
        out << httpClient.Fetch();
    } catch (const BaseException& exc) {
        LOG(Error, exc.GetMessage());
    }

    JsonStruct body(out.str());
    SSLOG(Info, out);
    HubHolder hub(body.Extract("domainID").toString(), body.Extract("homeID").toString(), body.Extract("hubID").toString(), body.Extract("brokerURL").toString());
    m_InclusionResult.SetHub(hub);
    m_InclusionResult.SetHubTag(hub.GetDomainID() + TOPIC_SEPARATOR + hub.GetHomeID() + TOPIC_SEPARATOR + "hub/" + hub.GetHubID());
}

void Device::InitialConnect() {
    LOG(Debug, "Connection initiated <u=" + DEFAULT_USERNAME + "> <p=" + DEFAULT_PASSWORD + ">");
    std::stringstream url;
    if (m_Secured) {
        url << "ssl://";
    } else {
        url << "tcp://";
    }
    url << m_InclusionResult.GetHub().GetLocation();

    std::string tag = m_InclusionResult.GetHub().GetDomainID() + TOPIC_SEPARATOR + m_InclusionResult.GetHub().GetHomeID() + TOPIC_SEPARATOR + "ipd/" + m_DeviceInfo.GetIdentifier();
    m_Messenger = new CMessenger("ipd/" + m_DeviceInfo.GetIdentifier(), tag, DEFAULT_USERNAME, DEFAULT_PASSWORD, url.str());

    m_Secured ? m_Messenger->SetSecureConfiguration(m_CertFile, "", "", "") : void();
    m_Messenger->RegisterResponder(this);
    m_Messenger->RegisterConnectionListener(this);

    try {
        m_Messenger->Connect();
        LOG(Info, "Successfully connected on <=" + url.str() + ">");
    } catch (const ProtocolException& exc) {
        Destroy();
        throw ProtocolException(exc);
    }
}

void Device::FinalConnect() {
    LOG(Debug, "Connection initiated <u=" + m_InclusionResult.GetUsername() + "> <p=" + m_InclusionResult.GetPassword() + ">");
    std::stringstream url;
    if (m_Secured) {
        url << "ssl://";
    } else {
        url << "tcp://";
    }
    url << m_InclusionResult.GetHub().GetLocation();

    std::string tag = m_InclusionResult.GetHub().GetDomainID() + TOPIC_SEPARATOR + m_InclusionResult.GetHub().GetHomeID() + TOPIC_SEPARATOR + "ipd/" + m_DeviceInfo.GetIdentifier();
    m_Messenger = new CMessenger("ipd/" + m_DeviceInfo.GetIdentifier(), tag, m_InclusionResult.GetUsername(), m_InclusionResult.GetPassword(), url.str());

    m_Secured ? m_Messenger->SetSecureConfiguration(m_CertFile, "", "", "") : void();
    m_Messenger->RegisterResponder(this);
    m_Messenger->RegisterConnectionListener(this);

    try {
        m_Messenger->Connect();
        LOG(Info, "Successfully connected on <u=" + url.str() + ">");

        if (m_ConnectedListener != NULL) {
            m_ConnectedListener->OnConnected(m_InclusionResult);
        }
    } catch (const ProtocolException& exc) {
        Destroy();
        LOG(Error, exc.GetMessage());
        throw ProtocolException(exc);
    } catch (const BaseException& exc) {
        Destroy();
        LOG(Error, exc.GetMessage());
        throw BaseException(exc);
    }
}

void Device::SendInclusionRequest() {
    RequestMessage req;
    req.SetName("inclusionReq");
    JsonStruct::Ptr params = new JsonStruct();

    params->Insert("id", m_DeviceInfo.GetIdentifier());
    params->Insert("manufacturer", m_DeviceInfo.GetManufacturer());
    params->Insert("model", m_DeviceInfo.GetModel());
    params->Insert("version", m_DeviceInfo.GetVersion());
    params->Insert("versionNumber", m_DeviceInfo.GetVersionNumber());
    params->Insert("protocol", 1);
    params->Insert("mac", m_DeviceInfo.GetMAC());
    params->Insert("name", m_DeviceInfo.GetName());

    req.SetParams(params);

    User::Ptr user = m_Messenger->GetUser(m_InclusionResult.GetHubTag());

    try {
        ResponseMessage::Ptr result = user->Request(req);
        if (!result.IsNull()) {
            if (result->GetCode() == 0) {
                std::string temp = result->GetParams()->Extract("inclusionID").toString();
                if (!temp.empty()) {
                    m_InclusionResult.SetInclusionID(temp);
                    LOG(Info, "Inclusion ID <i=" + temp + ">");
                }
            } else {
                throw ProtocolException(result->GetMessage(), result->GetCode());
            }
        }
    } catch (const BaseException& exc) {
        throw ProtocolException(exc.GetMessage(), exc.GetCode());
    }
}

void Device::SendChallengeReq() {
    RequestMessage req;
    req.SetName("challengeReq");

    User::Ptr user = m_Messenger->GetUser(m_InclusionResult.GetHubTag());

    for (unsigned int retry = 0; retry < m_ChallengeParameters.GetChallengeRetry(); ++retry) {
        try {
            ResponseMessage::Ptr result = user->Request(req, m_ChallengeParameters.GetChallengeTimeout());
            if (!result.IsNull()) {
                if (result->GetCode() == 0) {
                    std::string temp = result->GetParams()->Extract("challenge").toString();
                    if (!temp.empty()) {
                        m_ChallengeParameters.SetChallengeCode(temp);
                        LOG(Info, "Challenge code <c=" + temp +">");
                        break;
                    }
                } else {
                    throw ProtocolException(result->GetMessage(), result->GetCode());
                }
            }
        } catch (const TimeoutException& exc) {
            if (retry == 2) {
                throw ProtocolException(exc.GetMessage(), exc.GetCode());
            }
        } catch (const BaseException& exc) {
            throw ProtocolException(exc.GetMessage(), exc.GetCode());
        }
    }
}

void Device::SendCredentialsReq() {
    // SHA1
    LOG(Trace, OPENSSL_VERSION_TEXT);
    const EVP_MD* md = EVP_get_digestbyname("SHA1");
    unsigned char sha1Output[20];
    unsigned int sha1OutLen;
    std::string key;
    if (md != NULL) {
        EVP_MD_CTX *mdctx = EVP_MD_CTX_create();
        EVP_DigestInit(mdctx, md);
        EVP_DigestUpdate(mdctx, m_InclusionResult.GetInclusionID().c_str(), m_InclusionResult.GetInclusionID().length());
        EVP_DigestUpdate(mdctx, m_DeviceInfo.GetIdentifier().c_str(), m_DeviceInfo.GetIdentifier().length());
        EVP_DigestFinal_ex(mdctx, sha1Output, &sha1OutLen);
        std::ostringstream outss;
        for (int i = 0; i < 8; i++) {
            if ((int) sha1Output[i] < 16) {
                outss << "0";
            }
            outss << std::hex << (int) sha1Output[i];
        }
        key = outss.str();
        EVP_MD_CTX_destroy(mdctx);
    }

    LOG(Trace, "Key generated <k=" + key + ">");

    // AES with padding
    unsigned char aesOutput[32];
    int aesOutLen;
    const EVP_CIPHER* cipher = EVP_get_cipherbyname("aes-128-ecb");
    if (cipher != NULL) {
        EVP_CIPHER_CTX* cictx = EVP_CIPHER_CTX_new();
        EVP_CipherInit(cictx, cipher, (unsigned char* ) key.c_str(), 0, 1);
        std::string challengeCode = m_ChallengeParameters.GetChallengeCode();
        EVP_CipherUpdate(cictx, aesOutput, &aesOutLen, (unsigned char*) challengeCode.c_str(), challengeCode.length());
        EVP_CipherFinal_ex(cictx, aesOutput, &aesOutLen);
        aesOutput[aesOutLen] = '\0';
        EVP_CIPHER_CTX_free(cictx);
    }
    // Base64
    unsigned char base64Output[32] = {};
    int base64OutLen = EVP_EncodeBlock(base64Output, aesOutput, aesOutLen);
    base64Output[base64OutLen] = '\0';

    std::stringstream ss;
    ss << "Authorization code after crypting <c=" << (char*) base64Output << ">";
    SSLOG(Trace, ss);

    RequestMessage req;
    req.SetName("credentialsReq");
    JsonStruct::Ptr params = new JsonStruct();
    params->Insert("auth", (char*) base64Output);

    req.SetParams(params);
    User::Ptr user = m_Messenger->GetUser(m_InclusionResult.GetHubTag());

    try {
        ResponseMessage::Ptr result = user->Request(req);
        if (!result.IsNull()) {
            if (result->GetCode() == 0) {
                std::string tempUser = result->GetParams()->Extract("username").toString();
                std::string tempPass = result->GetParams()->Extract("password").toString();
                if (!tempPass.empty() || !tempUser.empty()) {
                    m_InclusionResult.SetUsername(tempUser);
                    m_InclusionResult.SetPassword(tempPass);
                    LOG(Debug, "Successfully finished authorization <u=" + tempUser + " p=" + tempPass + ">");
                }
            } else {
                throw ProtocolException(result->GetMessage(), result->GetCode());
            }
        }
    } catch (const BaseException& exc) {
        throw ProtocolException(exc.GetMessage());
    }
}

void Device::ActivateServices() {
    std::map<unsigned int, Group*>::const_iterator itr = m_GroupList.begin();
    Group::ConstIterator constItr;
    for (; itr != m_GroupList.end(); ++itr) {
        for (constItr = itr->second->Begin(); constItr != itr->second->End(); ++constItr) {
            m_Messenger->ActivateService(itr->first, constItr->first);
        }
    }
    LOG(Info, "Services activated");
}

void Device::WaitForHub(long _timeout) {
    try {
        if (_timeout == -1) {
            m_SyncSSDP.Wait();
        } else {
            m_SyncSSDP.Wait(_timeout * 1000);
        }
        StopDiscovery();
    } catch (const TimeoutException& exc) {
        StopDiscovery();
        throw TimeoutException(exc);
    }
}

void Device::WaitForJoinConfirmation() {
    try {
        m_SyncJoin.Wait(5 * 1000);
    } catch (const TimeoutException& exc) {
        throw TimeoutException(exc);
    }
}

void Device::AddService(unsigned int _groupId, BaseService* _service) {
    std::map<unsigned int, Group*>::iterator itr = m_GroupList.find(_groupId);
    if (itr == m_GroupList.end()) {
        Group* group = new Group(_groupId);
        group->AddService(_service);
        m_GroupList[_groupId] = group;
    } else {
        itr->second->AddService(_service);
    }
}

void Device::SetCertFile(const std::string &_cert) {
    m_CertFile = _cert;
}

void Device::SetKeyFile(const std::string &_key) {
    m_KeyFile = _key;
}

void Device::SetKeyPasscode(const std::string& _passcode) {
    m_KeyPasscode = _passcode;
}

void Device::SetSecure(bool _secure) {
    m_Secured = _secure;
}

void Device::SetChallengeParameters(ChallengeParameters _challengeParams) {
    m_ChallengeParameters = _challengeParams;
}

void Device::SetDeviceInfomation(DeviceInformation _deviceInformation) {
    m_DeviceInfo = _deviceInformation;
}

void Device::SetInclusionInformation(InclusionInformation _prevInclusionInformation) {
    m_InclusionResult = _prevInclusionInformation;
}

bool Device::IsInclusionInfoEmpty() const {
    return m_InclusionResult.IsEmpty();
}

void Device::SetInclusionTimeout(long _inclusionTimeout) {
    m_InclusionTimeout = _inclusionTimeout;
    m_InclusionTimer->SetTimeout(_inclusionTimeout * 1000);
}

void Device::SetSearchInterval(long _searchInterval) {
    m_DiscoveryObj.SetSearchInterval(_searchInterval * 1000);
}

std::string Device::GetVersion() {
    return WISE_VERSION;
}

std::string Device::GetPlatform() {
    return WISE_PLATFORM;
}

std::string Device::GetType() {
    return WISE_TYPE;
}

void Device::SetDescriptionProviderUrl(const std::string& _url) {
    m_DescProviderUrl = _url;
}

const Messenger::ResponseMessage::Ptr Device::Respond(const Messenger::RequestMessage::Ptr _request) {
    ResponseMessage::Ptr _response = new ResponseMessage();
    CallRequestHandler(_request->GetName(), _request, _response);
    return _response;
}

const ResponseMessage::Ptr Device::ServiceRespond(unsigned int _groupId, const std::string& _serviceType, const RequestMessage::Ptr _request) {
    ResponseMessage::Ptr _response = new ResponseMessage();
    CallServiceRequestHandler(_groupId, _serviceType, _request->GetName(), _request, _response);
    return _response;
}

void Device::OnConnectionLost(const std::string &_cause) {
    ScopedLock lock(m_ConnMutex);

    LOG(Info, "Connection lost <c=" + _cause + ">");
    m_ConnSync.Notify();
}

void Device::GetDeviceInfo(const Messenger::RequestMessage::Ptr _req, Messenger::ResponseMessage::Ptr _resp) {
    JsonStruct::Ptr params = new JsonStruct();
    JsonStruct toFill;
    ToJson(toFill);
    params->Insert("info", toFill);
    _resp->SetParams(params);
}

void Device::GetState(const RequestMessage::Ptr _req, ResponseMessage::Ptr _resp) {
    unsigned long long since = _req->GetParams()->Extract("since").toLongLong(1);
    if (since == 1) {
        _resp->SetCode(WISE_ARGUMENT_NOT_FOUND);
        std::stringstream ss;
        ss << "Arguments not provided";
        _resp->SetMessage(ss.str());
        SSLOG(Warn, ss);
    } else {
        JsonStruct::Ptr params = new JsonStruct();
        JsonStruct toFill;
        ToJson(toFill, since, true);
        params->Insert("state", toFill);
        _resp->SetParams(params);
    }
}

void Device::SetTime(const RequestMessage::Ptr _req, ResponseMessage::Ptr _resp) {
    unsigned long long timestamp = _req->GetParams()->Extract("timestamp").toLongLong(0);
    if (timestamp == 0) {
        _resp->SetCode(WISE_ARGUMENT_NOT_FOUND);
        std::stringstream ss;
        ss << "Arguments not provided";
        _resp->SetMessage(ss.str());
        SSLOG(Warn, ss);
    } else {
        if (m_TimeChangeListener != NULL) {
            m_TimeChangeListener->OnSetTime(timestamp);
        } else {
            _resp->SetCode(WISE_REQUEST_NOT_SUPPORTED);
            std::stringstream ss;
            ss << "Device does not support this request";
            _resp->SetMessage(ss.str());
            SSLOG(Warn, ss);
        }
    }
}

void Device::AdjustTime(const RequestMessage::Ptr _req, ResponseMessage::Ptr _resp) {
    unsigned long long diff = _req->GetParams()->Extract("diff").toLongLong(0);
    if (diff == 0) {
        _resp->SetCode(WISE_ARGUMENT_NOT_FOUND);
        std::stringstream ss;
        ss << "Arguments not provided";
        _resp->SetMessage(ss.str());
        SSLOG(Warn, ss);
    } else {
        if (m_TimeChangeListener != NULL) {
            m_TimeChangeListener->OnAdjustTime(diff);
        } else {
            _resp->SetCode(WISE_REQUEST_NOT_SUPPORTED);
            std::stringstream ss;
            ss << "Device does not support this request";
            _resp->SetMessage(ss.str());
            SSLOG(Warn, ss);
        }
    }
}

void Device::VerifyIdentity(const RequestMessage::Ptr _req, ResponseMessage::Ptr _resp) {
    std::string secret = _req->GetParams()->Extract("secret").toString("");
    if (secret.empty()) {
        _resp->SetCode(WISE_ARGUMENT_NOT_FOUND);
        std::stringstream ss;
        ss << "Arguments not provided";
        _resp->SetMessage(ss.str());
        SSLOG(Warn, ss);
    } else {
        if (m_KeyFile.empty()) {
            _resp->SetCode(WISE_REQUEST_NOT_SUPPORTED);
            std::stringstream ss;
            ss << "Request is not supported";
            _resp->SetMessage(ss.str());
            SSLOG(Warn, ss);
        } else {
            BIO* bufferBio = BIO_new(BIO_s_file());
            if (!bufferBio) {
                _resp->SetCode(WISE_RUNTIME_ERROR);
                std::stringstream ss;
                ss << "Cannot create BIO for reading private key";
                _resp->SetMessage(ss.str());
                SSLOG(Error, ss);
            } else {
                BIO_read_filename(bufferBio, m_KeyFile.c_str());

                RSA* rsaPrivateKey;
                if (m_KeyPasscode.empty()) {
                    rsaPrivateKey = PEM_read_bio_RSAPrivateKey(bufferBio, NULL, NULL, NULL);
                } else {
                    rsaPrivateKey = PEM_read_bio_RSAPrivateKey(bufferBio, NULL, NULL, (void*) m_KeyPasscode.c_str());
                }

                BIO_free(bufferBio);
                if (!rsaPrivateKey) {
                    RSA_free(rsaPrivateKey);
                    _resp->SetCode(WISE_RUNTIME_ERROR);
                    std::stringstream ss;
                    ss << "Failed to load private key from file";
                    _resp->SetMessage(ss.str());
                    SSLOG(Error, ss);
                } else {
                    unsigned char base64Output[128] = {};
                    int base64OutLen = EVP_DecodeBlock(base64Output, (unsigned char*) secret.c_str(), secret.length());
                    base64Output[base64OutLen] = '\0';

                    unsigned char rsaOutput[RSA_size(rsaPrivateKey)];
                    int rsaOutputSize = RSA_private_decrypt(RSA_size(rsaPrivateKey), base64Output, rsaOutput, rsaPrivateKey, RSA_PKCS1_PADDING);
                    if (rsaOutputSize == -1) {
                        char buffer[256] = {};
                        ERR_error_string(ERR_get_error(), buffer);
                        RSA_free(rsaPrivateKey);
                        _resp->SetCode(WISE_RUNTIME_ERROR);
                        std::stringstream ss;
                        ss << "Failed to decrypt: " << buffer;
                        _resp->SetMessage(ss.str());
                        SSLOG(Error, ss);
                    } else {
                        rsaOutput[rsaOutputSize] = '\0';
                        RSA_free(rsaPrivateKey);

                        std::stringstream ss;
                        ss << "Verify identity decrypt is successful <d=" << rsaOutput << ">";
                        SSLOG(Debug, ss);

                        JsonStruct::Ptr params = new JsonStruct();
                        params->Insert("answer", (char*) rsaOutput);
                        _resp->SetParams(params);
                    }
                }
            }
        }
    }
}


void Device::GetGroupList(const Messenger::RequestMessage::Ptr _req, Messenger::ResponseMessage::Ptr _resp) {
    JsonStruct::Ptr params = new JsonStruct();
    std::map<unsigned int, Group*>::iterator itr = m_GroupList.begin();
    std::vector<Jzon::Node> array;
    for (; itr != m_GroupList.end(); ++itr) {
        JsonStruct json;
        itr->second->ToJson(json);
        array.push_back(json.GetNode());
    }

    params->Insert("groups", array);
    _resp->SetParams(params);
}

void Device::GetServiceList(const Messenger::RequestMessage::Ptr _req, Messenger::ResponseMessage::Ptr _resp) {
    int groupId = _req->GetParams()->Extract("groupID").toInt(-1);
    if (groupId == -1) {
        _resp->SetCode(WISE_ARGUMENT_NOT_FOUND);
        std::stringstream ss;
        ss << "Arguments not provided";
        _resp->SetMessage(ss.str());
        SSLOG(Warn, ss);
    } else {
        std::map<unsigned int, Group*>::iterator itr = m_GroupList.find(groupId);
        if (itr != m_GroupList.end()) {
            JsonStruct::Ptr params = new JsonStruct();
            std::vector<Jzon::Node> array;
            Group::ConstIterator itrService = itr->second->Begin();
            for (; itrService != itr->second->End(); ++itrService) {
                JsonStruct json;
                itrService->second->ToJson(json);
                array.push_back(json.GetNode());
            }
            params->Insert("services", array);
            _resp->SetParams(params);
        } else {
            _resp->SetCode(WISE_INVALID_ARGUMENT);
            std::stringstream ss;
            ss << "Invalid arguments provided";
            _resp->SetMessage(ss.str());
            SSLOG(Warn, ss);
        }
    }
}

void Device::GetService(const Messenger::RequestMessage::Ptr _req, Messenger::ResponseMessage::Ptr _resp) {
    int groupId = _req->GetParams()->Extract("groupID").toInt(-1);
    std::string type = _req->GetParams()->Extract("type").toString();
    if (groupId == -1 || type.empty()) {
        _resp->SetCode(WISE_ARGUMENT_NOT_FOUND);
        std::stringstream ss;
        ss << "Arguments not provided";
        _resp->SetMessage(ss.str());
        SSLOG(Warn, ss);
    } else {
        std::map<unsigned int, Group*>::iterator itr = m_GroupList.find(groupId);
        if (itr != m_GroupList.end()) {
            JsonStruct::Ptr params = new JsonStruct();
            BaseService* service = itr->second->FindServiceByType(type);
            if (service != NULL) {
                JsonStruct json;
                service->ToJson(json);
                params->Insert("service", json);
                _resp->SetParams(params);
            } else {
                _resp->SetCode(WISE_SERVICE_NOT_SUPPORTED);
                std::stringstream ss;
                ss << "Service " + type + " not supported in Group object with id " << groupId;
                _resp->SetMessage(ss.str());
                SSLOG(Warn, ss);
            }
        } else {
            _resp->SetCode(WISE_INVALID_ARGUMENT);
            std::stringstream ss;
            ss << "Invalid arguments provided";
            _resp->SetMessage(ss.str());
            SSLOG(Warn, ss);
        }
    }

}

void Device::JoinConfirmation(const Messenger::RequestMessage::Ptr _req, Messenger::ResponseMessage::Ptr _resp) {
    int id = _req->GetParams()->Extract("id").toInt(-1);
    if (id == -1) {
        _resp->SetCode(WISE_ARGUMENT_NOT_FOUND);
        std::stringstream ss;
        ss << "Incomplete request argument list";
        _resp->SetMessage(ss.str());
        SSLOG(Warn, ss);
    } else {
        m_InclusionResult.SetID(id);
        JsonStruct::Ptr params = new JsonStruct();
        params->Insert("inclusionID", m_InclusionResult.GetInclusionID());
        _resp->SetParams(params);
        m_SyncJoin.Notify();
    }
}

void Device::GetPropertyList(unsigned int _groupId, const std::string &_serviceType, const RequestMessage::Ptr _req, ResponseMessage::Ptr _resp) {
    std::map<unsigned int, Group*>::iterator itr = m_GroupList.find(_groupId);
    if (itr != m_GroupList.end()) {
        BaseService* service = itr->second->FindServiceByType(_serviceType);
        if (service != NULL) {
            JsonStruct::Ptr params = new JsonStruct();
            std::vector<Jzon::Node> array;
            BaseService::PropertyIterator itrProperty = service->BeginProperty();
            for (; itrProperty != service->EndProperty(); ++itrProperty) {
                JsonStruct json;
                itrProperty->second.ToJson(json);
                array.push_back(json.GetNode());
            }

            params->Insert("properties", array);
            _resp->SetParams(params);
        }
    }
}

void Device::GetPropertyValue(unsigned int _groupId, const std::string& _serviceType, const Messenger::RequestMessage::Ptr _req, Messenger::ResponseMessage::Ptr _resp) {
    std::string name = _req->GetParams()->Extract("name").toString("");
    if (name.empty()) {
        _resp->SetCode(WISE_ARGUMENT_NOT_FOUND);
        std::stringstream ss;
        ss << "Arguments not provided";
        _resp->SetMessage(ss.str());
        SSLOG(Warn, ss);
    } else {
        std::map<unsigned int, Group*>::iterator itr = m_GroupList.find(_groupId);
        if (itr != m_GroupList.end()) {
            BaseService* service = itr->second->FindServiceByType(_serviceType);
            if (service != NULL) {
                try {
                    Property& prop = service->FindPropertyByName(name);
                    JsonStruct::Ptr param = new JsonStruct();
                    param->Insert("value", prop.GetValue());
                    _resp->SetParams(param);
                } catch (const BaseException& exc) {
                    _resp->SetCode(WISE_PROPERTY_NOT_SUPPORTED);
                    std::stringstream ss;
                    ss << "Property is not supported by " + _serviceType;
                    _resp->SetMessage(ss.str());
                    SSLOG(Warn, ss);
                }
            }
        }
    }
}

void Device::SetPropertyValue(unsigned int _groupId, const std::string& _serviceType, const Messenger::RequestMessage::Ptr _req, Messenger::ResponseMessage::Ptr _resp) {
    std::string name = _req->GetParams()->Extract("name").toString("");
    Jzon::Node value = _req->GetParams()->Extract("value");

    std::stringstream ss;
    ss << "Set <p=" << name << "> property of service <s=" + _serviceType + "> to value <v=" << value.toString() << ">";
    SSLOG(Info, ss);

    if (name.empty() || value.getType() == Jzon::Node::T_INVALID) {
        _resp->SetCode(WISE_ARGUMENT_NOT_FOUND);
        std::stringstream ss;
        ss << "Arguments not provided";
        _resp->SetMessage(ss.str());
        SSLOG(Warn, ss);
    } else {
        std::map<unsigned int, Group*>::iterator itr = m_GroupList.find(_groupId);
        if (itr != m_GroupList.end()) {
            BaseService* service = itr->second->FindServiceByType(_serviceType);
            if (service != NULL) {
                try {
                    Property& prop = service->FindPropertyByName(name);
                    if (!CheckPropertyFormat(value, prop, _resp)) {
                        LOG(Warn, _resp->GetMessage());
                        return;
                    }

                    if (prop.IsReadOnly()) {
                        _resp->SetCode(WISE_NOT_AUTHORIZED);
                        std::stringstream ss;
                        ss << "Property value is read-only";
                        _resp->SetMessage(ss.str());
                        SSLOG(Warn, ss);
                        return;
                    } else if (!prop.IsUsed()) {
                        _resp->SetCode(WISE_PROPERTY_NOT_SUPPORTED);
                        std::stringstream ss;
                        ss << "Property <" + name + "> is not set to be used";
                        SSLOG(Warn, ss);
                        _resp->SetMessage(ss.str());
                        return;
                    }
                    if (prop.GetPropertyFunction() != NULL) {
                        (*(prop.GetPropertyFunction()))(_groupId, _serviceType, prop.GetName(), value);
                    }
                    prop.SetValue(value);
                    SendPropertyChangedEvent(prop.GetName(), prop.GetValue(), _groupId, _serviceType);
                } catch (const BaseException& exc) {
                    _resp->SetCode(WISE_PROPERTY_NOT_SUPPORTED);
                    std::stringstream ss;
                    ss << "Property is not supported by " + _serviceType;
                    _resp->SetMessage(ss.str());
                    SSLOG(Warn, ss);
                }
            }
        }
    }
}

void Device::GetCommandList(unsigned int _groupId, const std::string &_serviceType, const RequestMessage::Ptr _req, ResponseMessage::Ptr _resp) {
    std::map<unsigned int, Group*>::iterator itr = m_GroupList.find(_groupId);
    if (itr != m_GroupList.end()) {
        BaseService* service = itr->second->FindServiceByType(_serviceType);
        if (service != NULL) {
            JsonStruct::Ptr params = new JsonStruct();
            std::vector<Jzon::Node> array;
            BaseService::CommandIterator itrCommand = service->BeginCommand();
            for (; itrCommand != service->EndCommand(); ++itrCommand) {
                JsonStruct json;
                itrCommand->second.ToJson(json);
                array.push_back(json.GetNode());
            }

            params->Insert("commands", array);
            _resp->SetParams(params);
        }
    }
}

void Device::ExecuteCommand(unsigned int _groupId, const std::string &_serviceType, const RequestMessage::Ptr _req, ResponseMessage::Ptr _resp) {
    std::string name = _req->GetParams()->Extract("name").toString();
    Jzon::Node input = _req->GetParams()->Extract("input");

    std::map<unsigned int, Group*>::iterator itr = m_GroupList.find(_groupId);
    if (itr != m_GroupList.end()) {
        BaseService* service = itr->second->FindServiceByType(_serviceType);
        if (service != NULL) {
            try {
                Command& command = service->FindCommandByName(name);
                const std::vector<InputParam>& inputParamVector = command.GetInputParams();
                std::vector<InputParam>::const_iterator constItr = inputParamVector.begin();

                if (input.getCount() != inputParamVector.size() && !command.HasParamAny()) {
                    _resp->SetCode(WISE_INVALID_FORMAT);
                    std::stringstream ss;
                    ss << "Invalid number of command param arguments <" + command.GetName() + ">";
                    SSLOG(Warn, ss);
                    _resp->SetMessage(ss.str());
                    return;
                } else {
                    for (; constItr != inputParamVector.end(); ++constItr) {
                        if (constItr->GetType() != InputParam::ANY) {
                            Jzon::Node param = input.get(constItr->GetName());
                            if (param.getType() == Jzon::Node::T_INVALID) {
                                _resp->SetCode(WISE_INVALID_FORMAT);
                                std::stringstream ss;
                                ss << "Command does not have <" + constItr->GetName() + "> as param";
                                SSLOG(Warn, ss);
                                _resp->SetMessage(ss.str());
                                return;
                            } else {
                                if (!CheckInputParamFormat(param, *constItr, _resp)) {
                                    LOG(Warn, _resp->GetMessage());
                                    return;
                                }
                            }
                        }
                    }
                }

                if (command.GetCommandFunction() != NULL) {
                    (*(command.GetCommandFunction()))(_groupId, _serviceType, input);
                }
            } catch (const BaseException& exc) {
                _resp->SetCode(WISE_COMMAND_NOT_SUPPORTED);
                std::stringstream ss;
                ss << "Command is not supported by " + _serviceType;
                _resp->SetMessage(ss.str());
                SSLOG(Warn, ss);
            }
        }
    }
}

bool Device::CheckInputParamFormat(const Jzon::Node& _param, const InputParam& _inputParam, ResponseMessage::Ptr _resp) const {
    Jzon::Node::Type paramType = _param.getType();
    InputParam::InputParamType inputParamType = _inputParam.GetType();

    bool retVal = false;
    if (inputParamType == InputParam::ATTRIBUTE) {
        unsigned int index = _param.toInt(Property::MAX_OPTION_SIZE);
        if ((index < Property::MAX_OPTION_SIZE) && (_inputParam.GetOptions().size() > index)) {
            retVal = true;
        }

        if (!retVal) {
            _resp->SetCode(WISE_INVALID_OPTION);
            std::stringstream ss;
            ss << "Input param <" + _inputParam.GetName() + "> doesn't contain attribute <" + _param.toString() + ">";
            _resp->SetMessage(ss.str());
        }
    } else if (inputParamType == InputParam::RANGE) {
        if (_param >= _inputParam.GetMin() && _param <= _inputParam.GetMax()) {
            retVal = true;
        } else {
            _resp->SetCode(WISE_INVALID_RANGE);
            std::stringstream ss;
            ss << "Input param <" + _inputParam.GetName() + "> is in range of " << _inputParam.GetMin().toString() << " - " << _inputParam.GetMax();
            _resp->SetMessage(ss.str());
        }
    } else if (paramType == Jzon::Node::T_STRING && inputParamType == InputParam::COLOR) {
        REGEX_OBJ regx("^[0-9A-Fa-f]{8}$");
        std::string stringToCheck = _param.toString();
        if (!stringToCheck.empty()) {
            retVal = REGEX_MATCH(stringToCheck, regx);
        }
        if (!retVal) {
            _resp->SetCode(WISE_INVALID_VALUE);
            _resp->SetMessage("Invalid value format.");
        }
    } else if ((paramType == Jzon::Node::T_NUMBER && inputParamType == InputParam::NUMBER) ||
              (paramType == Jzon::Node::T_BOOL && inputParamType == InputParam::BOOL) ||
              (paramType == Jzon::Node::T_STRING && inputParamType == InputParam::TEXT)) {
          retVal = true;
      } else {
          _resp->SetCode(WISE_INVALID_VALUE);
          _resp->SetMessage("Invalid value format.");
      }

    return retVal;
}

bool Device::CheckPropertyFormat(const Jzon::Node& _value, const Property& _property, ResponseMessage::Ptr _resp) const {
    Jzon::Node::Type valueType = _value.getType();
    Property::PropertyType propertyType = _property.GetType();

    bool retVal = false;
    if (propertyType == Property::ATTRIBUTE || propertyType == Property::CHANGABLE_ATTRIBUTE) {
        unsigned int index = _value.toInt(Property::MAX_OPTION_SIZE);
        if ((index < Property::MAX_OPTION_SIZE) && (_property.GetOptions().size() > index)) {
            retVal = true;
        }

        if (!retVal && !_resp.IsNull()) {
            _resp->SetCode(WISE_INVALID_OPTION);
            std::stringstream ss;
            ss << "Property <" + _property.GetName() + "> doesn't contain attribute <" + _value.toString() + ">";
            _resp->SetMessage(ss.str());
        }
    } else if (propertyType == Property::RANGE) {
        if (_value >= _property.GetOptions()[0] && _value <= _property.GetOptions()[1]) {
            retVal = true;
        } else if (!_resp.IsNull()) {
            _resp->SetCode(WISE_INVALID_RANGE);
            std::stringstream ss;
            ss << "Property <" + _property.GetName() + "> is in range of " << _property.GetOptions()[0].toString() << " - " << _property.GetOptions()[1].toString();
            _resp->SetMessage(ss.str());
        }
    } else if (valueType == Jzon::Node::T_STRING && propertyType == Property::COLOR) {
        REGEX_OBJ regx("^[0-9A-Fa-f]{8}$");
        std::string stringToCheck = _value.toString();
        if (!stringToCheck.empty()) {
            retVal = REGEX_MATCH(stringToCheck, regx);
        }
        if (!retVal && !_resp.IsNull()) {
            _resp->SetCode(WISE_INVALID_VALUE);
            _resp->SetMessage("Invalid value format.");
        }
    } else if ((valueType == Jzon::Node::T_NUMBER && propertyType == Property::NUMBER) ||
            (valueType == Jzon::Node::T_BOOL && propertyType == Property::BOOL) ||
            (valueType == Jzon::Node::T_STRING && propertyType == Property::TEXT)) {
        retVal = true;
    } else if (!_resp.IsNull()){
        _resp->SetCode(WISE_INVALID_VALUE);
        _resp->SetMessage("Invalid value format.");
    }

    return retVal;
}

bool Device::UpdatePropertyAttributeList(unsigned int _groupId, const std::string& _serviceType, const std::string& _name, const std::vector<Jzon::Node>& _attributeList) {
    std::stringstream ss;
     ss << "<Update property attribute " << _name << " list for service" << _serviceType << ">";
     SSLOG(Info, ss);
     std::map<unsigned int, Group*>::iterator itr = m_GroupList.find(_groupId);
     if (itr != m_GroupList.end()) {
         BaseService* service = itr->second->FindServiceByType(_serviceType);
         if (service != NULL) {
             try {
                 Property& prop = service->FindPropertyByName(_name);
                 if (prop.GetType()== Property::CHANGABLE_ATTRIBUTE){
                     prop.SetOptions(_attributeList);

                     if (prop.IsUsed()) {
                         prop.SetValue(0);

                         if (m_Messenger && m_Messenger->IsConnected()) {
                             SendPropertyChangedEvent(prop.GetName(), prop.GetValue(), _groupId, _serviceType);
                         }
                     }
                     return true;
                 } else {
                     std::stringstream ss;
                     ss << "Property <" + _name + "> can not be changed";
                     SSLOG(Warn, ss);
                     return false;
                 }
            } catch (const NotFoundException& exc) {
                 std::stringstream ss;
                 ss << "Service <" + _serviceType + "> doesn't own <" + _name + "> property";
                 SSLOG(Info, ss);
                 return false;
            }
        }

    }
    return false;
}

void Device::SetPropertyValueIntern(unsigned int _groupId, const std::string& _serviceType, const std::string& _name, const Jzon::Node& _value, bool _sendEvent) {
    std::stringstream ss;
    ss << "Set <p=" << _name << "> property of service <s=" + _serviceType + "> to value <v=" << _value.toString() << ">";
    SSLOG(Info, ss);
    std::map<unsigned int, Group*>::iterator itr = m_GroupList.find(_groupId);
    if (itr != m_GroupList.end()) {
        BaseService* service = itr->second->FindServiceByType(_serviceType);
        if (service != NULL) {
            try {
                Property& prop = service->FindPropertyByName(_name);
                if (prop.GetType() == Property::ATTRIBUTE || prop.GetType() == Property::CHANGABLE_ATTRIBUTE) {
                    unsigned int index = _value.toInt(Property::MAX_OPTION_SIZE);
                    if (!((index < Property::MAX_OPTION_SIZE) && (prop.GetOptions().size() > index))) {
                        std::stringstream ss;
                        ss << "Property <" + _name + "> doesn't contain attribute <" + _value.toString() + ">";
                        SSLOG(Warn, ss);
                        return;
                    }
                }

                if (!CheckPropertyFormat(_value, prop)) {
                    std::stringstream ss;
                    ss << "Invalid value format.";
                    SSLOG(Warn, ss);
                    return;
                }

                if (prop.GetType() == Property::RANGE) {
                    if (_value < prop.GetOptions()[0] || _value > prop.GetOptions()[1]) {
                        std::stringstream ss;
                        ss << "Property <" + _name + "> is in range of " << prop.GetOptions()[0].toString() << " - " << prop.GetOptions()[1].toString();
                        SSLOG(Warn, ss);
                        return;
                    }
                }
                if (prop.IsUsed()) {
                    prop.SetValue(_value);

                    if (_sendEvent && m_Messenger && m_Messenger->IsConnected()) {
                        SendPropertyChangedEvent(prop.GetName(), prop.GetValue(), _groupId, _serviceType);
                    }
                } else {
                    std::stringstream ss;
                    ss << "Property <" + _name + "> is not set to be used";
                    SSLOG(Warn, ss);
                }

            } catch (const NotFoundException& exc) {
                std::stringstream ss;
                ss << "Service <" + _serviceType + "> doesn't own <" + _name + "> property";
                SSLOG(Info, ss);
            }
        }
    }
}

Jzon::Node Device::GetPropertyValueIntern(unsigned int _groupId, const std::string& _serviceType, const std::string& _name) {
    std::map<unsigned int, Group*>::iterator itr = m_GroupList.find(_groupId);
    if (itr != m_GroupList.end()) {
        BaseService* service = itr->second->FindServiceByType(_serviceType);
        if (service != NULL) {
            try {
                Property& prop = service->FindPropertyByName(_name);
                return prop.GetValue();
            } catch (const BaseException& exc) {
                return Jzon::Node::T_INVALID;
            }
        } else {
            return Jzon::Node::T_INVALID;
        }
    } else {
        return Jzon::Node::T_INVALID;
    }
}

void Device::SendPropertyChangedEvent(const std::string& _name, const Jzon::Node& _value, unsigned int _groupId, const std::string &_serviceType) {
    EventMessage evt;
    evt.SetName("propertyChanged");
    JsonStruct::Ptr params = new JsonStruct();
    params->Insert("name", _name);
    params->Insert("value", _value);
    evt.SetParams(params);

    m_Messenger->SendEvent(evt, _groupId, _serviceType);
}

void Device::SendDeviceRebootedEvent() {
    EventMessage evt;
    evt.SetName("deviceRebooted");
    m_Messenger->SendEvent(evt, 0, "");
}

void Device::SendGenericEvent(const std::string& _name, const JsonStruct::Ptr _params) {
    EventMessage evt;
    evt.SetName(_name);
    evt.SetParams(_params);
    m_Messenger->SendEvent(evt, 0, SERVICE_TYPE_BASIC);
}

void Device::DisconnectWithoutStsMessage() {
    m_Messenger->DisconnectWithoutStsMessage();
}

void Device::RegisterDeviceExcludedListener(IDeviceExcludedListener *_deviceExcludedListener) {
    m_DeviceExcludedListener = _deviceExcludedListener;
}

void Device::UnregisterDeviceExcludedListener() {
    m_DeviceExcludedListener = NULL;
}

void Device::SetLogListener(ILogListener *_logListener) {
    if (m_Messenger != NULL) {
        m_Messenger->RegisterLogListener(_logListener);
    }
    m_DiscoveryObj.RegisterLogListener(_logListener);
    RegisterLogListener(_logListener);
}

void Device::RegisterConnectedListener(IConnectedListener *_connectedListener) {
    m_ConnectedListener = _connectedListener;
}

void Device::UnregisterConnectedListener() {
    m_ConnectedListener = NULL;
}

void Device::RegisterTimeChangeListener(ITimeChangeListener* _timeChangeListener) {
    m_TimeChangeListener = _timeChangeListener;
}

void Device::UnregisterTimeChangeListener() {
    m_TimeChangeListener = NULL;
}

void Device::ToJson(JsonStruct& _toFill, unsigned long long _timestamp, bool _stripped) const {
    _toFill.Insert("manufacturer", m_DeviceInfo.GetManufacturer());
    _toFill.Insert("model", m_DeviceInfo.GetModel());
    _toFill.Insert("id", m_InclusionResult.GetID());
    std::map<unsigned int, Group*>::const_iterator itr = m_GroupList.begin();
    std::vector<Jzon::Node> array;
    for (; itr != m_GroupList.end(); ++itr) {
        JsonStruct json;
        itr->second->ToJson(json, _timestamp, _stripped);
        if (!json.IsEmpty()) {
            array.push_back(json.GetNode());
        }
    }
    if (!array.empty()) {
        _toFill.Insert("groupList", array);
    }
}

void Device::Destroy() {
    if (m_Messenger != NULL) {
        if (m_Messenger->IsConnected()) {
            m_Messenger->Disconnect();
        }
        SAFE_DELETE_O(m_Messenger);
    }
}

void Device::InclusionTimerExpired(Utility::Timer& _timer) {
    LOG(Info, "Inclusion expired");
    StopDiscovery();
    Destroy();
    m_InclusionTimer->RestartTimer(0);
    m_InclusionTimerExpired = true;
}

RequestHandler::RequestHandler(Device *_obj) : m_Obj(_obj) {
}

void RequestHandler::AddRequestHandler(const std::string &_commandName, const RespondFunctionPtr _func) {
    ScopedLock lock(m_Mutex);
    RequestHandlerObj handler;
    handler.first = _commandName;
    handler.second = _func;

    m_RequestHandlers.insert(handler);
}

void RequestHandler::CallRequestHandler(const std::string &_commandName, const RequestMessage::Ptr _req, ResponseMessage::Ptr _rsp) {
    ScopedLock lock(m_Mutex);
    RequestHandlers::const_iterator itr = m_RequestHandlers.find(_commandName);
    if (itr != m_RequestHandlers.end()) {
        RespondFunctionPtr func = itr->second;
        (m_Obj->*func)(_req, _rsp);
    } else {
        _rsp->SetCode(WISE_REQUEST_NOT_SUPPORTED);
        std::stringstream ss;
        ss << "Request name doesn't match with supported Device/Group object requests";
        _rsp->SetMessage(ss.str());
        SSLOG(Warn, ss);
    }
}

void RequestHandler::AddServiceRequestHandler(const std::string &_commandName, const RespondServiceFunctionPtr _func) {
    ScopedLock lock(m_Mutex);
    RequestServiceHandlerObj handler;
    handler.first = _commandName;
    handler.second = _func;

    m_ServiceRequestHandlers.insert(handler);
}

void RequestHandler::CallServiceRequestHandler(unsigned int _groupId, const std::string& _serviceType, const std::string &_commandName, const RequestMessage::Ptr _req, ResponseMessage::Ptr _rsp) {
    ScopedLock lock(m_Mutex);
    ServiceRequestHandlers::const_iterator itr = m_ServiceRequestHandlers.find(_commandName);
    if (itr != m_ServiceRequestHandlers.end()) {
        RespondServiceFunctionPtr func = itr->second;
        (m_Obj->*func)(_groupId, _serviceType, _req, _rsp);
    } else {
        _rsp->SetCode(WISE_REQUEST_NOT_SUPPORTED);
        std::stringstream ss;
        ss << "Request name doesn't match with supported Service object requests";
        _rsp->SetMessage(ss.str());
        SSLOG(Warn, ss);
    }
}

DeviceInformation::DeviceInformation() {
}

void DeviceInformation::SetName(const std::string& _name) {
    m_Name = _name;
}

void DeviceInformation::SetManufacturer(const std::string &_manufacturer) {
    m_Manufacturer = _manufacturer;
}

void DeviceInformation::SetIdentifier(const std::string &_identifier) {
    m_Identifier = _identifier;
}

void DeviceInformation::SetModel(const std::string &_model) {
    m_Model = _model;
}

void DeviceInformation::SetVersion(const std::string &_version) {
    m_Version = _version;
}

void DeviceInformation::SetProtocolVersion(int _protocolVersion) {
    m_ProtocolVersion = _protocolVersion;
}

void DeviceInformation::SetMAC(const std::string &_mac) {
    m_MAC = _mac;
}

void DeviceInformation::SetVersionNumber(int _versionNumber) {
    m_VersionNumber = _versionNumber;
}

std::string DeviceInformation::GetName() const {
    return m_Name;
}

std::string DeviceInformation::GetManufacturer() const {
    return m_Manufacturer;
}

std::string DeviceInformation::GetIdentifier() const {
    return m_Identifier;
}

std::string DeviceInformation::GetModel() const {
    return m_Model;
}

std::string DeviceInformation::GetVersion() const {
    return m_Version;
}

int DeviceInformation::GetProtocolVersion() const {
    return m_ProtocolVersion;
}

std::string DeviceInformation::GetMAC() const {
    return m_MAC;
}

int DeviceInformation::GetVersionNumber() const {
    return m_VersionNumber;
}

InclusionInformation::InclusionInformation() {

}

void InclusionInformation::SetUsername(const std::string& _username) {
    m_Username = _username;
}

void InclusionInformation::SetPassword(const std::string& _password) {
    m_Password = _password;
}

void InclusionInformation::SetHub(const HubHolder &_hub) {
    m_Hub = _hub;
}

void InclusionInformation::SetInclusionID(const std::string& _inclusionID) {
    m_InclusionID = _inclusionID;
}

void InclusionInformation::SetID(int _id) {
    m_ID = _id;
}

void InclusionInformation::SetHubTag(const std::string& _hubTag) {
    m_HubTag = _hubTag;
}

std::string InclusionInformation::GetUsername() const {
    return m_Username;
}

std::string InclusionInformation::GetPassword() const {
    return m_Password;
}

SSDP::HubHolder InclusionInformation::GetHub() const {
    return m_Hub;
}

std::string InclusionInformation::GetInclusionID() const {
    return m_InclusionID;
}

int InclusionInformation::GetID() const {
    return m_ID;
}

std::string InclusionInformation::GetHubTag() const {
    return m_HubTag;
}

bool InclusionInformation::IsEmpty() const {
    if (m_Password.empty() || m_Username.empty() || m_InclusionID.empty() || m_ID == -1 || m_HubTag.empty()) {
        return true;
    } else {
        return false;
    }
}

ChallengeParameters::ChallengeParameters() {

}

unsigned int ChallengeParameters::GetChallengeRetry() const {
    return m_ChallengeRetry;
}

long ChallengeParameters::GetChallengeTimeout() const {
    return m_ChallengeTimeout;
}

std::string ChallengeParameters::GetChallengeCode() const {
    return m_ChallengeCode;
}

void ChallengeParameters::SetChallengeRetry(unsigned int _challengeRetry) {
    m_ChallengeRetry = _challengeRetry;
}

void ChallengeParameters::SetChallengeTimeout(long _challengeTimeout) {
    m_ChallengeTimeout = _challengeTimeout;
}

void ChallengeParameters::SetChallengeCode(const std::string &_challengeCode) {
    m_ChallengeCode = _challengeCode;
}
