#include "messenger/Client.h"

using namespace Messenger;
using namespace Utility;

AsyncClient::AsyncClient(const std::string &_clientId, const std::string &_serverUri, IDeliveryListener* _deliveryListener, ILogListener *_logListener) :
    ClientImpl(_serverUri, _clientId),
    m_DeliveryListener(_deliveryListener),
    m_ConnectionListener(NULL),
    m_LogListener(_logListener),
    m_SSLConfig(NULL),
    m_StatusConfig(NULL),
    m_NetworkProxy(*this) {
    RegisterListener(&m_NetworkProxy);
    m_ConnConfig = new ConnectConfig();
    m_ConnConfig->SetCleanSession(false);
}

AsyncClient::~AsyncClient() {

}

void AsyncClient::SetUsername(const std::string &_username) {
    m_ConnConfig->SetUsername(_username);
}

void AsyncClient::SetPassword(const std::string &_password) {
    m_ConnConfig->SetPassword(_password);
}

void AsyncClient::SetCleanSession(bool _cleanSession) {
    m_ConnConfig->SetCleanSession(_cleanSession);
}

void AsyncClient::SetKeepAliveInterval(int _keepAliveInterval) {
    m_ConnConfig->SetKeepAliveInterval(_keepAliveInterval);
}

bool AsyncClient::IsCleanSession() const {
    return m_ConnConfig->GetCleanSession();
}

void AsyncClient::SetConnectionTimeoutInterval(int _timeout) {
    m_ConnConfig->SetTimeout(_timeout);
}

void AsyncClient::SetStatusConfiguration(const std::string &_topic, const Message::Ptr _msg) {
    if (m_StatusConfig.IsNull()) {
        m_StatusConfig = new StatusConfig();
    }

    m_StatusConfig->SetTopic(_topic);
    m_StatusConfig->SetMessage(_msg);
    m_ConnConfig->SetStatus(m_StatusConfig);
}

void AsyncClient::DisableStatusMessage() {
    m_StatusConfig = NULL;
}

void AsyncClient::SetSecureConfiguration(const std::string& _trustStore, const std::string& _keyStore, const std::string& _privateKey, const std::string& _privateKeyPassword) {
    if (m_SSLConfig.IsNull()) {
        m_SSLConfig = new SSLConfig();
    }

    m_SSLConfig->SetTrustStore(_trustStore);
    m_SSLConfig->SetKeyStore(_keyStore);
    m_SSLConfig->SetPrivateKey(_privateKey);
    m_SSLConfig->SetPrivateKeyPassword(_privateKeyPassword);
    m_ConnConfig->SetSsl(m_SSLConfig);
}

void AsyncClient::Connect() {
    ClientImpl::Connect(m_ConnConfig);
}

void AsyncClient::Disconnect(long _timeout) {
    if (IsConnected()) {
        try {
            ClientImpl::Disconnect(_timeout);
        } catch (const BaseException& exc) {
        }
    }
}

void AsyncClient::Publish(const std::string &_topic, const Message::Ptr _msg) {
    m_NetworkProxy.Publish(_topic, _msg);
}

void AsyncClient::RegisterConnectionListener(IConnectionListener* _connectionListener) {
    m_ConnectionListener = _connectionListener;
}

void AsyncClient::UnregisterConnectionListener() {
    m_ConnectionListener = NULL;
}


void AsyncClient::OnMessageArrived(const std::string &_topic, const Message::Ptr _msg) {
    if (m_DeliveryListener != NULL) {
        m_DeliveryListener->OnMessageArrived(_topic, _msg);
    }
}

void AsyncClient::OnConnectionLost(const std::string &_cause) {
    if (m_ConnectionListener != NULL) {
        m_ConnectionListener->OnConnectionLost(_cause);
    }
}

void AsyncClient::OnPublishRequired(const std::string &_topic, const Message::Ptr _msg) {
    ScopedLock lock(m_PublishMutex);
    ClientImpl::Publish(_topic, _msg);
}

void AsyncClient::OnPublishFailed(int _code, const std::string &_topic, const Message::Ptr _msg) {
    if (m_DeliveryListener != NULL) {
        m_DeliveryListener->OnPublishFailed(_code, _topic, _msg);
    }
}

BuffClient::BuffClient(const std::string &_clientId, const std::string &_serverUri, IDeliveryListener &_deliveryListener, unsigned int _maxQueueSize, ILogListener *_logListener) :
    AsyncClient(_clientId, _serverUri, &_deliveryListener, _logListener),
    m_MaxQueueSize(_maxQueueSize),
    m_IgnoreSessionPresent(true) {
}

void BuffClient::Connect() {
    AsyncClient::Connect();

    if (m_IgnoreSessionPresent || !isSessionPresent()) {
        SubscribeStore();
        UnsubscribeStore();
    }

    PublishStore();
}

void BuffClient::Subscribe(const std::string &_topic, int _qos) {
    TopicCollection topics;
    topics.push_back(_topic);

    QoSCollection qoss;
    qoss.push_back(_qos);

    Subscribe(topics, qoss);
}

void BuffClient::Subscribe(const TopicCollection &_topics, const QoSCollection &_qoss) {
    try {
        ClientImpl::Subscribe(_topics, _qoss);
    } catch (const BaseException& exc) {
    }

    StoreSubscription(_topics, _qoss);
    RemoveUnsubscription(_topics);
}

void BuffClient::Unsubscribe(const TopicCollection &_topics) {
    try {
        ClientImpl::Unsubscribe(_topics);
    } catch (const BaseException& exc) {
    }

    StoreUnsubscription(_topics);
    RemoveSubscription(_topics);
}

void BuffClient::Unsubscribe(const std::string &_topic) {
    TopicCollection topic;
    topic.push_back(_topic);
    Unsubscribe(topic);
}

void BuffClient::ImmediatePublish(const std::string &_topic, const Message::Ptr _msg) {
    AsyncClient::OnPublishRequired(_topic, _msg);
}

void BuffClient::IgnoreSessionPresent(bool _ignore) {
    m_IgnoreSessionPresent = _ignore;
}

void BuffClient::OnPublishFailed(int _code, const std::string &_topic, const Message::Ptr _msg) {
    if (_code == MQTTCLIENT_DISCONNECTED && _msg->GetQoS() > Message::QOS0) {
        PublishHolder::Ptr pub = new PublishHolder(_topic, _msg);
        m_PublishMutex.Lock();

        if (m_Publications.size() >= m_MaxQueueSize) {
            m_Publications.pop();
        }

        m_Publications.push(pub);
        m_PublishMutex.Unlock();
    } else {
        AsyncClient::OnPublishFailed(_code, _topic, _msg);
    }
}

void BuffClient::StoreSubscription(const TopicCollection &_topics, const QoSCollection &_qoss) {
    ScopedLock lock(m_SubMutex);

    TopicCollection::const_iterator itrTop;
    QoSCollection::const_iterator itrQos;

    for (itrTop = _topics.begin(), itrQos = _qoss.begin(); itrTop != _topics.end(); ++itrTop, ++itrQos) {
        SubElement sub(*itrTop, *itrQos);
        m_Subs.push_back(sub);
    }
}

void BuffClient::StoreUnsubscription(const TopicCollection &_topics) {
    ScopedLock lock(m_UnsubMutex);

    TopicCollection::const_iterator itr;

    for (itr = _topics.begin(); itr != _topics.end(); ++itr) {
        SubElement sub(*itr, 0);
        m_Unsubs.push_back(sub);
    }
}

void BuffClient::RemoveSubscription(const TopicCollection &_topics) {
    ScopedLock lock(m_SubMutex);
    Remove(m_Subs, _topics);
}

void BuffClient::RemoveUnsubscription(const TopicCollection &_topics) {
    ScopedLock lock(m_UnsubMutex);
    Remove(m_Unsubs, _topics);
}

void BuffClient::Remove(SubscriptionBuffer &_buffer, const TopicCollection &_topics) {
    SubscriptionBuffer::iterator itr;
    SubscriptionBuffer::iterator nextItr;
    std::string topic;

    for (itr = _buffer.begin(); itr != _buffer.end(); ++itr) {
        if (itr != _buffer.end()) {
            nextItr = itr + 1;
        }

        topic = itr->first;

        if (std::find(_topics.begin(), _topics.end(), topic) != _topics.end()) {
            _buffer.erase(itr);
        }

        itr = nextItr;
    }
}

void BuffClient::PublishStore() {
    ScopedLock lock(m_PublishMutex);

    while(!m_Publications.empty()) {
        PublishHolder::Ptr elem = m_Publications.front();
        m_Publications.pop();

        const std::string topic = elem->GetTopic();
        const Message::Ptr msg = elem->GetMessage();

        Publish(topic, msg);
    }
}

void BuffClient::SubscribeStore() {
    TopicCollection topics;
    QoSCollection qoss;

    m_SubMutex.Lock();

    while(!m_Subs.empty()) {
        SubElement elem = m_Subs.front();
        m_Subs.pop_front();

        const std::string topic = elem.first;
        const int qos = elem.second;

        topics.push_back(topic);
        qoss.push_back(qos);

    }

    m_SubMutex.Unlock();
    Subscribe(topics, qoss);
}

void BuffClient::UnsubscribeStore() {
    TopicCollection topics;

    m_UnsubMutex.Lock();

    while(!m_Unsubs.empty()) {
        SubElement elem = m_Unsubs.front();
        m_Unsubs.pop_front();

        const std::string topic = elem.first;

        topics.push_back(topic);
    }

    m_UnsubMutex.Unlock();
    Unsubscribe(topics);
}
