#include "messenger/MQTTClient.h"

using namespace Messenger;
using namespace Utility;

char** GetPahoTopics(const ClientImpl::TopicCollection& _topics) {
    ClientImpl::TopicCollection::const_iterator itr;
    char** topics = (char **) malloc(_topics.size() * sizeof(char*));
    int i = 0;

    for (itr = _topics.begin(); itr != _topics.end(); ++itr, ++i) {
        std::string stringTopic = (*itr);
        char* charTopic = strndup(stringTopic.c_str(), stringTopic.size());
        topics[i] = charTopic;
    }

    return topics;
}

void FreePahoTopics(char** _topics, int _cnt) {
    for (int i = 0; i < _cnt; i++) {
        char* topic = _topics[i];

        if (topic) {
            free(topic);
        }
    }
    if (_topics) {
        free(_topics);
    }
}

int* GetPahoQoss(const ClientImpl::QoSCollection& _qos) {
    ClientImpl::QoSCollection::const_iterator itr;
    int* qoss = (int*) malloc(_qos.size() * sizeof(int));
    int i = 0;

    for (itr = _qos.begin(); itr != _qos.end(); ++itr, ++i) {
        qoss[i] = (*itr);
    }

    return qoss;
}

void FreePahoQoss(int* _qoss) {
    if (_qoss) {
        free(_qoss);
    }
}

ClientImpl::ClientImpl(const std::string& _serverUri, const std::string& _clientId) : m_ServerUri(_serverUri), m_ClientId(_clientId), m_StatusListener(NULL) {
    MQTTClient_create(&m_Client, m_ServerUri.c_str(), m_ClientId.c_str(), 0, NULL);
}

ClientImpl::~ClientImpl() {
    MQTTClient_destroy(&m_Client);
}

void ClientImpl::Connect(ConnectConfig::Ptr _config) {
    if (!IsConnected()) {
        m_SessionPresent = false;
        MQTTClient_connectOptions* options;

        if (!_config->GetSsl().IsNull() && m_ServerUri.compare(0, 6, "ssl://") != 0) {
            throw ProtocolException("URI must start with <ssl://>");
        }

        options = _config->PackConfig();

        int rc = MQTTClient_connect(m_Client, options);
        bool sessionPresent = options->returned.sessionPresent;
        ConnectConfig::FreeConfig(options);

        if (rc != MQTTCLIENT_SUCCESS) {
            MQTTClient_disconnect(m_Client, 300);
            throw ProtocolException("Connecting failed", rc);
        }

        m_SessionPresent = sessionPresent;
    }
}

bool ClientImpl::isSessionPresent() const {
    return m_SessionPresent;
}

void ClientImpl::Disconnect(long _timeout) {
    if (IsConnected()) {
        int rc = MQTTClient_disconnect(m_Client, _timeout);

        if (rc != MQTTCLIENT_SUCCESS) {
            throw ProtocolException("Disconnecting failed", rc);
        }
    }
}

std::string ClientImpl::GetServerUri() const {
    return m_ServerUri;
}

std::string ClientImpl::GetClientId() const {
    return m_ClientId;
}

bool ClientImpl::IsConnected() const {
    return (MQTTClient_isConnected(m_Client)) != 0;
}

void ClientImpl::Publish(const std::string& _topic, const Message::Ptr _msg) {
    MQTTClient_message* message;
    _msg->WriteMessage(&message);

    int rc = MQTTClient_publishMessage(m_Client, _topic.c_str(), message, NULL);
    Message::FreeMessage(&message);
    if (rc != MQTTCLIENT_SUCCESS) {
        m_StatusListener->PublishFailed(rc, _topic, _msg);
    }
}

void ClientImpl::RegisterListener(IStatusListener* _listener) {
    ScopedLock lock(m_Mutex);

    m_StatusListener = _listener;
    int rc = MQTTClient_setCallbacks(m_Client, this, &ClientImpl::OnConnectionLost, &ClientImpl::OnMessageArrived, &ClientImpl::OnDeliveryComplete);
    if (rc != MQTTCLIENT_SUCCESS) {
        throw ProtocolException("Registering listener failed", rc);
    }
}

void ClientImpl::Subscribe(const TopicCollection& _topics, const QoSCollection& _qoss) {
    unsigned int n = _topics.size();

    if (n == 0) {
        return;
    }

    if (n != _qoss.size()) {
        throw ProtocolException("Collection sizes doesn't match");
    }

    char** topics = GetPahoTopics(_topics);
    int* qoss = GetPahoQoss(_qoss);

    int rc = MQTTClient_subscribeMany(m_Client, n, topics, qoss);

    FreePahoTopics(topics, n);
    FreePahoQoss(qoss);
    if (rc != MQTTCLIENT_SUCCESS) {
        throw ProtocolException("Subscribing many failed", rc);
    }
}

void ClientImpl::Subscribe(const std::string& _topic, int _qos) {
    int rc = MQTTClient_subscribe(m_Client, _topic.c_str(), _qos);
    if (rc != MQTTCLIENT_SUCCESS) {
        throw ProtocolException("Subscribing failed", rc);
    }
}

void ClientImpl::Unsubscribe(const TopicCollection& _topics) {
    unsigned int n = _topics.size();

    if (n == 0) {
        return;
    }

    char** topics = GetPahoTopics(_topics);
    int rc = MQTTClient_unsubscribeMany(m_Client, n, topics);
    FreePahoTopics(topics, n);

    if (rc != MQTTCLIENT_SUCCESS) {
        throw ProtocolException("Unsubscribing many failed", rc);
    }
}

void ClientImpl::Unsubscribe(const std::string& _topic) {
    int rc = MQTTClient_unsubscribe(m_Client, _topic.c_str());
    if (rc != MQTTCLIENT_SUCCESS) {
        throw ProtocolException("Unsubscribing failed", rc);
    }
}

void ClientImpl::OnConnectionLost(void* _context, char* _cause) {
    if (_context) {
        ClientImpl* cli = static_cast<ClientImpl*>(_context);
        IStatusListener* cb = cli->GetStatusListener();

        if (cb) {
            cb->ConnectionLost(_cause ? std::string(_cause) : std::string());
        }
    }
}

int ClientImpl::OnMessageArrived(void* _context, char* _topic, int _topicLen, MQTTClient_message* _message) {
    if (_context) {
        ClientImpl* cli = static_cast<ClientImpl*>(_context);
        IStatusListener* cb = cli->GetStatusListener();

        if (cb) {
            std::string topic(_topic);

            Message::Ptr msg = Message::ParseMessage(_message);
            cb->MessageArrived(topic, msg);
        }
    }

    MQTTClient_freeMessage(&_message);
    MQTTClient_free(_topic);
    return (!0);
}

void ClientImpl::OnDeliveryComplete(void* _context, MQTTClient_token _token) {
}

IStatusListener* ClientImpl::GetStatusListener() {
    ScopedLock lock(m_Mutex);
    return m_StatusListener;
}

