#include "messenger/MQTTConfiguration.h"

using namespace Messenger;

const MQTTClient_connectOptions& InitConnectionConfig() {
    static MQTTClient_connectOptions connConfig = MQTTClient_connectOptions_initializer;
    return connConfig;
}

const MQTTClient_SSLOptions& InitSecureConnectionConfig() {
    static MQTTClient_SSLOptions secConfig = MQTTClient_SSLOptions_initializer;
    return secConfig;
}

const MQTTClient_willOptions& InitStatusConfig() {
    static MQTTClient_willOptions statusConfig = MQTTClient_willOptions_initializer;
    return statusConfig;
}

std::string StatusConfig::GetTopic() const {
    return m_Topic;
}

void StatusConfig::SetTopic(const std::string& _topic) {
    m_Topic = _topic;
}

Message::Ptr StatusConfig::GetMessage() const {
    return m_Message;
}

void StatusConfig::SetMessage(const Message::Ptr _msg) {
    m_Message = _msg;
}

MQTTClient_willOptions* StatusConfig::PackConfig() const {
    MQTTClient_willOptions* config;
    config = new MQTTClient_willOptions(InitStatusConfig());
    config->topicName = Utility::DuplicateString(m_Topic);
    config->message = Utility::DuplicateString(m_Message->GetPayload());
    config->retained = m_Message->IsRetained() ? 1 : 0;
    config->qos = (int) m_Message->GetQoS();
    return config;
}

void StatusConfig::FreeConfig(MQTTClient_willOptions* _config) {
    MQTTClient_willOptions* config = _config;
    SAFE_DELETE_ARRAY_O(config->message);
    SAFE_DELETE_ARRAY_O(config->topicName);
    SAFE_DELETE_O(config);
}

SSLConfig::SSLConfig(): m_VerifyServer(true) {
}

std::string SSLConfig::GetTrustStore() const {
    return m_TrustStore;
}

void SSLConfig::SetTrustStore(const std::string& _trustStore) {
    m_TrustStore = _trustStore;
}

std::string SSLConfig::GetKeyStore() const {
    return m_KeyStore;
}

void SSLConfig::SetKeyStore(const std::string& _keyStore) {
    m_KeyStore = _keyStore;
}

std::string SSLConfig::GetPrivateKey() const {
    return m_PrivateKey;
}

void SSLConfig::SetPrivateKey(const std::string& _privateKey) {
    m_PrivateKey = _privateKey;
}

std::string SSLConfig::GetPrivateKeyPassword() const {
    return m_PrivateKeyPassword;
}

void SSLConfig::SetPrivateKeyPassword(const std::string& _privateKeyPassword) {
    m_PrivateKeyPassword = _privateKeyPassword;
}

bool SSLConfig::GetVerifyServer() const {
    return m_VerifyServer;
}

void SSLConfig::SetVerifyServer(bool _verifyServer) {
    m_VerifyServer = _verifyServer;
}

MQTTClient_SSLOptions* SSLConfig::PackConfig() const {
    MQTTClient_SSLOptions* config;
    config = new MQTTClient_SSLOptions(InitSecureConnectionConfig());
    config->trustStore = m_TrustStore.empty() ? NULL : Utility::DuplicateString(m_TrustStore);
    config->keyStore = m_KeyStore.empty() ? NULL : Utility::DuplicateString(m_KeyStore);
    config->privateKey = m_PrivateKey.empty() ? NULL : Utility::DuplicateString(m_PrivateKey);
    config->privateKeyPassword = m_PrivateKeyPassword.empty() ? NULL : Utility::DuplicateString(m_PrivateKeyPassword);
    config->enableServerCertAuth = m_VerifyServer ? 1 : 0;
    return config;
}

void SSLConfig::FreeConfig(MQTTClient_SSLOptions* _config) {
    MQTTClient_SSLOptions* config = _config;
    SAFE_DELETE_ARRAY_O(config->privateKeyPassword);
    SAFE_DELETE_ARRAY_O(config->privateKey);
    SAFE_DELETE_ARRAY_O(config->keyStore);
    SAFE_DELETE_ARRAY_O(config->trustStore);
    SAFE_DELETE_O(config);
}

ConnectConfig::ConnectConfig(const std::string& _username,
                               const std::string& _password,
                               bool _cleanSession,
                               int _keepAliveInterval,
                               int _timeout,
                               int _retryinterval,
                               bool _reliable):
    m_Status(NULL),
    m_SSL(NULL) {
    m_Username = _username;
    m_Password = _password;
    m_CleanSession = _cleanSession;
    m_KeepAliveInterval = _keepAliveInterval;
    m_Timeout = _timeout;
    m_RetryInterval = _retryinterval;
    m_Reliable = _reliable;
}

int ConnectConfig::GetKeepAliveInterval() const {
    return m_KeepAliveInterval;
}

void ConnectConfig::SetKeepAliveInterval(int _keepAliveInterval) {
    m_KeepAliveInterval = _keepAliveInterval;
}

bool ConnectConfig::GetCleanSession() const {
    return m_CleanSession;
}

void ConnectConfig::SetCleanSession(bool _cleanSession) {
    m_CleanSession = _cleanSession;
}

bool ConnectConfig::IsReliable() const {
    return m_Reliable;
}

void ConnectConfig::SetReliable(bool _reliable) {
    m_Reliable = _reliable;
}

StatusConfig::Ptr ConnectConfig::GetStatus() const {
    return m_Status;
}

void ConnectConfig::SetStatus(const StatusConfig::Ptr& _status) {
    m_Status = _status;
}

std::string ConnectConfig::GetUsername() const {
    return m_Username;
}

void ConnectConfig::SetUsername(const std::string& _username) {
    m_Username = _username;
}

std::string ConnectConfig::GetPassword() const {
    return m_Password;
}

void ConnectConfig::SetPassword(const std::string& _password) {
    m_Password = _password;
}

int ConnectConfig::GetTimeout() const {
    return m_Timeout;
}

void ConnectConfig::SetTimeout(int _timeout) {
    m_Timeout = _timeout;
}

int ConnectConfig::GetRetryInterval() const {
    return m_RetryInterval;
}

void ConnectConfig::SetRetryInterval(int _retryInterval) {
    m_RetryInterval = _retryInterval;
}

SSLConfig::Ptr ConnectConfig::GetSsl() const {
    return m_SSL;
}

void ConnectConfig::SetSsl(const SSLConfig::Ptr& _ssl) {
    m_SSL = _ssl;
}

MQTTClient_connectOptions* ConnectConfig::PackConfig() const {
    MQTTClient_connectOptions* config;
    config = new MQTTClient_connectOptions(InitConnectionConfig());
    config->MQTTVersion = MQTTVERSION_3_1_1;
    config->keepAliveInterval = m_KeepAliveInterval;
    config->cleansession = m_CleanSession ? 1 : 0;
    config->reliable = m_Reliable ? 1 :0;

    if ( !m_Status.IsNull() ) {
        config->will = m_Status->PackConfig();
    }

    config->username = Utility::DuplicateString(m_Username);
    config->password = Utility::DuplicateString(m_Password);
    config->connectTimeout = m_Timeout;
    config->retryInterval = m_RetryInterval;

    if ( !m_SSL.IsNull() ) {
        config->ssl = m_SSL->PackConfig();
    }

    return config;
}

void ConnectConfig::FreeConfig(MQTTClient_connectOptions* _config) {
    MQTTClient_connectOptions* config = _config;

    if ( config->ssl ) {
        SSLConfig::FreeConfig(config->ssl);
    }

    SAFE_DELETE_ARRAY_O(config->password);
    SAFE_DELETE_ARRAY_O(config->username);

    if ( config->will ) {
        StatusConfig::FreeConfig(config->will);
    }

    SAFE_DELETE_O(config);
}
