#include "messenger/Messages.h"

using namespace Messenger;
using namespace Utility;

const std::string MessageImpl::PARAMS = "params";
const std::string MessageImpl::ID = "uid";
const std::string MessageImpl::TS = "ts";
MessageImpl::MessageImpl() : m_Params(NULL), m_Id(rand()) {
    SetTimestamp();
}

MessageImpl::MessageImpl(const MessageImpl &_obj) : m_Params(_obj.m_Params), m_Id(_obj.m_Id), m_Timestamp(_obj.m_Timestamp) {
}

unsigned int MessageImpl::GetId() const {
    return m_Id;
}

const JsonStruct::Ptr MessageImpl::GetParams() const {
    return m_Params;
}

unsigned long long MessageImpl::GetTimestamp() const {
    return m_Timestamp;
}

void MessageImpl::SetId(unsigned int _id) {
    m_Id = _id;
}

void MessageImpl::SetParams(JsonStruct::Ptr _params) {
    m_Params = _params;
}

void MessageImpl::SetTimestamp() {
    struct timeval timeVal;
    gettimeofday(&timeVal, NULL);
    m_Timestamp = timeVal.tv_sec * 1000ULL + timeVal.tv_usec / 1000ULL;
}

void MessageImpl::SetTimestamp(unsigned long long _timestamp) {
    m_Timestamp = _timestamp;
}

std::string MessageImpl::ToString() const {
    JsonStruct ds;
    ds.Insert(ID, m_Id);
    ds.Insert(TS, m_Timestamp);

    if (!m_Params.IsNull()) {
        ds.Insert(PARAMS, *m_Params);
    }

    InflateJson(ds);

    std::string temp = ds.ToString();
    return temp;
}

void MessageImpl::Parse(MessageImpl::Ptr _messageToFill, JsonStruct::Ptr _messageToParse) {
    JsonStruct::Ptr jsonStruct = _messageToParse->ExtractJson(PARAMS);

    _messageToFill->SetId(_messageToParse->Extract(ID).toInt());
    _messageToFill->SetParams(jsonStruct);
    _messageToFill->SetTimestamp(_messageToParse->Extract(TS).toLongLong());
}

const std::string EventMessage::NAME = "name";
const std::string EventMessage::ETS = "ets";
EventMessage::EventMessage() {
    m_Expiration = 0;
}

EventMessage::EventMessage(const EventMessage& _event) : MessageImpl(_event), m_Name(_event.m_Name), m_Expiration(_event.m_Expiration) {

}

std::string EventMessage::GetName() const {
    return m_Name;
}

unsigned long long EventMessage::GetExpiration() const {
    return m_Expiration;
}

void EventMessage::SetName(const std::string& _name) {
    m_Name = _name;
}

void EventMessage::SetExpiration(unsigned long long _expiration) {
    m_Expiration = _expiration;
}

EventMessage::Ptr EventMessage::Parse(Utility::JsonStruct::Ptr _message) {
    EventMessage::Ptr msg(new EventMessage());

    MessageImpl::Parse(msg, _message);
    std::string name = _message->Extract(NAME).toString("");
    if (name.empty()) {
        throw ParseException("Error while parsing event message");
    }
    msg->SetName(name);
    unsigned long ets = _message->Extract(ETS).toLongLong(0);
    if (ets != 0) {
        msg->SetExpiration(ets);
    }

    return msg;
}

void EventMessage::InflateJson(JsonStruct& _obj) const {
    _obj.Insert(EventMessage::NAME, m_Name);
    if (m_Expiration != 0) {
        _obj.Insert(EventMessage::ETS,  m_Expiration);
    }
}

const std::string ResponseMessage::CODE = "code";
const std::string ResponseMessage::MESSAGE = "message";
ResponseMessage::ResponseMessage() : m_Code(0) {
}

ResponseMessage::ResponseMessage(const ResponseMessage &_response) : MessageImpl(_response), m_Code(_response.m_Code), m_Message(_response.m_Message) {
}

int ResponseMessage::GetCode() const {
    return m_Code;
}

std::string ResponseMessage::GetMessage() const {
    return m_Message;
}

void ResponseMessage::SetCode(int _code) {
    m_Code = _code;
}

void ResponseMessage::SetMessage(const std::string& _message) {
    m_Message = _message;
}

ResponseMessage::Ptr ResponseMessage::Parse(Utility::JsonStruct::Ptr _message) {
    ResponseMessage::Ptr msg(new ResponseMessage());

    MessageImpl::Parse(msg, _message);
    int code = _message->Extract(CODE).toInt(9999);
    if (code == 9999) {
        throw ParseException("Error while parsing response message");
    }
    msg->SetCode(code);
    std::string message = _message->Extract(MESSAGE).toString();
    if (!message.empty()) {
        msg->SetMessage(message);
    }

    return msg;
}

void ResponseMessage::InflateJson(JsonStruct& obj) const {
    obj.Insert(ResponseMessage::CODE, m_Code);
    if (!m_Message.empty()) {
        obj.Insert(ResponseMessage::MESSAGE, m_Message);
    }
}

const std::string RequestMessage::SENDER = "sender";
const std::string RequestMessage::NAME = "name";
const std::string RequestMessage::ETS = "ets";
RequestMessage::RequestMessage() : m_Expiration(0), m_CompletionListener(NULL) {
}

RequestMessage::RequestMessage(const RequestMessage &_obj) :
    MessageImpl(_obj),
    m_Sender(_obj.m_Sender),
    m_Name(_obj.m_Name),
    m_Expiration(_obj.m_Expiration),
    m_CompletionListener(_obj.m_CompletionListener) {

}

std::string RequestMessage::GetName() const {
    return m_Name;
}

std::string RequestMessage::GetSender() const {
    return m_Sender;
}

unsigned long long RequestMessage::GetExpiration() const {
    return m_Expiration;
}

void RequestMessage::SetName(const std::string& _name) {
    m_Name = _name;
}

void RequestMessage::SetSender(const std::string& _sender) {
    m_Sender = _sender;
}

void RequestMessage::SetExpiration(unsigned long long _expiration) {
    m_Expiration = _expiration;
}

void RequestMessage::InflateJson(JsonStruct& obj) const {
    obj.Insert(RequestMessage::NAME, m_Name);
    obj.Insert(RequestMessage::SENDER, m_Sender);
    if (m_Expiration != 0) {
        obj.Insert(RequestMessage::ETS, m_Expiration);
    }
}

void RequestMessage::RegisterCompletionListener(ICompletionListener *_completionListener) {
    m_CompletionListener = _completionListener;
}

void RequestMessage::ReportResponse(const ResponseMessage::Ptr _response) {
    if (m_CompletionListener != NULL) {
        m_CompletionListener->OnRequestCompleted(GetId());
    }
}

void RequestMessage::ReportFailure() {
    if (m_CompletionListener != NULL) {
        m_CompletionListener->OnRequestCompleted(GetId());
    }
}

RequestMessage::Ptr RequestMessage::Parse(Utility::JsonStruct::Ptr _message) {
    RequestMessage::Ptr msg(new RequestMessage());

    MessageImpl::Parse(msg, _message);
    std::string name = _message->Extract(NAME).toString("");
    std::string sender = _message->Extract(SENDER).toString("");
    if (name.empty() || sender.empty()) {
        throw ParseException("Error while parsing request message");
    }
    msg->SetName(name);
    msg->SetSender(sender);
    unsigned long long ets = _message->Extract(ETS).toLongLong(0);
    if (ets != 0) {
        msg->SetExpiration(ets);
    }

    return msg;

}

SyncRequest::SyncRequest(const RequestMessage &_request) :
    RequestMessage(_request), m_Response(NULL) {
}

const ResponseMessage::Ptr SyncRequest::GetResponse() const {
    return m_Response;
}

void SyncRequest::Wait(unsigned int _timeout) {
    m_SyncObject.Wait(_timeout);
}

void SyncRequest::ReportResponse(const ResponseMessage::Ptr _response) {
    m_Response = _response;
    m_SyncObject.Notify();
    RequestMessage::ReportResponse(_response);
}

void SyncRequest::ReportFailure() {
    m_SyncObject.Notify();
    RequestMessage::ReportFailure();
}

AsyncRequest::AsyncRequest(const RequestMessage &_request, RequestDelegate::Ptr _delegate) : RequestMessage(_request), m_RequestDelegate(_delegate) {
}

void AsyncRequest::ReportResponse(const ResponseMessage::Ptr _response) {
    if (!_response.IsNull() && !m_RequestDelegate.IsNull()) {
        m_RequestDelegate->ReportSuccess(GetId(), *_response);
    }
    RequestMessage::ReportResponse(_response);
}

void AsyncRequest::ReportFailure() {
    if (!m_RequestDelegate.IsNull()) {
        m_RequestDelegate->ReportFailure(GetId());
    }
    RequestMessage::ReportFailure();
}

RequestDelegate::RequestDelegate() : m_Object(NULL), m_StubSucc(NULL), m_StubFail(NULL) {
}

void RequestDelegate::ReportSuccess(unsigned int _requestId, const ResponseMessage &_response) const {
    (*m_StubSucc)(m_Object, _requestId, _response, m_AutoDelete);
}

void RequestDelegate::ReportFailure(unsigned int _requestId) const {
    (*m_StubFail)(m_Object, _requestId, m_AutoDelete);
}

const std::string StatusMessage::ONLINE = "online";
StatusMessage::StatusMessage() : m_Online(false) {
}

StatusMessage::StatusMessage(const StatusMessage &_status) : MessageImpl(_status), m_Online(_status.m_Online) {
}

bool StatusMessage::GetStatus() const {
    return m_Online;
}

void StatusMessage::SetStatus(bool _online) {
    m_Online = _online;
}

std::string StatusMessage::GenerateWillPayload() {
    StatusMessage msg;
    msg.SetStatus(false);
    msg.SetTimestamp(0);
    return msg.ToString();
}

StatusMessage::Ptr StatusMessage::Parse(Utility::JsonStruct::Ptr _message) {
    StatusMessage::Ptr msg(new StatusMessage());

    MessageImpl::Parse(msg, _message);
    msg->SetStatus(_message->Extract(ONLINE).toBool());
    return msg;
}

void StatusMessage::InflateJson(JsonStruct& obj) const {
    obj.Insert(StatusMessage::ONLINE, m_Online);
}

