#include "messenger/Messenger.h"

using namespace Messenger;
using namespace Utility;

const std::string CMessenger::REQUEST_TOPIC_ELEMENT = "/req";
const std::string CMessenger::RESPONSE_TOPIC_ELEMENT = "/rsp";
const std::string CMessenger::EVENT_TOPIC_ELEMENT = "/evt";
const std::string CMessenger::STS_TOPIC_ELEMENT = "/sts";
const std::string CMessenger::TOPIC_SEPARATOR = "/";
const std::string CMessenger::EMPTY_STRING = "";

Cleaner::CleanupElement::CleanupElement(unsigned int _id, int _timeout) {
    m_Id = _id;
    m_Timeout = _timeout;
}

Cleaner::CleanupElement::CleanupElement(const CleanupElement &_elem) {
    m_Id = _elem.m_Id;
    m_Timeout = _elem.m_Timeout;
}

Cleaner::Cleaner() {
    m_Timer = new Timer(1000);
    m_Timer->StartTimer(Callback<Cleaner>(*this, &Cleaner::OnCleanup));
}

Cleaner::Cleaner(const Cleaner &_cleaner) : m_CleanupList(_cleaner.m_CleanupList) {
    m_Timer = new Timer(1000);
    m_Timer->StartTimer(Callback<Cleaner>(*this, &Cleaner::OnCleanup));
}

void Cleaner::StopCleaning() {
    m_Timer->StopTimer();
    SAFE_DELETE_O(m_Timer);
    CleanupAll();
}

void Cleaner::ScheduleCleanup(unsigned int _id, unsigned int _time) {
    ScopedLock lock(m_Mutex);
    CleanupList::iterator itr;
    CleanupElement::Ptr tmpElem;

    int diff = _time;
    int prevDiff = diff;

    for (itr = m_CleanupList.begin(); itr != m_CleanupList.end(); ++itr) {
        tmpElem = *itr;
        diff -= tmpElem->m_Timeout;

        if (diff <= 0) {
            tmpElem->m_Timeout = abs(diff);
            m_CleanupList.insert(itr, new CleanupElement(_id, prevDiff));
            break;
        }

        prevDiff = diff;
    }

    if (itr == m_CleanupList.end()) {
        m_CleanupList.push_back(new CleanupElement(_id, prevDiff));
    }
}

void Cleaner::OnCleanup(Timer &_timer) {
    ScopedLock lock(m_Mutex);
    CleanupList::iterator itr;
    CleanupElement::Ptr elem;

    int diff = 1;

    for (itr = m_CleanupList.begin(); itr != m_CleanupList.end(); ++itr) {
        elem = *itr;
        elem->m_Timeout = elem->m_Timeout - diff;
        if (elem->m_Timeout < 0) {
            CleanRequest(elem->m_Id);
        } else {
            break;
        }
    }

    m_CleanupList.erase(m_CleanupList.begin(), itr);
}

void Cleaner::CleanupAll() {
    ScopedLock lock(m_Mutex);
    CleanupList::iterator itr;
    CleanupElement::Ptr elem;

    for (itr = m_CleanupList.begin(); itr != m_CleanupList.end(); ++itr) {
        elem = *itr;
        CleanRequest(elem->m_Id);
    }
}

CMessenger::CMessenger(const std::string& _clientId, const std::string &_tag, const std::string &_username, const std::string &_password, const std::string &_uri) :
    m_UserTag(_tag),
    m_ProcessingMsg(false),
    m_Responder(NULL),
    m_DefaultListener(NULL),
    m_Client(_clientId, _uri, *this) {

    m_Client.SetUsername(_username);
    m_Client.SetPassword(_password);

    Message::Ptr sts = new Message();
    sts->SetPayload(StatusMessage::GenerateWillPayload());
    sts->SetQoS(Message::QOS1);
    m_Client.SetStatusConfiguration(m_UserTag + STS_TOPIC_ELEMENT, sts);

    std::string topic = m_UserTag + RESPONSE_TOPIC_ELEMENT;
    m_Client.Subscribe(topic, Message::QOS1);
}

CMessenger::~CMessenger() {
    StopCleaning();
    try {
        if (IsConnected()) {
            Disconnect();
        }
    } catch (const BaseException& exc) {
    }
    CleanRequests();
}

void CMessenger::SetSecureConfiguration(const std::string &_trustStore, const std::string &_keyStore, const std::string &_privateKey, const std::string &_privateKeyPassword) {
    m_Client.SetSecureConfiguration(_trustStore, _keyStore, _privateKey, _privateKeyPassword);
}

void CMessenger::Connect() {
    try {
        m_Client.Connect();

        StatusMessage sts;
        sts.SetStatus(true);
        sts.SetTimestamp();
        std::string payload = sts.ToString();
        Message::Ptr pub = new Message(payload);
        pub->SetQoS(Message::QOS1);

        m_Client.Publish(m_UserTag + STS_TOPIC_ELEMENT, pub);
    } catch (const ProtocolException& exc) {
        throw ProtocolException(exc);
    }
}

void CMessenger::Reconnect() {
    if (m_ProcessingMsg) {
        m_ProcessMsgSync.Reset();
        m_ProcessMsgSync.Wait();
    }

    if (m_Client.IsConnected()) {
        StatusMessage stsMsg;
        stsMsg.SetStatus(false);
        stsMsg.SetTimestamp();
        std::string payload = stsMsg.ToString();
        Message::Ptr sts = new Message(payload);
        sts->SetQoS(Message::QOS1);
        m_Client.ImmediatePublish(m_UserTag + STS_TOPIC_ELEMENT, sts);
        m_Client.Disconnect();
    }

    m_Client.SetCleanSession(true);
    StatusMessage sts;
    sts.SetStatus(true);
    sts.SetTimestamp();
    std::string payload = sts.ToString();
    Message::Ptr pub = new Message(payload);
    pub->SetQoS(Message::QOS1);
    m_Client.Connect();
    m_Client.Publish(m_UserTag + STS_TOPIC_ELEMENT, pub);
}

void CMessenger::Disconnect() {
    if (m_ProcessingMsg) {
        m_ProcessMsgSync.Reset();
        m_ProcessMsgSync.Wait();
    }

    if (m_Client.IsConnected()) {
        StatusMessage stsMsg;
        stsMsg.SetStatus(false);
        stsMsg.SetTimestamp();
        std::string payload = stsMsg.ToString();
        Message::Ptr sts = new Message(payload);
        sts->SetQoS(Message::QOS1);
        m_Client.ImmediatePublish(m_UserTag + STS_TOPIC_ELEMENT, sts);
        m_Client.Disconnect();
    }
}

void CMessenger::DisconnectWithoutStsMessage() {
    if (m_ProcessingMsg) {
        m_ProcessMsgSync.Reset();
        m_ProcessMsgSync.Wait();
    }

    if (m_Client.IsConnected()) {
        m_Client.Disconnect();
    }
}

bool CMessenger::IsConnected() const {
    return m_Client.IsConnected();
}

void CMessenger::IgnoreSessionPresent(bool _ignore) {
    m_Client.IgnoreSessionPresent(_ignore);
}

void CMessenger::RegisterResponder(IResponder *_responder) {
    if (m_Responder != NULL) {
        m_Responder = _responder;
    } else {
        m_Responder = _responder;
        std::string topic = m_UserTag + REQUEST_TOPIC_ELEMENT;
        try {
            m_Client.Subscribe(topic, Message::QOS1);
        } catch (const BaseException& exc) {
            throw ProtocolException("Error while registering responder");
        }
    }
}

void CMessenger::UnregisterResponder() {
    m_Responder = NULL;
    std::string topic = m_UserTag + REQUEST_TOPIC_ELEMENT;

    try {
        m_Client.Unsubscribe(topic);
    } catch (const BaseException& exc) {
        throw ProtocolException("Error while cleaning responder");
    }
}

void CMessenger::ActivateService(unsigned int _groupId, const std::string& _serviceType) {
    std::stringstream ss;
    ss << m_UserTag + TOPIC_SEPARATOR << _groupId << TOPIC_SEPARATOR + _serviceType + REQUEST_TOPIC_ELEMENT;
    std::string topic = ss.str();

    try {
        m_Client.Subscribe(topic, Message::QOS1);
    } catch (const BaseException& exc) {
        throw ProtocolException("Error while activating service");
    }

}

void CMessenger::DeactivateService(unsigned int _groupId, const std::string& _serviceType) {
    std::stringstream ss;
    ss << m_UserTag + TOPIC_SEPARATOR << _groupId << TOPIC_SEPARATOR + _serviceType + REQUEST_TOPIC_ELEMENT;
    std::string topic = ss.str();

    try {
        m_Client.Unsubscribe(topic);
    } catch (const BaseException& exc) {
        throw ProtocolException("Error while deactivating service");
    }
}

void CMessenger::SendEvent(const EventMessage &_event, unsigned int _groupId, const std::string &_serviceType) {
    LOG(Debug, "Sending event: " + _event.ToString());
    std::string topic;

    if (_serviceType.empty()) {
        topic = m_UserTag + EVENT_TOPIC_ELEMENT;
    } else {
        std::stringstream ss;
        ss << m_UserTag + TOPIC_SEPARATOR << _groupId << TOPIC_SEPARATOR + _serviceType + EVENT_TOPIC_ELEMENT;
        topic = ss.str();
    }

    std::string payload = _event.ToString();
    Message::Ptr pub = new Message(payload);
    pub->SetQoS(Message::QOS1);
    try {
        m_Client.Publish(topic, pub);
    } catch (const BaseException& exc) {
        throw ProtocolException("Error while sending event");
    }
}

void CMessenger::RegisterForEvents(IEventListener *_eventListener, const std::string &_tag, unsigned int _groupId, const std::string& _serviceType) {
    ScopedLock lock(m_EventStatusMutex);

    std::string topic;
    if (_serviceType.empty()) {
        topic = m_UserTag + EVENT_TOPIC_ELEMENT;
    } else {
        std::stringstream ss;
        ss << m_UserTag + TOPIC_SEPARATOR << _groupId << TOPIC_SEPARATOR + _serviceType + EVENT_TOPIC_ELEMENT;
        topic = ss.str();
    }
    EventListenersList listeners = m_EventListeners[topic];

    if (listeners.empty()) {
        try {
            m_Client.Subscribe(topic, Message::QOS1);
        } catch (const BaseException& exc) {
            throw ProtocolException("Error while registering for events");
        }
    }
    listeners.push_back(_eventListener);
    m_EventListeners[topic] = listeners;
}

void CMessenger::UnregisterFromEvents(IEventListener *_eventListener, const std::string &_tag, unsigned int _groupId, const std::string &_serviceType) {
    std::string tag;

    if (_serviceType.empty()) {
        tag = _tag + EVENT_TOPIC_ELEMENT;
    } else {
        std::stringstream ss;
        ss << _tag + TOPIC_SEPARATOR << _groupId << TOPIC_SEPARATOR + _serviceType + EVENT_TOPIC_ELEMENT;
        tag = ss.str();
    }

    EventListenersList::iterator itr;
    ScopedLock lock(m_EventStatusMutex);
    EventListenersList listeners = m_EventListeners[tag];
    itr = std::find(listeners.begin(), listeners.end(), _eventListener);

    if (itr != listeners.end()) {
        listeners.erase(itr);
    }

    if (listeners.empty()) {
        try {
            m_Client.Unsubscribe(tag);
        } catch (const BaseException& exc) {
            throw ProtocolException("Error while unregistering for events");
        }
    }

    m_EventListeners[tag] = listeners;
}

void CMessenger::RegisterForStatusMsg(IEventListener *_eventListener, const std::string &_tag) {
    ScopedLock lock(m_EventStatusMutex);
    EventListenersList listeners = m_StatusListeners[_tag];

    if (listeners.empty()) {
        try {
            m_Client.Subscribe(_tag + STS_TOPIC_ELEMENT, Message::QOS1);
        } catch (const BaseException& exc) {
            throw ProtocolException("Error while registering for status message");
        }
    }

    listeners.push_back(_eventListener);
    m_StatusListeners[_tag] = listeners;
}

void CMessenger::UnregisterFromStatusMsg(IEventListener* _eventListener, const std::string& _tag) {
    EventListenersList::iterator itr;
    ScopedLock lock(m_EventStatusMutex);
    EventListenersList listeners = m_StatusListeners[_tag];
    itr = std::find(listeners.begin(), listeners.end(), _eventListener);

    if (itr != listeners.end() ) {
        listeners.erase(itr);
    }

    if (listeners.empty()) {
        try {
            m_Client.Unsubscribe(_tag + STS_TOPIC_ELEMENT);
        } catch (const BaseException& exc) {
            throw ProtocolException("Error while unregistering for status message");
        }
    }

    m_StatusListeners[_tag] = listeners;
}

void CMessenger::Request(RequestMessage::Ptr _request, const std::string &_tag, unsigned int _groupId, const std::string &_serviceType) {
    if (m_Client.IsConnected()) {
        LOG(Debug, "Sending request: " + _request->ToString());
        _request->SetSender(m_UserTag);
        std::string topic;
        if (_serviceType.empty()) {
            topic = _tag + REQUEST_TOPIC_ELEMENT;
        } else {
            std::stringstream ss;
            ss << _tag + TOPIC_SEPARATOR << _groupId << TOPIC_SEPARATOR + _serviceType + REQUEST_TOPIC_ELEMENT;
            topic = ss.str();
        }
        std::string payload = _request->ToString();

        Message::Ptr pub = new Message(payload);
        pub->SetQoS(Message::QOS0);

        ScopedLock lock(m_RequestMutex);

        try {
            m_Client.Publish(topic, pub);
        } catch (const BaseException& exc) {
            throw ProtocolException("Error while sending request");
        }

        CleanRequest(_request->GetId());
        m_Requests[_request->GetId()] = _request;
    } else {
        throw InvalidStateException("Illegal state. Not connected");
    }
}

void CMessenger::CleanRequest(unsigned int _reqId) {
    RequestsMap::iterator itr;
    RequestMessage::Ptr request;

    ScopedLock lock(m_RequestMutex);

    itr = m_Requests.find(_reqId);
    if (itr != m_Requests.end()) {
        request = itr->second;
        m_Requests.erase(itr);
        request->ReportFailure();
    }
}

void CMessenger::CleanRequests() {
    ScopedLock lock(m_RequestMutex);
    m_Requests.clear();
}

User::Ptr CMessenger::GetUser(const std::string &_tag) {
    User::Ptr obj = new User(_tag);
    obj->SetMessenger(this);
    return obj;
}

void CMessenger::RegisterConnectionListener(IConnectionListener *_connectionListener) {
    m_Client.RegisterConnectionListener(_connectionListener);
}

void CMessenger::UnregisterConnectionListener() {
    m_Client.UnregisterConnectionListener();
}

void CMessenger::RegisterDefaultListener(IEventListener *_listener) {
    m_DefaultListener = _listener;
}

void CMessenger::RegisterLogListener(ILogListener *_logListener) {
    Logger::RegisterLogListener(_logListener);
}

void CMessenger::UnregisterLogListener() {
    Logger::UnregisterLogListener();
}

void CMessenger::ProcessEvent(const std::string &_senderTag, const JsonStruct::Ptr _payload) {
    EventListenersMap::iterator itr;
    ScopedLock lock(m_EventStatusMutex);
    itr = m_EventListeners.find(_senderTag);
    EventListenersList listeners;

    if (itr != m_EventListeners.end()) {
        listeners = itr->second;
    } else if (m_DefaultListener) {
        listeners.push_back(m_DefaultListener);
    }

    if (!listeners.empty()) {
        EventMessage::Ptr event;
        event = EventMessage::Parse(_payload);
        EventListenersList::iterator eventItr;

        for (eventItr = listeners.begin(); eventItr != listeners.end(); ++eventItr) {
            IEventListener* lsnr = *eventItr;
            lsnr = lsnr ? lsnr : m_DefaultListener;

            if (lsnr) {
                lsnr->OnEvent(_senderTag, event);
            }
        }
    }
}

void CMessenger::ProcessStatus(const std::string &_senderTag, const JsonStruct::Ptr _payload) {
    EventListenersMap::iterator itr;
    ScopedLock lock(m_EventStatusMutex);
    itr = m_EventListeners.find(_senderTag);
    EventListenersList listeners;

    if (itr != m_EventListeners.end()) {
        listeners = itr->second;
    } else if (m_DefaultListener) {
        listeners.push_back(m_DefaultListener);
    }

    if (!listeners.empty()) {
        StatusMessage::Ptr sts;
        sts = StatusMessage::Parse(_payload);
        EventListenersList::iterator statusItr;

        for (statusItr = listeners.begin(); statusItr != listeners.end(); ++statusItr) {
            IEventListener* lsnr = *statusItr;
            lsnr = lsnr ? lsnr : m_DefaultListener;

            if (lsnr) {
                lsnr->OnStatusMessage(_senderTag, sts);
            }
        }
    }
}

void CMessenger::ProcessRequest(const JsonStruct::Ptr _payload, unsigned int _groupId, const std::string &_serviceType) {
    if (m_Responder != NULL) {
        RequestMessage::Ptr req;
        req = RequestMessage::Parse(_payload);
        ResponseMessage::Ptr rsp = NULL;

        LOG(Debug, "Request received: " + req->ToString());
        if (_serviceType.empty()) {
            rsp = m_Responder->Respond(req);
        } else {
            rsp = m_Responder->ServiceRespond(_groupId, _serviceType, req);
        }

        if (!rsp.IsNull()) {
            rsp->SetId(req->GetId());
            LOG(Debug, "Sending response: " + rsp->ToString());
            SendResponse(req->GetSender(), rsp);
        }
    }
}

void CMessenger::ProcessFailedRequest(const JsonStruct::Ptr _payload) {
    RequestMessage::Ptr req;
    req = RequestMessage::Parse(_payload);
    CleanRequest(req->GetId());
}

void CMessenger::ProcessResponse(const JsonStruct::Ptr _payload) {
    ResponseMessage::Ptr rsp;
    rsp = ResponseMessage::Parse(_payload);
    LOG(Debug, "Response received: " + rsp->ToString());
    m_RequestMutex.Lock();
    RequestsMap::iterator itr = m_Requests.find(rsp->GetId());

    if (itr != m_Requests.end()) {
        RequestMessage::Ptr req = itr->second;
        m_Requests.erase(itr);
        m_RequestMutex.Unlock();
        req->ReportResponse(rsp);
    } else {
        m_RequestMutex.Unlock();
    }
}

void CMessenger::SendResponse(const std::string &_receiverTag, const ResponseMessage::Ptr _response) {
    std::string topic;

    if (!_response.IsNull()) {
        topic = _receiverTag + RESPONSE_TOPIC_ELEMENT;
        std::string payload = _response->ToString();
        Message::Ptr pubmsg = new Message(payload);
        pubmsg->SetQoS(Message::QOS0);

        try {
            m_Client.Publish(topic, pubmsg);
        } catch (const BaseException& exc) {
            throw BaseException("Error while sending response");
        }
    }
}

bool StartsWith(const std::string& _string, const std::string& _starting) {
    unsigned int stringSize = _string.length();
    unsigned int startingSize = _starting.length();

    if (stringSize >= startingSize) {
        return (0 == _string.compare(0, startingSize, _starting));
    } else {
        return false;
    }
}

bool EndsWith(const std::string& _string, const std::string& _ending) {
    unsigned int stringSize = _string.length();
    unsigned int endingSize = _ending.length();

    if (stringSize >= endingSize) {
        return (0 == _string.compare(stringSize - endingSize, endingSize, _ending));
    } else {
        return false;
    }
}

std::string Replace(std::string& _str, const std::string& _from, const std::string& _to) {
    size_t position = 0;

    // std::string result(_str);

    while((position = _str.find(_from)) != std::string::npos) {
        _str.replace(position, _from.size(), _to);
    }

    return _str;
}

void CMessenger::OnMessageArrived(const std::string &_topic, const Message::Ptr _msg) {
    std::string topic = _topic;
    JsonStruct::Ptr payload;

    payload = ParseObject(_msg->GetPayload());
    m_ProcessingMsg = true;

    try {
        if (EndsWith(topic, EVENT_TOPIC_ELEMENT)) {
            topic = Replace(topic, EVENT_TOPIC_ELEMENT, EMPTY_STRING);
            ProcessEvent(topic, payload);
        } else if (EndsWith(topic, REQUEST_TOPIC_ELEMENT)) {
            topic = Replace(topic, REQUEST_TOPIC_ELEMENT, EMPTY_STRING);
            if (StartsWith(topic, m_UserTag)) {
                std::string serviceFunctionality = Replace(topic, m_UserTag, EMPTY_STRING);
                unsigned int groupId = 0;
                std::string serviceType;
                if (!serviceFunctionality.empty()) {
                    serviceFunctionality.replace(0, 1, EMPTY_STRING);
                    groupId = StoI(serviceFunctionality.substr(0, serviceFunctionality.find(TOPIC_SEPARATOR)));
                    serviceType = serviceFunctionality.substr(serviceFunctionality.find(TOPIC_SEPARATOR) + 1);
                }

                ProcessRequest(payload, groupId, serviceType);
            }
        } else if (EndsWith(topic, RESPONSE_TOPIC_ELEMENT)) {
            topic = Replace(topic, RESPONSE_TOPIC_ELEMENT, EMPTY_STRING);

            if (topic == m_UserTag) {
                ProcessResponse(payload);
            }
        } else if (EndsWith(topic, STS_TOPIC_ELEMENT)) {
            topic = Replace(topic, STS_TOPIC_ELEMENT, EMPTY_STRING);

            ProcessStatus(m_UserTag, payload);
        } else {
            return;
        }
    } catch (const BaseException& exc) {
    }

    m_ProcessingMsg = false;
    m_ProcessMsgSync.Set();
}

void CMessenger::OnPublishFailed(int _code, const std::string &_topic, const Message::Ptr _msg) {
    if (EndsWith(_topic, REQUEST_TOPIC_ELEMENT)) {
        JsonStruct::Ptr payload;

        payload = ParseObject(_msg->GetPayload());
        try {
            ProcessFailedRequest(payload);
        } catch (const BaseException& exc) {
        }
    } else {
    }
}

JsonStruct::Ptr CMessenger::ParseObject(const std::string &_object) {
    JsonStruct::Ptr temp = new JsonStruct(_object);

    return temp;
}
