#include "messenger/User.h"
#include "messenger/Messenger.h"

using namespace Messenger;
using namespace Utility;

ServiceFunctionality::ServiceFunctionality(unsigned int _groupId, const std::string& _serviceType) : m_GroupId(_groupId), m_ServiceType(_serviceType), m_User(NULL) {
}

void ServiceFunctionality::SetOwner(User *_user) {
    m_User = _user;
}

void ServiceFunctionality::RegisterForEvents(IEventListener *_eventListener) {
    if (m_User != NULL) {
        m_User->RegisterForEvents(_eventListener, m_GroupId, m_ServiceType);
    } else {
        throw InvalidStateException("Message user is not set");
    }
}

void ServiceFunctionality::UnregisterFromEvents() {
    if (m_User != NULL) {
        m_User->UnregisterFromEvents(m_GroupId, m_ServiceType);
    } else {
        throw InvalidStateException("Message user is not set");
    }
}

const ResponseMessage::Ptr ServiceFunctionality::Request(const RequestMessage& _request, unsigned int _timeout) {
    if (m_User != NULL) {
        return m_User->Request(_request, m_GroupId, m_ServiceType, _timeout);
    } else {
        throw InvalidStateException("Message user is not set");
    }
}

void ServiceFunctionality::Request(const RequestMessage& _request, const RequestDelegate::Ptr _requestDelegate, unsigned int _timeout) {
    if (m_User != NULL) {
        m_User->Request(_request, _requestDelegate, m_GroupId, m_ServiceType, _timeout);
    } else {
        throw InvalidStateException("Message user is not set");
    }
}

User::User(const std::string &_tag) : m_Tag(_tag) {

}

User::User(const User &_obj) : m_Tag(_obj.m_Tag), m_Messenger(_obj.m_Messenger), m_EventListener(_obj.m_EventListener) {
}

User::~User() {
}

void User::SetMessenger(CMessenger *_messenger) {
    m_Messenger = _messenger;
}

std::string User::GetTag() const {
    return m_Tag;
}

void User::RegisterForEvents(IEventListener *_eventListener) {
    RegisterForEvents(_eventListener, 0, "");
}

void User::UnregisterFromEvents() {
    UnregisterFromEvents(0, "");
}

void User::RegisterForStatusMsg(IEventListener *_statusMsgListener) {
    if (m_Messenger != NULL) {
        m_Messenger->RegisterForStatusMsg(_statusMsgListener, m_Tag);
        m_StatusMsgListener = _statusMsgListener;
    } else {
        throw InvalidStateException("Messenger pointer is not valid");
    }
}

void User::UnregisterFromStatusMsg() {
    if (m_Messenger != NULL) {
        m_Messenger->UnregisterFromStatusMsg(m_StatusMsgListener, m_Tag);
    } else {
        throw InvalidStateException("Messenger pointer is not valid");
    }
}

void User::RegisterForEvents(IEventListener *_eventListener, unsigned int _groupId, const std::string &_serviceType) {
    if (m_Messenger != NULL) {
        m_Messenger->RegisterForEvents(_eventListener, m_Tag, _groupId, _serviceType);
        m_EventListener = _eventListener;
    } else {
        throw InvalidStateException("Messenger pointer is not valid");
    }
}

void User::UnregisterFromEvents(unsigned int _groupId, const std::string &_serviceType) {
    if (m_Messenger != NULL) {
        m_Messenger->UnregisterFromEvents(m_EventListener, m_Tag, _groupId, _serviceType);
    } else {
        throw InvalidStateException("Messenger pointer is not valid");
    }
}

const ResponseMessage::Ptr User::Request(const RequestMessage &_request, unsigned int _timeout) {
    return Request(_request, 0, "", _timeout);
}

const ResponseMessage::Ptr User::Request(const RequestMessage &_request, unsigned int _groupId, const std::string &_serviceType, unsigned int _timeout) {
    if (m_Messenger != NULL) {
        SyncRequest::Ptr request = new SyncRequest(_request);

        m_Messenger->Request(request, m_Tag, _groupId, _serviceType);

        try {
            request->Wait(_timeout * 1000);
        } catch (const BaseException& exc) {
            m_Messenger->CleanRequest(request->GetId());
            throw TimeoutException("Time limit exceeded");
        }

        const ResponseMessage::Ptr response = request->GetResponse();

        if (response.IsNull()) {
            throw NullPointerException("Response failed");
        }

        m_Messenger->CleanRequest(request->GetId());
        return response;
    } else {
        throw InvalidStateException("Messenger pointer is not valid");
    }
}

void User::Request(const RequestMessage& _request, const RequestDelegate::Ptr _requestDelegate, unsigned int _timeout) {
    return Request(_request, _requestDelegate, 0, "", _timeout);
}

void User::Request(const RequestMessage& _request, const RequestDelegate::Ptr _requestDelegate, unsigned int _groupId, const std::string &_serviceType, unsigned int _timeout) {
    if (m_Messenger != NULL) {
        if (!_requestDelegate.IsNull()) {
            AsyncRequest::Ptr request = new AsyncRequest(_request, _requestDelegate);
            m_Messenger->Request(request, m_Tag, _groupId, _serviceType);
            m_Messenger->ScheduleCleanup(request->GetId(), _timeout);
        } else {
            throw InvalidArgumentException("Delegate argument is not valid");
        }
    } else {
        throw InvalidStateException("Messenger pointer is not valid");
    }
    return Request(_request, _requestDelegate, _groupId, _serviceType, _timeout);
}

void User::SendResponse(unsigned int _requestId, ResponseMessage::Ptr _response) {
    if (m_Messenger != NULL) {
        if (!_response.IsNull()) {
            _response->SetId(_requestId);
            m_Messenger->SendResponse(m_Tag, _response);
        } else {
            throw NullPointerException("Response message is null");
        }
    } else {
        throw InvalidStateException("Messenger pointer is not valid");
    }
}

ServiceFunctionality::Ptr User::GetServiceFunctionality(unsigned int _groupId, const std::string& _serviceType) {
    ServiceFunctionality::Ptr serviceFunct = new ServiceFunctionality(_groupId, _serviceType);
    serviceFunct->SetOwner(this);
    return serviceFunct;
}
