#include "ssdp/SSDPClient.h"

using namespace rapidxml;
using namespace SSDP;
using namespace Utility;

const int SSDPClient::SSDP_PORT = 1900;
const int SSDPClient::MAX_BUFFER_LENGTH = 2048;
const std::string SSDPClient::HUB_URN = "urn:wise-com:device:WISEhub:1";
const int SSDPClient::TIME_FOR_DATA = 5;

SSDPClient::SSDPClient(long _searchInterval) {
    m_IsTerminated = true;
    m_ShouldRun = false;
    m_Processing = false;

    SocketAddress sa("", Socket::MULTICAST_PORT);
    try {
        m_Socket = new Socket(sa);
        m_Socket->ReuseSocket(true);
        m_Socket->SetNonBlocking(true);
        m_Socket->IncludeLoopback(true); // false
        SocketAddress msa(Socket::MULTICAST_ADDR, Socket::MULTICAST_PORT);
        m_Socket->JoinGroup(msa);
        m_Socket->BindSocket();
    } catch (const InvalidStateException& exc) {
        exc.Rethrow();
    }

    m_SearchTimer = new Timer(_searchInterval);
}

SSDPClient::~SSDPClient() {
    ScopedLock lock(m_Mutex);
    m_ShouldRun = false;
    if (!m_IsTerminated) {
        m_Thread.Join();
    }

    m_Socket->Close();
    SAFE_DELETE_O(m_Socket);
    SAFE_DELETE_O(m_SearchTimer);
}

void SSDPClient::StartSSDP(Mode _mode, const std::string& _hubId) {
    ScopedLock lock(m_Mutex);
    std::stringstream ss;
    ss << "Starting SSDP client <m=" << _mode << " h=" << _hubId << "> ...";
    SSLOG(Debug, ss);
    m_Mode = _mode;
    if (m_Mode == SEARCH) {
        m_SearchTimer->StartTimer(Callback<SSDPClient>(*this, &SSDPClient::SendMSearch));
    }
    m_HubId = _hubId;
    m_ShouldRun = true;
    m_IsTerminated = false;
    m_Thread.Start(*this);
    LOG(Debug, "SSDP client started.");
}

void SSDPClient::StopSSDP() {
    ScopedLock lock(m_Mutex);
    std::stringstream ss;
    ss << "Stopping SSDP client <m=" << m_Mode << "> ...";
    SSLOG(Debug, ss);
    if (m_Mode == SEARCH) {
        m_SearchTimer->StopTimer();
    }
    m_ShouldRun = false;
    if (!m_IsTerminated) {
        m_Thread.Join();
    }
    m_IsTerminated = true;
    LOG(Debug, "SSDP client stopped.");
}

void SSDPClient::SetSearchInterval(long _searchInterval) {
    m_SearchTimer->SetTimeout(_searchInterval);
}

void SSDPClient::RegisterListener(ISearchListener *_listener) {
    m_SearchListener = _listener;
}

void SSDPClient::UnregisterListener() {
    m_SearchListener = NULL;
}

void SSDPClient::RegisterLogListener(ILogListener *_logListener) {
    Logger::RegisterLogListener(_logListener);
}

void SSDPClient::UnregisterLogListener() {
    Logger::UnregisterLogListener();
}

void SSDPClient::ParseDeviceXml(char* _xml, const std::string& _location) {
    xml_document<> descriptionXml;
    descriptionXml.parse<0>(_xml);
    xml_node<>* rootNode = descriptionXml.first_node();
    if (rootNode != NULL) {
        xml_node<>* deviceNode = rootNode->first_node("device");
        if (deviceNode != NULL) {
            xml_node<>* domainId = deviceNode->first_node("domainID");
            xml_node<>* homeId = deviceNode->first_node("homeID");
            xml_node<>* hubId = deviceNode->first_node("hubID");
            xml_node<>* brokerPort = deviceNode->first_node("brokerPort");
            if (domainId != NULL && homeId != NULL && hubId != NULL && brokerPort != NULL) {
                if (!m_HubId.empty()) {
                    if (m_HubId != hubId->value()) {
                        return;
                    }
                }
                std::stringstream ss;
                ss << _location << ":" << brokerPort->value();
                HubHolder hubHolder(domainId->value(), homeId->value(), hubId->value(), ss.str());

                std::stringstream ssprint;
                ssprint << "Gateway found: <s=" << hubHolder.GetHubID() << " l=" << hubHolder.GetLocation() << ">";
                SSLOG(Info, ssprint);
                m_SearchListener->OnHubDetected(hubHolder);
                if (m_Mode == SEARCH) {
                    m_SearchTimer->StopTimer();
                }
            }
        }
    }
    m_Processing = false;
}

void SSDPClient::SendMSearch(Timer &_timer) {
    static const std::string searchMsg =
            "M-SEARCH * HTTP/1.1\r\n"
            "HOST: " + Socket::MULTICAST_ADDR + ":" + ItoS(SSDP_PORT) + "\r\n"
            "MAN: \"ssdp:discover\"\r\n"
            "MX: 2\r\n"
            "ST: " + HUB_URN + "\r\n"
            "\r\n";
    std::stringstream ss;
    ss << "Sending MSEARCH <u=" << HUB_URN << ">";
    SSLOG(Debug, ss);
    if (!m_Processing) {
        try {
            m_Socket->SendMulticast((void *) searchMsg.c_str(), searchMsg.length(), SSDP_PORT);
        } catch (const BaseException& exc) {
            LOG(Error, exc.GetMessage());
        }
    }
}

void SSDPClient::ProcessMessage(const std::string& _msg) {
    if ((_msg.find("NOTIFY * HTTP/1.1") != std::string::npos &&
         _msg.find("ssdp:alive") != std::string::npos &&
         m_Mode == LISTEN &&
         _msg.find(HUB_URN) != std::string::npos) ||
         (_msg.find("HTTP/1.1 200 OK") != std::string::npos &&
         m_Mode == SEARCH &&
         _msg.find(HUB_URN) != std::string::npos)) {
        LOG(Trace, _msg);

        m_Processing = true;
        std::string location = _msg.substr(_msg.find("LOCATION: ") + 10);
        location = location.substr(0, location.find_first_of("\n\r"));

        std::stringstream ss;
        ss << "Message arrived <l=" << location << ">";
        SSLOG(Debug, ss);

        std::stringstream out;
        try {
            HTTPClient httpClient(location);
            out << httpClient.Fetch();
        } catch (const BaseException& exc) {
            std::stringstream ss;
            ss << "Failed to send HTTP GET on <l=" << location << " r=" << exc.GetMessage() << ">";
            SSLOG(Error, ss)
            m_Processing = false;
            return;
        }

        std::string xml;
        xml = out.str();
        char* c_xml = new char[xml.length() + 1];
        strcpy(c_xml, xml.c_str());

        std::string url = location;
        url = url.substr(url.find("http://") + 7);
        url = url.substr(0, url.find(":"));
        ParseDeviceXml(c_xml, url);
        delete[] c_xml;
    }
}

void SSDPClient::Run() {
    if (m_Socket->GetSocketFD() != -1) {
        while(m_ShouldRun) {
            static char buffer[MAX_BUFFER_LENGTH];
            memset(buffer, 0, sizeof(buffer));
            std::string sender;
            unsigned short senderPort;

            if (m_Socket->OnDataRead(TIME_FOR_DATA) > 0) {
                if (!m_ShouldRun) {
                    break;
                }
                if (m_Socket->RecvDataGram((void *)buffer, sizeof(buffer), sender, senderPort) != -1) {
                    std::string msg(buffer);
                    LOG(Trace, "Message arrived, processing...");
                    ProcessMessage(msg);
                }
            }
        }
    }
}

HubHolder::HubHolder() {

}

HubHolder::HubHolder(const std::string &_domainId, const std::string &_homeId, const std::string &_hubId, const std::string &_location) :
    m_DomainID(_domainId),
    m_HomeID(_homeId),
    m_HubID(_hubId),
    m_Location(_location) {

}

std::string HubHolder::GetDomainID() const {
    return m_DomainID;
}

std::string HubHolder::GetHomeID() const {
    return m_HomeID;
}

std::string HubHolder::GetHubID() const {
    return m_HubID;
}

std::string HubHolder::GetLocation() const {
    return m_Location;
}

void HubHolder::SetDomainID(const std::string &_domainId) {
    m_DomainID = _domainId;
}

void HubHolder::SetHomeID(const std::string &_homeId) {
    m_HomeID = _homeId;
}

void HubHolder::SetHubID(const std::string &_hubId) {
    m_HubID = _hubId;
}

void HubHolder::SetLocation(const std::string &_location) {
    m_Location = _location;
}
