#include "ssdp/Socket.h"

#include <errno.h>

using namespace SSDP;
using namespace Utility;

const std::string Socket::MULTICAST_ADDR = "239.255.255.250";
const int Socket::MULTICAST_PORT = 1900;

Socket::Socket() {
}

Socket::Socket(const SocketAddress &_address) : m_Family(AF_INET), m_Addr(_address), m_Closed(false) {
    m_Socket = socket(m_Family, SOCK_DGRAM, 0);
    if (m_Socket < 0) {
        m_Socket = -1;
        throw InvalidStateException("Error creating socket: " + std::string(std::strerror(errno)), errno);
    }
}

Socket::~Socket() {
    if (!m_Closed) {
        close(m_Socket);
        m_Closed = true;
    }
    m_Socket = -1;
}

int Socket::GetSocketFD() const {
    return m_Socket;
}

void Socket::Close() {
    close(m_Socket);
    m_Closed = true;
    m_Socket = -1;
}

void Socket::BindSocket() {
    if (bind(m_Socket, m_Addr.GetSocketAddress(), m_Addr.GetSize()) < 0) {
        Close();
        throw InvalidStateException("Binding socket failed: " + std::string(std::strerror(errno)), errno);
    }
}

void Socket::ReuseSocket(bool _reuse) {
    int enable = _reuse ? 1 : 0;
    if (setsockopt(m_Socket, SOL_SOCKET, SO_REUSEADDR, &enable, sizeof(enable)) < 0) {
        Close();
        throw InvalidStateException("Reuse socket failed: " + std::string(std::strerror(errno)), errno);
    }
}

void Socket::JoinGroup(const SocketAddress& _address) {
    struct ip_mreq mreq;

    mreq.imr_multiaddr.s_addr = _address.GetAddress().s_addr;
    mreq.imr_interface.s_addr = htonl(INADDR_ANY);

    if (setsockopt(m_Socket, IPPROTO_IP, IP_ADD_MEMBERSHIP, &mreq, sizeof(mreq)) < 0) {
        Close();
        throw InvalidStateException("Join group failed: " + std::string(std::strerror(errno)), errno);
    }
}

void Socket::IncludeLoopback(bool _include) {
    int enable = _include ? 1 : 0;
    if (setsockopt(m_Socket, IPPROTO_IP, IP_MULTICAST_LOOP, &enable, sizeof(enable)) < 0) {
        Close();
        throw InvalidStateException("Include loopback failed: " + std::string(std::strerror(errno)), errno);
    }
}

int Socket::OnDataRead(unsigned long _timeToWait) {
    fd_set master;
    fd_set read_fds;

    int fdmax;

    FD_ZERO(&master);
    FD_ZERO(&read_fds);

    FD_SET(m_Socket, &master);
    fdmax = m_Socket;

    read_fds = master;

    struct timeval timeout;
    timeout.tv_sec = _timeToWait;
    timeout.tv_usec = 0;
    int rc = select(fdmax + 1, &read_fds, NULL, NULL, &timeout);

    if (rc == -1) {
        //Data exception
    } else if (rc > 0) {
        //Data arrived
    } else if (rc == 0) {
        //Data timed out
    }

    return rc;
}

void Socket::SendDataGram(const void* _buffer, int _bufferLen, const std::string& _foreignAddress, unsigned short _foreignPort) {
    sockaddr_in destAddr = {};

    size_t n;
    destAddr.sin_addr.s_addr = inet_addr(_foreignAddress.c_str());
    destAddr.sin_port = htons(_foreignPort);
    destAddr.sin_family = m_Family;

    if ((n = sendto(m_Socket, (void*) _buffer, _bufferLen, 0, (sockaddr*) &destAddr, sizeof(destAddr))) != (unsigned int) _bufferLen) {
        throw InvalidStateException("Error sending datagram: " + std::string(std::strerror(errno)));
    }
}

int Socket::RecvDataGram(void* _buffer, int _bufferLen, std::string& _sourceAddress, unsigned short& _sourcePort) {
    sockaddr_in sourceAddr = {};

    socklen_t sourceAddrLen = sizeof(sockaddr_in);
    int bytes;
    if ((bytes = recvfrom(m_Socket, (void*) _buffer, _bufferLen, 0, (sockaddr*) &sourceAddr, &sourceAddrLen)) < 0 ) {
        return -1;
    }

    _sourceAddress = inet_ntoa(sourceAddr.sin_addr);
    _sourcePort = ntohs(sourceAddr.sin_port);
    char* receivedData = static_cast<char *>(_buffer);
    receivedData[bytes] = '\0';
    return bytes;
}

void Socket::SendMulticast(const void* _buffer, int _bufferLen, unsigned short _foreignPort) {
    SendDataGram(_buffer, _bufferLen, MULTICAST_ADDR, _foreignPort);
}

void Socket::Send(const void *_buffer, int _bufferLen) {
    if ((send(m_Socket, (void* ) _buffer, _bufferLen, 0)) < 0) {
        return;
    }
}

int Socket::Recv(void *_buffer, int _bufferLen) {
    int bytes;

    if ((bytes = recv(m_Socket, (void *) _buffer, _bufferLen, 0)) < 0) {
        return -1;
    }
    char* dataArrived = static_cast<char *>(_buffer);
    dataArrived[bytes] = '\0';
    return bytes;
}

void Socket::SetNonBlocking(bool _blocking) {
    int opts;

    opts = fcntl(m_Socket, F_GETFL);

    if (opts < 0)
        return;

    if (_blocking) {
        opts = (opts | O_NONBLOCK);
    } else {
        opts = (opts & ~O_NONBLOCK);
    }

    fcntl(m_Socket, F_SETFL, opts);
}
